/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

/* Benchmark concurrency within directories in the R/CW protocol */

#include <math.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>

#include "pasis_config.h"
#include "pasis_debug.h"
#include "pasis_error.h"
#include "pasis_timer.h"

#include "pasisio.h"
#include "pasisio_config_parse.h"
#include "pasisio_md_protocol.h"
#include "pasisio_drives.h"
#include "pasisio_select.h"

#include "pasis_md_prot.h"

#define BLOCK_SIZE 16384

static pasis_config_server_list_t md_server_list;
static pasis_md_protocol_t        md_protocol;
static pasis_md_linkage_t         md_linkages[PASIS_MD_MAX_LINKAGES];

static unsigned int               measure_time=5;
static unsigned int               warmup_time=1;
static unsigned int               req_count=0;

static unsigned int               samples=0;
static p_uint64                   time_total_resp=0;
static p_uint64                   time_total_sq_resp=0;

/* A type definition of a pointer to an NFS test function  */
typedef void (*op_func_t)(int setup_state);



static void
do_unlink_internal(pasis_md_fid_t *parent_dir, pasis_md_fid_t *fid, char* name)
{
    pasis_md_attr_t dir_attr_out;
    pasis_error_t error;
    
    error = PasisIO_MDUnlink (&md_protocol,
                              md_linkages,
                              md_server_list.num_servers,
                              parent_dir,
                              fid,
                              name,
                              &dir_attr_out);
    PASIS_DASSERT(error == PASIS_SUCCESS);
}


void do_create_internal (pasis_md_fid_t *fid, char* filename, char *symlink_name,
                         p_uint32 mode)
{
    pasis_md_attr_t name_attr_in, parent_attr_out, name_attr_out;
    pasis_error_t error;

    /* Setup create attributes */
    name_attr_in.mode = mode;
    name_attr_in.uid = 314;
    name_attr_in.gid = 159;
    name_attr_in.fileid = 20202020;
    memset(&name_attr_in.mtime, 0, sizeof(name_attr_in.mtime));
    memset(&name_attr_in.atime, 0, sizeof(name_attr_in.atime));
    memset(&name_attr_in.ctime, 0, sizeof(name_attr_in.ctime));

    error = PasisIO_MDCreate (&md_protocol,
                              md_linkages,
                              md_server_list.num_servers,
                              (pasis_md_fid_t*) &PASIS_MD_ROOT_FID_C,
                              fid,
                              filename,
                              symlink_name, /* symbolic link */
                              &name_attr_in,
                              &md_protocol,
                              md_linkages,
                              md_server_list.num_servers,
                              &parent_attr_out,
                              &name_attr_out);
    PASIS_DASSERT(error == PASIS_SUCCESS);
}


void do_create_test (int setup_state)
{
    static int need_to_delete_file = 0;
    char* filename = "test-create-file";
    pasis_md_fid_t fid, *parent_dir = (pasis_md_fid_t*)&PASIS_MD_ROOT_FID_C;
    memset (fid.data, 1, PASISIO_FILEID_SIZE);
    
    /* Initialization */
    if (setup_state) {
        if (need_to_delete_file) {
            do_unlink_internal(parent_dir, &fid, filename);
        }
    } else {
        do_create_internal(&fid, filename, NULL, MD_ATTR_MODE_REG);
        need_to_delete_file = 1;
    }
}


void do_link_test (int setup_state)
{
    static int need_to_delete_file = 0, need_to_create_original_fid = 1;
    pasis_md_attr_t dir_attr_out, file_attr_out;
    pasis_md_fid_t fid, *parent_dir = (pasis_md_fid_t*)&PASIS_MD_ROOT_FID_C;
    char* orig_filename = "orig-file", *link_filename = "link-file";
    memset (fid.data, 1, PASISIO_FILEID_SIZE);

    /* Initialization */
    if (setup_state) {
        if (need_to_create_original_fid) {
            do_create_internal(&fid, orig_filename, NULL, MD_ATTR_MODE_REG);
            need_to_create_original_fid = 0;
        }

        if (need_to_delete_file) {
            do_unlink_internal(parent_dir, &fid, link_filename);
        }
    } else {
        pasis_error_t error;
        error = PasisIO_MDLink (&md_protocol,
                                md_linkages,
                                md_server_list.num_servers,
                                &fid,
                                parent_dir,
                                link_filename,
                                &file_attr_out,
                                &dir_attr_out);
        PASIS_DASSERT(error == PASIS_SUCCESS);
        need_to_delete_file = 1;
    }
}


void do_unlink_test (int setup_state)
{
    pasis_md_attr_t dir_attr_out, file_attr_out;
    pasis_md_fid_t fid, *parent_dir = (pasis_md_fid_t*)&PASIS_MD_ROOT_FID_C;
    char* filename = "unlink-filename";
    memset (fid.data, 1, PASISIO_FILEID_SIZE);

    /* Initialization */
    if (setup_state) {
        do_create_internal(&fid, filename, NULL, MD_ATTR_MODE_REG);
    } else {
        do_unlink_internal(parent_dir, &fid, filename);
    }
}


/* used 3 objects */
void do_rename_test (int setup_state)
{
    static int need_to_delete_file = 0, need_to_create_other_dir = 1;
    char* filename = "test-rename-file", *other_dir_name = "other-directory";;
    pasis_md_fid_t fid, other_dir_fid, bad_fid;
    memset (fid.data, 1, PASISIO_FILEID_SIZE);
    memset (other_dir_fid.data, 2, PASISIO_FILEID_SIZE);
    memset (bad_fid.data, 13, PASISIO_FILEID_SIZE);
    
    /* Initialization */
    if (setup_state) {
        if (need_to_create_other_dir) {
            do_create_internal(&other_dir_fid, other_dir_name, NULL, MD_ATTR_MODE_DIR);
            need_to_create_other_dir = 0;
        }

        if (need_to_delete_file) {
            do_unlink_internal(&other_dir_fid, &fid, filename);
        }

        do_create_internal(&fid, filename, NULL, MD_ATTR_MODE_REG);
    } else {
        pasis_md_attr_t from_dir_attr_out, to_dir_attr_out;
        pasis_error_t error;

        error = PasisIO_MDRename (&md_protocol,
                                  md_linkages,
                                  md_server_list.num_servers,
                                  (pasis_md_fid_t*)&PASIS_MD_ROOT_FID_C,
                                  &other_dir_fid,
                                  &fid,
                                  &bad_fid, /* to_fid; unused */
                                  filename, /* from_name */
                                  filename, /* to_name */
                                  0, /* to_name_valid */
                                  &from_dir_attr_out,
                                  &to_dir_attr_out);
        PASIS_DASSERT(error == PASIS_SUCCESS);

        need_to_delete_file = 1;
    }
}


void do_setattr_test (int setup_state)
{
    /* Initialization */
    if (setup_state) {
        /* No initialization */
    } else {
        pasis_md_sattr_t sattr;
        pasis_md_attr_t attr_out;
        p_uint32 guard_ctime[2] = {0, 0};
        pasis_error_t error;

        sattr.set_uid = sattr.set_gid = sattr.set_size = 1;
        sattr.set_mode = sattr.set_atime = sattr.set_mtime = 0;
        sattr.uid = 271;
        sattr.gid = 828;
        sattr.size = 123456789;

        error = PasisIO_MDSetattr (&md_protocol,
                                   md_linkages,
                                   md_server_list.num_servers,
                                   (pasis_md_fid_t*) &PASIS_MD_ROOT_FID_C,
                                   &sattr,
                                   guard_ctime,
                                   &attr_out);

        PASIS_DASSERT(error == PASIS_SUCCESS);
    }
}


void do_getattr_test (int setup_state)
{
    /* Initialization */
    if (setup_state) {
        /* No initialization */
    } else {
        pasis_md_attr_t attr_out;
        pasis_md_loc_t distribution_out;
        pasis_error_t error;

        error = PasisIO_MDGetattr (&md_protocol,
                                   md_linkages,
                                   md_server_list.num_servers,
                                   (pasis_md_fid_t*) &PASIS_MD_ROOT_FID_C,
                                   &attr_out,
                                   &distribution_out);

        PASIS_DASSERT(error == PASIS_SUCCESS);
    }
}


void do_lookup_test (int setup_state)
{
    char* filename = "test-lookup-file";
    pasis_md_fid_t fid, fid_out;
    memset (fid.data, 1, PASISIO_FILEID_SIZE);

    /* Initialization */
    if (setup_state) {
        static int first_time = 1;
        if (first_time) {
            do_create_internal(&fid, filename, NULL, MD_ATTR_MODE_REG);
        }
        first_time = 0;
    } else {
        pasis_md_attr_t dir_attr_out;
        pasis_md_loc_t distribution_out;
        pasis_error_t error;

        error = PasisIO_MDLookup (&md_protocol,
                                  md_linkages,
                                  md_server_list.num_servers,
                                  (pasis_md_fid_t*) &PASIS_MD_ROOT_FID_C,
                                  filename,
                                  &fid_out,
                                  &dir_attr_out,
                                  &distribution_out);

        PASIS_DASSERT(error == PASIS_SUCCESS);
    }
}


void do_readlink_test (int setup_state)
{
    char* filename = "test-readlink", *symlink_filename = "symlink";
    pasis_md_fid_t symlink, target;
    memset (symlink.data, 1, PASISIO_FILEID_SIZE); 
    memset (target.data, 2, PASISIO_FILEID_SIZE); 
    
    /* Initialization */
    if (setup_state) {
        static int first_time = 1;
        if (first_time) {
            do_create_internal(&target, filename, NULL, MD_ATTR_MODE_REG);
            do_create_internal(&symlink, symlink_filename, filename, MD_ATTR_MODE_LNK);
        }
        first_time = 0;
    } else {
        char path[1024];
        pasis_md_attr_t fid_attr_out;
        pasis_error_t error;
        error = PasisIO_MDReadlink (&md_protocol,
                                    md_linkages,
                                    md_server_list.num_servers,
                                    &symlink,
                                    path,
                                    &fid_attr_out);
        PASIS_DASSERT(error == PASIS_SUCCESS);
    }
}


void do_readdir_test (int setup_state)
{
    /* Initialization */
    if (setup_state) {
        /* No setup necessary */
    } else {
        p_byte buffer[4096];
        pasis_md_attr_t dir_attr_out;
        pasis_error_t error;
        error = PasisIO_MDReaddir (&md_protocol,
                                   md_linkages,
                                   md_server_list.num_servers,
                                   (pasis_md_fid_t*)&PASIS_MD_ROOT_FID_C,
                                   0, /* block # */
                                   buffer,
                                   &dir_attr_out);
        PASIS_DASSERT(error == PASIS_SUCCESS);
    }
}


static inline double
calc_stdv (double sum, double sq_sum, double samples)
{
    double local_stdv;

    local_stdv = sqrt (sq_sum/samples - ((sum/samples) * (sum/samples)));

    return local_stdv;
}


static void
issue_req (op_func_t op, int measure)
{
    pasis_timer_t timer;
    pasis_timespec_t  ts;
    p_uint64 time;
    pasis_error_t error;

    /* Setup operation state */
    op (1);

    /* Execute and measure operation */
    PASIS_TIMER_START_TM (&timer);
    op (0);
    PASIS_TIMER_STOP_TM (&timer);
    PASIS_TIMER_ELAPSED_TS (&timer, &ts);

    if (measure) {
        samples++;
        time = ts.ts_sec * 1000000 + ts.ts_nsec / 1000;
        time_total_resp += time;
        time_total_sq_resp += (time * time);
    }
}


static void
issue_requests (op_func_t op) 
{
    int     end_time;
    double  stdv;
    double  mean;

    /* Warmup phase first */
    end_time = st_time () + warmup_time;
    while (st_time () < end_time) {
        issue_req (op, 0);
        req_count++;
    }

    /* Calibration */

    /* Measurement */
    end_time = st_time () + measure_time;
    while (st_time () < end_time) {
        issue_req (op, 1);
        req_count++;
    }
        
    /* Slowdown */
    end_time = st_time () + warmup_time;
    while (st_time () < end_time) {
        issue_req (op, 0);
        req_count++;
    }


    stdv = calc_stdv (time_total_resp, time_total_sq_resp, samples);
    mean = time_total_resp / samples;

    printf ("--- Results ---\n");
    printf ("Issued %d reqs in %d seconds\n", req_count, 
            2 * warmup_time + measure_time);
    printf ("Samples during measure:   %d\n", samples); 
    printf ("Mean response time    :   %.4fms\n", mean / 1000.0);
    printf ("Standard deviation    :   %.4fms\n\n", stdv / 1000.0);
}


static int
init (void)
{
    pasis_error_t error;

    st_init ();
    st_timecache_set (1);

    Pasis_TimeInit ();

    /* Initialize PASIS */
    error = PasisIO_Init (1, BLOCK_SIZE);
    if (error) {
        PASIS_PRINT_ERR ("PasisIO_Init error: %s\n", PASIS_STRERROR (error));
        return -1;
    }

    /* Add storage nodes to PASIS drives module */
    error = PasisIO_ConnectToStorageNodes (&md_server_list, 
                                           PASISIO_DRIVE_TYPE_MD_INDEX,
                                           md_linkages, NULL);
    if (error) {
        PASIS_PRINT_ERR ("IOInit Error while connecting to md drives\n");
        exit (1);
    }
    
    return 0;
}


static void
usage (char *name)
{
    PASIS_PRINT_INFO ("Usage: %s -r <file> -o <op> [-m #]\n", name);
    PASIS_PRINT_INFO (" -r <file> root config file (for protocol/linkages)\n");
    PASIS_PRINT_INFO (" -o <op>   operation (e.g. create, unlink, etc...)\n");
    PASIS_PRINT_INFO (" -m <time> measurement time\n");
    PASIS_PRINT_INFO (" -w <time> warmup/slowdown time\n");
}


int
main (int argc, char *argv[])
{
    pasis_error_t       error;
    char               *md_cfg_file = NULL;
    int                 i, do_format=0;
    char               *operation = NULL;
    op_func_t           op_func = NULL;
    char                c;

    pasis_debug_print_enable = 0;

    while ((c = getopt (argc, argv, "w:m:r:t:o:")) != -1) {
        switch (c) {
        case 'r':
            md_cfg_file = optarg;
            break;
        case 'm':
            sscanf (optarg, "%u", &measure_time);
            break;
        case 'w':
            sscanf (optarg, "%u", &warmup_time);
            break;
        case 'o':
            operation = optarg;
            break;
        default:
            usage (argv[0]);
            exit (1);
        }
    }

    if (md_cfg_file == NULL || operation == NULL) {
        usage (argv[0]);
        exit (1);
    }

    /* Parse config file, which contains protocol/scheme/storage node list */
    if (PasisIO_ParseConfigurationFile (md_cfg_file, &md_protocol, &md_server_list,
                                        NULL, NULL)) {
        PASIS_PRINT_ERR ("Error: configuration file parsing failed\n");
        exit (1);
    }

    if (init ()) {
        exit (1);
    }

    if (!strcmp (operation, "create")) {
        op_func = do_create_test;
    } else if (!strcmp (operation, "setattr")) {
        op_func = do_setattr_test;
    } else if (!strcmp (operation, "getattr")) {
        op_func = do_getattr_test;
    } else if (!strcmp (operation, "lookup")) {
        op_func = do_lookup_test;
    } else if (!strcmp (operation, "rename")) {
        op_func = do_rename_test;
    } else if (!strcmp (operation, "link")) {
        op_func = do_link_test;
    } else if (!strcmp (operation, "unlink")) {
        op_func = do_unlink_test;
    } else if (!strcmp (operation, "readlink")) {
        op_func = do_readlink_test;
    } else if (!strcmp (operation, "readdir")) {
        op_func = do_readdir_test;
#if 0
    } else if (!strcmp (operation, "timer")) {
        op_func = do_timer_test_cputicks;
#endif
    } else {
        printf ("Bad operation name\n");
        exit (-1);
    }

    issue_requests (op_func);

    PasisIO_Shutdown ();

    exit (0);
}
