/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

/*
 * file: pasis_freelist.h
 * author: Garth Goodson
 *
 *
 * This file provides routines for manipulating freelists (pools of 
 * preallocated memory)
 * Routines in this file are based on the NASD source.
 *
 */

#ifndef __PASIS_FREELIST_H__
#define __PASIS_FREELIST_H__

#include "pasis_config.h"
#include "pasis_debug.h"

#include "pasis_config.h"
#include "pasis_error.h"

#ifdef __cplusplus
extern "C" {
#endif
#if 0
}
#endif

#if PASIS_FREELIST_STATS > 0

#include <string.h>                /* for memset */

typedef struct pasis_freelist_stats_s {
    char *file;
    int  line;
    int  allocations;
    int  frees;
    int  maxFree;
    int  grows;
    int  outstanding;
    int  maxOutstanding;
    int  primeCount;
    int  empty;
} pasis_freelist_stats_t;

#endif /* PASIS_FREELIST_STATS */

typedef struct pasis_freelist_obj_s {
    struct pasis_freelist_obj_s *next;
} pasis_freelist_obj_t;


typedef struct pasis_freelist_s {
    pasis_freelist_obj_t *head;   /* Head of object list */
    pasis_freelist_obj_t *tail;   /* Tail of object list */
    int freeCnt;                  /* Free object count */
    int maxCnt;                   /* Max object count - before freeing */
    int objInc;                   /* Number of objects to allocate at a time */
    int objSize;                  /* Size of each object */
    void (*initFunc)(void *arg);  /* Constructor -- called for each elem */
    void (*destroyFunc)(void *arg); /* Destructor -- called for each elem */
#if PASIS_FREELIST_STATS > 0
    pasis_freelist_stats_t stats;
#endif /* PASIS_FREELIST_STATS */
} pasis_freelist_t;


#if PASIS_FREELIST_STATS > 0

static inline void
PasisFreelist_StatsInit(pasis_freelist_t *freelist, 
                          char *file, 
                          int line) 
{
    memset(&freelist->stats, 0, sizeof(pasis_freelist_stats_t));

    if (freelist->freeCnt > freelist->stats.maxFree) {
        freelist->stats.maxFree = freelist->freeCnt;
    }

    freelist->stats.primeCount = freelist->freeCnt;
    freelist->stats.file = file;
    freelist->stats.line = line;

    return;
}

static inline void
PasisFreelist_StatsEmpty(pasis_freelist_t *freelist)
{
    freelist->stats.empty++;
    return;
}

static inline void
PasisFreelist_StatsGet(pasis_freelist_t *freelist, 
                         int allocs) 
{
    if (allocs > 0) {
        freelist->stats.grows++;
    }
    freelist->stats.allocations += allocs;
    freelist->stats.outstanding++;
    if (freelist->stats.outstanding > freelist->stats.maxOutstanding) {
        freelist->stats.maxOutstanding = freelist->stats.outstanding;
    }
    if (freelist->freeCnt > freelist->stats.maxFree) {
        freelist->stats.maxFree = freelist->freeCnt;
    }

    return;
}

static inline void
PasisFreelist_StatsPut(pasis_freelist_t *freelist)
{
    freelist->stats.frees++;
    freelist->stats.outstanding--;
    if (freelist->freeCnt > freelist->stats.maxFree) {
        freelist->stats.maxFree = freelist->freeCnt;
    }

    return;
}

static inline void
PasisFreelist_StatsDestroy(pasis_freelist_t *freelist)
{
    PASIS_PRINT_INFO("\n---Freelist Stats---\n");
    PASIS_PRINT_INFO("Created in      : %s on line %d\n", freelist->stats.file, 
                     freelist->stats.line);
    PASIS_PRINT_INFO("Allocations     : %d\n", freelist->stats.allocations);
    PASIS_PRINT_INFO("Frees           : %d\n", freelist->stats.frees);
    PASIS_PRINT_INFO("Max Free Count  : %d\n", freelist->stats.maxFree);
    PASIS_PRINT_INFO("Grows (incr=%03d): %d\n", freelist->objInc, 
                     freelist->stats.grows);
    PASIS_PRINT_INFO("Outstanding     : %d\n", freelist->stats.outstanding);
    PASIS_PRINT_INFO("Max Outstanding : %d\n", freelist->stats.maxOutstanding);
    PASIS_PRINT_INFO("Gets of NULL (empty): %d\n", freelist->stats.empty);
    return;
}

#define PasisFreelist_CreateInit(objSize, maxCnt, objInc, primeCnt, initFunc, destroyFunc) \
  PasisFreelist_Create2(objSize, maxCnt, objInc, primeCnt, initFunc, destroyFunc, __FILE__, __LINE__)

#define PasisFreelist_CreateInitFixed(objSize, maxCnt, initFunc, destroyFunc) \
  PasisFreelist_Create2(objSize, maxCnt, 0, maxCnt, initFunc, destroyFunc, __FILE__, __LINE__)

#define PasisFreelist_Create(objSize, maxCnt, objInc, primeCnt) \
  PasisFreelist_Create2(objSize, maxCnt, objInc, primeCnt, NULL, NULL, __FILE__, __LINE__)

#define PasisFreelist_CreateFixed(objSize, maxCnt) \
  PasisFreelist_Create2(objSize, maxCnt, 0, maxCnt, NULL, NULL, __FILE__, __LINE__)

#else /* #if PASIS_FREELIST_STATS > 0 */

#define PasisFreelist_CreateInit(objSize, maxCnt, objInc, primeCnt, initFunc, destroyFunc) \
  PasisFreelist_Create2(objSize, maxCnt, objInc, primeCnt, initFunc, destroyFunc)

#define PasisFreelist_CreateInitFixed(objSize, maxCnt, initFunc, destroyFunc) \
  PasisFreelist_Create2(objSize, maxCnt, 0, maxCnt, initFunc, destroyFunc)

#define PasisFreelist_Create(objSize, maxCnt, objInc, primeCnt) \
  PasisFreelist_Create2(objSize, maxCnt, objInc, primeCnt, NULL, NULL)

#define PasisFreelist_CreateFixed(objSize, maxCnt) \
  PasisFreelist_Create2(objSize, maxCnt, 0, maxCnt, NULL, NULL)

#endif /* ! #if PASIS_FREELIST_STATS > 0 */

/* 
 * PasisFreelist_Create
 *   Takes parameters of freelist to create and creates a freelist 
 *   Returns created freelist
 */

static inline pasis_freelist_t *
PasisFreelist_Create2(unsigned int objSize,  /* Size of each object */ 
                      int maxCnt,   /* Max number of elements */
                      int objInc,   /* Number of elements to allocate at a time */
                      int primeCnt, /* Number of elements to prime list with */
                      void (*initFunc)(void *arg),    /* Constructor */
                      void (*destroyFunc)(void *arg) /* Destructor */
#if PASIS_FREELIST_STATS > 0
                      ,char *file,
                      int line
#endif /* PASIS_FREELIST_STATS > 0 */
                      )
{
    pasis_freelist_t *freelist;
    int i;
  
    if (objInc < 0 || objSize == 0) {
        PASIS_PRINT_ERR("Freelist_Create - Error: objInc < 0 or objSize == 0\n");
        return NULL;
    }

    PASIS_DASSERT(objSize >= sizeof(pasis_freelist_obj_t));

    freelist = (pasis_freelist_t *)malloc(sizeof(pasis_freelist_t));
    if (freelist == NULL) {
        return NULL;
    }

    freelist->maxCnt = maxCnt;
    freelist->objSize = objSize;
    freelist->objInc = objInc;
    freelist->head = NULL;
    freelist->tail = NULL;
    freelist->freeCnt = 0;
    freelist->initFunc = initFunc;
    freelist->destroyFunc = destroyFunc;
  
    if (primeCnt) {
        pasis_freelist_obj_t *obj;
        for (i = 0; i < primeCnt; i++) {
            obj = (pasis_freelist_obj_t *)malloc(objSize +
                                                   sizeof(pasis_freelist_obj_t));
            if (initFunc != NULL) {
                initFunc((void *)((char *)(obj)+sizeof(pasis_freelist_obj_t)));
            }
            PASIS_DASSERT(obj);
            obj->next = freelist->head;
            freelist->head = obj;
            if (freelist->tail == NULL) {
                freelist->tail = obj;
            }
            freelist->freeCnt++;
        }
    }


#if PASIS_FREELIST_STATS > 0
    PasisFreelist_StatsInit(freelist, file, line);
#endif /* PASIS_FREELIST_STATS > 0 */
  
    return freelist;
}


/*
 * PasisFreelist_Get
 *  Takes freelist
 *  Returns object off of freelist (cast to void *)
 */

static inline void *
PasisFreelist_Get(pasis_freelist_t *freelist)
{
    pasis_freelist_obj_t *obj=NULL;
    char *data;
    int i;

    if (freelist->head == NULL) {
        if (freelist->objInc == 0) {
#if PASIS_FREELIST_STATS > 0
            PasisFreelist_StatsEmpty(freelist);
#endif /* PASIS_FREELIST_STATS */
            return NULL;
        }
        for (i = 0; i < freelist->objInc; i++) {
            obj = (pasis_freelist_obj_t *)malloc(freelist->objSize +
                                                   sizeof(pasis_freelist_obj_t));
            if (freelist->initFunc) {
                freelist->initFunc((void *)((char *)(obj)+sizeof(pasis_freelist_obj_t)));
            }
            obj->next = freelist->head;
            freelist->head = obj;
            if (freelist->tail == NULL) {
                freelist->tail = obj;
            }
            freelist->freeCnt++;
        }
#if PASIS_FREELIST_STATS > 0
        PasisFreelist_StatsGet(freelist, freelist->objInc);
#endif  /* PASIS_FREELIST_STATS > 0 */
    } else {
#if PASIS_FREELIST_STATS > 0
        PasisFreelist_StatsGet(freelist, 0);
#endif  /* PASIS_FREELIST_STATS > 0 */
    }
 
    obj = freelist->head;
    freelist->head = obj->next;
    if (freelist->tail == obj) {
        freelist->tail = NULL;
    }
    freelist->freeCnt--;

    data = (char *)obj;
    data += sizeof(pasis_freelist_obj_t);
    
    PASIS_DASSERT(((char *)(obj) + sizeof(pasis_freelist_obj_t)) == data);

    return data;
}


/*
 * PasisFreelist_Put
 *   Takes freelist (puts object back onto freelist)
 *   Returns void
 */

static inline void
PasisFreelist_Put(pasis_freelist_t *freelist, 
                    void *obj)
{
    pasis_freelist_obj_t *fObj;

    if (freelist->freeCnt >= freelist->maxCnt) {
        if (freelist->destroyFunc) {
            freelist->destroyFunc(obj);
        }
        free ((char *)(obj)-sizeof(pasis_freelist_obj_t));
    } else {
        fObj = (pasis_freelist_obj_t *)((char *)(obj)-sizeof(pasis_freelist_obj_t));
        fObj->next = NULL;
        if (freelist->tail) {
            freelist->tail->next = fObj;
        }
        freelist->tail = fObj;
        if (freelist->head == NULL) {
            freelist->head = fObj;
        }
        freelist->freeCnt++;
    }

#if PASIS_FREELIST_STATS > 0
    PasisFreelist_StatsPut(freelist);
#endif  /* PASIS_FREELIST_STATS > 0 */
  
    return;
}


/* 
 * PasisFreelist_Destroy
 *   Takes freelist (and deallocates all objects and itself)
 *   Returns void
 */

static inline void
PasisFreelist_Destroy(pasis_freelist_t *freelist)
{
    pasis_freelist_obj_t *obj, *objNext;

#if PASIS_FREELIST_STATS > 0
    PasisFreelist_StatsDestroy(freelist);
#endif  /* PASIS_FREELIST_STATS > 0 */
    
    for (obj = freelist->head; obj != NULL; obj = objNext) {
        objNext = obj->next;
        
        if (freelist->destroyFunc) {
            freelist->destroyFunc((void *)((char *)(obj)+sizeof(pasis_freelist_obj_t)));
        }
        
        free (obj);
    }

    free (freelist);
  
    return;
}


#ifdef __cplusplus
}
#endif

#endif /* __PASIS_FREELIST_H__ */

/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */  

