/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

/*
 * pasis_lists.h
 *
 * This file implements a variety of routines
 * for manipulating lists.  It should be used
 * to build other, more complicated linked structures
 * like stacks, queues, or hash tables.
 */

#ifndef __PASIS_LISTS_H__
#define __PASIS_LISTS_H__


#ifdef __cplusplus
extern "C" {
#endif

/*
 * Put this declaration inside the structure
 * you want to use as a list node.
 */

#define PASIS_LL_PTRS(listName, type) \
struct { \
   type *next; \
   type *prev; \
} listName;

#define PASIS_LL_HEAD(type, head) \
   type *head

#define PASIS_LL_TAIL(type, tail) \
   type *tail


#define PASIS_LL_HEAD_INIT(head) \
do{ \
   head = NULL; \
}while(0)

#define PASIS_LL_TAIL_INIT(tail) \
do{ \
   tail = NULL; \
}while(0)

#define PASIS_LL_NODE_INIT(listName, nodePtr) \
do{ \
    (nodePtr)->listName.next = NULL; \
    (nodePtr)->listName.prev = NULL; \
}while(0)

#define PASIS_LL_ADD_NODE_AFTER(listName, newNodePtr, oldNodePtr)     \
do{                                                                \
    (newNodePtr)->listName.next = (oldNodePtr)->listName.next;     \
    (newNodePtr)->listName.prev = (oldNodePtr);                    \
    (oldNodePtr)->listName.next = (newNodePtr);                    \
    if ((newNodePtr)->listName.next != NULL) {                     \
        (newNodePtr)->listName.next->listName.prev = (newNodePtr); \
    }                                                              \
}while(0)

#define PASIS_LL_ADD_NODE_AFTER_WITH_TAIL(listName, newNodePtr, oldNodePtr, tail)\
do{                                                                \
    (newNodePtr)->listName.next = (oldNodePtr)->listName.next;     \
    (newNodePtr)->listName.prev = (oldNodePtr);                    \
    (oldNodePtr)->listName.next = (newNodePtr);                    \
    if ((oldNodePtr) == (tail)) {                                  \
        (tail) = (newNodePtr);                                     \
    } else {                                                       \
        (newNodePtr)->listName.next->listName.prev = (newNodePtr); \
    }                                                              \
}while(0)

#define PASIS_LL_ADD_NODE_BEFORE(listName, newNodePtr, oldNodePtr, head)   \
do{                                                                     \
    (newNodePtr)->listName.next = (oldNodePtr);                         \
    if (oldNodePtr != NULL) {                                           \
        (newNodePtr)->listName.prev = (oldNodePtr)->listName.prev;      \
        (oldNodePtr)->listName.prev = (newNodePtr);                     \
    } else {                                                            \
        (newNodePtr)->listName.prev = NULL;                             \
    }                                                                   \
    if ((head) == (oldNodePtr)) {                                       \
        (head) = (newNodePtr);                                          \
    } else {                                                            \
        (newNodePtr)->listName.prev->listName.next = (newNodePtr);      \
    }                                                                   \
}while(0)

/* Adds a newNode before the node where the comparison first fails */
#define PASIS_LL_ADD_NODE_CMP(listName, newNodePtr, head, cmp, type, elem)       \
do{                                                                           \
    type *tmpNode = head;                                                     \
    type *tmpPrev = NULL;                                                     \
                                                                              \
    while (tmpNode != NULL && ((newNodePtr)->elem) cmp tmpNode->elem) {       \
        tmpPrev = tmpNode;                                                    \
        tmpNode = tmpNode->listName.next;                                     \
    }                                                                         \
                                                                              \
    if (tmpPrev == NULL) {     /* List was empty, add to head */              \
        (head) = (newNodePtr);                                                \
        (newNodePtr)->listName.next = NULL;                                   \
        (newNodePtr)->listName.prev = NULL;                                   \
    } else if (tmpNode) {     /* Add before tmpNode */                        \
        newNodePtr->listName.next = tmpNode;                                  \
        newNodePtr->listName.prev = tmpNode->listName.prev;                   \
        tmpNode->listName.prev = newNodePtr;                                  \
        if (head == tmpNode) {                                                \
            head = newNodePtr;                                                \
        } else {                                                              \
            newNodePtr->listName.prev->listName.next = newNodePtr;            \
        }                                                                     \
    } else {  /* tmpNode is Null so insert after tmpPrev */                   \
        (newNodePtr)->listName.next = tmpPrev;                                \
        (newNodePtr)->listName.prev = tmpPrev->listName.prev;                 \
        tmpPrev->listName.prev = newNode;                                     \
        if (head == tmpPrev) {                                                \
            head = newNodePtr;                                                \
        } else {                                                              \
            (newNodePtr)->listName.prev->listName.next = newNodePtr;          \
        }                                                                     \
    }                                                                         \
}while(0)



/*
 * PASIS_LL_REMOVE_NODE(listName, node, head);
 *   Removes a node from the list.  The user
 * is responsible for freeing memory.
 */

#define PASIS_LL_REMOVE_NODE(listName, nodePtr, head) \
do{ \
    if((head) == (nodePtr)){ \
        (head) = (nodePtr)->listName.next; \
    } else { \
        (nodePtr)->listName.prev->listName.next = (nodePtr)->listName.next; \
    } \
    if((nodePtr)->listName.next != NULL){ \
        (nodePtr)->listName.next->listName.prev = (nodePtr)->listName.prev; \
    } \
}while(0)

#define PASIS_LL_REMOVE_NODE_WITH_TAIL(listName, nodePtr, head, tail)          \
do{                                                                           \
    if ((head) == (nodePtr)) {                                              \
        (head) = (nodePtr)->listName.next;                                  \
    } else {                                                                \
        (nodePtr)->listName.prev->listName.next = (nodePtr)->listName.next; \
    }                                                                       \
    if ((tail) == (nodePtr)) {                                              \
        (tail) = (nodePtr)->listName.prev;                                  \
    } else {                                                                \
        (nodePtr)->listName.next->listName.prev = (nodePtr)->listName.prev; \
    }                                                                       \
}while(0)

#define PASIS_LL_IS_EMPTY(head) ((head) == NULL)
#define PASIS_LL_NEXT(listName, nodePtr) ((nodePtr)->listName.next)
#define PASIS_LL_PREV(listName, nodePtr) ((nodePtr)->listName.prev)

/* 
 * Linked list functions which could be used as Q_'s 
 */

#define PASIS_LL_ADD_TAIL(listName, nodePtr, head, tail) \
do{ \
    if((tail) == NULL){ \
        /* add the first element */ \
        (tail) = (nodePtr); \
        (head) = (nodePtr); \
        (nodePtr)->listName.next = NULL; \
        (nodePtr)->listName.prev = NULL; \
    }else{ \
        /* normal case for adding a node */ \
        (tail)->listName.next = (nodePtr); \
        (nodePtr)->listName.prev = (tail); \
        (nodePtr)->listName.next = NULL; \
        (tail) = nodePtr; \
    } \
}while(0)

#define PASIS_LL_REMOVE_HEAD(listName, head, tail) \
do{ \
    if((head) != NULL) { \
        /* normal case */ \
        (head) = (head)->listName.next;  /* adjust list */ \
        if((head) == NULL){ \
          /* this was the only element, set tail to null */ \
          (tail) = NULL; \
        } else { \
          (head)->listName.prev = NULL; \
        } \
    } \
}while(0)

#define PASIS_LL_ADD_HEAD(listName, nodePtr, head, tail) \
do{ \
    if((head) == NULL){ \
        /* first element */ \
        (tail) = (nodePtr); \
    }else{ \
         (head)->listName.prev = (nodePtr); \
    } \
    (nodePtr)->listName.next = (head); \
    (nodePtr)->listName.prev = NULL; \
    (head) = (nodePtr); \
}while(0)

#define PASIS_LL_REMOVE_TAIL(listName, head, tail) \
do{ \
     if((tail) != NULL){ \
         /* normal case */ \
         (tail) = (tail)->listName.prev;  /* detach from back of queue */ \
         if((tail) == NULL){ \
           /* was last element on the list */ \
           (head) = NULL; \
         } else { \
           (tail)->listName.next = NULL; \
         }  \
    } \
}while(0)

#define PASIS_LL_GET_HEAD(head) head
#define PASIS_LL_GET_TAIL(tail) tail


#define PASIS_LL_PEEK_HEAD(nodePtr, head)     \
do{                                             \
    (nodePtr) = (head);                         \
} while(0)

#define PASIS_LL_PEEK_TAIL(nodePtr, tail)     \
do{                                             \
    (nodePtr) = (tail);                         \
} while(0)


/** Simplified Queue Structures based on the generic lists **/
#define PASIS_Q_PTRS(type) \
struct {                \
    type *next;         \
    type *prev;         \
}

#define PASIS_Q_STRUCT(listName, type) \
struct listName {                      \
    PASIS_LL_HEAD(type, head);       \
    PASIS_LL_TAIL(type, tail);       \
}

#define PASIS_Q_INIT(q)            \
do {                            \
    PASIS_LL_HEAD_INIT((q).head);  \
    PASIS_LL_TAIL_INIT((q).tail);  \
} while(0)
    
#define PASIS_Q_FIRST(q) ((q).head)
#define PASIS_Q_LAST(q) ((q).tail)

#define PASIS_Q_EMPTY(q) ((q).head == NULL)

#define PASIS_Q_ADD_HEAD(q, listName, node) \
  PASIS_LL_ADD_HEAD(listName, node, (q).head, (q).tail)

#define PASIS_Q_ADD_TAIL(q, listName, node) \
  PASIS_LL_ADD_TAIL(listName, node, (q).head, (q).tail)

#define PASIS_Q_REMOVE(q, listName, node) \
  PASIS_LL_REMOVE_NODE_WITH_TAIL(listName, node, (q).head, (q).tail)


#define PASIS_Q_NEXT PASIS_LL_NEXT
#define PASIS_Q_PREV PASIS_LL_PREV

#ifdef __cplusplus
}
#endif

#endif /*!__PASIS_LISTS_H__ */

/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */

