/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

/*
 * pasis_timer.h
 *
 * author: Garth Goodson
 *
 * Timer functions to access i386 performance counters giving nanosecond timing
 * precision
 */

#ifndef __PASIS_TIMER_H__
#define __PASIS_TIMER_H__

#include "pasis_config.h"
#include "pasis_debug.h"
#include "pasis_types.h"

#include <unistd.h>

#ifdef __cplusplus
extern "C" {
#endif
#if 0
}
#endif

/* Usage is as follows:
 *
 * timer_state = Pasis_TimerInit (name, number of timers)
 * - Sets up state, allocates space for the number of timers specified
 * - Returns a pointer to a timer_state structure
 *
 * Pasis_TimerRegister (timer_state, index, name)
 * - Call it for each timer to register (number of timers specified in init)
 * - Pass in the timer_state, the index (starting at 0), and name for the timer
 *
 * PASIS_TIMER_START ()
 * - At the beginning of the call to start a timer
 *
 * PASIS_TIMER_STOP (timer_state, index)
 * - Stops the timer (must be in the same function as the start, pass in the
 *   index for the registered timer)
 *
 * Pasis_TimerShutdown (timer_state)
 * - Dumps all timer information for a particular timer_state
 * - Deallocates/cleans up all memory
 *
 */

typedef struct pasis_timer_s {
    p_uint64     start;
    p_uint64     stop;
    char         *start_file;
    char         *stop_file;
    int           start_line;
    int           stop_line;
} pasis_timer_t;

typedef struct pasis_timespec_s {
    p_uint32 ts_sec;
    p_uint32 ts_nsec;
} pasis_timespec_t;

typedef struct pasis_timer_set_s {
    pasis_timespec_t ts;
    pasis_timespec_t *ts_history;
    unsigned int     count;
    unsigned int     history_size;
    char            *name;
} pasis_time_unit_t;

typedef struct pasis_timer_state_s {
    pasis_time_unit_t *timer_set;
    unsigned int       timer_num;
    char              *timer_set_name;
} pasis_timer_state_t;

extern p_uint64           pasis_timer_ticks;


void Pasis_TimeInit (void);

pasis_timer_state_t * Pasis_TimerInit (char *name, unsigned int num_timers, 
                                       unsigned int history_size);

void Pasis_TimerPrintAndReset (pasis_timer_state_t *timer_state);

void Pasis_TimerShutdown (pasis_timer_state_t *timer_state);


/* Starts the timer - takes a pasis_timer_t */
#define PASIS_TIMER_START_TM(_tm_) Pasis_TimerStart(_tm_, __FILE__, __LINE__)
/* Stops a timer - takes a pasis_timer_t */ 
#define PASIS_TIMER_STOP_TM(_tm_) Pasis_TimerStop(_tm_, __FILE__, __LINE__)
/* Gives elapsed time - takes a pasis_timer_t, and a pasis_timespec_t */
#define PASIS_TIMER_ELAPSED_TS(_tm_,_tsp_) Pasis_TimerElapsed(_tm_, _tsp_)
/* Pretty prints a timer report - takes a pasis_timer_t */
#define PASIS_TIMER_REPORT(_tm_) Pasis_TimerReport(_tm_)
/* Add two timespecs together */
#define PASIS_TIMER_ADD_TS(_tsp1_, _tsp2_, _result_) \
Pasis_TimerAddTs(_tsp1_, _tsp2_, _result_)

#define PASIS_TIMER_START() \
  pasis_timer_t __timer; \
  pasis_timespec_t __ts; \
  PASIS_TIMER_START_TM (&__timer);

#define PASIS_TIMER_STOP(timer_state, index) \
  PASIS_TIMER_STOP_TM (&__timer); \
  PASIS_TIMER_ELAPSED_TS (&__timer, &__ts); \
  PASIS_TIMER_ADD_TS (&__ts, &(timer_state)->timer_set[index].ts, &(timer_state)->timer_set[index].ts); \
  (timer_state)->timer_set[index].count++; \
  if( (timer_state)->timer_set[index].history_size > 0 ) {(timer_state)->timer_set[index].ts_history[ (timer_state)->timer_set[index].count % (timer_state)->timer_set[index].history_size ] = __ts; }

#define PASIS_TIMER_START_2(timer) PASIS_TIMER_START_TM (timer)

#define PASIS_TIMER_STOP_2(timer_state, index, __timer, __ts) \
  PASIS_TIMER_STOP_TM (__timer); \
  PASIS_TIMER_ELAPSED_TS (__timer, __ts); \
  PASIS_TIMER_ADD_TS (__ts, &(timer_state)->timer_set[index].ts, &(timer_state)->timer_set[index].ts); \
  (timer_state)->timer_set[index].count++; \
  if( (timer_state)->timer_set[index].history_size > 0 ) {(timer_state)->timer_set[index].ts_history[ (timer_state)->timer_set[index].count % (timer_state)->timer_set[index].history_size ] = *__ts; }


#if defined (__i386__)

static inline p_uint64
Pasis_TimerRPPC64 (void)
{
    volatile p_uint32 high, low;
    asm volatile ("rdtsc" : "=a" (low), "=d" (high));
    return ((p_uint64)(high) << 32) | low;
}
#else
/* Use gettimeofday to emulate the Pentium performance counters. */
static inline p_uint64
Pasis_TimerRPPC64 (void)
{
    struct timeval tv;
    gettimeofday (&tv, NULL);
    return tv.tv_sec * 1000000 + tv.tv_usec;
}
#endif


static inline void
Pasis_TimerAddTs(pasis_timespec_t *tsp1, pasis_timespec_t *tsp2, 
                 pasis_timespec_t *result)
{
    result->ts_sec = tsp1->ts_sec + tsp2->ts_sec;
    result->ts_nsec = tsp1->ts_nsec + tsp2->ts_nsec;
    if (result->ts_nsec >= 1000000000) {
         ++result->ts_sec;
         result->ts_nsec -= 1000000000;
    }
}


static inline void
Pasis_TimerRegister (pasis_timer_state_t *timer_state, 
                     unsigned int index, char *name)
{
    PASIS_DASSERT (index < timer_state->timer_num);
    timer_state->timer_set[index].name = (char *) malloc (strlen (name) + 1);
    strcpy (timer_state->timer_set[index].name, name);
}


static inline void
Pasis_TimerStart(pasis_timer_t *tm, char *file, int line)
{
    PASIS_DASSERT (pasis_timer_ticks != 0);

    tm->start_file = file;
    tm->start_line = line;
    tm->start = Pasis_TimerRPPC64();
}


static inline void
Pasis_TimerStop(pasis_timer_t *tm, char *file, int line)
{
    tm->stop = Pasis_TimerRPPC64();
    tm->stop_file = file;
    tm->stop_line = line;
}


static inline void
Pasis_TimerElapsed(pasis_timer_t *tm, pasis_timespec_t *tsp)
{
    p_uint64 ticks, diff;

    ticks = pasis_timer_ticks;
    diff = tm->stop - tm->start;
    tsp->ts_sec = (p_uint32)(diff / ticks);
    tsp->ts_nsec = (p_uint32)(((diff % ticks) * 1000000000) / ticks);
}


static inline void
Pasis_TimerReport(pasis_timer_t *tm)
{
    p_uint32 secs, nsecs;
    p_uint64 ticks, diff;

    ticks = pasis_timer_ticks;
    PASIS_PRINT_INFO("----- Timer\n");
    diff = tm->stop - tm->start;
    secs = (p_uint32)(diff / ticks);
    nsecs = (p_uint32)(((diff % ticks) * 1000000000) / ticks);

    PASIS_PRINT_INFO("%u:%09u (secs:nsecs) (%llu ticks)\n", secs, nsecs, diff);
    PASIS_PRINT_INFO(" start %s:%d\n", tm->start_file, tm->start_line);
    PASIS_PRINT_INFO(" end %s:%d\n", tm->stop_file, tm->stop_line);
    PASIS_PRINT_INFO(" (%llu..%llu) (ticks)\n", tm->start, tm->stop);

    PASIS_PRINT_INFO("-----\n");
}


#ifdef __cplusplus
}
#endif

#endif /* __PASIS_TIMER_H__ */

/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
