/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"
#include "pasis_debug.h"
#include "pasis_types.h"

#include "pasis_timer.h"

p_uint64           pasis_timer_ticks;

void
Pasis_TimeInit (void)
{
    struct timeval stv, etv;

    p_uint64 spcc, epcc;
    p_uint32 elap, elap_cc;

    if(!pasis_timer_ticks) {
        gettimeofday (&stv, NULL);
        spcc = Pasis_TimerRPPC64();
        usleep(100000);             /* 100ms */
        epcc = Pasis_TimerRPPC64();
        gettimeofday (&etv, NULL);
    
        elap = (((etv.tv_sec - stv.tv_sec) * 1000000) + 
                (etv.tv_usec - stv.tv_usec));
        elap_cc = epcc - spcc;

        PASIS_DASSERT (elap != 0);

        pasis_timer_ticks = ((p_uint64)elap_cc * 1000000) / elap;
    }
}

pasis_timer_state_t *
Pasis_TimerInit (char *name, 
                 unsigned int num_timers, 
                 unsigned int history_size )
{
    pasis_timer_state_t *timer_state;

    unsigned int i;
  
    Pasis_TimeInit ();

    timer_state = (pasis_timer_state_t *)malloc (sizeof (pasis_timer_state_t));
    memset (timer_state, 0, sizeof (pasis_timer_state_t));

    timer_state->timer_num = num_timers;
    timer_state->timer_set_name = (char *) malloc (strlen (name) + 1);
    strcpy (timer_state->timer_set_name, name);

    timer_state->timer_set = (pasis_time_unit_t *) 
        malloc (sizeof (pasis_time_unit_t) * num_timers);
    
    memset (timer_state->timer_set, 0, sizeof(pasis_time_unit_t) * num_timers);

    for( i = 0; i < num_timers; ++i ) {
        timer_state->timer_set[ i ].history_size = history_size;
        if (history_size) {
            timer_state->timer_set[ i ].ts_history = ( pasis_timespec_t * )
                malloc( sizeof (pasis_timespec_t) * history_size );
            memset( timer_state->timer_set[ i ].ts_history, 0, 
                sizeof( pasis_timespec_t ) * history_size );
        } else {
             timer_state->timer_set[ i ].ts_history = NULL;
        }
    }

    return timer_state;
}


static void
PrintTimers (pasis_timer_state_t *timer_state)
{
    unsigned int i, j;
    pasis_timespec_t total_time={0,0};
    unsigned int count=0;

    PASIS_PRINT_INFO ("\n--- Timers: %s ---\n", timer_state->timer_set_name);
    
    for (i = 0; i < timer_state->timer_num; i++) {
        double req_time;
        if (timer_state->timer_set[i].name) {
            
            if (timer_state->timer_set[i].count == 0) {
                req_time = 0;
            } else {
                req_time = ((timer_state->timer_set[i].ts.ts_sec * 1000 + 
                           (timer_state->timer_set[i].ts.ts_nsec/1000000)) /
                            (double)timer_state->timer_set[i].count);
            }
            
            PASIS_PRINT_INFO ("%-18s :  Count=%-12u  Req Time = %.3fms  Total = %u.%.3us\n",
                              timer_state->timer_set[i].name,
                              timer_state->timer_set[i].count,
                              req_time,
                              timer_state->timer_set[i].ts.ts_sec,
                              timer_state->timer_set[i].ts.ts_nsec/1000000);

            if( timer_state->timer_set[i].history_size > 0 ) {
                PASIS_PRINT_INFO("History (%d) [in seconds]: \n",
                                 timer_state->timer_set[i].history_size ); 
                for( j = 0; j < timer_state->timer_set[i].history_size; ++j ) {
                    PASIS_PRINT_INFO("%u.%.6u ", 
                                     timer_state->timer_set[i].
                                     ts_history[j].ts_sec,
                                     timer_state->timer_set[i].
                                     ts_history[j].ts_nsec/1000 );
                }
                PASIS_PRINT_INFO("\n");
            }

            count += timer_state->timer_set[i].count;
            PASIS_TIMER_ADD_TS (&total_time, &timer_state->timer_set[i].ts, 
                                &total_time);
        }
    }

    PASIS_PRINT_INFO ("------\n");
    PASIS_PRINT_INFO ("%-18s :  Count=%-12u  Total Time = %u.%.3us\n\n",
                      "Totals", count, total_time.ts_sec,
                      total_time.ts_nsec/1000000);
}


void
Pasis_TimerPrintAndReset (pasis_timer_state_t *timer_state)
{
    unsigned int i;

    /* First, print the timers */
    PrintTimers (timer_state);

    /* Second, reset the timers */
    for (i = 0; i < timer_state->timer_num; ++i) {
        timer_state->timer_set[i].ts.ts_sec = 0;
        timer_state->timer_set[i].ts.ts_nsec = 0;
        timer_state->timer_set[i].count = 0;
        /* I don't handle reseting the history, but it's unused... */
        PASIS_DASSERT (timer_state->timer_set[i].history_size == 0);
    }

}


void
Pasis_TimerShutdown (pasis_timer_state_t *timer_state)
{
    unsigned int i;

    /* First, print the timers */
    PrintTimers (timer_state);

    /* Second, deallocate the timers */
    for (i = 0; i < timer_state->timer_num; i++) {
        if (timer_state->timer_set[i].name) {
            free (timer_state->timer_set[i].ts_history);
            free (timer_state->timer_set[i].name);
        }
    }

    free (timer_state->timer_set_name);
    free (timer_state->timer_set);
    free (timer_state);
}

/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
