#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>

#include <sys/socket.h>
#include <arpa/inet.h>

#include "pasis_config.h"
#include "pasis_types.h"
#include "pasis_debug.h"
#include "pasis_timer.h"

#if USE_KERBEROS
#include "krb_box.h"
#include "pasisio.h"
#endif

#include "rpc_common.h"

#define RPC_CLIENT_TIMER_NUM    8

#define RPC_CLIENT_TIMER_REQ    0
#define RPC_CLIENT_TIMER_WRITEV 1
#define RPC_CLIENT_TIMER_WAIT   2
#define RPC_CLIENT_TIMER_RECV   3
#define RPC_CLIENT_TIMER_RECV1  4
#define RPC_CLIENT_TIMER_RECV2  5
#define RPC_CLIENT_TIMER_RECV2A 6
#define RPC_CLIENT_TIMER_RECV3  7


/* Common defines */
#ifndef MIN
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#endif

#ifndef MAX
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#endif

extern p_byte rpc_dummy_buffer[RPC_MAX_PAYLOAD];

void * rpc_poll_tcp_client_thread (void *arg);
void * rpc_poll_tcp_server_thread (void *arg);

static inline
char *dump_time (void)
{
    struct timeval tv;
    static char buffer[64];

    gettimeofday (&tv, NULL);
    
    snprintf (buffer, 63, "%lu.%09lu", tv.tv_sec, (long unsigned int)tv.tv_usec);

    return buffer;
}

static inline
int init_tcp_code (void)
{
    struct protoent *ent;
    ent = getprotobyname ("tcp");
    PASIS_DASSERT (ent != NULL);
    return ent->p_proto;
}

static inline
int get_tcp_code (void)
{
    static int tcp_code = 0;
    if (tcp_code == 0) {
        tcp_code = init_tcp_code ();
    }
    return tcp_code;
}


/***** Early Termination Helper Functions *****/

static inline rpc_req_table_t *
rpc_get_table_entry (rpc_handle_t *handle)
{
    rpc_req_table_t *entry;
    int i;
    
    while (1) {

        for (i = 0; i < handle->max_outstanding; i++) {
            entry = &handle->req_table[i];
            if (entry->state == RPC_REQ_STATE_INVALID) {
                entry->state = RPC_REQ_STATE_VALID;
                entry->terminate = 0;
                return entry;
            }
        }

        st_cond_wait (handle->req_table_cond);
    }

    return NULL;
}


static inline void
rpc_put_table_entry (rpc_handle_t    *handle,
                     rpc_req_table_t *entry)
{
    entry->state = RPC_REQ_STATE_INVALID;
    st_cond_signal (handle->req_table_cond);
}


#if USE_KERBEROS

static pasis_error_t 
rpc_auth_validate (rpc_handle_type_t  type,
                   p_uint32           local_ip,
                   p_uint32           remote_ip,
                   p_byte           **buffers,
                   unsigned int      *buffer_lengths,
                   unsigned int       num_buffers,
                   p_byte            *signature)
{
    p_byte signature_gen[PASIS_HASH_SIZE];
    octet_string* sh_key = NULL;

    /* Get shared key, a little different depending on who is calling it */
    if (type == RPC_HANDLE_CLIENT) {
        sh_key = krb_get_sh_key_cli(DEFAULT_PRINCIPAL, local_ip, remote_ip);
    } else {
        sh_key = krb_get_sh_key_srv(DEFAULT_PRINCIPAL, remote_ip, local_ip);
    }
    
    if (sh_key == NULL) {
        PASIS_PRINT_ERR ("Error: shared key not found (%s)\n", __FUNCTION__);
        return PASIS_ERROR;
    }

    /* Sign header -- generate signature */
    ED_Buffers_Keyed_Hash ((const p_byte *)sh_key->data, 
                           MIN (sh_key->length, PASIS_HASH_SIZE),
                           (const p_byte **) buffers, buffer_lengths, 
                           num_buffers, signature_gen);

    free(sh_key->data);
    free(sh_key);

    if (memcmp (signature_gen, signature, PASIS_HASH_SIZE) != 0) {
        PASIS_PRINT_ERR ("Error: signature validation failed (%s)\n",
                         __FUNCTION__);
        return PASIS_ERROR;
    }

    return PASIS_SUCCESS;
}


static pasis_error_t 
rpc_auth_generate (rpc_handle_type_t  type,
                   p_uint32           local_ip,
                   p_uint32           remote_ip,
                   p_byte           **buffers,
                   unsigned int      *buffer_lengths,
                   unsigned int       num_buffers,
                   p_byte            *signature)
{
    octet_string* sh_key = NULL;

    /* Get shared key, a little different depending on who is calling it */
    if (type == RPC_HANDLE_CLIENT) {
        sh_key = krb_get_sh_key_cli(DEFAULT_PRINCIPAL, local_ip, remote_ip);
    } else {
        sh_key = krb_get_sh_key_srv(DEFAULT_PRINCIPAL, remote_ip, local_ip);
    }

    if (sh_key == NULL) {
        PASIS_PRINT_ERR ("Error: shared key not found (%s)\n", __FUNCTION__);
        return PASIS_ERROR;
    }

    /* Sign header -- generate signature */
    ED_Buffers_Keyed_Hash ((const p_byte *)sh_key->data,
                           MIN (sh_key->length, PASIS_HASH_SIZE),
                           (const p_byte **)buffers, buffer_lengths, 
                           num_buffers, signature);

    free(sh_key->data);
    free(sh_key);

    return PASIS_SUCCESS;
}

#endif /* USE_KERBEROS */

/******************************  TCP Client ********************************/

static inline void
rpc_tcp_client_disconnect (rpc_handle_t *handle)
{
    if (handle->sock_fd != NULL) {
        close (st_netfd_fileno (handle->sock_fd));
        /*
          st_mutex_lock (handle->sock_lock);
          st_netfd_close (handle->sock_fd);
          handle->sock_fd = NULL;
          st_mutex_unlock (handle->sock_lock);
        */
    }
}


void
rpc_encode_tcp_call (rpc_handle_t *handle,
                     rpc_req_t *req,
                     rpc_tcp_call_hdr_t *hdr)
{
    int i;
    unsigned int payload_len=0;

#if USE_KERBEROS
    p_byte        *buffers[2];
    unsigned int   buffer_lens[2];
    pasis_error_t  error;

    hdr->hash_region_len = htonl (req->req_iov[1].iov_len);
#endif 

    hdr->xid = htonl (req->rpc_xid);
    hdr->msg_type = htonl (RPC_CALL);
    hdr->prog = htonl (req->rpc_prog);
    hdr->vers = htonl (req->rpc_vers);
    hdr->proc = htonl (req->rpc_proc);

    /* Start at 1 since 0 contains the rpc header info */
    for (i = 1; i < req->req_iov_len; i++) {
        payload_len += req->req_iov[i].iov_len;
    }

    hdr->payload_len = htonl (payload_len);

#if USE_KERBEROS
    buffers[0] = ((p_byte *)hdr) + PASIS_HASH_SIZE;
    buffers[1] = req->req_iov[1].iov_base;
    buffer_lens[0] = sizeof (rpc_tcp_call_hdr_t) - PASIS_HASH_SIZE;
    buffer_lens[1] = req->req_iov[1].iov_len;

    error = rpc_auth_generate (handle->handle_type, handle->local_ip, 
                               handle->remote_ip, buffers, buffer_lens, 
                               2, hdr->signature);

    PASIS_DASSERT (error == PASIS_SUCCESS);
#endif
}


int
rpc_transmit_tcp_req (rpc_handle_t *client,
                      rpc_req_t    *req)
{
    int rc, i, real_errno;
    rpc_req_table_t *req_entry;

    pasis_timer_t timer_trans, timer_writev, timer_wait;
    pasis_timespec_t ts;

    PASIS_TIMER_START_2 (&timer_trans);

    req->wait_flag = 1;
    req->transmit_flag = 0;
    req->wait_status = RPC_STATUS_SUCCESS;
    inprog_req_queue (client, req);                 /* Add to wait queue */

    /* Check that connection is established */
    if (client->sock_fd == NULL) {
        return -1;
        PASIS_DASSERT (client->sock_fd != NULL);
    }

    /* Get a rpc table request entry */
    req_entry = rpc_get_table_entry (client);
    PASIS_DASSERT (req_entry != NULL);
    
    req_entry->thread_id = st_thread_self ();
    req_entry->req = req;
    req_entry->state = RPC_REQ_STATE_TRANSMIT;

    req->req_entry = req_entry;

    PASIS_DPRINTF (PASIS_DEBUG_RPC, 4, "Transmitting: xid=%u\n", req->rpc_xid);
    for (i = 0; i < req->req_iov_len; i++) {
        PASIS_DPRINTF (PASIS_DEBUG_RPC, 4, "- %d: len=%u\n", 
                       i, (unsigned int) req->req_iov[i].iov_len);
    }

    PASIS_TIMER_START_2 (&timer_writev);

    st_mutex_lock (client->sock_lock);
    rc = st_writev (client->sock_fd, req->req_iov, req->req_iov_len, 
                    RPC_REQ_TIMEOUT);
    real_errno = errno; /* save errno since st_mutex_unlock will clear it */
    st_mutex_unlock (client->sock_lock);

    PASIS_TIMER_STOP_2 (client->timers, RPC_CLIENT_TIMER_WRITEV,
                        &timer_writev, &ts);

    if (rc < 0) {
        /* Close connection */
        PASIS_PRINT_ERR ("Error: writev failed (%d)\n", real_errno);

        if (req->wait_status == RPC_STATUS_CONN) {
            /* Request already dequeued in rpc_poll_tcp_client_thread */
        } else {
            inprog_req_dequeue (client, req);
        }

        rpc_put_table_entry (client, req_entry);

        //rpc_tcp_client_disconnect (client);
        close (st_netfd_fileno (client->sock_fd));
        printf ("%s: line %d\n", __FUNCTION__, __LINE__);

        return -1;
    }

    client->bytes_written += rc;
    client->num_requests++;

    req_entry->state = RPC_REQ_STATE_WAIT;

    PASIS_TIMER_START_2 (&timer_wait);

    /* Wait for response */
    while (req->wait_flag && req_entry->terminate == 0) {
        rc = st_cond_timedwait (req->wait_cond, RPC_REQ_TIMEOUT);
        //rc = st_cond_wait (req->wait_cond);
        
        if (rc == -1 && req->wait_flag == 1) {
            if (errno == ETIME && req->wait_flag == 1) {
                /* Timed out */
                //inprog_req_dequeue (client, req);
                rpc_put_table_entry (client, req_entry);

                /* For now close connection */
                close (st_netfd_fileno (client->sock_fd));
                //rpc_tcp_client_disconnect (client);

                PASIS_TIMER_STOP_2 (client->timers, RPC_CLIENT_TIMER_WAIT, 
                                    &timer_wait, &ts);
                PASIS_TIMER_STOP_2 (client->timers, RPC_CLIENT_TIMER_REQ, 
                                    &timer_trans, &ts);

                return RPC_STATUS_TIMEOUT;
            }
        }
    }
    PASIS_TIMER_STOP_2 (client->timers, RPC_CLIENT_TIMER_WAIT, 
                        &timer_wait, &ts);

    if (req->wait_status == RPC_STATUS_CONN) {
        /* Request already dequeued in rpc_poll_tcp_client_thread */
    } else {
        inprog_req_dequeue (client, req);
    }
    
    rc = req->wait_status;
    if (req_entry->terminate) {
        rc = RPC_STATUS_TERM;
    }

    rpc_put_table_entry (client, req_entry);

    PASIS_TIMER_STOP_2 (client->timers, RPC_CLIENT_TIMER_REQ, 
                        &timer_trans, &ts);

    return rc;
}


static int
rpc_handle_tcp_reply (rpc_handle_t *client)
{
    rpc_req_t *req;
    rpc_tcp_reply_hdr_t hdr;
    int rc, decode_err, i;
    unsigned int payload_len, read_len;
    
    pasis_timer_t timer_recv;
    pasis_timespec_t ts;

#if USE_KERBEROS
    p_byte              *buffers[2];
    unsigned int         buffer_lens[2];
    rpc_tcp_reply_hdr_t  hdr_nbo;  /* Header in network byte order */
    pasis_error_t        error;
#endif

    PASIS_TIMER_START ();
 
    PASIS_TIMER_START_2 (&timer_recv);

    rc = st_read_fully (client->sock_fd, &hdr, 
                        sizeof (rpc_tcp_reply_hdr_t), -1);
    if (rc <= 0) {
        /* Error on connection */
        PASIS_PRINT_ERR ("Error: rpc_handle_tcp_reply, connection closed\n");
        printf ("%s: %d\n", __FUNCTION__, __LINE__);
        return -1;
    }

    client->bytes_read += rc;

#if USE_KERBEROS
    /* Copy pre-decoded header for use in validation */
    memcpy (&hdr_nbo, &hdr, sizeof (rpc_tcp_reply_hdr_t));
#endif

    decode_err = rpc_decode_tcp_reply (&hdr);

    PASIS_TIMER_STOP_2 (client->timers, RPC_CLIENT_TIMER_RECV1, 
                        &timer_recv, &ts);

    PASIS_DPRINTF (PASIS_DEBUG_RPC, 4, 
                   "%s: found xid=%u, payload_len=%u (%d)\n", 
                   __FUNCTION__, hdr.xid, hdr.payload_len, decode_err);

    PASIS_TIMER_START_2 (&timer_recv);

    /* Find the request */
    req = client->inprog_queue_hd;
    while (req) {
        if (req->rpc_xid == hdr.xid) {
            break;
        }
        req = req->next;
    }

    if ((decode_err || req == NULL) && hdr.payload_len) {
        rc = st_read_fully (client->sock_fd, rpc_dummy_buffer, 
                            hdr.payload_len, -1);
        if (rc <= 0) {
            rc = errno;
            printf ("%s: %d\n", __FUNCTION__, __LINE__);
            return rc;
        }

        client->bytes_read += rc;
    }

    if (req == NULL) {
        /* Request not found */
        PASIS_DPRINTF (PASIS_DEBUG_RPC, 4, "%s: invalid xid=%u\n",
                       __FUNCTION__, hdr.xid);

        PASIS_TIMER_STOP_2 (client->timers, RPC_CLIENT_TIMER_RECV2A, 
                            &timer_recv, &ts);
        return 0;
    } else {
        /* Change the request's state from WAIT to RECEIVE.  This way, the early
         * termination code (see rpc_early_terminate) will _not_ signal the
         * waiting thread to terminate the request.  We don't want this to
         * happen, as we're reading into the request buffer and thus we want to
         * lock the request.  If the waiting thread was to awaken, it will
         * immediately give up the request and another thread can use it, and we
         * would have data corruption as two threads are reading into the same
         * request. */
        req->req_entry->state = RPC_REQ_STATE_RECEIVE;
    }

    req->reply_status = hdr.reply_status;
    req->reply_accepted_status = hdr.accepted_status;

    if (decode_err) {
        /* No connection error, just rpc error */
        /* Wake up the waiter */
        req->wait_flag = 0;
        req->wait_status = RPC_STATUS_SUCCESS;
        st_cond_signal (req->wait_cond);
        fprintf (stderr, "%s: decode error %d\n", __FUNCTION__, __LINE__);
        return 0;
    }
    
    PASIS_TIMER_STOP_2 (client->timers, RPC_CLIENT_TIMER_RECV2, 
                        &timer_recv, &ts);

    /* Play a little trick here, it is possible for waiter to time out while
       reads are blocking, to prevent bad things from happening set wait_flag
       to 2, this prevents time outs, while still blocking waiter */

    //req->wait_flag = 2;

    PASIS_TIMER_START_2 (&timer_recv);

    payload_len = hdr.payload_len;

    /* Read up to max of buffer space allocated and payload length */
    for (i = 0; i < req->reply_iov_len && payload_len > 0; i++) {

        read_len = MIN (payload_len, req->reply_iov[i].iov_len);

        PASIS_DPRINTF (PASIS_DEBUG_RPC, 4, 
                       "%s: xid=%u, [%d] waiting to read %u/%u\n",
                       __FUNCTION__, hdr.xid, i,
                       (unsigned int)req->reply_iov[i].iov_len, payload_len);

        rc = st_read_fully (client->sock_fd, req->reply_iov[i].iov_base, 
                            read_len, -1);
        if (rc <= 0) {
            req->wait_flag = 1;
            rc = errno;
            printf ("%s: %d\n", __FUNCTION__, __LINE__);
            return rc;
        }

        client->bytes_read += rc;

        payload_len -= read_len;
    }

    if (payload_len != 0) {
        PASIS_PRINT_ERR ("Warning: rpc payload not completely consumed (%d)\n",
                         payload_len);
        st_read (client->sock_fd, rpc_dummy_buffer, payload_len, -1);
    }

#if USE_KERBEROS
    buffers[0] = ((p_byte *)&hdr_nbo) + PASIS_HASH_SIZE;
    buffers[1] = req->reply_iov[0].iov_base;
    buffer_lens[0] = sizeof (rpc_tcp_reply_hdr_t) - PASIS_HASH_SIZE;
    buffer_lens[1] = req->reply_iov[0].iov_len;

    error = rpc_auth_validate (client->handle_type, client->local_ip, 
                               client->remote_ip, buffers, buffer_lens, 
                               2, hdr.signature);
    PASIS_DASSERT (error == PASIS_SUCCESS);
#endif

    PASIS_TIMER_STOP_2 (client->timers, RPC_CLIENT_TIMER_RECV3, 
                        &timer_recv, &ts);

    /* Wake up the waiter */
    req->wait_flag = 0;
    req->wait_status = RPC_STATUS_SUCCESS;
    st_cond_signal (req->wait_cond);

    PASIS_TIMER_STOP (client->timers, RPC_CLIENT_TIMER_RECV);

    return 0;
}


int
rpc_decode_tcp_reply (rpc_tcp_reply_hdr_t *hdr)
                      
{
#if USE_KERBEROS
    hdr->hash_region_len = ntohl (hdr->hash_region_len);
#endif

    hdr->xid = ntohl (hdr->xid);
    hdr->msg_type = ntohl (hdr->msg_type);
    hdr->reply_status = ntohl (hdr->reply_status);
    hdr->auth_status = ntohl (hdr->auth_status);
    hdr->accepted_status = ntohl (hdr->accepted_status);
    hdr->payload_len = ntohl (hdr->payload_len);

    PASIS_DASSERT (hdr->msg_type == RPC_REPLY);

    if (hdr->reply_status != RPC_MSG_ACCEPTED) {
        PASIS_PRINT_ERR("Error: rpc_req_decode_tcp_reply, not accepted\n");
        if (hdr->auth_status != RPC_AUTH_OK) {
            PASIS_PRINT_ERR ("Error: authentication failure\n");
        }

        if (hdr->accepted_status != RPC_ACCEPT_SUCCESS) {
            PASIS_PRINT_ERR("Error: accept failed %d\n", hdr->accepted_status);
                           
        }
        return -1;
    }

    return 0;
}


/***** Client Connection Management *****/

int
rpc_client_reconnect (rpc_handle_t *client)
{
    int fd;
    int flag = 1;
    st_netfd_t sock_fd;

    fd = socket (AF_INET, SOCK_STREAM, 0);
    if (fd < 0) {
        return -1;
    }

    setsockopt (fd, get_tcp_code (), TCP_NODELAY, &flag, sizeof (int));

    sock_fd = st_netfd_open_socket (fd);

    if (st_connect (sock_fd, (struct sockaddr *)&client->sock_addr, 
                    sizeof (struct sockaddr_in), RPC_CONNECT_TIMEOUT) < 0) {
        PASIS_DPRINTF (PASIS_DEBUG_RPC, 4, 
                       "%s: failed to connect (errno=%d)\n",
                       __FUNCTION__, errno);
        st_netfd_close (sock_fd);
        return -1;
    }

    st_sleep (4);

    PASIS_DPRINTF (PASIS_DEBUG_RPC, 4,
                   "%s: reconnected (%s)\n", __FUNCTION__, dump_time ());

    client->sock_fd = sock_fd;
    st_cond_signal (client->reconnect_cond);
    
    return 0;
}


void *
rpc_poll_tcp_client_thread (void *arg)
{
    int rc;
    rpc_handle_t *handle = (rpc_handle_t *)arg;
    rpc_req_t *req;

    while (!handle->req_poll_shutdown) {
        rc = st_netfd_poll (handle->sock_fd, POLLIN, -1);

        if (rc == 0) {
            rc = rpc_handle_tcp_reply (handle);
        }

        if (rc != 0) {
            if (errno == EINTR) {
                PASIS_PRINT_INFO ("%s: shutting down; closing socket to "
                                  "host %s\n", __FUNCTION__,
                                  pasis_ip_to_hostname (handle->remote_ip));
            } else {
                /* Error occured, close connection and quit thread */
                PASIS_PRINT_ERR ("%s: socket closed (errno = %d, host = %s)\n",
                                 __FUNCTION__, errno,
                                 pasis_ip_to_hostname (handle->remote_ip));
            }

            /* Wake up all requests waiting on this handle */
            while ((req = handle->inprog_queue_hd)) {
                req->wait_flag = 0;
                req->wait_status = RPC_STATUS_CONN;
                st_cond_signal (req->wait_cond);
                inprog_req_dequeue (handle, req);
            }

            /* Close and free netfd -- lock to protect writev */
            if (handle->sock_fd != NULL) {
                st_mutex_lock (handle->sock_lock);
                close (st_netfd_fileno (handle->sock_fd));
                st_netfd_free (handle->sock_fd);
                handle->sock_fd = NULL;
                st_mutex_unlock (handle->sock_lock);
            }

            PASIS_DASSERT (handle->inprog_queue_hd == NULL);
            PASIS_DASSERT (handle->inprog_queue_tl == NULL);

            /* Notify handler of disconnection */
            if (handle->disconnect_handler) {
                handle->disconnect_handler (handle->disconnect_arg);
            }
            
            while (handle->sock_fd == NULL && !handle->req_poll_shutdown) {
                st_cond_wait (handle->reconnect_cond);
            }


            if (!handle->req_poll_shutdown) {
                PASIS_PRINT_INFO ("%s: reconnected (%s)\n", 
                                  __FUNCTION__, dump_time ());
            }
        }
    }

    return NULL;
}


rpc_handle_t *
rpc_client_create_tcp (p_uint32 server_ip,
                       unsigned int max_outstanding,
                       unsigned long program,
                       unsigned long version,
                       void *disconnect_arg,
                       void (*disconnect_handler)(void *))
{
    rpc_handle_t *client;
    int flag = 1, ret;
    int fd;

    fd = socket (AF_INET, SOCK_STREAM, 0);

    if (fd < 0) {
        PASIS_PRINT_ERR ("%s: error creating socket (%d)\n", __FUNCTION__,
                         errno);
        return NULL;
    }

#if 1
    ret = setsockopt (fd, get_tcp_code (), TCP_NODELAY, &flag, sizeof (int));

    if (ret < 0) {
        PASIS_PRINT_ERR ("%s: error when turning off Nagle's algorithm (%d)\n",
                         __FUNCTION__, errno);
        return NULL;
    }
#else
    ret = 1;
#endif

    client = rpc_client_create_common (server_ip, max_outstanding, 
                                       program, version, IPPROTO_TCP);

    if (client == NULL) {
        close (fd);
        return NULL;
    }

    client->sock_lock = st_mutex_new ();
    client->sock_fd = st_netfd_open_socket (fd);
    client->proto = IPPROTO_TCP;

    if (st_connect (client->sock_fd, (struct sockaddr *)&client->sock_addr, 
                    sizeof (struct sockaddr_in), RPC_CONNECT_TIMEOUT) < 0) {
        PASIS_PRINT_ERR ("%s: failed to connect to %s (%u)\n", __FUNCTION__,
                         inet_ntoa(*(struct in_addr*)&server_ip), errno);
        st_netfd_close (client->sock_fd);
        client->sock_fd = NULL;
        return NULL;
    }
    
    client->req_table = malloc (max_outstanding * sizeof (rpc_req_table_t));
    memset (client->req_table, 0, max_outstanding * sizeof (rpc_req_table_t));
    client->req_table_cond = st_cond_new ();

    client->disconnect_arg = disconnect_arg;
    client->disconnect_handler = disconnect_handler;
    client->reconnect_cond = st_cond_new ();
    client->req_poll_thread = st_thread_create (rpc_poll_tcp_client_thread, 
                                                client, 1, 0);

    client->timers = Pasis_TimerInit ("RPC_Client_Timers", 
                                      RPC_CLIENT_TIMER_NUM, 0);
    Pasis_TimerRegister (client->timers, RPC_CLIENT_TIMER_REQ,
                         "rpc_client_req");
    Pasis_TimerRegister (client->timers, RPC_CLIENT_TIMER_WRITEV,
                         "rpc_client_writev");
    Pasis_TimerRegister (client->timers, RPC_CLIENT_TIMER_WAIT,
                         "rpc_client_wait");
    Pasis_TimerRegister (client->timers, RPC_CLIENT_TIMER_RECV,
                         "rpc_client_recv");
    Pasis_TimerRegister (client->timers, RPC_CLIENT_TIMER_RECV1,
                         "rpc_client_recv1");
    Pasis_TimerRegister (client->timers, RPC_CLIENT_TIMER_RECV2A,
                         "rpc_client_recv2a");
    Pasis_TimerRegister (client->timers, RPC_CLIENT_TIMER_RECV2,
                         "rpc_client_recv2b");
    Pasis_TimerRegister (client->timers, RPC_CLIENT_TIMER_RECV3,
                         "rpc_client_recv3");

    return client;
}


/*******************************  TCP Server ********************************/


static pasis_error_t
rpc_handle_tcp_request (rpc_handle_t *server, 
                        st_netfd_t netfd,
                        p_uint32 remote_ip)
{
    rpc_req_t *req;
    p_byte    *buffer;
    int        rc;
    unsigned int payload_len;
    rpc_tcp_call_hdr_t hdr;

#if USE_KERBEROS
    p_byte             *buffers[2];
    unsigned int        buffer_lens[2];
    rpc_tcp_call_hdr_t  hdr_nbo; /* Header in network byte order */
    pasis_error_t       error;
#endif

    buffer = PasisFreelist_Get (server->buffer_freelist);
    PASIS_DASSERT (buffer != NULL);

    req = rpc_req_get (server);
    PASIS_DASSERT (req != NULL);

    /* Read in header */
    rc = st_read_fully (netfd, &hdr, sizeof (rpc_tcp_call_hdr_t), 
                        RPC_REQ_TIMEOUT);
    if (rc <= 0) {
        if (rc == 0) {
            rc = EPIPE; /* indicate to caller that other party closed socket */
        } else {
            PASIS_PRINT_ERR ("Error: 1. st_read (%d = %s)\n",
                             errno, PASIS_STRERROR (errno));
            rc = errno;
        }
        PasisFreelist_Put (server->buffer_freelist, buffer);
        rpc_req_put (server, req);
        close (st_netfd_fileno (netfd));
        return rc;
    }

    server->bytes_read += rc;

#if USE_KERBEROS
    /* Copy pre-decoded rpc header */
    memcpy (&hdr_nbo, &hdr, sizeof (rpc_tcp_call_hdr_t));
#endif

    req->reply_netfd = netfd;
    req->reply_ip = remote_ip;
    req->rpc_hdr = (u_int32_t *)buffer;
    req->data_buffer = buffer;
    payload_len = rpc_decode_tcp_call (req, &hdr, remote_ip);
    
    PASIS_DPRINTF (PASIS_DEBUG_RPC, 4, 
                   "handle_tcp_request: xid=%u, payload_len=%u\n", 
                   req->rpc_xid, payload_len);

    /* Read in data and args */
    PASIS_DASSERT (payload_len <= RPC_MAX_PAYLOAD);
    rc = st_read_fully (netfd, buffer, payload_len, RPC_REQ_TIMEOUT);
    if (rc <= 0) {
        if (rc == 0) {
            rc = EPIPE; /* indicate to caller that other party closed socket */
        } else {
            PASIS_PRINT_ERR ("Error: 2. st_read (%s)\n", PASIS_STRERROR (errno));
            rc = errno;
        }
        rpc_req_put (server, req);
        close (st_netfd_fileno (netfd));
        return rc;
    }

    server->bytes_read += rc;

    PASIS_DPRINTF (PASIS_DEBUG_RPC, 4, "handle_tcp_request: read data\n");


#if USE_KERBEROS
    buffers[0] = ((p_byte *)&hdr_nbo) + PASIS_HASH_SIZE;
    buffers[1] = buffer;
    buffer_lens[0] = sizeof (rpc_tcp_call_hdr_t) - PASIS_HASH_SIZE;
    buffer_lens[1] = hdr.hash_region_len;
    
    error = rpc_auth_validate (server->handle_type, server->local_ip, 
                               remote_ip, buffers, buffer_lens, 2, 
                               hdr.signature);

    PASIS_DASSERT (error == PASIS_SUCCESS);
#endif

    /* Call callback */
    server->callback (req, buffer, rc);

    return PASIS_SUCCESS;
}


int
rpc_decode_tcp_call (rpc_req_t *req,
                     rpc_tcp_call_hdr_t *hdr,
                     p_uint32 remote_ip)
{
#if USE_KERBEROS
    hdr->hash_region_len = ntohl (hdr->hash_region_len);
#endif

    req->rpc_xid = ntohl (hdr->xid);
    req->rpc_prog = ntohl (hdr->prog);
    req->rpc_vers = ntohl (hdr->vers);
    req->rpc_proc = ntohl (hdr->proc);
    
    if (ntohl (hdr->msg_type) != RPC_CALL) {
        PASIS_PRINT_ERR ("Error: rpc_decode_tcp_call, msg type mismatch\n");
        PASIS_DASSERT (0);
        return -1;
    }

    return ntohl (hdr->payload_len);
}


void
rpc_encode_tcp_reply (rpc_handle_t *handle,
                      rpc_req_t *req,
                      rpc_tcp_reply_hdr_t *hdr)
{
    int i;
    unsigned int payload_len=0;

#if USE_KERBEROS
    p_byte        *buffers[2];
    unsigned int   buffer_lens[2];
    pasis_error_t  error;

    hdr->hash_region_len = htonl (req->reply_iov[1].iov_len);
#endif

    hdr->xid = htonl (req->rpc_xid);
    hdr->msg_type = htonl (RPC_REPLY);
    hdr->reply_status = htonl (RPC_MSG_ACCEPTED);
    hdr->accepted_status = htonl (RPC_ACCEPT_SUCCESS);
    hdr->auth_status = htonl (RPC_AUTH_OK);

    /* Start at 1, since 0 contains the rpc hdr info */
    for (i = 1; i < req->reply_iov_len; i++) {
        payload_len += req->reply_iov[i].iov_len;
    }

    hdr->payload_len = htonl (payload_len);

    
#if USE_KERBEROS
    buffers[0] = req->reply_iov[0].iov_base + PASIS_HASH_SIZE;
    buffers[1] = req->reply_iov[1].iov_base;
    buffer_lens[0] = req->reply_iov[0].iov_len - PASIS_HASH_SIZE;
    buffer_lens[1] = req->reply_iov[1].iov_len;

    error = rpc_auth_generate (handle->handle_type, handle->local_ip, 
                               req->reply_ip, buffers, buffer_lens, 2, 
                               hdr->signature);

    PASIS_DASSERT (error == PASIS_SUCCESS);
#endif    

    return;
}


int
rpc_transmit_tcp_reply (rpc_handle_t *server,
                        rpc_req_t *req)
{
    int rc, i, real_errno;

    PASIS_DPRINTF (PASIS_DEBUG_RPC, 4, "Transmitting: xid=%u\n", req->rpc_xid);
    for (i = 0; i < req->reply_iov_len; i++) {
        PASIS_DPRINTF (PASIS_DEBUG_RPC, 4, "- %d: len=%u\n", 
                       i, (p_uint32)req->reply_iov[i].iov_len);
    }

    st_mutex_lock (server->sock_lock);
    rc = st_writev (req->reply_netfd, req->reply_iov, 
                    req->reply_iov_len, RPC_REQ_TIMEOUT);
    real_errno = errno; /* save errno since st_mutex_unlock will clear it */
    st_mutex_unlock (server->sock_lock);
    
    /* Handle closed or error condition */
    if (rc <= 0) {
        PASIS_PRINT_INFO ("%s: writev error, closed fd (error: %s, host: %s)\n",
                          __FUNCTION__,  strerror(real_errno),
                          pasis_ip_to_hostname (req->reply_ip));
        close (st_netfd_fileno (req->reply_netfd));
    }

    server->bytes_written += rc;

    PASIS_DPRINTF (PASIS_DEBUG_RPC, 4, "Transmitted: %d\n", rc);

    return rc;
}


void *
rpc_poll_tcp_server_thread (void *arg) 
{
    int            rc, i, err;
    rpc_handle_t  *handle = (rpc_handle_t *)arg;
    struct pollfd *fds;
    st_netfd_t    *netfds;
    unsigned int   pollfd_num;
    p_uint32      *ips;

    fds = malloc (RPC_MAX_CONNECTIONS * sizeof (struct pollfd));
    netfds = malloc (RPC_MAX_CONNECTIONS * sizeof (st_netfd_t));
    ips = malloc (RPC_MAX_CONNECTIONS * sizeof (p_uint32));

    fds[0].fd = st_netfd_fileno (handle->sock_fd);
    fds[0].events = POLLIN;
    
    netfds[0] = handle->sock_fd;
    
    pollfd_num = 1;

    while (!handle->req_poll_shutdown) {
        rc = st_poll (fds, pollfd_num, -1);
        if (rc == -1) {
            continue;
        }

        /* Server */
        for (i = 1; i < pollfd_num && rc > 0; i++) {

            if (fds[i].revents & POLLIN) {
                err = rpc_handle_tcp_request (handle, netfds[i], ips[i]);
                rc--;
            } else if (fds[i].revents & POLLERR || fds[i].revents & POLLHUP ||
                       fds[i].revents & POLLNVAL) {
                err = fds[i].revents & (POLLERR | POLLHUP | POLLNVAL);
                rc--;
            } else {
                err = 0;
            }
        
            /* Close connection */
            if (err) { 
                if (err == EPIPE) {
                    PASIS_PRINT_INFO ("%s: %s closed socket; closing "
                                      "connection\n", __FUNCTION__,
                                      pasis_ip_to_hostname (ips[i]));
                } else {
                    PASIS_PRINT_ERR ("%s: Closing connection (error: %d, "
                                     "host: %s)\n", __FUNCTION__, err,
                                     pasis_ip_to_hostname (ips[i]));
                }

                st_mutex_lock (handle->sock_lock);
                close (st_netfd_fileno (netfds[i]));
                st_netfd_free (netfds[i]);
                st_mutex_unlock (handle->sock_lock);        

                memmove (&fds[i], &fds[i+1], (pollfd_num - i - 1) * 
                         sizeof (struct pollfd));
                memmove (&netfds[i], &netfds[i+1], 
                         (pollfd_num - i - 1) * sizeof (st_netfd_t));
                memmove (&ips[i], &ips[i+1], 
                         (pollfd_num - i - 1) * sizeof (p_uint32));
                pollfd_num--;
                i--;
            }
        
        }
    
        /* Handle new connection */
        if (fds[0].revents & POLLIN) {
            struct sockaddr_in addr;
            int addrlen = sizeof (addr);

            netfds[pollfd_num] = st_accept (netfds[0], 
                                            (struct sockaddr *)&addr, 
                                            &addrlen, RPC_ACCEPT_TIMEOUT);
            if (netfds[pollfd_num] != NULL) {
                PASIS_PRINT_INFO ("%s: accepting new connection from %s\n",
                                  __FUNCTION__,
                                  pasis_ip_to_hostname (addr.sin_addr.s_addr));
                fds[pollfd_num].fd = st_netfd_fileno (netfds[pollfd_num]);
                fds[pollfd_num].events = POLLIN;
                ips[pollfd_num] = addr.sin_addr.s_addr;
                pollfd_num++;
            }
        }
    }

    for (i = 1; i < pollfd_num; i++) {
        PASIS_DASSERT (netfds[i] != NULL);
        st_netfd_close (netfds[i]);
    }

    free (fds);
    free (netfds);
    free (ips);
    
    return NULL;
}


rpc_handle_t *
rpc_server_create_tcp (unsigned int max_outstanding,
                       unsigned short port,
                       unsigned long  program,
                       unsigned long  version,
                       void (*callback)(rpc_req_t *, p_byte *, unsigned int))
{
    rpc_handle_t *server;
    int fd, rc, flag=1, ret;

    fd = socket (AF_INET, SOCK_STREAM, 0);
    if (fd < 0) {
        PASIS_PRINT_INFO ("%s: error creating socket (%d)\n", __FUNCTION__,
                          errno);
        return NULL;
    }
    
#if 1
    ret = setsockopt (fd, get_tcp_code (), TCP_NODELAY, &flag, sizeof (int));
#else
    ret = 1;
#endif
    setsockopt (fd, SOL_SOCKET, SO_REUSEADDR, &flag, sizeof (int));

    if (ret < 0) {
        PASIS_PRINT_ERR ("%s: error when turning off Nagle's algorithm (%d)\n",
                         __FUNCTION__, errno);
        return NULL;
    }


    server = rpc_server_create_common (max_outstanding, port, program, 
                                       version, IPPROTO_TCP, callback);
    if (server == NULL) {
        close (fd);
        return NULL;
    }

    rc = bind (fd, (struct sockaddr *)&server->sock_addr, 
               sizeof (server->sock_addr));
    if (rc < 0) {
        PASIS_PRINT_ERR ("%s: bind failed (%s)\n", __FUNCTION__,
                         PASIS_STRERROR (errno));
        free (server);
        close (fd);
        return NULL;
    }

    rc = listen (fd, 8);
    if (rc < 0) {
        PASIS_PRINT_ERR ("%s: listen failed %d\n", __FUNCTION__,
                         errno);
        free (server);
        close (fd);
        return NULL;
    }

    server->sock_lock = st_mutex_new ();
    server->sock_fd = st_netfd_open_socket (fd);
    server->proto = IPPROTO_TCP;
    server->req_poll_thread = st_thread_create (rpc_poll_tcp_server_thread, 
                                                server, 1, 0);
    
    return server;
}

/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
