#include <netinet/in.h>
#include <netinet/tcp.h>

#include "pasis_config.h"
#include "pasis_types.h"
#include "pasis_debug.h"
#include "pasis_timer.h"

#include "rpc_common.h"

#ifndef MAX_AUTH_BYTES
#define MAX_AUTH_BYTES  400
#endif

#ifndef RPC_MSG_VERSION
#define RPC_MSG_VERSION         ((u_long) 2)
#endif

extern p_byte rpc_dummy_buffer[RPC_MAX_PAYLOAD];

void * rpc_poll_udp_thread (void *arg);


static inline int
__rpc_transmit (rpc_handle_t *client,
                rpc_req_t *req)
{
    struct msghdr msg;
    int rc;

    msg.msg_name = &client->sock_addr;
    msg.msg_namelen = sizeof (client->sock_addr);
    msg.msg_iov = req->req_iov;
    msg.msg_iovlen = req->req_iov_len;
    msg.msg_control = NULL;
    msg.msg_controllen = 0;
    msg.msg_flags = 0;

    req->transmit_flag = 1;

    rc = st_sendmsg (client->sock_fd, &msg, 0, -1);
    PASIS_DASSERT (rc > 0);

    PASIS_DPRINTF (PASIS_DEBUG_RPC, 4, "%s: sendmsg (xid=%u, rc=%d)\n", 
                   __FUNCTION__, req->rpc_xid, rc);

    return rc;
}



static void
rpc_handle_request (rpc_handle_t *server)
{
    rpc_req_t *req;
    p_byte    *buffer, *data;
    int        rc, addr_size;

    buffer = PasisFreelist_Get (server->buffer_freelist);
    PASIS_DASSERT (buffer != NULL);

    req = rpc_req_get (server);
    PASIS_DASSERT (req != NULL);

    addr_size = sizeof (server->sock_addr);
    rc = st_recvfrom (server->sock_fd, buffer, RPC_MAX_PAYLOAD, 
                      (struct sockaddr *)&req->reply_addr, &addr_size, -1);

    if (rc < 0) {
        PASIS_PRINT_ERR ("Error: st_recvfrom (%s)\n", PASIS_STRERROR (errno));
        PasisFreelist_Put (server->buffer_freelist, buffer);
        rpc_req_put (server, req);
        return;
    }

    req->rpc_hdr = (u_int32_t *)buffer;
    req->data_buffer = buffer;
    data = (p_byte *)rpc_decode_udp_call (req, req->rpc_hdr);
    
    /* Call callback */
    server->callback (req, data, rc - (int)((buffer - data)));
}




static void
rpc_handle_reply (rpc_handle_t *client)
{
    rpc_req_t *req, req_buf;
    int error;
    socklen_t from_len=0;
    struct msghdr msg;
    int hdr_len=0, msg_len;
    char reply_hdr_buf[RPC_REPLY_HDR_SIZE << 2];
    rpc_req_t *t_req=NULL;

    PASIS_DPRINTF (PASIS_DEBUG_RPC, 5, "%s: %d\n", __FUNCTION__, __LINE__);

    /* Peek at the response rpc header for demuxing */
    /* Not exactly using st threads here, but it should work */
    error = recvfrom (st_netfd_fileno(client->sock_fd), reply_hdr_buf, 
                      (RPC_REPLY_HDR_SIZE << 2), MSG_PEEK, NULL, &from_len);
    if (error <= 0) {
        PASIS_PRINT_INFO ("%s: recvfrom error (errno=%d)\n", 
                          __FUNCTION__, errno);
        if (errno == EAGAIN || errno == EINTR) {
            return;
        }
    } else {
        req_buf.rpc_xid = 0;
        hdr_len = rpc_decode_udp_reply (&req_buf, (u_int32_t *)reply_hdr_buf);
    }

    PASIS_DPRINTF (PASIS_DEBUG_RPC, 4, "%s: Recv XID=%u\n", 
                   __FUNCTION__, req_buf.rpc_xid);

    PASIS_DPRINTF (PASIS_DEBUG_RPC, 5, "%s: %d, hdr_len=%d\n", 
                   __FUNCTION__, __LINE__, hdr_len);

    /* Find the request */
    req = client->inprog_queue_hd;
    while (req) {
        if (req->rpc_xid == req_buf.rpc_xid) {
            break;
        }
        req = req->next;
    }

    if (req == NULL) { /* Couldn't find a request, throw data away */
        PASIS_DASSERT (0);
        PASIS_PRINT_INFO ("%s: invalid xid=%u\n", 
                          __FUNCTION__, req_buf.rpc_xid);
        /* Again, not using st_recvfrom, but there should be a packet... */
        recvfrom (st_netfd_fileno (client->sock_fd), rpc_dummy_buffer, 
                  RPC_MAX_PAYLOAD, 0, NULL, &from_len);

        return;
    }

    /* Fixup reply header length for read */
    req->reply_iov[0].iov_len = hdr_len;

    /* Read in the response */
    msg.msg_name = &client->sock_addr;
    msg.msg_namelen = sizeof (client->sock_addr);
    msg.msg_iov = req->reply_iov;
    msg.msg_iovlen = req->reply_iov_len;
    msg.msg_control = NULL;
    msg.msg_controllen = 0;
    msg.msg_flags = 0;

    /* Again not using st_recvfrom... */
    msg_len = recvmsg (st_netfd_fileno (client->sock_fd), &msg, 0);

    PASIS_DPRINTF (PASIS_DEBUG_RPC, 4, "rpc_req_recv (%d)\n", msg_len);
    if (msg_len < 0) {
        PASIS_PRINT_INFO ("%s: recvmsg error (errno=%d)\n", 
                          __FUNCTION__, errno);
    }

    req->reply_msg_len = msg_len;

    /* Wake up the waiter */
    req->wait_flag = 0;
    req->wait_status = RPC_STATUS_SUCCESS;
    st_cond_signal (req->wait_cond);

    /* Check if more requests can be transmitted */
    client->trans_inprog_count--;

    if (client->trans_inprog_count < RPC_TRANSMIT_HIGH_WATER) {
        t_req = client->trans_queue_hd;
        if (t_req) {
            transmit_req_dequeue (client, t_req);
            client->trans_inprog_count++;
        }
    }

    if (t_req) {
        PASIS_DPRINTF (PASIS_DEBUG_RPC, 4, 
                       "%s: HIGH WATER Transmit: XID: %u\n", 
                       pasis_ip_to_hostname (client->remote_ip), t_req->rpc_xid);

        __rpc_transmit (client, t_req);
    }

}


int
rpc_decode_udp_reply (rpc_req_t *req,
                      u_int32_t *rpc_hdr)
{
    u_int32_t xid, tmp;
    u_int32_t *base=rpc_hdr;

    /* Get XID */
    xid = ntohl (*rpc_hdr++);

    if (req->rpc_xid == 0) {
        req->rpc_xid = xid;
    }

    if (xid != req->rpc_xid) {
        PASIS_DPRINTF (PASIS_DEBUG_RPC, 3, 
                       "Error: rpc_reply_decode xid mismatch\n");
        return -1;
    }

    /* Check this is a reply */
    if ((tmp = ntohl(*rpc_hdr++)) != RPC_REPLY) {
        PASIS_DPRINTF (PASIS_DEBUG_RPC, 3, 
                       "Error: not an RPC reply (%#x)\n", tmp);
        return -1;
    }

    /* Check reply status */
    if ((req->reply_status = ntohl (*rpc_hdr++)) != RPC_MSG_ACCEPTED) {
        req->reply_rejected_status = ntohl (*rpc_hdr++);
        if (req->reply_rejected_status == RPC_AUTH_ERROR) {
            req->reply_rejected_auth_status = ntohl (*rpc_hdr++);
            PASIS_DPRINTF (PASIS_DEBUG_RPC, 3, 
                           "Error: rpc -- reply rejected status=%u\n",
                           req->reply_rejected_auth_status);
        }
        
        return ((rpc_hdr - base) << 2);
    }

    /* Decode auth info */
    tmp = ntohl (*rpc_hdr++);  /* Verf type */
    if (tmp != RPC_AUTH_NULL && tmp != RPC_AUTH_UNIX && 
        tmp != RPC_AUTH_SHORT) {
        PASIS_DPRINTF (PASIS_DEBUG_RPC, 3, 
                       "Error: invalid reply verf flavor (%#x)\n", tmp);
        return -1;
    }

    if ((tmp = ntohl (*rpc_hdr++)) > MAX_AUTH_BYTES) {  /* Verf len */
        PASIS_DPRINTF (PASIS_DEBUG_RPC, 3, 
                       "Error: invalid reply verf size (%u)\n", tmp);
        return -1;
    }

    rpc_hdr += (tmp >> 2);

    /* Get accepted status */
    req->reply_accepted_status = ntohl (*rpc_hdr++);
    if (req->reply_accepted_status != RPC_ACCEPT_SUCCESS) {
        PASIS_DPRINTF (PASIS_DEBUG_RPC, 3, 
                       "Error: rpc -- reply accepted status=%u\n",
                       req->reply_accepted_status);
    }


    return ((rpc_hdr - base) << 2);
}

u_int32_t *
rpc_decode_udp_call (rpc_req_t *req,
                     u_int32_t *rpc_hdr)
{
    u_int32_t tmp;

    req->rpc_xid = ntohl (*rpc_hdr++);

    tmp = ntohl (*rpc_hdr++);
    PASIS_DASSERT (tmp == RPC_CALL);
    
    tmp = ntohl (*rpc_hdr++);
    PASIS_DASSERT (tmp == RPC_MSG_VERSION);

    req->rpc_prog = ntohl (*rpc_hdr++);
    req->rpc_vers = ntohl (*rpc_hdr++);
    req->rpc_proc = ntohl (*rpc_hdr++);

    tmp = ntohl (*rpc_hdr++);  /* Cred flavor */
    PASIS_DASSERT (tmp == RPC_AUTH_UNIX);

    tmp = ntohl (*rpc_hdr++);   /* Cred length */

    tmp = ntohl (*rpc_hdr++);  /* Time */
    tmp = ntohl (*rpc_hdr++);  /* Hostname length in bytes */
    rpc_hdr += (tmp + 3) >> 2;

    req->rpc_auth.uid = htonl (*rpc_hdr++);
    req->rpc_auth.gid = htonl (*rpc_hdr++);

    tmp = htonl (*rpc_hdr++);  /* GID array length */
    /* Skip over GID array for now */
    rpc_hdr += tmp;

    tmp = ntohl (*rpc_hdr++);  /* Verf flavor */
    PASIS_DASSERT (tmp == RPC_AUTH_NULL);
    tmp = ntohl (*rpc_hdr++);/* Verf length */
    PASIS_DASSERT (tmp == 0);
    
    return rpc_hdr;
}

u_int32_t *
rpc_encode_udp_call (rpc_handle_t *client, 
                     rpc_req_t *req, 
                     u_int32_t *rpc_hdr)
{
    u_int32_t *base_auth;
    time_t secs;
    int n;

    secs = time(NULL);

    *rpc_hdr++ = htonl (req->rpc_xid);
    *rpc_hdr++ = htonl (RPC_CALL);
    *rpc_hdr++ = htonl (RPC_MSG_VERSION);
    *rpc_hdr++ = htonl (req->rpc_prog);
    *rpc_hdr++ = htonl (req->rpc_vers);
    *rpc_hdr++ = htonl (req->rpc_proc);

    *rpc_hdr++ = htonl(RPC_AUTH_UNIX);
    base_auth = rpc_hdr++;
    *rpc_hdr++ = htonl(secs);
    
    n = client->local_hostlen;
    *rpc_hdr++ = htonl(n);
    memcpy(rpc_hdr, client->local_hostname, n);
    rpc_hdr += (n + 3) >> 2;
    
    *rpc_hdr++ = htonl((u_int32_t) req->rpc_auth.uid);
    *rpc_hdr++ = htonl((u_int32_t) req->rpc_auth.gid);
    
    *rpc_hdr++ = htonl((u_int32_t) 1);                  /* gid array length */
    *rpc_hdr++ = htonl((u_int32_t) req->rpc_auth.gid);

    *base_auth = htonl((rpc_hdr - base_auth - 1) << 2);     /* cred length */
    *rpc_hdr++ = htonl(RPC_AUTH_NULL);
    *rpc_hdr++ = htonl(0);

    return rpc_hdr;
}

int
rpc_transmit_udp_reply (rpc_handle_t *server,
                        rpc_req_t *req)
{
    int rc;
    struct msghdr msg;

    msg.msg_name = &req->reply_addr;
    msg.msg_namelen = sizeof (req->reply_addr);
    msg.msg_iov = req->reply_iov;
    msg.msg_iovlen = req->reply_iov_len;
    msg.msg_control = NULL;
    msg.msg_controllen = 0;
    msg.msg_flags = 0;

    rc = st_sendmsg (server->sock_fd, &msg, 0, -1);
    PASIS_DASSERT (rc > 0);

    return rc;
}






int
rpc_transmit_udp_req (rpc_handle_t *client, 
                      rpc_req_t *req)
{
    int rc;

    req->wait_flag = 1;
    req->transmit_flag = 0;
    inprog_req_queue (client, req);                 /* Add to wait queue */
    
    PASIS_DPRINTF (PASIS_DEBUG_RPC, 5, "%s: XID=%u\n",
                   __FUNCTION__, req->rpc_xid);

    if (client->trans_inprog_count >= RPC_TRANSMIT_HIGH_WATER) {
        /* Add to queue and wait for transmit */

        PASIS_DPRINTF (PASIS_DEBUG_RPC, 5, 
                       "%s: HIGH WATER Block: %u, XID: %u\n", 
                       pasis_ip_to_hostname (client->remote_ip),
                       client->trans_inprog_count,
                       req->rpc_xid);

        transmit_req_queue (client, req);
        return 1;
    }

    /* Do the transmit now */
    client->trans_inprog_count++;

    rc = __rpc_transmit (client, req);
    if (rc <= 0) {
        PASIS_PRINT_INFO ("Error: rpc_req_transmit (errno=%d)\n", errno);
    }

    PASIS_DPRINTF (PASIS_DEBUG_RPC, 4, "rpc_req_transmit (%d)\n", rc);

    return rc;
}


int
rpc_req_udp_wait (rpc_handle_t *client,
                  rpc_req_t *req)
{
    static int timeouts=0;
    int error=0, rc;
    unsigned int retry_count=0;
    st_utime_t timeout_secs;

    PASIS_DPRINTF (PASIS_DEBUG_RPC, 5, "%s: XID=%u\n",
                   __FUNCTION__, req->rpc_xid);

    timeout_secs = RPC_TIMEOUT_BASE;

    while (req->wait_flag) {             /* Wait for response */
        error = st_cond_timedwait (req->wait_cond, timeout_secs);

        if (error == -1) {
            PASIS_DASSERT (req->wait_flag == 1);
            if (errno == ETIME && req->transmit_flag) {
                timeouts++;
                PASIS_PRINT_INFO ("RPC RETRANSMIT!!! %s rpc XID=%u (TOs=%d, "
                                  "Sec=%llu, InP=%u)\n",
                                  pasis_ip_to_hostname (client->remote_ip), 
                                  req->rpc_xid, timeouts, timeout_secs, 
                                  client->trans_inprog_count);

                if (retry_count >= RPC_RETRIES) {
                    break;
                }

                retry_count++;
                timeout_secs *= RPC_TIMEOUT_MULT;

                rc = __rpc_transmit (client, req);
                if (rc < 0) {
                    PASIS_DPRINTF (PASIS_DEBUG_RPC, 5, 
                                   "%s: retransmit error XID=%u\n", 
                                   __FUNCTION__, req->rpc_xid);
                }
            }
        }
    }

    PASIS_DPRINTF (PASIS_DEBUG_RPC, 4, "%s: Waking XID=%u (wf=%d, err=%d)\n",
                   __FUNCTION__, req->rpc_xid, req->wait_flag, error);

    inprog_req_dequeue (client, req);     /* Remove from wait queue */

    return error;
}


void *
rpc_poll_udp_thread (void *arg)
{
    int rc;
    rpc_handle_t *handle = (rpc_handle_t *)arg;

    while (!handle->req_poll_shutdown) {
        rc = st_netfd_poll (handle->sock_fd, POLLIN, -1);
        if (rc == -1) {
            continue;
        }

        if (handle->handle_type == RPC_HANDLE_CLIENT) {
            rpc_handle_reply (handle);
        } else {
            rpc_handle_request (handle);
        }
    }

    return NULL;
}


rpc_handle_t *
rpc_client_create_udp (p_uint32 server_ip,
                       unsigned int max_outstanding,
                       unsigned long program,
                       unsigned long version)
{
    rpc_handle_t *client;
    int fd;

    fd = socket (AF_INET, SOCK_DGRAM, 0);

    if (fd < 0) {
        PASIS_PRINT_INFO ("Error: creating socket (%d)\n", errno);
        return NULL;
    }

    client = rpc_client_create_common (server_ip, max_outstanding, 
                                       program, version, IPPROTO_UDP);

    if (client == NULL) {
        close (fd);
        return NULL;
    }

    client->sock_fd = st_netfd_open_socket (fd);
    client->proto = IPPROTO_UDP;
    client->req_poll_thread = st_thread_create (rpc_poll_udp_thread, 
                                                client, 1, 0);

    return client;
}


rpc_handle_t *
rpc_server_create_udp (unsigned int max_outstanding,
                       unsigned short port,
                       unsigned long  program,
                       unsigned long  version,
                       void (*callback)(rpc_req_t *, p_byte *, unsigned int))
{
    rpc_handle_t *server;
    int fd, rc, flag=1;

    fd = socket (AF_INET, SOCK_DGRAM, 0);
    if (fd < 0) {
        PASIS_PRINT_INFO ("Error: creating socket (%d)\n", errno);
        return NULL;
    }
    
    setsockopt (fd, SOL_SOCKET, SO_REUSEADDR, &flag, sizeof (int));

    server = rpc_server_create_common (max_outstanding, port, program,
				       version, IPPROTO_UDP, callback);
    if (server == NULL) {
        close (fd);
        return NULL;
    }

    rc = bind (fd, (struct sockaddr *)&server->sock_addr, 
               sizeof (server->sock_addr));
    if (rc < 0) {
        PASIS_PRINT_ERR ("Error: rpc_server create, bind failed (%s)\n",
                         PASIS_STRERROR (errno));
        free (server);
        close (fd);
        return NULL;
    }

    server->sock_fd = st_netfd_open_socket (fd);
    server->proto = IPPROTO_UDP;
    server->req_poll_thread = st_thread_create (rpc_poll_udp_thread, 
                                                server, 1, 0);
    
    return server;
}
