/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#ifndef __PASISFS_CACHE_H__
#define __PASISFS_CACHE_H__

#include "pasis_config.h"
#include "pasis_debug.h"
#include "pasis_types.h"

#include "pasis_lists.h"

#ifdef __cplusplus
extern "C" {
#endif
#if 0
}
#endif


/* Buffer flags */
#define PASISFS_B_NONE     0x0000  /* No flags, free and clean */
#define PASISFS_B_DIRTY    0x0004  /* Dirty block */
#define PASISFS_B_IO       0x0008  /* I/O pending */
#define PASISFS_B_Q_NONE   0x0010  /* Buffer not on any list */
#define PASISFS_B_Q_FREE   0x0020  /* Buffer on free list */
#define PASISFS_B_Q_CLEAN  0x0040  /* Buffer on clean list */
#define PASISFS_B_Q_DIRTY  0x0080  /* Buffer on dirty list */
#define PASISFS_B_Q_LOCKED 0x0100  /* Buffer on locked list */
#define PASISFS_B_HASHED   0x0200  /* Buffer is in hash table */

#define PASISFS_B_IS_DIRTY(bp) \
  (((bp)->b_flags & PASISFS_B_DIRTY) ? PASIS_TRUE : PASIS_FALSE)

#define PASISFS_B_IS_PINNED(bp) \
  ((bp)->b_pin_count > 0)

#define PASISFS_B_IS_HASHED(bp) \
  (((bp)->b_flags & PASISFS_B_HASHED) ? PASIS_TRUE : PASIS_FALSE)

struct pasisfs_bcache_s;

typedef struct pasisfs_buf_s {
    /* Internal cache management */
    unsigned int     b_id;
    unsigned int     b_hash;

    struct pasisfs_bcache_s *b_cache; /*< Cache */

    PASIS_Q_PTRS (struct pasisfs_buf_s) b_hash_chain;
    PASIS_Q_PTRS (struct pasisfs_buf_s) b_queue;

    p_uint32         b_flags;
    p_uint32         b_pin_count;

    p_byte          *b_tag;

    /* Externally accessed */
    st_cond_t        b_io_cond;

    void            *b_data;

    PASIS_Q_PTRS (struct pasisfs_buf_s) b_list;
} pasisfs_buf_t;


typedef PASIS_Q_STRUCT(pasisfs_buf_queue_s, pasisfs_buf_t)
     pasisfs_buf_queue_t;

typedef struct pasisfs_bcache_s {
    char *name;

    unsigned int count;          /*< Number of elements */
    unsigned int buckets;        /*< Number of hash buckets */
    unsigned int tag_size;       /*< Size of buffer tag */

    /*< Function pointers for hashing and comparing buffers */
    unsigned int (* hash_fp)(void *tag, unsigned buckets);
    int (* compare_fp)(void *tag1, void *tag2);
    pasis_error_t (* write_fp)(pasisfs_buf_t *bp);
    void (* evict_fp)(pasisfs_buf_t *bp);
    void (* dealloc_fp)(void *data);

    pasisfs_buf_queue_t *hash_array;

    /*< Buffer queues */
    pasisfs_buf_queue_t free_buf_queue;
    pasisfs_buf_queue_t clean_buf_queue;
    pasisfs_buf_queue_t dirty_buf_queue;
    pasisfs_buf_queue_t locked_buf_queue;
   
    p_byte *metadata_region;

    /*< Stats */
    unsigned int         in_core_hits;
    unsigned int         in_core_misses;
    unsigned int         releases;
    unsigned int         dirty_releases;
    unsigned int         free_gets;
    unsigned int         clean_gets;
    unsigned int         dirty_gets;
    unsigned int         get_retries;
    
} pasisfs_bcache_t;


pasisfs_buf_t *PasisFS_BufCacheGetEmpty (pasisfs_bcache_t *cache);

pasis_error_t PasisFS_BufCacheRelease (pasisfs_buf_t *bp);

void PasisFS_BufCacheMarkDirty (pasisfs_buf_t *bp);

void PasisFS_BufCacheMarkClean (pasisfs_buf_t *bp);

pasis_error_t PasisFS_BufCacheInsert (pasisfs_buf_t *bp);

void PasisFS_BufCacheUnpin (pasisfs_buf_t *bp);

pasis_error_t PasisFS_BufCachePin (pasisfs_buf_t *bp);

pasisfs_buf_t *PasisFS_BufCacheFindInCore (pasisfs_bcache_t *cache,
                                           void *tag);

void PasisFS_BufCacheShutdown (pasisfs_bcache_t *cache);

pasisfs_bcache_t *PasisFS_BufCacheInit (char *cache_name,
                                        unsigned int count, 
                                        unsigned int tag_size, 
                                        unsigned int (* hash_fp)(void *tag, 
                                                        unsigned int buckets),
                                        int (* compare_fp)(void *tag1, 
                                                           void *tag2),
                                        pasis_error_t 
                                        (* write_fp)(pasisfs_buf_t *),
                                        void (* evict_fp) (pasisfs_buf_t *),
                                        void *(* alloc_fp)(unsigned int idx),
                                        void (* dealloc_fp)(void *data));

              
/* Verify cache integrity -- assumes no pinned or dirty buffers */
void PasisFS_BufCacheVerify (pasisfs_bcache_t *cache, int allow_locked);

#ifdef __cplusplus
}
#endif

/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */  


#endif /* __PASISFS_CACHE_H__ */
