/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"
#include "pasis_debug.h"
#include "pasis_error.h"
#include "pasis_timer.h"

#include "pasis_md_prot.h"

#include "mount3_prot.h"
#include "nfs3_prot.h"

#include "fe_rpc.h"
#include "fe_xdr.h"
#include "xdr_nfs3.h"

#include "nfs3_fe.h"


/* Use NFSv2 definition of max pathlen -- it is up to the server */
#ifndef NFS_MAXPATHLEN
#define NFS_MAXPATHLEN 1024
#endif

#define START_FUNCTION                                                  \
do {                                                                    \
    PASISFS_STATS_INCR (proc_counts[rpc_state->rpc_proc]);              \
    PASIS_TIMER_START_TM (&timer);                                      \
} while (0)

#define END_FUNCTION                                                    \
do {                                                                    \
    PASIS_TIMER_STOP_TM (&timer);                                       \
    PASIS_TIMER_ELAPSED_TS (&timer, &time_spec);                        \
    PASIS_TIMER_ADD_TS (&time_spec, &statistics.proc_timers[rpc_state->rpc_proc], &statistics.proc_timers[rpc_state->rpc_proc]);                        \
} while (0)

#define PASISFS_STATS_INCR(var) (statistics.var++)

#define RPC_IN_ARGS(rpc_state) \
  ((u_int32_t *)((rpc_state)->in_iov[1].iov_base))

#define RPC_OUT_ARGS(rpc_state) \
  ((u_int32_t *)((rpc_state)->out_iov[1].iov_base))

#define RPC_SET_OUT_LEN(rpc_state, p) \
  ((rpc_state)->out_iov[1].iov_len = ((char *)p - (char *)rpc_state->out_iov[1].iov_base))


static char *nfs_procs[22]={"NFS3_NULL", "NFS3_GETATTR", "NFS3_SETATTR",
                            "NFS3_LOOKUP", "NFS3_ACCESS", "NFS3_READLINK",
                            "NFS3_READ", "NFS3_WRITE", "NFS3_CREATE",
                            "NFS3_MKDIR", "NFS3_SYMLINK", "NFS3_MKNOD",
                            "NFS3_REMOVE", "NFS3_RMDIR", "NFS3_RENAME",
                            "NFS3_LINK", "NFS3_READDIR", "NFS3_READDIRPLUS",
                            "NFS3_FSSTAT", "NFS3_FSINFO", "NFS3_PATHCONF",
                            "NFS3_COMMIT"};

typedef struct pasisfs_nfs3_stats_s {
    unsigned int proc_counts[22];
    pasis_timespec_t proc_timers[22];
} pasisfs_nfs3_stats_t;

static pasisfs_nfs3_stats_t statistics;

void
NFS3Init (void)
{
    Pasis_TimeInit ();
    memset(&statistics, 0, sizeof (pasisfs_nfs3_stats_t));
}


void
NFS3Shutdown (void)
{
    int i;
    unsigned int total_count=0;
    pasis_timespec_t total_time={0,0};

    PASIS_PRINT_INFO ("---- NFSv3 Statistics ----\n");
    for (i = 0; i < 22; i++) {
        PASIS_PRINT_INFO ("%-17s :  Count=%-12u  Time=%u.%.3us\n", 
                          nfs_procs[i], statistics.proc_counts[i],
                          statistics.proc_timers[i].ts_sec,
                          statistics.proc_timers[i].ts_nsec/1000000);
        total_count += statistics.proc_counts[i];
        PASIS_TIMER_ADD_TS (&total_time, &statistics.proc_timers[i], 
                              &total_time);
    }
    PASIS_PRINT_INFO ("-------------------\n");
    PASIS_PRINT_INFO ("Totals            :  Count=%-12u  Time=%u.%.9us\n\n",
                      total_count, total_time.ts_sec, total_time.ts_nsec);

    PasisFS_NFS3FeShutdown ();
}


static bool_t
nfsproc3_null (rpc_state_t *rpc_state)
{
    PASISFS_STATS_INCR(proc_counts[rpc_state->rpc_proc]);
    return PASIS_TRUE;
}

static bool_t
nfsproc3_getattr (rpc_state_t *rpc_state)
{
    nfs_fh3     arg;
    getattr3res result;
    u_int32_t  *p;
    
    pasis_timer_t timer;
    pasis_timespec_t time_spec;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, "\n--%s\n", __FUNCTION__);

    START_FUNCTION;

    xdr_decode_fh (RPC_IN_ARGS (rpc_state), &arg);

    result.status = PasisFS_NFS3GetAttr (rpc_state->rpc_auth_uid, &arg, 
                                         &result.getattr3res_u.attributes);

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, "%s: m:%o u:%d g:%d s:%llu\n",
                   __FUNCTION__, result.getattr3res_u.attributes.mode,
                   result.getattr3res_u.attributes.uid, 
                   result.getattr3res_u.attributes.gid,
                   result.getattr3res_u.attributes.size);

    p = xdr_encode_getattr (RPC_OUT_ARGS (rpc_state), &result);
    RPC_SET_OUT_LEN (rpc_state, p);

    END_FUNCTION;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 2, "%s: status=%d\n\n", __FUNCTION__,
                   result.status);

    return PASIS_TRUE;
}

static bool_t
nfsproc3_setattr (rpc_state_t *rpc_state)
{
    setattr3args arg;
    wccstat3     result;
    u_int32_t   *p;

    pasis_timer_t timer;
    pasis_timespec_t time_spec;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, "\n--%s\n", __FUNCTION__);

    START_FUNCTION;

    xdr_decode_setattr (RPC_IN_ARGS (rpc_state), &arg);

    result.status = PasisFS_NFS3SetAttr (rpc_state->rpc_auth_uid, 
                                          &arg.object, 
                                          &arg.new_attributes, &arg.guard, 
                                          &result.wccstat3_u.wcc);

    p = xdr_encode_wccstat (RPC_OUT_ARGS (rpc_state), &result);
    RPC_SET_OUT_LEN (rpc_state, p);

    END_FUNCTION;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 2, "%s: status=%d\n\n", __FUNCTION__,
                   result.status);

    return PASIS_TRUE;
}

static bool_t
nfsproc3_lookup (rpc_state_t *rpc_state)
{
    diropargs3 arg;
    lookup3res result;
    char       fh[NFS3_FHSIZE];
    u_int32_t *p;

    pasis_timer_t timer;
    pasis_timespec_t time_spec;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, "\n--%s\n", __FUNCTION__);

    START_FUNCTION;

    xdr_decode_dirop (RPC_IN_ARGS (rpc_state), &arg);
    
    result.lookup3res_u.resok.object.data.data_val = fh;

    result.status = PasisFS_NFS3Lookup (rpc_state->rpc_auth_uid, &arg.dir, 
                                   arg.name, &result.lookup3res_u.resok.object,
                                   &result.lookup3res_u.resok.obj_attributes,
                                   &result.lookup3res_u.resok.dir_attributes,
                                   &result.lookup3res_u.resfail);

    p = xdr_encode_lookup (RPC_OUT_ARGS (rpc_state), &result);
    RPC_SET_OUT_LEN (rpc_state, p);

    END_FUNCTION;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 2, "%s: status=%d\n\n", __FUNCTION__,
                   result.status);

    return PASIS_TRUE;
}

static bool_t
nfsproc3_access (rpc_state_t *rpc_state)
{
    access3args arg;
    access3res  result;
    u_int32_t  *p;

    pasis_timer_t timer;
    pasis_timespec_t time_spec;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, "\n--%s\n", __FUNCTION__);

    START_FUNCTION;

    xdr_decode_access (RPC_IN_ARGS (rpc_state), &arg);

    result.status = PasisFS_NFS3Access (rpc_state->rpc_auth_uid, &arg.object, 
                                    arg.access,
                                    &result.access3res_u.resok.obj_attributes,
                                    &result.access3res_u.resok.access);

    p = xdr_encode_access (RPC_OUT_ARGS (rpc_state), &result);
    RPC_SET_OUT_LEN (rpc_state, p);

    END_FUNCTION;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 2, "%s: status=%d\n\n", __FUNCTION__,
                   result.status);

    return PASIS_TRUE;
}

static bool_t
nfsproc3_readlink (rpc_state_t *rpc_state)
{
    nfs_fh3      arg;
    readlink3res result;
    char         path[NFS_MAXPATHLEN+1];  /* This is as much as 1KB */
    u_int32_t   *p;

    pasis_timer_t timer;
    pasis_timespec_t time_spec;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, "\n--%s\n", __FUNCTION__);

    START_FUNCTION;

    xdr_decode_fh (RPC_IN_ARGS (rpc_state), &arg);

    result.readlink3res_u.resok.data = path;
    result.status = PasisFS_NFS3Readlink (rpc_state->rpc_auth_uid, &arg, 
                             result.readlink3res_u.resok.data, 
                             &result.readlink3res_u.resok.symlink_attributes);

    p = xdr_encode_readlink (RPC_OUT_ARGS (rpc_state), &result);
    RPC_SET_OUT_LEN (rpc_state, p);

    END_FUNCTION;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 2, "%s: status=%d\n\n", __FUNCTION__,
                   result.status);

    return PASIS_TRUE;
}

/* Max size of read3res structure -- used to calculate where to place data */
#define READ3RES_MAXSIZE ((1 + 22 + 2 + 1) << 2)

static bool_t
nfsproc3_read (rpc_state_t *rpc_state)
{
    read3args arg;
    read3res  result;
    u_int32_t *p;

    pasis_timer_t timer;
    pasis_timespec_t time_spec;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, "\n--%s\n", __FUNCTION__);

    START_FUNCTION;

    xdr_decode_read (RPC_IN_ARGS (rpc_state), &arg);

    /* Setup place for data */
    rpc_state->out_iov[2].iov_base = (rpc_state->out_iov[1].iov_base + 
                                      READ3RES_MAXSIZE);

    result.read3res_u.resok.data.data_val = rpc_state->out_iov[2].iov_base;

    result.status = PasisFS_NFS3Read (rpc_state->rpc_auth_uid, &arg.file, 
                                     arg.offset, arg.count,
                                     result.read3res_u.resok.data.data_val,
                                     &result.read3res_u.resok.count,
                                     &result.read3res_u.resok.eof,
                                     &result.read3res_u.resok.file_attributes);

    result.read3res_u.resok.data.data_len = result.read3res_u.resok.count;
    rpc_state->out_iov[2].iov_len = result.read3res_u.resok.count;

    p = xdr_encode_read (RPC_OUT_ARGS (rpc_state), &result);
    RPC_SET_OUT_LEN (rpc_state, p);

    END_FUNCTION;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 2, "%s: status=%d\n\n", __FUNCTION__,
                   result.status);

    return PASIS_TRUE;
}

static bool_t
nfsproc3_write (rpc_state_t *rpc_state)
{
    write3args arg;
    write3res  result;
    u_int32_t *p;

    pasis_timer_t timer;
    pasis_timespec_t time_spec;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, "\n--%s\n", __FUNCTION__);

    START_FUNCTION;

    xdr_decode_write (RPC_IN_ARGS (rpc_state), &arg);

    result.status = PasisFS_NFS3Write (rpc_state->rpc_auth_uid, &arg.file, 
                                       arg.offset, arg.count,
                                       arg.data.data_val, arg.stable,
                                       &result.write3res_u.resok.count,
                                       &result.write3res_u.resok.file_wcc,
                                       &result.write3res_u.resok.committed,
                                       result.write3res_u.resok.verf);

    p = xdr_encode_write (RPC_OUT_ARGS (rpc_state), &result);
    RPC_SET_OUT_LEN (rpc_state, p);

    END_FUNCTION;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 2, "%s: status=%d\n\n", __FUNCTION__,
                   result.status);

    return PASIS_TRUE;
}

static bool_t
nfsproc3_create (rpc_state_t *rpc_state)
{
    create3args arg;
    diropres3   result;
    char        fh[NFS3_FHSIZE];
    u_int32_t  *p;

    pasis_timer_t timer;
    pasis_timespec_t time_spec;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, "\n--%s\n", __FUNCTION__);

    START_FUNCTION;

    xdr_decode_create (RPC_IN_ARGS (rpc_state), &arg);

    result.diropres3_u.resok.obj.post_op_fh3_u.handle.data.data_val = fh;
    result.status = PasisFS_NFS3Create (rpc_state->rpc_auth_uid, 
                                        rpc_state->rpc_auth_gid, 
                                        &arg.where.dir, 
                                        arg.where.name, arg.how.mode, 
                                        &arg.how.createhow3_u.obj_attributes,
                                        &result.diropres3_u.resok);

    if (result.status != NFS3_OK) {
        memcpy (&result.diropres3_u.resfail, 
                &result.diropres3_u.resok.dir_wcc, sizeof (wcc_data));
    }

    p = xdr_encode_dirop (RPC_OUT_ARGS (rpc_state), &result);
    RPC_SET_OUT_LEN (rpc_state, p);

    END_FUNCTION;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 2, "%s: status=%d\n\n", __FUNCTION__,
                   result.status);

    return PASIS_TRUE;
}

static bool_t
nfsproc3_mkdir (rpc_state_t *rpc_state)
{
    mkdir3args arg;
    diropres3  result;
    char       fh[NFS3_FHSIZE];
    u_int32_t *p;

    pasis_timer_t timer;
    pasis_timespec_t time_spec;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, "\n--%s\n", __FUNCTION__);

    START_FUNCTION;
    
    xdr_decode_mkdir (RPC_IN_ARGS (rpc_state), &arg);

    result.diropres3_u.resok.obj.post_op_fh3_u.handle.data.data_val = fh;
    result.status = PasisFS_NFS3Mkdir (rpc_state->rpc_auth_uid, 
                                        rpc_state->rpc_auth_gid, 
                                        &arg.where.dir,
                                        arg.where.name, &arg.attributes,
                                        &result.diropres3_u.resok);
                                        
    if (result.status != NFS3_OK) {
        memcpy (&result.diropres3_u.resfail, 
                &result.diropres3_u.resok.dir_wcc, sizeof (wcc_data));
    }

    p = xdr_encode_dirop (RPC_OUT_ARGS (rpc_state), &result);
    RPC_SET_OUT_LEN (rpc_state, p);

    END_FUNCTION;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 2, "%s: status=%d\n\n", __FUNCTION__,
                   result.status);

    return PASIS_TRUE;
}

static bool_t
nfsproc3_symlink (rpc_state_t *rpc_state)
{
    symlink3args arg;
    diropres3    result;
    char         fh[NFS3_FHSIZE];
    u_int32_t   *p;

    pasis_timer_t timer;
    pasis_timespec_t time_spec;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, "\n--%s\n", __FUNCTION__);

    START_FUNCTION;

    xdr_decode_symlink (RPC_IN_ARGS (rpc_state), &arg);

    result.diropres3_u.resok.obj.post_op_fh3_u.handle.data.data_val = fh;
    result.status = PasisFS_NFS3Symlink (rpc_state->rpc_auth_uid, 
                                         rpc_state->rpc_auth_gid, 
                                         &arg.where.dir, arg.where.name, 
                                         arg.symlink.symlink_data,
                                         &arg.symlink.symlink_attributes,
                                         &result.diropres3_u.resok);
                                          
    if (result.status != NFS3_OK) {
        memcpy (&result.diropres3_u.resfail, 
                &result.diropres3_u.resok.dir_wcc, sizeof (wcc_data));
    }

    p = xdr_encode_dirop (RPC_OUT_ARGS (rpc_state), &result);
    RPC_SET_OUT_LEN (rpc_state, p);

    END_FUNCTION;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 2, "%s: status=%d\n\n", __FUNCTION__,
                   result.status);

    return PASIS_TRUE;
}

static bool_t
nfsproc3_mknod (rpc_state_t *rpc_state)
{
    mknod3args arg;
    diropres3  result;
    char       fh[NFS3_FHSIZE];
    u_int32_t *p;

    pasis_timer_t timer;
    pasis_timespec_t time_spec;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, "\n--%s\n", __FUNCTION__);

    START_FUNCTION;
    
    xdr_decode_mknod (RPC_IN_ARGS (rpc_state), &arg);

    result.diropres3_u.resok.obj.post_op_fh3_u.handle.data.data_val = fh;
    result.status = PasisFS_NFS3Mknod (rpc_state->rpc_auth_uid, 
                                        rpc_state->rpc_auth_gid, 
                                        &arg.where.dir, 
                                        arg.where.name, &arg.what,
                                        &result.diropres3_u.resok);

    if (result.status != NFS3_OK) {
        memcpy (&result.diropres3_u.resfail, 
                &result.diropres3_u.resok.dir_wcc, sizeof (wcc_data));
    }

    p = xdr_encode_dirop (RPC_OUT_ARGS (rpc_state), &result);
    RPC_SET_OUT_LEN (rpc_state, p);

    END_FUNCTION;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 2, "%s: status=%d\n\n", __FUNCTION__,
                   result.status);

    return PASIS_TRUE;
}


static bool_t
nfsproc3_remove (rpc_state_t *rpc_state)
{
    diropargs3 arg;
    wccstat3   result;
    u_int32_t *p;

    pasis_timer_t timer;
    pasis_timespec_t time_spec;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, "\n--%s\n", __FUNCTION__);

    START_FUNCTION;

    xdr_decode_dirop (RPC_IN_ARGS (rpc_state), &arg);

    result.status = PasisFS_NFS3Unlink (rpc_state->rpc_auth_uid, &arg.dir, 
                                         arg.name, &result.wccstat3_u.wcc);

    p = xdr_encode_wccstat (RPC_OUT_ARGS (rpc_state), &result);
    RPC_SET_OUT_LEN (rpc_state, p);

    END_FUNCTION;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 2, "%s: status=%d\n\n", __FUNCTION__,
                   result.status);

    return PASIS_TRUE;
}

static bool_t
nfsproc3_rmdir (rpc_state_t *rpc_state)
{
    diropargs3 arg;
    wccstat3   result;
    u_int32_t *p;

    pasis_timer_t timer;
    pasis_timespec_t time_spec;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, "\n--%s\n", __FUNCTION__);

    START_FUNCTION;

    xdr_decode_dirop (RPC_IN_ARGS (rpc_state), &arg);

    result.status = PasisFS_NFS3Rmdir (rpc_state->rpc_auth_uid, &arg.dir, 
                                        arg.name, &result.wccstat3_u.wcc);

    p = xdr_encode_wccstat (RPC_OUT_ARGS (rpc_state), &result);
    RPC_SET_OUT_LEN (rpc_state, p);

    END_FUNCTION;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 2, "%s: status=%d\n\n", __FUNCTION__,
                   result.status);

    return PASIS_TRUE;
}

static bool_t
nfsproc3_rename (rpc_state_t *rpc_state)
{
    rename3args arg;
    rename3res  result;
    u_int32_t  *p;

    pasis_timer_t timer;
    pasis_timespec_t time_spec;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, "\n--%s\n", __FUNCTION__);

    START_FUNCTION;

    xdr_decode_rename (RPC_IN_ARGS (rpc_state), &arg);

    result.status = PasisFS_NFS3Rename (rpc_state->rpc_auth_uid, 
                                         &arg.from.dir, arg.from.name,
                                         &arg.to.dir, arg.to.name,
                                         &result.rename3res_u.res.fromdir_wcc,
                                         &result.rename3res_u.res.todir_wcc);

    p = xdr_encode_rename (RPC_OUT_ARGS (rpc_state), &result);
    RPC_SET_OUT_LEN (rpc_state, p);

    END_FUNCTION;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 2, "%s: status=%d\n\n", __FUNCTION__,
                   result.status);

    return PASIS_TRUE;
}

static bool_t
nfsproc3_link (rpc_state_t *rpc_state)
{
    link3args  arg;
    link3res   result;
    u_int32_t *p;

    pasis_timer_t timer;
    pasis_timespec_t time_spec;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, "\n--%s\n", __FUNCTION__);

    START_FUNCTION;

    xdr_decode_link (RPC_IN_ARGS (rpc_state), &arg);

    result.status = PasisFS_NFS3Link (rpc_state->rpc_auth_uid, &arg.file, 
                                       &arg.link.dir, arg.link.name, 
                                       &result.link3res_u.res.file_attributes,
                                       &result.link3res_u.res.linkdir_wcc);

    p = xdr_encode_link (RPC_OUT_ARGS (rpc_state), &result);
    RPC_SET_OUT_LEN (rpc_state, p);

    END_FUNCTION;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 2, "%s: status=%d\n\n", __FUNCTION__,
                   result.status);

    return PASIS_TRUE;
}

#define READDIR3RES_MAXSIZE ((1 + 22 + 2) << 2)

static bool_t
nfsproc3_readdir (rpc_state_t *rpc_state)
{
    readdir3args arg;
    readdir3res  result;
    u_int32_t   *p;

    pasis_timer_t timer;
    pasis_timespec_t time_spec;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, "\n--%s\n", __FUNCTION__);

    START_FUNCTION;

    xdr_decode_readdir (RPC_IN_ARGS (rpc_state), &arg);

    /* Setup memory to place entries */
    rpc_state->out_iov[2].iov_base = (rpc_state->out_iov[1].iov_base +
                                      READDIR3RES_MAXSIZE);

    result.status = PasisFS_NFS3Readdir (rpc_state->rpc_auth_uid, &arg.dir, 
                                   arg.count, arg.cookie, arg.cookieverf,
                                   &result.readdir3res_u.resok.dir_attributes,
                                   result.readdir3res_u.resok.cookieverf,
                                   &rpc_state->out_iov[2]);

    p = xdr_encode_readdir (RPC_OUT_ARGS (rpc_state), &result);
    RPC_SET_OUT_LEN (rpc_state, p);

    END_FUNCTION;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 2, "%s: status=%d\n\n", __FUNCTION__,
                   result.status);

    return PASIS_TRUE;
}

#define READDIRPLUS3RES_MAXSIZE ((1 + 22 + 2) << 2)

static bool_t
nfsproc3_readdirplus (rpc_state_t *rpc_state)
{
    readdirplus3args arg;
    readdirplus3res  result;
    u_int32_t       *p;
    
    pasis_timer_t timer;
    pasis_timespec_t time_spec;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, "\n--%s\n", __FUNCTION__);

    START_FUNCTION;

    xdr_decode_readdirplus (RPC_IN_ARGS (rpc_state), &arg);

    /* Setup memory to place entries */
    rpc_state->out_iov[2].iov_base = (rpc_state->out_iov[1].iov_base +
                                      READDIRPLUS3RES_MAXSIZE);

    result.status = PasisFS_NFS3Readdirplus (rpc_state->rpc_auth_uid, 
                               &arg.dir, arg.maxcount,
                               arg.dircount, arg.cookie, arg.cookieverf,
                               &result.readdirplus3res_u.resok.dir_attributes,
                               result.readdirplus3res_u.resok.cookieverf,
                               &rpc_state->out_iov[2]);

    p = xdr_encode_readdirplus (RPC_OUT_ARGS (rpc_state), &result);
    RPC_SET_OUT_LEN (rpc_state, p);

    END_FUNCTION;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 2, "%s: status=%d\n\n", __FUNCTION__,
                   result.status);

    return PASIS_TRUE;
}

static bool_t
nfsproc3_fsstat (rpc_state_t *rpc_state)
{
    nfs_fh3    arg;
    fsstat3res result;
    u_int32_t *p;

    pasis_timer_t timer;
    pasis_timespec_t time_spec;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, "\n--%s\n", __FUNCTION__);

    START_FUNCTION;

    xdr_decode_fh (RPC_IN_ARGS (rpc_state), &arg);

    result.status = PasisFS_NFS3FSstat (rpc_state->rpc_auth_uid, &arg, 
                                        &result.fsstat3res_u.resok);

    p = xdr_encode_fsstat (RPC_OUT_ARGS (rpc_state), &result);
    RPC_SET_OUT_LEN (rpc_state, p);

    END_FUNCTION;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 2, "%s: status=%d\n\n", __FUNCTION__,
                   result.status);

    return PASIS_TRUE;
}

static bool_t
nfsproc3_fsinfo (rpc_state_t *rpc_state)
{
    nfs_fh3    arg;
    fsinfo3res result;
    u_int32_t *p;

    pasis_timer_t timer;
    pasis_timespec_t time_spec;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, "\n--%s\n", __FUNCTION__);

    START_FUNCTION;

    xdr_decode_fh (RPC_IN_ARGS (rpc_state), &arg);

    result.status = PasisFS_NFS3FSinfo (rpc_state->rpc_auth_uid, &arg, 
                                        &result.fsinfo3res_u.resok);

    p = xdr_encode_fsinfo (RPC_OUT_ARGS (rpc_state), &result);
    RPC_SET_OUT_LEN (rpc_state, p);

    END_FUNCTION;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 2, "%s: status=%d\n\n", __FUNCTION__,
                   result.status);

    return PASIS_TRUE;
}

static bool_t
nfsproc3_pathconf (rpc_state_t *rpc_state)
{
    nfs_fh3      arg;
    pathconf3res result;
    u_int32_t   *p;

    pasis_timer_t timer;
    pasis_timespec_t time_spec;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, "\n--%s\n", __FUNCTION__);

    START_FUNCTION;

    xdr_decode_fh (RPC_IN_ARGS (rpc_state), &arg);

    result.status = PasisFS_NFS3Pathconf (rpc_state->rpc_auth_uid, &arg, 
                                           &result.pathconf3res_u.resok);

    p = xdr_encode_pathconf (RPC_OUT_ARGS (rpc_state), &result);
    RPC_SET_OUT_LEN (rpc_state, p);

    END_FUNCTION;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 2, "%s: status=%d\n\n", __FUNCTION__,
                   result.status);

    return PASIS_TRUE;
}

static bool_t
nfsproc3_commit (rpc_state_t *rpc_state)
{
    commit3args arg;
    commit3res  result;
    u_int32_t  *p;

    pasis_timer_t timer;
    pasis_timespec_t time_spec;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 1, "\n--%s\n", __FUNCTION__);

    START_FUNCTION;

    xdr_decode_commit (RPC_IN_ARGS (rpc_state), &arg);

    result.status = PasisFS_NFS3Commit (rpc_state->rpc_auth_uid, &arg.file, 
                                         arg.offset, arg.count, 
                                         &result.commit3res_u.resok.file_wcc,
                                         result.commit3res_u.resok.verf);

    p = xdr_encode_commit (RPC_OUT_ARGS (rpc_state), &result);
    RPC_SET_OUT_LEN (rpc_state, p);

    END_FUNCTION;

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 2, "%s: status=%d\n\n", __FUNCTION__,
                   result.status);

    return PASIS_TRUE;
}

void 
nfs_program_3(rpc_state_t *rpc_state)
{
    switch (rpc_state->rpc_proc) {
    case NFSPROC3_NULL:
        nfsproc3_null (rpc_state);
        break;
    case NFSPROC3_GETATTR:
        nfsproc3_getattr (rpc_state);
        break;
    case NFSPROC3_SETATTR:
        nfsproc3_setattr (rpc_state);
        break;
    case NFSPROC3_LOOKUP:
        nfsproc3_lookup (rpc_state);
        break;
    case NFSPROC3_ACCESS:
        nfsproc3_access (rpc_state);
        break;
    case NFSPROC3_READLINK:
        nfsproc3_readlink (rpc_state);
        break;
    case NFSPROC3_READ:
        nfsproc3_read (rpc_state);
        break;
    case NFSPROC3_WRITE:
        nfsproc3_write (rpc_state);
        break;
    case NFSPROC3_CREATE:
        nfsproc3_create (rpc_state);
        break;
    case NFSPROC3_MKDIR:
        nfsproc3_mkdir (rpc_state);
        break;
    case NFSPROC3_SYMLINK:
        nfsproc3_symlink (rpc_state);
        break;
    case NFSPROC3_MKNOD:
        nfsproc3_mknod (rpc_state);
        break;
    case NFSPROC3_REMOVE:
        nfsproc3_remove (rpc_state);
        break;
    case NFSPROC3_RMDIR:
        nfsproc3_rmdir (rpc_state);
        break;
    case NFSPROC3_RENAME:
        nfsproc3_rename (rpc_state);
        break;
    case NFSPROC3_LINK:
        nfsproc3_link (rpc_state);
        break;
    case NFSPROC3_READDIR:
        nfsproc3_readdir (rpc_state);
        break;
    case NFSPROC3_READDIRPLUS:
        nfsproc3_readdirplus (rpc_state);
        break;
    case NFSPROC3_FSSTAT:
        nfsproc3_fsstat (rpc_state);
        break;
    case NFSPROC3_FSINFO:
        nfsproc3_fsinfo (rpc_state);
        break;
    case NFSPROC3_PATHCONF:
        nfsproc3_pathconf (rpc_state);
        break;
    case NFSPROC3_COMMIT:
        nfsproc3_commit (rpc_state);
        break;
    default:
        PASIS_PRINT_ERR ("%s: error unknown nfsv3 RPC proc=%d\n",
                         __FUNCTION__, rpc_state->rpc_proc);
        break;
    }
}


/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
