/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"
#include "pasis_macosx_compat.h"

#include "pasisio.h"
#include "pasisio_config_parse.h"
#include "pasisio_drives.h"
#include "pasisio_md_protocol.h"

#include "fe.h"
#include "nfs3_fe.h"
#include "nfs3_svc.h"

#include "pasisfs_io.h"
#include "pasisfs_data_cache.h"

#include "pasis_md_prot.h"




/* NFS worker thread number */
static unsigned int nfs_outstanding = PASIS_NFS_DEFAULT_WORKERS;

/* Buffer cache size */
static unsigned int buffer_pool_size = PASIS_NFS_DEFAULT_BUFPOOL << 20;

/* Shutdown pipe */
static st_netfd_t shutdown_pipe[2];


#ifndef MIN
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#endif

static void
SigHandler(int signum)
{
    static int signalled = 0;

    if (!signalled) {
        char dummy;
        signalled=1;
        st_write (shutdown_pipe[1], &dummy, 1, -1);
    }
}


static void *
MainShutdownThread (void *arg)
{
    char dummy;
    int rc;

    /* Blocking read until shutdown is ready to occur */
    do {
        rc = st_read_fully (shutdown_pipe[0], &dummy, 1, -1);
    } while (rc != 1);
    
    return NULL;
}


static void
MainShutdown (void)
{
    pasis_info_print_enable = 0;

    FE_Shutdown ();

    PasisFS_NFS3FeShutdown ();

    NFS3Shutdown ();
    
    PasisIO_Shutdown ();

    pasis_info_print_enable = 1;
}


void 
usage (char *name)
{
    PASIS_PRINT_INFO ("Usage: %s [-w # -B #] -r <file> [-D] [-d <file>]\n", name);
    PASIS_PRINT_INFO (" -w #       # of nfs worker threads (Default %d)\n",
                      nfs_outstanding);
    PASIS_PRINT_INFO (" -B #       Buffer cache size in MB (Default %d MB)\n",
                      buffer_pool_size >> 20);
    PASIS_PRINT_INFO (" -r <file>  root config file (for protocol/linkages)\n");
    PASIS_PRINT_INFO (" -D         store data using Q/U synthetic objects\n");
    PASIS_PRINT_INFO (" -d <file>  data config file (for protocol/linkages)\n");
    PASIS_PRINT_INFO ("            This is optional; if not specified, \n");
    PASIS_PRINT_INFO ("            data will be stored at /tmp/pasisfs, if -D isn't specified\n");
}


int
main(int argc, char *argv[])
{
    char                      *md_cfg_file = NULL, *data_cfg_file = NULL;
    st_thread_t                shutdown_thread;
    pasis_error_t              error;
    void                      *buffer_pool;
    char                       c;
    int                        tmp_pipe[2], store_data_locally_flag = 0;
    int                        store_data_using_qu = 0;
    pasis_config_server_list_t md_server_list;
    pasis_md_linkage_t         md_linkages[PASIS_MD_MAX_LINKAGES];
    pasis_md_protocol_t        md_proto;
    pasis_config_server_list_t data_server_list;
    pasis_md_linkage_t         data_linkages[PASIS_MD_MAX_LINKAGES];
    pasis_md_protocol_t        data_proto;


    while ((c = getopt (argc, argv, "w:B:r:Dd:")) != -1) {
        switch (c) {
        case 'w':
            sscanf (optarg, "%u", &nfs_outstanding);
            break;
        case 'B':
            sscanf (optarg, "%u", &buffer_pool_size);

            /* In megabytes -- translate to bytes */
            buffer_pool_size = buffer_pool_size << 20;
            break;
        case 'r':
            md_cfg_file = optarg;
            break;
        case 'd':
            data_cfg_file = optarg;
            break;
        case 'D':
            store_data_using_qu = 1;
            break;
        default:
            usage (argv[0]);
            exit (0);
        }
    }

    if (md_cfg_file == NULL) {
        usage (argv[0]);
        exit (1);
    }

    if (store_data_using_qu) {
        PASIS_PRINT_INFO ("Using Q/U synthetic objects to store data.\n");
    } else if (data_cfg_file == NULL) {
        PASIS_PRINT_INFO ("Using local filesystem at /tmp/pasisfs to store "
                          "data.\n");
        store_data_locally_flag = 1;
    } else {
        PASIS_PRINT_INFO ("Using R/W to store data.\n");
    }

    /* Parse config file, which contains protocol/scheme/storage node list */
    if (PasisIO_ParseConfigurationFile (md_cfg_file, &md_proto, &md_server_list,
                                        NULL, NULL)) {
        PASIS_PRINT_ERR ("Error: configuration file parsing failed\n");
        exit (1);
    }


    if (store_data_locally_flag || store_data_using_qu) {
        memset (&data_proto, 0, sizeof(data_proto));
        memset (&data_server_list, 0, sizeof(data_server_list));
        memset (&data_linkages, 0, sizeof(data_linkages));
    } else {
        /* Parse config file, it contains protocol/scheme/storage node list */
        error = PasisIO_ParseConfigurationFile (data_cfg_file, &data_proto,
                                                &data_server_list, NULL, NULL);
        if (error) {                                        
            PASIS_PRINT_ERR ("Error: configuration file parsing failed\n");
            exit (1);
        }
    }

    /* Initialize state threads library */
    st_init ();

    /* Setup buffer pool -- needs to be done before PasisIO_Init (for ED) */
    buffer_pool_size = PAGE_ALIGN (buffer_pool_size);

    buffer_pool =  mmap (0, buffer_pool_size, PROT_READ | PROT_WRITE, 
                             MAP_ANONYMOUS | MAP_SHARED, -1, 0);
    if (buffer_pool == MAP_FAILED) {
        PASIS_PRINT_ERR ("Error: failed mapping buffer pool (%s)\n",
                         PASIS_STRERROR (errno));
        return 1;
    }

    error = PasisIO_Init (nfs_outstanding, PASIS_NFS_DATA_BLKSIZE);
    if (error) {
        PASIS_PRINT_ERR ("IOInit Error: %s\n", PASIS_STRERROR(error));
        munmap (buffer_pool, buffer_pool_size);
        exit (1);
    }

    error = PasisFS_IOInit (store_data_locally_flag, store_data_using_qu);
    if (error) {
        PASIS_PRINT_ERR ("IOInit Error: %s\n", PASIS_STRERROR(error));
        munmap (buffer_pool, buffer_pool_size);
        exit (1);
    }

    error = PasisFS_DataCacheInit (buffer_pool, buffer_pool_size);
    if (error) {
        PASIS_PRINT_ERR ("BufCacheInit Error: %s\n",
                           PASIS_STRERROR(error));
        munmap (buffer_pool, buffer_pool_size);
        exit (1);
    }

    /* Add storage nodes to PASIS drives module */
    error = PasisIO_ConnectToStorageNodes (&md_server_list, 
                                           PASISIO_DRIVE_TYPE_MD_INDEX,
                                           md_linkages, NULL);
    if (error) {
        PASIS_PRINT_ERR ("IOInit Error while connecting to md drives\n");
        munmap (buffer_pool, buffer_pool_size);
        exit (1);
    }

    if (!store_data_locally_flag) {
#if 0
        error = PasisIO_ConnectToStorageNodes (&data_server_list, 
                                               PASISIO_DRIVE_TYPE_S4_INDEX,
                                               data_linkages, NULL);
        if (error) {
            PASIS_PRINT_ERR ("IOInit Error while connecting to data drives\n");
            munmap (buffer_pool, buffer_pool_size);
            exit (1);
        }
#endif
    }

    NFS3Init ();

    error = PasisFS_NFS3FeInit (&md_proto, md_server_list.num_servers, 
                                md_linkages, &data_proto, 
                                data_server_list.num_servers,
                                data_linkages);
    if (error) {
        PASIS_PRINT_ERR ("Error initing NFS3 FE\n");
        exit (1);
    }

    error = FE_Startup (nfs_outstanding);
    if (error) {
        PASIS_PRINT_ERR ("FE Error: %s\n", PASIS_STRERROR (error));
        exit (1);
    }


    /* Setup shutdown pipe and shutdown thread */
    pipe (tmp_pipe);
    shutdown_pipe[0] = st_netfd_open (tmp_pipe[0]);
    shutdown_pipe[1] = st_netfd_open (tmp_pipe[1]);

    shutdown_thread = st_thread_create (MainShutdownThread, NULL, 1, 0);

    if (signal (SIGINT, SigHandler) == SIG_ERR) {
        MainShutdown ();
        munmap (buffer_pool, buffer_pool_size);
        PASIS_PRINT_ERR ("Error: couldn't setup signal handler\n");
        exit (1);
    }

    st_thread_join (shutdown_thread, NULL);

    PASIS_DPRINTF (PASIS_DEBUG_FS_NFS, 4, 
                   "%s: joined with shutdown thread, shutting down...\n",
                   __FUNCTION__);

    st_netfd_close (shutdown_pipe[0]);
    st_netfd_close (shutdown_pipe[1]);

    MainShutdown ();

    munmap (buffer_pool, buffer_pool_size);

    return 0;
}


/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
