/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"

#include "fe.h"
#include "pasisfs_data_cache.h"
#include "pasisfs_buf_cache.h"
#include "pasisfs_io.h"

/******************************* Data Cache **********************************/

pasisfs_bcache_t        *data_cache=NULL;
static p_byte           *buffer_pool=NULL;
static unsigned int      buffer_pool_size;
static unsigned int      buffer_pool_offset=0;


static inline char *
dump_fid (pasis_md_fid_t *f)
{
#if PASIS_DEBUG > 0
    static char buf[PASISIO_FILEID_SIZE*2 + 1];
    int i;
    char *bp = buf;
    for (i=0; i<PASISIO_FILEID_SIZE; i++) {
        snprintf(bp, 3, "%02X", (unsigned char)f->data[i]);
        bp +=2;
    }

    return buf;
#else 
    return NULL;
#endif
}


static unsigned int
DataCacheHash (void *vtag, unsigned int buckets)
{
    pasisfs_dcache_tag_t *tag = (pasisfs_dcache_tag_t *)(vtag);

    return (tag->blkno ^ (*(unsigned int *)tag->fid.data)) % buckets;
}


static int
DataCacheCompare (void *vtag1, void *vtag2)
{
    pasisfs_dcache_tag_t *tag1 = (pasisfs_dcache_tag_t *)(vtag1);
    pasisfs_dcache_tag_t *tag2 = (pasisfs_dcache_tag_t *)(vtag2);

    if (tag1->blkno == tag2->blkno) {
        if (memcmp (&tag1->fid, &tag2->fid, sizeof (pasis_md_fid_t)) == 0) {
            return 1;
        }
    }

    return 0;
}


static void *
DataCacheAlloc (unsigned int idx)
{
    pasisfs_dcache_data_t *data;

    data = malloc (sizeof (pasisfs_dcache_data_t));

    data->evict_time = 0;
    data->data = (p_byte *)(buffer_pool + buffer_pool_offset);

    buffer_pool_offset += PASIS_NFS_DATA_BLKSIZE;
    PASIS_DASSERT (buffer_pool_offset <= buffer_pool_size);

    return data;
}


static void
DataCacheDealloc (void *ptr)
{
    pasisfs_dcache_data_t *data;

    data = (pasisfs_dcache_data_t *)ptr;
    free (data);
}

/* Eviction callback handler -- called from within BufCacheRelease */
static void
DataCacheEvict (pasisfs_buf_t *bp)
{
    return;
}


pasisfs_buf_t *
PasisFS_DataCacheGetEmpty (void)
{
    pasisfs_buf_t *bp;

    bp = PasisFS_BufCacheGetEmpty (data_cache);
    PASIS_DASSERT (bp);

    memset (bp->b_tag, 0, sizeof (pasisfs_dcache_tag_t));

    return bp;
}


pasis_error_t
PasisFS_DataCacheInsert (pasis_md_fid_t *fid,
                         pasis_md_loc_t *loc,
                         p_uint64 blkno,
                         pasisfs_buf_t *bp)
{
    pasisfs_dcache_data_t *data;
    pasisfs_dcache_tag_t  *tag;

    tag = PASISFS_DCACHE_TAG (bp);
    data = PASISFS_DCACHE_DATA (bp);

    PASIS_DPRINTF (PASIS_DEBUG_FS_FILE, 4, 
                   "%s: inserting %s, bp=%p (%d) blkno=%llu\n",
                   __FUNCTION__, dump_fid (fid), bp, bp->b_id, blkno);

    /* Setup tag */
    memcpy (&tag->fid, fid, sizeof (pasis_md_fid_t));
    tag->blkno = blkno;    

    /* Do a lookup and make sure it doesn't exist */
    PASIS_DASSERT (PasisFS_BufCacheFindInCore (data_cache, &tag) == NULL);

    /* Update eviction time */
    data->evict_time = time (NULL) + 10;

    memcpy (&data->loc, loc, sizeof (pasis_md_loc_t));

    /* Insert buffer into cache */
    PasisFS_BufCacheInsert (bp);

    return PASIS_SUCCESS;
}

/* Either: Callback for writing dirty blocks that are being evicted or
 * Called by Sync */
pasis_error_t
PasisFS_DataCacheWrite (pasisfs_buf_t *bp)
{
    pasisfs_dcache_tag_t  *tag;
    pasisfs_dcache_data_t *data;
    pasis_error_t          error;

    tag = PASISFS_DCACHE_TAG (bp);
    data = PASISFS_DCACHE_DATA (bp);

    PASIS_DPRINTF (PASIS_DEBUG_FS_FILE, 3, 
                   "%s: fid=%s, blkno=%llu, bp=%p (%d)\n",
                   __FUNCTION__, dump_fid (&tag->fid), tag->blkno, 
                   bp, bp->b_id);

    PasisFS_BufCachePin (bp);
    if (!PASISFS_B_IS_DIRTY (bp)) {
        PasisFS_BufCacheUnpin (bp);
        return PASIS_SUCCESS;
    }

    /* Mark as I/O in progress */
    bp->b_flags |= PASISFS_B_IO;

    /* Do write I/O */
    error = PasisFS_IOWrite (&tag->fid, &data->loc,
                             tag->blkno, data->data);
    if (error) {
        PASIS_PRINT_ERR ("%s: Error: %d\n", __FUNCTION__, error);
        PASIS_DASSERT (error == PASIS_SUCCESS); /* XXX For now */
    }
        
    PasisFS_BufCacheMarkClean (bp);

    bp->b_flags &= ~PASISFS_B_IO;
    PasisFS_BufCacheUnpin (bp);

    /* Signal any waiters */
    if (bp->b_pin_count > 0) {
        st_cond_broadcast (bp->b_io_cond);
    }

    return PASIS_SUCCESS;
}


pasisfs_buf_t *
PasisFS_DataCacheRead (pasis_md_fid_t   *fid,
                       pasis_md_loc_t   *loc,
                       pasisfs_blkaddr_t blkno)
{
    pasisfs_dcache_tag_t   tag;
    pasisfs_dcache_data_t *data;
    pasisfs_buf_t         *bp;
    pasis_error_t          error;

    PASIS_DPRINTF (PASIS_DEBUG_FS_FILE, 3, 
                   "%s: fid=%s, blkno=%llu\n", __FUNCTION__, 
                   dump_fid (fid), blkno);
    
    memcpy (&tag.fid, fid, sizeof (pasis_md_fid_t));
    tag.blkno = blkno;

    /* First check incore buffers */
    bp = PasisFS_BufCacheFindInCore (data_cache, &tag);
    if (bp) {
        data = PASISFS_DCACHE_DATA(bp);
        
        if (data->evict_time < time (NULL)) {
            /* Evict from cache */
            PasisFS_BufCacheRelease (bp);
        } else {
            PASIS_DPRINTF (PASIS_DEBUG_FS_FILE, 4, "%s: read cache hit\n", 
                           __FUNCTION__);
            return bp;
        }
    }

    /* Get empty buffer and initialize it */
    bp = PasisFS_BufCacheGetEmpty (data_cache);
    PASIS_DASSERT (bp);

    data = PASISFS_DCACHE_DATA(bp);
    memcpy (&data->loc, loc, sizeof (pasis_md_loc_t));
    memcpy (bp->b_tag, &tag, sizeof (pasisfs_dcache_tag_t));

    /* Pin buffer */
    PasisFS_BufCachePin (bp);

    /* Mark as I/O in progress */
    bp->b_flags |= PASISFS_B_IO;

    /* Insert buffer into cache */
    PasisFS_BufCacheInsert (bp);

    /* Issue IO to read in data block */
    error = PasisFS_IORead (fid, &data->loc, blkno, data->data);
    if (error == PASIS_SUCCESS) {
        /* Success; do nothing */
    } else if (error == PASIS_ERROR_PROTO_NOWRITE) {
        /* Block doesn't exist; occurs if we're reading a non-existant block
         * in a sparse file. We'll return NULL and let our caller handle this.
         * But first, clean up */
        bp->b_flags &= ~PASISFS_B_IO;
        PasisFS_BufCacheUnpin (bp);
        PasisFS_BufCacheRelease (bp);
        return NULL;
    } else {
        /* fatal error */
        PASIS_PRINT_ERR ("%s: Error: %d\n", __FUNCTION__, error);
        PASIS_DASSERT (error == PASIS_SUCCESS);
        return NULL;
    }

    bp->b_flags &= ~PASISFS_B_IO;

    /* Unpin buffer */
    PasisFS_BufCacheUnpin (bp);

    /* Signal any waiters */
    if (bp->b_pin_count > 0) {
        st_cond_broadcast (bp->b_io_cond);
    }

    return bp;
}


pasisfs_buf_t *
PasisFS_DataCacheLookup (pasis_md_fid_t *fid,
                         pasisfs_blkaddr_t blkno)
{
    pasisfs_dcache_tag_t tag;
    pasisfs_buf_t *bp;

    memcpy (&tag.fid, fid, sizeof (pasis_md_fid_t));
    tag.blkno = blkno;

    /* Find in core */
    bp = PasisFS_BufCacheFindInCore (data_cache, &tag);

    return bp;
}


pasis_error_t
PasisFS_DataCacheInit (p_byte      *data_pool,
                       unsigned int data_pool_size)
{
    unsigned int num_buffers;

    /* Store args for use in DataCacheAlloc */
    buffer_pool = data_pool;
    buffer_pool_size = data_pool_size;
    buffer_pool_offset = 0;

    num_buffers = data_pool_size / PASIS_NFS_DATA_BLKSIZE;

    data_cache = PasisFS_BufCacheInit ("Data cache", num_buffers,
                                       sizeof(pasisfs_dcache_tag_t),
                                       DataCacheHash, DataCacheCompare, 
                                       PasisFS_DataCacheWrite, DataCacheEvict, 
                                       DataCacheAlloc, DataCacheDealloc);

    return PASIS_SUCCESS;
}


void
PasisFS_DataCacheShutdown (void)
{
    PasisFS_BufCacheShutdown (data_cache);
}



/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */  

