/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"
#include "pasis_debug.h"
#include "pasis_error.h"
#include "nfs3_prot.h"
#include "mount3_prot.h"
#include "mount1_prot.h"
#include "fe_rpc.h"

#define UID_NOBODY (-2)
#define GID_NOBODY (-2)



void
rpc_encode_reply(rpc_state_t *rpc_state)
{
    u_int32_t *rpc_hdr;

    /* Place header after data region */
    rpc_hdr = (u_int32_t *)(rpc_state->out_iov[0].iov_base);

    /* Encode XID */
    *rpc_hdr++ = htonl (rpc_state->rpc_xid);

    /* Make a reply */
    *rpc_hdr++ = htonl (RPC_REPLY);

    /* Put reply status */
    *rpc_hdr++ = htonl(rpc_state->reply_status);
    if (rpc_state->reply_status != RPC_MSG_ACCEPTED) {
        *rpc_hdr++ = htonl(rpc_state->reply_rejected_status);
        
        if (rpc_state->reply_rejected_status == RPC_AUTH_ERROR) {
            *rpc_hdr++ = htonl(rpc_state->reply_rejected_auth_status);
            
            PASIS_DPRINTF (PASIS_DEBUG_RPC, 3, 
                           "Error: rpc -- reply rejected status=%u\n",
                           rpc_state->reply_rejected_auth_status);
        }
        /* XXX not sure what to do here */
    }

    /* Auth verf type */
    *rpc_hdr++ = htonl (RPC_AUTH_NONE);  /* XXX For now hardcode */

    /* Auth verf length */
    *rpc_hdr++ = 0;

    /* Accepted status */
    *rpc_hdr++ = htonl(rpc_state->reply_accepted_status);

    rpc_state->out_iov[0].iov_len = ((char *)rpc_hdr) - 
        (char *)rpc_state->out_iov[0].iov_base;

    return;
}


pasis_error_t
rpc_decode_call (rpc_state_t *rpc_state)
{
    unsigned int msg_len = rpc_state->in_iov[0].iov_len;
    u_int32_t *rpc_hdr = (u_int32_t *)rpc_state->in_iov[0].iov_base;
    u_int32_t tmp;
    u_int32_t auth_flavor;

    rpc_state->rpc_xid = ntohl (*rpc_hdr++);

    /* Check if this is a CALL massage */
    tmp = ntohl (*rpc_hdr++);
    if( tmp != RPC_CALL ) { 
        rpc_state->reply_status = RPC_MSG_DENIED;
        rpc_state->reply_rejected_status = RPC_REJECT_MISMATCH;
        goto rejected_request;
    }

    /* RPC Version must be two */
    tmp = ntohl (*rpc_hdr++);
    if( tmp != RPC_MESG_VERSION ) { 
        rpc_state->reply_status = RPC_MSG_DENIED;
        rpc_state->reply_rejected_status = RPC_REJECT_MISMATCH;
        goto rejected_request;
    }

    /* Make sure this is the right RPC program, proc, and version */
    rpc_state->rpc_prog = ntohl (*rpc_hdr++);
    rpc_state->rpc_vers = ntohl (*rpc_hdr++);
    rpc_state->rpc_proc = ntohl (*rpc_hdr++);

    switch( rpc_state->rpc_prog ) { 
        
    case NFS_PROGRAM: 
        
        if( rpc_state->rpc_vers != NFS_V3 ) { 
            rpc_state->reply_status = RPC_MSG_ACCEPTED;
            rpc_state->reply_accepted_status = RPC_PROG_MISMATCH;
            goto rejected_request;
        }
        break;

    case MOUNT_PROGRAM: 
        if( rpc_state->rpc_vers != MOUNT_V3 && rpc_state->rpc_vers != MOUNTVERS ) { 
            rpc_state->reply_status = RPC_MSG_ACCEPTED;
            rpc_state->reply_accepted_status = RPC_PROG_MISMATCH;
            goto rejected_request;
        }
        break;

    default: 
        rpc_state->reply_status = RPC_MSG_ACCEPTED;
        rpc_state->reply_accepted_status = RPC_PROG_UNAVAIL;
        break;
    }
            

    auth_flavor = ntohl (*rpc_hdr++);  /* Cred flavor */
    tmp = ntohl (*rpc_hdr++);   /* Cred length */

    /* Parse through fields for AUTH_NONE and AUTH_UNIX */
    switch( auth_flavor ) { 

    case RPC_AUTH_UNIX: 
        tmp = ntohl (*rpc_hdr++);  /* Time */
        tmp = ntohl (*rpc_hdr++);  /* Hostname length in bytes */
        rpc_hdr += (tmp + 3) >> 2;

        rpc_state->rpc_auth_uid = htonl (*rpc_hdr++);
        rpc_state->rpc_auth_gid = htonl (*rpc_hdr++);

        tmp = htonl (*rpc_hdr++);  /* GID array length */

        /* Skip over GID array for now */
        rpc_hdr += tmp;

        break;

    case RPC_AUTH_NONE: 

        /* just set to NOBODY (-2) for now */
        rpc_state->rpc_auth_uid = UID_NOBODY;
        rpc_state->rpc_auth_gid = GID_NOBODY;
        
        break;

    default: 
        /* Reject the message because we don't support
         * the authentication protocol used by the client
         */
        rpc_state->reply_status = RPC_MSG_DENIED;
        rpc_state->reply_rejected_status = RPC_AUTH_ERROR;
        rpc_state->reply_rejected_auth_status = RPC_AUTH_REJECTEDCRED;
        goto rejected_request;

        break;
    }

    /* Decode the verifier -- ASSERT these for now *
     * These asserts will have go go away once we
     * start supporting AUTH_SHORT credentials
     */
    tmp = ntohl (*rpc_hdr++);  /* Verf flavor */
    PASIS_DASSERT (tmp == RPC_AUTH_NULL);
    tmp = ntohl (*rpc_hdr++);/* Verf length */
    PASIS_DASSERT (tmp == 0);


    rpc_state->reply_status = RPC_MSG_ACCEPTED;
    rpc_state->reply_accepted_status = RPC_ACCEPT_SUCCESS;


    /* Setup input iovec (0 - rpc header, 1 - data) */
    rpc_state->in_iov[0].iov_base = rpc_state->in_buffer;
    rpc_state->in_iov[0].iov_len = ((char *)rpc_hdr) - rpc_state->in_buffer;

    rpc_state->in_iov[1].iov_base = (char *)rpc_hdr;
    rpc_state->in_iov[1].iov_len = msg_len - rpc_state->in_iov[0].iov_len;

    return PASIS_SUCCESS;

 rejected_request: 
    
    rpc_state->in_iov[0].iov_base = NULL;
    rpc_state->in_iov[0].iov_len = 0;
    
    rpc_state->in_iov[1].iov_base = NULL;
    rpc_state->in_iov[1].iov_len = 0;

    return PASIS_SUCCESS;
}


pasis_error_t
rpc_read_udp_request (rpc_state_t *rpc_state)
{
    int rc, addr_size;

    addr_size = sizeof (rpc_state->rpc_addr);
    rc = st_recvfrom (rpc_state->rpc_fd, rpc_state->in_buffer, RPC_MAX_PAYLOAD,
                      (struct sockaddr *)&rpc_state->rpc_addr, &addr_size, -1);

    if (rc < 0) {
        PASIS_PRINT_ERR ("Error: st_recvfrom (%s)\n", PASIS_STRERROR (errno));
        return PASIS_ERROR;
    }

    rpc_state->in_iov[0].iov_base = rpc_state->in_buffer;
    rpc_state->in_iov[0].iov_len = rc;

    return PASIS_SUCCESS;
}


pasis_error_t
rpc_send_udp_reply (rpc_state_t *rpc_state, st_mutex_t write_lock)
{
    struct msghdr msg;
    int rc;

    msg.msg_name = &rpc_state->rpc_addr;
    msg.msg_namelen = sizeof (rpc_state->rpc_addr);
    msg.msg_iov = rpc_state->out_iov;
    msg.msg_iovlen = (rpc_state->out_iov[2].iov_len > 0) ? 3 : 2;
    msg.msg_control = NULL;
    msg.msg_controllen = 0;
    msg.msg_flags = 0;

    st_mutex_lock (write_lock);
    rc = st_sendmsg (rpc_state->rpc_fd, &msg, 0, -1);
    st_mutex_unlock (write_lock);

    if (rc <= 0) {
        PASIS_PRINT_ERR ("%s: error in st_sendmsg (%d)\n", 
                         __FUNCTION__, errno);
        return PASIS_ERROR;
    }

    PASIS_DPRINTF (PASIS_DEBUG_RPC, 4, 
                   "%s: sendmsg (iovlen=%d [%d|%d|%d], xid=%u, rc=%d)\n", 
                   __FUNCTION__, msg.msg_iovlen, (int)msg.msg_iov[0].iov_len, 
                   (int)msg.msg_iov[1].iov_len, (int)msg.msg_iov[2].iov_len, 
                   rpc_state->rpc_xid, rc);

    return PASIS_SUCCESS;
}



pasis_error_t
rpc_read_tcp_request(rpc_state_t *rpc_state)
{
    u_int32_t temp_hdr;
    u_int32_t nbytes;
    u_int32_t total_bytes = 0;
    int rc;

    do {
        st_read_fully(rpc_state->rpc_fd, &temp_hdr, 4, RPC_REQ_TIMEOUT);

        temp_hdr = ntohl(temp_hdr);

        /* MSB is a flag indicating last fragment */
        nbytes = temp_hdr - (temp_hdr & (1<<31));

        PASIS_DASSERT (total_bytes + nbytes <= RPC_MAX_PACKET);

        rc = st_read_fully (rpc_state->rpc_fd, rpc_state->in_buffer + 
                            total_bytes, nbytes, RPC_REQ_TIMEOUT);
            
        if (rc < 0) {
            return PASIS_ERROR;
        }

        total_bytes += nbytes;

    } while (temp_hdr & (1<<31));
  
    rpc_state->in_iov[0].iov_base = rpc_state->in_buffer;
    rpc_state->in_iov[0].iov_len = total_bytes;

    return PASIS_SUCCESS;
}


pasis_error_t 
rpc_send_tcp_reply(rpc_state_t *rpc_state, st_mutex_t write_lock)
{
    struct iovec iov[4];
    
    u_int32_t temp_hdr;
    u_int32_t nbytes = 0;
  
    int i, rc, iov_cnt=0;
 
    /* Go through output iovecs and figure out what to send */
    for (i = 1; i < 4; i++) {
        if (rpc_state->out_iov[i-1].iov_len > 0) {
            iov[i].iov_base = rpc_state->out_iov[i-1].iov_base;
            iov[i].iov_len = rpc_state->out_iov[i-1].iov_len;
            nbytes += iov[i].iov_len;
            iov_cnt++;
        }
    }

    /* Set fragment header, most-significant bit is 1 indicating last frag */
    temp_hdr = htonl (nbytes | (1 << 31));
  
    iov[0].iov_base = &temp_hdr;
    iov[0].iov_len = 4;

    st_mutex_lock (write_lock);
    rc = st_writev (rpc_state->rpc_fd, iov, 3, RPC_REQ_TIMEOUT);
    st_mutex_unlock (write_lock);

    if (rc <=0) {
	PASIS_PRINT_INFO ("%s: writev error, closed fd\n", __FUNCTION__);
        
        /* XXX figure out how to close this properly */
	close (st_netfd_fileno (rpc_state->rpc_fd));
	return PASIS_ERROR;
    }

    return PASIS_SUCCESS;
}

/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
