/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"

#include "nfs3_prot.h"
#include "fe_xdr.h"

#include "xdr_nfs3.h"


/* Decoding functions */
static inline u_int32_t *
xdr_decode_set_uint32 (u_int32_t *p, set_uint32_t *set)
{
    if (*p++) {
        set->set = 1;
        set->set_uint32_t_u.val = ntohl (*p++);
    } else {
        set->set = 0;
    }

    return p;
}

static inline u_int32_t *
xdr_decode_set_uint64 (u_int32_t *p, set_uint64_t *set)
{
    if (*p++) {
        set->set = 1;
        p = xdr_decode_hyper (p, &set->set_uint64_t_u.val);
    } else {
        set->set = 0;
    }

    return p;
}

static inline u_int32_t *
xdr_decode_set_time (u_int32_t *p, set_time *set)
{
    set->set = ntohl (*p++);
    if (set->set == SET_TO_CLIENT_TIME) {
        set->set_time_u.time.seconds = ntohl (*p++);
        set->set_time_u.time.nseconds = ntohl (*p++);
    }

    return p;
}

static inline u_int32_t *
xdr_decode_sattr (u_int32_t *p, sattr3 *sattr)
{
    p = xdr_decode_set_uint32 (p, &sattr->mode);
    p = xdr_decode_set_uint32 (p, &sattr->uid);
    p = xdr_decode_set_uint32 (p, &sattr->gid);
    p = xdr_decode_set_uint64 (p, &sattr->size);
    p = xdr_decode_set_time (p, &sattr->atime);
    p = xdr_decode_set_time (p, &sattr->mtime);
    return p;
}

u_int32_t *
xdr_decode_dirop (u_int32_t *p, diropargs3 *dirop)
{
    p = xdr_decode_fh (p, &dirop->dir);
    p = xdr_decode_string (p, &dirop->name);
    return p;   
}

u_int32_t *
xdr_decode_fh (u_int32_t *p, nfs_fh3 *fh)
{
    fh->data.data_len = ntohl (*p++);
    fh->data.data_val = (char *)p;
    
    return p + XDR_QUADLEN (fh->data.data_len);
}



/* setattr */
u_int32_t *
xdr_decode_setattr (u_int32_t *p, setattr3args *args)
{
    p = xdr_decode_fh (p, &args->object);
    p = xdr_decode_sattr (p, &args->new_attributes);
    if (*p++) {
        args->guard.check = 1;
        args->guard.sattrguard3_u.ctime.seconds = ntohl (*p++);
        args->guard.sattrguard3_u.ctime.nseconds = ntohl (*p++);
    } else {
        args->guard.check = 0;
    }

    return p;
}

u_int32_t *
xdr_decode_access (u_int32_t *p, access3args *args)
{
    p = xdr_decode_fh (p, &args->object);
    args->access = ntohl (*p++);
    return p;
}

u_int32_t *
xdr_decode_read (u_int32_t *p, read3args *args)
{
    p = xdr_decode_fh (p, &args->file);
    p = xdr_decode_hyper (p, &args->offset);
    args->count = ntohl (*p++);

    return p;
}

u_int32_t *
xdr_decode_write (u_int32_t *p, write3args *args)
{
    p = xdr_decode_fh (p, &args->file);
    p = xdr_decode_hyper (p, &args->offset);
    args->count = ntohl (*p++);
    args->stable = ntohl (*p++);
    args->data.data_len = ntohl (*p++);
    args->data.data_val = (char *)p;

    return (p + XDR_QUADLEN (args->data.data_len));
}

u_int32_t *
xdr_decode_create (u_int32_t *p, create3args *args)
{
    p = xdr_decode_dirop (p, &args->where);
    args->how.mode = ntohl (*p++);
    if (args->how.mode == GUARDED || args->how.mode == UNCHECKED) {
        p = xdr_decode_sattr (p, &args->how.createhow3_u.obj_attributes);
    } else if (args->how.mode == EXCLUSIVE) {
        memcpy (args->how.createhow3_u.verf, p, NFS3_CREATEVERFSIZE);
        p += 2;
    }
    return p;
}

u_int32_t *
xdr_decode_mkdir (u_int32_t *p, mkdir3args *args)
{
    p = xdr_decode_dirop (p, &args->where);
    p = xdr_decode_sattr (p, &args->attributes);
    return p;
}

u_int32_t *
xdr_decode_symlink (u_int32_t *p, symlink3args *args)
{
    p = xdr_decode_dirop (p, &args->where);
    p = xdr_decode_sattr (p, &args->symlink.symlink_attributes);
    p = xdr_decode_string (p, &args->symlink.symlink_data);
    return p;
}

u_int32_t *
xdr_decode_mknod (u_int32_t *p, mknod3args *args)
{
    p = xdr_decode_dirop (p, &args->where);
    args->what.type = ntohl (*p++);
    if (args->what.type == NF3BLK || args->what.type == NF3CHR) {
        p = xdr_decode_sattr (p, 
                              &args->what.mknoddata3_u.device.dev_attributes);
        args->what.mknoddata3_u.device.spec.major = ntohl (*p++);
        args->what.mknoddata3_u.device.spec.minor = ntohl (*p++);
    } else if (args->what.type == NF3SOCK || args->what.type == NF3FIFO) {
        p = xdr_decode_sattr (p, &args->what.mknoddata3_u.pipe_attributes); 
    }
    return p;
}

u_int32_t *
xdr_decode_rename (u_int32_t *p, rename3args *args)
{
   p = xdr_decode_dirop (p, &args->from);
   p = xdr_decode_dirop (p, &args->to);
   return p;
}

u_int32_t *
xdr_decode_link (u_int32_t *p, link3args *args)
{
    p = xdr_decode_fh (p, &args->file);
    p = xdr_decode_dirop (p, &args->link);
    return p;
}

u_int32_t *
xdr_decode_readdir (u_int32_t *p, readdir3args *args)
{
    p = xdr_decode_fh (p, &args->dir);
    p = xdr_decode_hyper (p, &args->cookie);
    memcpy (args->cookieverf, p, NFS3_COOKIEVERFSIZE);
    p += 2;
    args->count = ntohl (*p++);
    return p;
}


u_int32_t *
xdr_decode_readdirplus (u_int32_t *p, readdirplus3args *args)
{
    p = xdr_decode_fh (p, &args->dir);
    p = xdr_decode_hyper (p, &args->cookie);
    memcpy (args->cookieverf, p, NFS3_COOKIEVERFSIZE);
    p += 2;
    args->dircount = ntohl (*p++);
    args->maxcount = ntohl (*p++);
    return p;
}

u_int32_t *
xdr_decode_commit (u_int32_t *p, commit3args *args)
{
    p = xdr_decode_fh (p, &args->file);
    p = xdr_decode_hyper (p, &args->offset);
    args->count = ntohl (*p++);
    return p;
}


/* Encoding */
static inline u_int32_t *
xdr_encode_nfstime (u_int32_t *p, nfstime3 *time)
{
    *p++ = htonl (time->seconds);
    *p++ = htonl (time->nseconds);
    return p;
}

static inline u_int32_t *
xdr_encode_fattr (u_int32_t *p, fattr3 *fattr)
{
    *p++ = htonl (fattr->type);
    *p++ = htonl (fattr->mode);
    *p++ = htonl (fattr->nlink);
    *p++ = htonl (fattr->uid);
    *p++ = htonl (fattr->gid);
    p = xdr_encode_hyper (p, &fattr->size);
    p = xdr_encode_hyper (p, &fattr->used);
    *p++ = htonl (fattr->rdev.major);
    *p++ = htonl (fattr->rdev.minor);
    p = xdr_encode_hyper (p, &fattr->fsid);
    p = xdr_encode_hyper (p, &fattr->fileid);
    p = xdr_encode_nfstime (p, &fattr->atime);
    p = xdr_encode_nfstime (p, &fattr->mtime);
    p = xdr_encode_nfstime (p, &fattr->ctime);
    return p;
}


static inline u_int32_t *
xdr_encode_postopattr (u_int32_t *p, post_op_attr *pattr)
{
    *p++ = htonl (pattr->present);
    if (pattr->present) {
        p = xdr_encode_fattr (p, &pattr->post_op_attr_u.attributes);
    }
    return p;
}

static inline u_int32_t *
xdr_encode_wccattr (u_int32_t *p, wcc_attr *wattr)
{
    p = xdr_encode_hyper (p, &wattr->size);
    p = xdr_encode_nfstime (p, &wattr->mtime);
    p = xdr_encode_nfstime (p, &wattr->ctime);
    return p;
}

static inline u_int32_t *
xdr_encode_preopattr (u_int32_t *p, pre_op_attr *pattr)
{
    *p++ = htonl (pattr->present);
    if (pattr->present) {
        p = xdr_encode_wccattr (p, &pattr->pre_op_attr_u.attributes);
    }
    return p;
}

static inline u_int32_t *
xdr_encode_wccdata (u_int32_t *p, wcc_data *wdata) 
{
    p = xdr_encode_preopattr (p, &wdata->before);
    p = xdr_encode_postopattr (p, &wdata->after);
    return p;
}

static inline u_int32_t *
xdr_encode_fh (u_int32_t *p, nfs_fh3 *fh)
{
    *p++ = htonl (fh->data.data_len);
    memcpy ((char *)p, fh->data.data_val, fh->data.data_len);
    return p + XDR_QUADLEN (fh->data.data_len);
}

static inline u_int32_t *
xdr_encode_postopfh (u_int32_t *p, post_op_fh3 *pfh)
{
    *p++ = htonl (pfh->present);
    if (pfh->present) {
        p = xdr_encode_fh (p, &pfh->post_op_fh3_u.handle);
    }
    return p;
}

u_int32_t *
xdr_encode_wccstat (u_int32_t *p, wccstat3 *res)
{
    *p++ = htonl (res->status);
    if (res->status == NFS3_OK) {
        p = xdr_encode_wccdata (p, &res->wccstat3_u.wcc);
    }
    return p;
}

u_int32_t *
xdr_encode_dirop (u_int32_t *p, diropres3 *res)
{
    *p++ = htonl (res->status);
    if (res->status == NFS3_OK) {
        p = xdr_encode_postopfh (p, &res->diropres3_u.resok.obj);
        p = xdr_encode_postopattr (p, &res->diropres3_u.resok.obj_attributes);
        p = xdr_encode_wccdata (p, &res->diropres3_u.resok.dir_wcc);
    } else {
        p = xdr_encode_wccdata (p, &res->diropres3_u.resfail);
    }
    return p;
}

u_int32_t *
xdr_encode_getattr (u_int32_t *p, getattr3res *res)
{
    *p++ = htonl (res->status);
    if (res->status == NFS3_OK) {
        p = xdr_encode_fattr (p, &res->getattr3res_u.attributes);
    }
    return p;
}

u_int32_t *
xdr_encode_lookup (u_int32_t *p, lookup3res *res)
{
   *p++ = htonl (res->status);
    if (res->status == NFS3_OK) {
        p = xdr_encode_fh (p, &res->lookup3res_u.resok.object);
        p = xdr_encode_postopattr (p, &res->lookup3res_u.resok.obj_attributes);
        p = xdr_encode_postopattr (p, &res->lookup3res_u.resok.dir_attributes);
    } else {
        p = xdr_encode_postopattr (p, &res->lookup3res_u.resfail);
    }
    return p;
}

u_int32_t *
xdr_encode_access (u_int32_t *p, access3res *res)
{
   *p++ = htonl (res->status);
    if (res->status == NFS3_OK) {
        p = xdr_encode_postopattr (p, &res->access3res_u.resok.obj_attributes);
        *p++ = htonl (res->access3res_u.resok.access);
    }
    return p;
}

u_int32_t *
xdr_encode_readlink (u_int32_t *p, readlink3res *res)
{
    *p++ = htonl (res->status);
    if (res->status == NFS3_OK) {
        p = xdr_encode_postopattr (p, 
                               &res->readlink3res_u.resok.symlink_attributes);
        p = xdr_encode_string (p, res->readlink3res_u.resok.data);
    } else {
        p = xdr_encode_postopattr (p, &res->readlink3res_u.resfail);
    }
    return p;
}

u_int32_t *
xdr_encode_read (u_int32_t *p, read3res *res)
{
    *p++ = htonl (res->status);
    if (res->status == NFS3_OK) {
        p = xdr_encode_postopattr (p, &res->read3res_u.resok.file_attributes);
        *p++ = htonl (res->read3res_u.resok.count);
        *p++ = htonl (res->read3res_u.resok.eof);
        *p++ = htonl (res->read3res_u.resok.data.data_len);
        /* data goes here */
    } else {
        p = xdr_encode_postopattr (p, &res->read3res_u.resfail);        
    }
    return p;
}

u_int32_t *
xdr_encode_write (u_int32_t *p, write3res *res)
{
    *p++ = htonl (res->status);
    if (res->status == NFS3_OK) {
        p = xdr_encode_wccdata (p, &res->write3res_u.resok.file_wcc);
        *p++ = htonl (res->write3res_u.resok.count);
        *p++ = htonl (res->write3res_u.resok.committed);
        memcpy ((char *)p, res->write3res_u.resok.verf, NFS3_WRITEVERFSIZE);
        p += 2;
    } else {
        p = xdr_encode_wccdata (p, &res->write3res_u.resfail);
    }
    return p;
}


u_int32_t *
xdr_encode_rename (u_int32_t *p, rename3res *res)
{
    *p++ = htonl (res->status);
    if (res->status == NFS3_OK) {
        p = xdr_encode_wccdata (p, &res->rename3res_u.res.fromdir_wcc);
        p = xdr_encode_wccdata (p, &res->rename3res_u.res.todir_wcc);
    }
    return p;
}


u_int32_t *
xdr_encode_link (u_int32_t *p, link3res *res)
{
    *p++ = htonl (res->status);
    if (res->status == NFS3_OK) {
        p = xdr_encode_postopattr (p, &res->link3res_u.res.file_attributes);
        p = xdr_encode_wccdata (p, &res->link3res_u.res.linkdir_wcc);
    }
    return p;
}

u_int32_t *
xdr_encode_readdir_end (u_int32_t *p, int eof)
{
    *p++ = 0; /* Marks end of linked list */
    if (eof) {
        *p++ = htonl (1);  /* bool_t eof marker found in dirlist(plus)3 */
    } else {
        *p++ = 0;
    }
    return p;
}

u_int32_t *
xdr_encode_entry (u_int32_t *p, entry3 *entry)
{
    *p++ = htonl (1);
    p = xdr_encode_hyper (p, &entry->fileid);
    p = xdr_encode_string (p, entry->name);
    p = xdr_encode_hyper (p, &entry->cookie);
    return p;
}

u_int32_t *
xdr_encode_readdir (u_int32_t *p, readdir3res *res)
{
    *p++ = htonl (res->status);
    if (res->status == NFS3_OK) {
        p = xdr_encode_postopattr (p, 
                                   &res->readdir3res_u.resok.dir_attributes);
        memcpy (p, res->readdir3res_u.resok.cookieverf, NFS3_COOKIEVERFSIZE);
        p += 2;
        /* Entries go here -- encoded by xdr_encode_entry, 
           followed by EOF flag -- encoded by xdr_encode_readdir_end */
    } else {
        p = xdr_encode_postopattr (p, &res->readdir3res_u.resfail);
    }
    return p;
}


u_int32_t *
xdr_encode_entryplus (u_int32_t *p, entryplus3 *entry)
{
    /* printf ("%s: name=%s, namelen=%d, fileid=%llu\n", __FUNCTION__,
     *       entry->name, (int)strlen (entry->name), entry->fileid); */

    *p++ = htonl (1);
    p = xdr_encode_hyper (p, &entry->fileid);
    p = xdr_encode_string (p, entry->name);
    p = xdr_encode_hyper (p, &entry->cookie);
    p = xdr_encode_postopattr (p, &entry->name_attributes);
    p = xdr_encode_postopfh (p, &entry->name_handle);
    return p;
}

u_int32_t *
xdr_encode_readdirplus (u_int32_t *p, readdirplus3res *res)
{
    *p++ = htonl (res->status);
    if (res->status == NFS3_OK) {
        p = xdr_encode_postopattr (p, 
                                &res->readdirplus3res_u.resok.dir_attributes);
        memcpy (p, res->readdirplus3res_u.resok.cookieverf, 
                NFS3_COOKIEVERFSIZE);
        p += 2; 
        /* Entries go here, followed by EOF flag */
    } else {
        p = xdr_encode_postopattr (p, &res->readdirplus3res_u.resfail);    
    }
    return p;
}

u_int32_t *
xdr_encode_fsstat (u_int32_t *p, fsstat3res *res)
{
    *p++ = htonl (res->status);
    if (res->status == NFS3_OK) {
        p = xdr_encode_postopattr (p, &res->fsstat3res_u.resok.obj_attributes);
        p = xdr_encode_hyper (p, &res->fsstat3res_u.resok.tbytes);
        p = xdr_encode_hyper (p, &res->fsstat3res_u.resok.fbytes);
        p = xdr_encode_hyper (p, &res->fsstat3res_u.resok.abytes);
        p = xdr_encode_hyper (p, &res->fsstat3res_u.resok.tfiles);
        p = xdr_encode_hyper (p, &res->fsstat3res_u.resok.ffiles);
        p = xdr_encode_hyper (p, &res->fsstat3res_u.resok.afiles);
        *p++ = htonl (res->fsstat3res_u.resok.invarsec);
    } else {
        p = xdr_encode_postopattr (p, &res->fsstat3res_u.resfail);
    }
    return p;
}

u_int32_t *
xdr_encode_fsinfo (u_int32_t *p, fsinfo3res *res)
{
    *p++ = htonl (res->status);
    if (res->status == NFS3_OK) {
        p = xdr_encode_postopattr (p, &res->fsinfo3res_u.resok.obj_attributes);
        *p++ = htonl (res->fsinfo3res_u.resok.rtmax);
        *p++ = htonl (res->fsinfo3res_u.resok.rtpref);
        *p++ = htonl (res->fsinfo3res_u.resok.rtmult);
        *p++ = htonl (res->fsinfo3res_u.resok.wtmax);
        *p++ = htonl (res->fsinfo3res_u.resok.wtpref);
        *p++ = htonl (res->fsinfo3res_u.resok.wtmult);
        *p++ = htonl (res->fsinfo3res_u.resok.dtpref);
        p = xdr_encode_hyper (p, &res->fsinfo3res_u.resok.maxfilesize);
        p = xdr_encode_nfstime (p, &res->fsinfo3res_u.resok.time_delta);
        *p++ = htonl (res->fsinfo3res_u.resok.properties);
    } else {
        p = xdr_encode_postopattr (p, &res->fsinfo3res_u.resfail);
    }
    return p;
}

u_int32_t *
xdr_encode_pathconf (u_int32_t *p, pathconf3res *res)
{
    *p++ = htonl (res->status);
    if (res->status == NFS3_OK) {
        p = xdr_encode_postopattr (p, 
                                   &res->pathconf3res_u.resok.obj_attributes);
        *p++ = htonl (res->pathconf3res_u.resok.linkmax);
        *p++ = htonl (res->pathconf3res_u.resok.name_max);
        *p++ = htonl (res->pathconf3res_u.resok.no_trunc);
        *p++ = htonl (res->pathconf3res_u.resok.chown_restricted);
        *p++ = htonl (res->pathconf3res_u.resok.case_insensitive);
        *p++ = htonl (res->pathconf3res_u.resok.case_preserving);
    } else {
        p = xdr_encode_postopattr (p, &res->pathconf3res_u.resfail);
    }
    return p;
}

u_int32_t *
xdr_encode_commit (u_int32_t *p, commit3res *res)
{
    *p++ = htonl (res->status);
    if (res->status == NFS3_OK) {
        p = xdr_encode_wccdata (p, &res->commit3res_u.resok.file_wcc);
        memcpy (p, res->commit3res_u.resok.verf, NFS3_WRITEVERFSIZE);
        p += 2;
    } else {
        p = xdr_encode_wccdata (p, &res->commit3res_u.resfail);
    }
    return p;
}

/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */

