/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"
#include "pasis_lists.h"
#include "pasis_types.h"

#include "pasis_md_prot.h"
#include "pasis_md_clnt.h"
#include "pasis_md_hist_matrix.h"

#include "rpc_common.h"

#include "pasisio_drives.h"
#include "pasisio_drive_md.h"
#include "pasisio_md_protocol.h"

#ifndef MIN
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#endif

#ifndef MAX
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#endif


#define TIMER_RPC_READ_OP       0
#define TIMER_RPC_WRITE_OP      1

#define TIMER_NUM               2
static int md_instances = 0;


static int
md_clnt_read_op (pasisio_drive_req_t     *req,
                 pasisio_drive_md_t      *priv,
                 rpc_auth_t              *auth,
                 pasisio_md_proto_data_t *proto_data)
{
    pasis_md_read_op_args_t  args;
    pasis_md_read_op_res_t  *res;
    pasis_error_t            error = PASIS_SUCCESS;
    int rc, i;

    memcpy (&args, &proto_data->read_op_args, 
            sizeof (pasis_md_read_op_args_t));

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "%s: part=%u, op_type=%u\n", 
                   __FUNCTION__,  priv->partition, args.op_type);

    args.part = priv->partition;

    res = &proto_data->read_op_res[req->share_index];

    /* Even for queries, we still send an OHS sometimes.  This happens when a
     * repair occurs, and the client needs to send an OHS to convince a server
     * that it should object-sync the repairable version if necessary.
     * -mabdelmalek 19/3/2005 */
    /* Set the authenticators for this storage-node */
    for (i = 0; i < args.history_matrices_num; i++) {
        int node_id = req->share_index;
        PasisIO_MDHistCacheSetAuths (proto_data->auth_to_send[i],
                                     proto_data->auth_to_send_num[i],
                                     &args.history_matrices[i], 
                                     node_id);
    }

    rc = pasis_md_clnt_read_op (priv->md_client, auth, &args, res);
    if (rc) {
        if (rc == RPC_STATUS_TERM) {
            PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, 
                           "md_clnt_read: early termination\n");
            error = MD_RPC_TERM;
        } else {
            PASIS_PRINT_ERR ("Error: md_clnt_read failed (%d,%d)\n", 
                             rc, res->status);
            error = MD_SERVER_ERROR;
        }
    }
    /* MD_RPC_ERROR never used now */

    /*
    if (args.op_type == PASIS_MD_READ_EC) {
        printf ("%s: error=%d\n - data (%p)=%#llx\n - checksums(%p)=%llx\n", 
                __FUNCTION__, error, req->share_list, 
                *(p_uint64 *)req->share_list, req->checksums, 
                *(p_uint64 *)req->checksums);
    }
    */
    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, 
                   "%s: op_type=%u, flags=%#x, error=%u\n", __FUNCTION__, 
                   args.op_type, args.read_flags, error);
    
    if (DO_NOT_UPDATE_HISTORY) {
        /* printf("%s: NOT updating history information\n", __FUNCTION__); */
    } else {
        /* Update history cache */
        if (error == PASIS_SUCCESS) {
            PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, 
                           "%s: updating history cache (proto_index=%u, id=%u)\n",
                           __FUNCTION__, req->share_index,
                           req->req_hdr->parent->id);
            for (i = 0; i < res->history_vec_num; i++) {
                PasisIO_MDHistCacheUpdate (&res->history_vectors[i],
                                           proto_data->fid_hist_cache[i]);
            }
        }
    }
    
    return error;
}


static int
md_clnt_write_op (pasisio_drive_req_t     *req,
                  pasisio_drive_md_t      *priv,
                  rpc_auth_t              *auth,
                  pasisio_md_proto_data_t *proto_data)
{
    pasis_md_write_op_args_t  args;
    pasis_md_write_op_res_t  *res;
    pasis_error_t             error = PASIS_SUCCESS;
    int rc, i;

    /* Copy global args into local copy */
    memcpy (&args, &proto_data->write_op_args, 
            sizeof (pasis_md_write_op_args_t));;

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "%s: part=%u, op_type=%u\n",
                   __FUNCTION__, priv->partition, args.op_type);
    
    args.part = priv->partition;

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, 
                   "%s: shuffle=%u  share_index=%u  ip=%u (node=%u)\n", 
                   __FUNCTION__, proto_data->shuffle, 
                   req->share_index, priv->ip_address, 
                   (proto_data->protocol->N - proto_data->shuffle + 
                    req->share_index) % proto_data->protocol->N);
            
    /* Set the authenticators for this storage-node */
    for (i = 0; i < args.history_matrices_num; i++) {
        int node_id = req->share_index;
        PasisIO_MDHistCacheSetAuths (proto_data->auth_to_send[i],
                                     proto_data->auth_to_send_num[i],
                                     &args.history_matrices[i], 
                                     node_id);
    }

    /* Set results */
    res = &proto_data->write_op_res[req->share_index];

    rc = pasis_md_clnt_write_op (priv->md_client, auth, &args, res);
    
    if (rc) {
        if (rc == RPC_STATUS_TERM) {
            PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, 
                           "md_clnt_write: early termination\n");
            error = MD_RPC_TERM;
        } else {
            PASIS_PRINT_ERR ("Error: md_clnt_write failed (%d,%d)\n", 
                             rc, res->status);
            error =  MD_SERVER_ERROR;
        }
    }
    /* MD_RPC_ERROR never used now */
    
    /* XXX: proto structure is freed after early termination, for now we cannot
     * rely on the results being accurate */

    if (DO_NOT_UPDATE_HISTORY) {
        /* printf("%s: NOT updating history information\n", __FUNCTION__); */
    } else {
        /* Update history cache */
        if (error == PASIS_SUCCESS) {
            PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, 
                           "%s: updating history cache (proto_index=%u)\n",
                           __FUNCTION__, req->share_index);

            for (i = 0; i < res->history_vec_num; i++) {
                PasisIO_MDHistCacheUpdate (&res->history_vectors[i],
                                           proto_data->fid_hist_cache[i]);
            }
        }
    }

    return error;
}


static pasis_error_t
md_req_handler (pasisio_drive_instance_t *instance,
                pasisio_drive_req_t *req)
{
    pasisio_drive_md_t      *priv = (pasisio_drive_md_t *)instance->priv;
    pasisio_md_proto_data_t *proto_data;
    rpc_auth_t               auth;
    int                      rc;

    auth.uid = req->req_hdr->parent->uid;
    auth.gid = MD_DEFAULT_GID;

    proto_data = req->req_hdr->parent->proto_data;

    priv->stats.in_progress++;
    priv->stats.max_in_progress = MAX (priv->stats.in_progress, 
                                       priv->stats.max_in_progress);

    switch (req->op_type) {

    /* Protocol specific op codes */
    
    case PASISIO_MD_DRIVE_P_WRITE_OP: {
        PASIS_TIMER_START ();

        PASIS_DASSERT (proto_data != NULL);
        rc = md_clnt_write_op (req, priv, &auth, proto_data);
                            
        PASIS_TIMER_STOP (priv->stats.timers, TIMER_RPC_WRITE_OP);
        break;
    }
        
    case PASISIO_MD_DRIVE_P_READ_OP: {
        PASIS_TIMER_START ();
        
        PASIS_DASSERT (proto_data != NULL);
        
        rc = md_clnt_read_op (req, priv, &auth, proto_data);

        PASIS_TIMER_STOP (priv->stats.timers, TIMER_RPC_READ_OP);
        break;
    }

    default:
        PASIS_PRINT_ERR ("Error: %s, op=%#x not supported\n", 
                         __FUNCTION__, req->op_type);
        return PASISIO_DRIVE_REQ_OPERERR;
    }

    priv->stats.in_progress--;

    /* Check rc... */
    switch (rc) {
    case MD_RPC_ERROR:            /* Operation error */
        PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4,
                       "Error: MD op %d to %s failed (RPC_ERROR)\n", 
                       req->op_type, pasis_ip_to_hostname (priv->ip_address));
        return PASISIO_DRIVE_REQ_OPERERR;
        
    case MD_SERVER_ERROR:         /* Server error */
        PASIS_PRINT_ERR ("Error: MD op %d to %s failed (SERVER_ERROR)\n",
                         req->op_type, pasis_ip_to_hostname (priv->ip_address));
        instance->status = PASISIO_DRIVE_STATUS_DOWN;
        return PASISIO_DRIVE_REQ_SERVERR;  

    case MD_RPC_TERM:
        return PASISIO_DRIVE_REQ_TERM;

    }

    return PASISIO_DRIVE_REQ_SUCCESS;
}


static void *
md_reconnect_thread (void *arg)
{
    pasisio_drive_instance_t *instance = (pasisio_drive_instance_t *)arg;
    pasisio_drive_md_t *priv = instance->priv;

    /* Sleep some random amount of time to begin with */
    st_cond_timedwait (priv->shutdown_cond, 1000000LL * 
                       ((rand () % PASISIO_DRIVE_ALIVE_TIMEOUT) + 5));

    while (priv->shutdown_flag == 0) {

        /* If instance was down, try to reconnect */
        if (instance->status == PASISIO_DRIVE_STATUS_DOWN) {

            //PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, 
            PASIS_PRINT_INFO (
                           "%s: attempting reconnect to %s\n", 
                           __FUNCTION__, pasis_ip_to_hostname (priv->ip_address));

            if (rpc_client_reconnect (priv->md_client) == 0) {
                instance->status = PASISIO_DRIVE_STATUS_UP;
                //PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, 
                PASIS_PRINT_INFO (
                                  "%s: setting instance status up %s\n", 
                                  __FUNCTION__, pasis_ip_to_hostname (priv->ip_address));
            }
        }

        /* Sleep */
        st_cond_timedwait (priv->shutdown_cond, 1000000LL * 
                           PASISIO_DRIVE_ALIVE_TIMEOUT);
    }

    return NULL;
}


static void
md_disconnect_handler (void *arg)
{
    pasisio_drive_instance_t *drive_instance = (pasisio_drive_instance_t *)arg;

    //PASIS_PRINT_INFO (
    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, 
                      "%s: setting instance status down %s\n",
                      __FUNCTION__, 
                      pasis_ip_to_hostname(((pasisio_drive_md_t *)(drive_instance->priv))
                               ->ip_address));

    drive_instance->status = PASISIO_DRIVE_STATUS_DOWN;
    
    return;
}


static pasis_error_t
md_create_instance (p_uint32 remote_ip, char *path,
                    pasisio_drive_instance_t** instance)
{
    pasisio_drive_type_t     *drive_type;
    pasisio_drive_instance_t *drive_instance;

    pasisio_drive_md_t       *priv;
    unsigned int              partition;

    char                      timer_name[255];

    if (path == NULL || strcmp (path, "/") == 0) {
        PASIS_PRINT_ERR ("%s: no partition id specified\n", __FUNCTION__);
        return PASIS_ERROR;
    }

    sscanf (path, "/%u", &partition);

    if (partition > PASIS_MD_MAX_PARTITION || 
        partition < PASIS_MD_MIN_PARTITION) {
        PASIS_PRINT_ERR ("%s: partition id out of range (%d)\n", __FUNCTION__,
                         partition);
        return PASIS_ERROR;
    }

    drive_type = PasisIO_DriveFindTypeIndex (PASISIO_DRIVE_TYPE_MD_INDEX);
    if (drive_type == NULL) {
        PASIS_PRINT_ERR ("Error: MD drive type never created\n");
        return PASIS_ERROR;
    }

    /* Setup private drive instance structure */
    priv = (pasisio_drive_md_t *) malloc (sizeof (pasisio_drive_md_t));

    priv->partition = partition;
    priv->ip_address = remote_ip;

    drive_instance = PasisIO_DriveCreateInstance (drive_type, priv->ip_address,
                                                  path, MD_WORKER_THREADS);
    if (drive_instance == NULL) {
        PASIS_PRINT_ERR ("Error: drive instance %s exists\n", 
                         pasis_ip_to_hostname(remote_ip));
        return PASIS_ERROR;
    }

    drive_instance->priv = priv;

    /* Set up RPC client mount handle */
    priv->md_client = rpc_client_create_tcp (remote_ip,
                                             MD_WORKER_THREADS, 
                                             PASIS_MD_RPC_PROGNUM, 
                                             PASIS_MD_RPC_VERSION,
                                             drive_instance,
                                             md_disconnect_handler);
    
    if (priv->md_client == NULL) {
        PASIS_PRINT_ERR ("Error: md rpc_client_create failed (%s:/%d)\n",
                         pasis_ip_to_hostname(priv->ip_address), priv->partition);

        drive_instance->init_complete = 2; /* signals init didn't succeed */
        PasisIO_DriveDestroyInstance (drive_instance);
        free (priv);
        drive_instance->priv = NULL;

        return PASIS_ERROR;
    }

    /* Setup stats */
    memset (&priv->stats, 0, sizeof (pasisio_drive_md_stats_t));
    
    snprintf (timer_name, 255, "Drive MD: %s:%s", 
              pasis_ip_to_hostname (priv->ip_address), path);
    priv->stats.timers = Pasis_TimerInit (timer_name, TIMER_NUM, 0);
    Pasis_TimerRegister (priv->stats.timers, TIMER_RPC_WRITE_OP, 
                         "rpc_write_op"); 
    Pasis_TimerRegister (priv->stats.timers, TIMER_RPC_READ_OP, 
                         "rpc_read_op"); 

    /* Setup shutdown code */
    priv->shutdown_flag = 0;
    priv->shutdown_cond = st_cond_new ();
    PASIS_DASSERT (priv->shutdown_cond != NULL);

    /* Create reconnect thread */
    priv->reconnect_thread = st_thread_create (md_reconnect_thread,
                                               drive_instance, 1, 0);
    md_instances++;

    drive_instance->status = PASISIO_DRIVE_STATUS_UP;

    if (instance) {
        *instance = drive_instance;
    }

    drive_instance->init_complete = 1;
    st_cond_broadcast (drive_instance->init_cond);

    return PASIS_SUCCESS;
}


static void
md_destroy_instance (pasisio_drive_instance_t *instance)
{
    pasisio_drive_md_t *priv;

    priv = (pasisio_drive_md_t *)instance->priv;
    
    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "%s: shutting down\n", __FUNCTION__);
        
    while (!instance->init_complete) {
        st_cond_wait (instance->init_cond);
    }

    priv->shutdown_flag = 1;

    if (instance->init_complete != 2) { /* =2 if failure during init */
        st_cond_signal (priv->shutdown_cond);
        st_thread_join (priv->reconnect_thread, NULL);

        PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "%s: joined reconnect thread\n",
                   __FUNCTION__);

        st_cond_destroy (priv->shutdown_cond);

        PASIS_PRINT_INFO ("\nStatistics:  PasisIO_Drive_MD\n");
        PASIS_PRINT_INFO ("-----------------------------\n");
   
        Pasis_TimerShutdown (priv->stats.timers);

        PASIS_PRINT_INFO ("--- Other Stats ---\n");
        PASIS_PRINT_INFO ("Max In Progress    :  %u\n", 
                          priv->stats.max_in_progress);
        PASIS_PRINT_INFO ("-----------------------------\n\n");

        rpc_signal_shutdown (priv->md_client);
        rpc_client_destroy (priv->md_client);

        free (priv);
    }

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "%s: exiting\n", __FUNCTION__);

    md_instances--;
}


static void
md_req_terminate (pasisio_drive_instance_t *instance,
                  st_thread_t thread_id)
{
    pasisio_drive_md_t *priv;

    priv = (pasisio_drive_md_t *)instance->priv;

    rpc_early_terminate (priv->md_client, thread_id);
}


/* Init MD drive */
pasis_error_t
PasisIO_DriveInitMD (void)
{
    pasisio_drive_type_t *drive_type;

    /* Create or setup drive type */
    drive_type = PasisIO_DriveCreateType (PASISIO_DRIVE_TYPE_MD_INDEX, 
                                          PASISIO_DRIVE_TYPE_MD_NAME,
                                          md_req_handler,
                                          md_create_instance,
                                          md_destroy_instance,
                                          md_req_terminate);

    return PASIS_SUCCESS;
}


/* Shutdown MD drive */
void
PasisIO_DriveShutdownMD (void)
{
    pasisio_drive_type_t *drive_type;

    if (md_instances) {
        PASIS_PRINT_ERR ("%s: Error %d instances alive\n", 
                         __FUNCTION__, md_instances);
    }

    drive_type = PasisIO_DriveFindTypeIndex (PASISIO_DRIVE_TYPE_MD_INDEX);
    PasisIO_DriveDestroyType (drive_type);
}


/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */ 
