/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"
#include "pasis_types.h"
#include "pasis_error.h"
#include "pasis_lists.h"
#include "pasis_freelist.h"
#include "pasis_timer.h"

#include "pasisio.h"
#include "pasisio_internal.h"
#include "pasisio_drives.h"
#include "pasisio_select.h"

#include "pasisio_drive_md.h"

#include <netdb.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define DRIVE_TIMER_REQUEST    0
#define DRIVE_TIMER_OPERATION  1

#define DRIVE_TIMER_NUM        2

#define RANDOMIZE_REQUEST_ISSUES 0

typedef struct drive_stats_s {
    pasis_timer_state_t *timers;
    unsigned int         operations;
    unsigned int         enqueue_queued;
    unsigned int         enqueues;
    unsigned int         enqueue_blocked;
    unsigned int         threads_in_prog;
    unsigned int         reqs_term;
    unsigned int         reqs_term_failed;
} drive_stats_t;

static drive_stats_t              stats;

static pasis_freelist_t          *drive_req_freelist;
static pasis_freelist_t          *drive_req_hdr_freelist;

static pasisio_drive_type_t      *drive_type_array[PASISIO_DRIVE_TYPE_NUM];

static void * PasisIO_DriveThreadDispatch (void *arg);
void driveInitReqHdr (void *arg);
void driveDestReqHdr (void *arg);

void
driveInitReqHdr (void *arg)
{
    pasisio_drive_req_hdr_t *hdr = (pasisio_drive_req_hdr_t *)arg;
    hdr->cond = st_cond_new ();
}


void
driveDestReqHdr (void *arg)
{
    pasisio_drive_req_hdr_t *hdr = (pasisio_drive_req_hdr_t *)arg;
    st_cond_destroy (hdr->cond);
}


static inline pasisio_drive_req_t *
PasisIO_DriveQueueDequeue (pasisio_drive_instance_t *instance)
{
    pasisio_drive_req_t *req;

    req = PASIS_Q_FIRST (instance->dispatch_req_queue);
    if (req == NULL) {
        return NULL;
    }

    instance->dispatch_count--;
    PASIS_DASSERT (instance->dispatch_count >= 0);
    PASIS_Q_REMOVE (instance->dispatch_req_queue, queue, req);

    return req;
}


static inline void
PasisIO_DriveQueueEnqueue (pasisio_drive_instance_t *instance,
                           pasisio_drive_req_t *req)
{
    pasisio_drive_thread_t *worker;

    stats.enqueue_queued += instance->dispatch_count;
    stats.threads_in_prog += instance->threads_in_prog;

    PASIS_Q_ADD_TAIL (instance->dispatch_req_queue, queue, req);
    instance->dispatch_count++;

    instance->reqs_queued++;
    
    worker = PASIS_Q_FIRST (instance->thread_queue);
    if (worker != NULL) {
        PASIS_Q_REMOVE (instance->thread_queue, queue, worker);
        worker->signal = 1;
        st_cond_signal (worker->cond);
    } else {
        stats.enqueue_blocked++;

        PASIS_DPRINTF (PASIS_DEBUG_IO_DRIVE, 4, 
                       "DriveQueueEnqueue: no workers (queued=%u)\n",
                       instance->dispatch_count);
    }

    return;
}


static void
early_terminate_outstanding (pasisio_drive_req_hdr_t *req_hdr)
{
    pasisio_select_t   *select;
    int                 i;

    select = req_hdr->parent->drive_select;

    for (i = 0; i < select->count; i++) {
        if (select->flags[i] == PASISIO_SELECT_IO_PENDING) {

            req_hdr->req_state[i].early_terminate = 1;

            PASIS_DPRINTF (PASIS_DEBUG_IO_DRIVE, 4, 
                           "%s: req_hdr=%p, share=%d\n", 
                           __FUNCTION__, req_hdr, i);

            /* Call early terminate instance function */
            if (req_hdr->req_state[i].thread_id != NULL &&
                select->instances[i]->drive->req_terminate) {
                select->instances[i]->drive->req_terminate (
                        select->instances[i], req_hdr->req_state[i].thread_id);
                                                          
            }
        }
    }
}


static void
dispatch_do_work (pasisio_drive_instance_t *instance)
{
    pasisio_drive_req_hdr_t  *req_hdr;
    pasisio_drive_req_t      *req;
    pasisio_select_t         *select;
    pasisio_drive_type_t     *drive;

    PASIS_TIMER_START ();

    PASIS_DPRINTF (PASIS_DEBUG_IO_DRIVE, 4, 
                   "DriveThreadDispatch: found work...\n");
    
    drive = instance->drive;

    /* Dequeue and dispatch to worker thread */
    req = PasisIO_DriveQueueDequeue (instance);
    PASIS_DASSERT (req != NULL);

    req_hdr = req->req_hdr;
    select = req_hdr->parent->drive_select;
    
    /* Check for early termination */
    if (req_hdr->req_state[req->share_index].early_terminate) {
        stats.reqs_term_failed++;
    }

    req_hdr->req_state[req->share_index].thread_id = st_thread_self ();

    /* Do work... */
    req->status = drive->req_handler (instance, req);
    
    /* Set status */
    switch (req->status) {

    case PASISIO_DRIVE_REQ_SUCCESS:
        PASIS_DPRINTF (PASIS_DEBUG_IO_DRIVE, 4, "completion success\n");
        
        select->flags[req->share_index] = PASISIO_SELECT_IO_SUCCESS;
        req->req_hdr->successful++;
        break;
        
    case PASISIO_DRIVE_REQ_SERVERR:
        PASIS_DPRINTF (PASIS_DEBUG_IO_DRIVE, 4, 
                       "completion failure: server failure\n");
        
        select->flags[req->share_index] = PASISIO_SELECT_IO_SERVERR;
        break;
        
    case PASISIO_DRIVE_REQ_OPERERR:
        PASIS_DPRINTF (PASIS_DEBUG_IO_DRIVE, 4, 
                       "completion failure: operation failure\n");
        
        select->flags[req->share_index] = PASISIO_SELECT_IO_OPERR;
        break;
        
    case PASISIO_DRIVE_REQ_TERM:
        PASIS_DPRINTF (PASIS_DEBUG_IO_DRIVE, 4, 
                       "completion: request terminated early\n");
        select->flags[req->share_index] = PASISIO_SELECT_IO_TERM;
        stats.reqs_term++;
        break;
    }
    
    req_hdr->completed++;

    PASIS_DPRINTF (PASIS_DEBUG_IO_DRIVE, 4, 
                   "%s: complete %d/%d req_hdr=%p, share=%d\n", 
                   __FUNCTION__, req_hdr->completed, req_hdr->issued, req_hdr, 
                   req->share_index);
    
    /* Wait for all shares to complete */
    if (req_hdr->completed == req_hdr->issued) {
        
        PASIS_DPRINTF (PASIS_DEBUG_IO_DRIVE, 4, "completed drive req\n");
        
        req_hdr->signal = 1;
        st_cond_signal (req_hdr->cond);
    } else if (req_hdr->successful == req_hdr->early_term_threshold) {
        PASIS_DPRINTF (PASIS_DEBUG_IO_DRIVE, 4, 
                       "early termination reached, squashing outstanding\n");
        early_terminate_outstanding (req_hdr);
    }
    
    PASIS_TIMER_STOP (stats.timers, DRIVE_TIMER_REQUEST);

    return;
}

#if PASIS_DEBUG > 0
static void
check_queue (pasisio_drive_instance_t *instance,
             pasisio_drive_thread_t   *worker)
{
    pasisio_drive_thread_t *tmp;
    tmp = PASIS_Q_FIRST (instance->thread_queue);
    while (tmp) {
        if (tmp == worker) return;
        tmp = PASIS_Q_NEXT (queue, tmp);
    }

    PASIS_DASSERT (0);
}
#endif

void *
PasisIO_DriveThreadDispatch (void *arg)
{
    pasisio_drive_instance_t *instance;
    pasisio_drive_thread_t   *worker;
    unsigned int reqs_serviced=0;

    worker = (pasisio_drive_thread_t *)arg;
    instance = worker->instance;

    while (1) {

        /* Wait for work */
        while (instance->dispatch_count == 0 && !instance->dispatch_shutdown) {
#if PASIS_DEBUG > 0
            check_queue (instance, worker);
#endif
            st_cond_wait (worker->cond);

            /* If this happens add worker back to queue, 
               someone else serviced this request */
            if (instance->dispatch_count == 0 && worker->signal) {
                PASIS_Q_ADD_TAIL (instance->thread_queue, queue, worker);
            } else {
                instance->threads_in_prog++;
            }

            worker->signal = 0;
        }

        if (instance->dispatch_shutdown) {
            break;
        }

        reqs_serviced++;
        dispatch_do_work (instance);

        /* Relies on the fact that we go around the loop & immediately wait */
        if (instance->dispatch_count == 0 && !instance->dispatch_shutdown) {
            /* Add worker back to work queue */
            PASIS_Q_ADD_TAIL (instance->thread_queue, queue, worker);
            instance->threads_in_prog--;
        }
    }

    PASIS_DPRINTF (PASIS_DEBUG_IO_DRIVE, 4, 
                   "worker done: %p:%d, serviced=%u\n", 
                   instance, worker->id, reqs_serviced);

    return NULL;
}


/* Returns number succeeded */
int
PasisIO_DriveEnqueueShares (pasisio_drive_op_t drive_op,
                            pasisio_request_t *p_req,
                            unsigned int to_issue,
                            unsigned int early_term_threshold)
{
    pasisio_drive_req_t *req[PASIS_MAX_NUM_SERVERS];
    pasisio_drive_req_hdr_t *req_hdr;
    int i, count=0;
#if RANDOMIZE_REQUEST_ISSUES > 0
    int index;
#endif

    PASIS_TIMER_START ();

    stats.operations++;

    req_hdr = PasisFreelist_Get (drive_req_hdr_freelist);
    PASIS_DASSERT (req_hdr != NULL);

    memset (req_hdr->req_state, 0, sizeof (pasisio_drive_req_state_t) *
            PASIS_MAX_NUM_SERVERS);
    req_hdr->successful = 0;
    req_hdr->completed = 0;
    req_hdr->issued = to_issue;
    req_hdr->signal = 0;
    req_hdr->parent = p_req;
    req_hdr->early_term_threshold = early_term_threshold;

    for (i = 0; i < p_req->drive_select->count; i++) {

        /* Only first share block contains valid flag info */
        if (p_req->drive_select->flags[i] == PASISIO_SELECT_IO_ISSUE) {

            /* Get request -- one per backend storage node. */

            req[i] = PasisFreelist_Get (drive_req_freelist);
            PASIS_DASSERT (req[i] != NULL);

            p_req->drive_select->flags[i] = PASISIO_SELECT_IO_PENDING;

            /* Setup drive request structure */
            req[i]->op_type = drive_op;
            
            /* Setup request's share block */
            req[i]->share_index = i;

            req[i]->block_num = p_req->block_num;

            req[i]->req_hdr = req_hdr;

            stats.enqueues++;

#if RANDOMIZE_REQUEST_ISSUES == 0
            PasisIO_DriveQueueEnqueue (p_req->drive_select->instances[i], 
                                       req[i]);
#endif
        } else {
            req[i] = NULL;
        }
    }

#if RANDOMIZE_REQUEST_ISSUES > 0
    /* Randomize the order in which we send out requests.  This reduces the
     * chance that some servers are more heavily loaded than others.  */
    index = rand () % p_req->drive_select->count;
    for (i = 0; i < p_req->drive_select->count; i++) {
        if (p_req->drive_select->flags[index] == PASISIO_SELECT_IO_PENDING) {
            PasisIO_DriveQueueEnqueue (p_req->drive_select->instances[index],
                                       req[index]);
        }
        index = (index + 1) % p_req->drive_select->count;
    }
#endif


    /* Wait for completion -- cleanup of req bufs handled in dispatch thread */
    while (req_hdr->signal == 0) {
        PASIS_DPRINTF (PASIS_DEBUG_IO_DRIVE, 4, 
                       "waiting for io request (%d)\n", req_hdr->signal);

        st_cond_wait (req_hdr->cond);
    }

    PASIS_DPRINTF (PASIS_DEBUG_IO_DRIVE, 4, "done with io...\n");

    /* Cleanup */
    for (i = 0; i < p_req->drive_select->count; i++) {
        if (req[i] != NULL) {
            req[i]->req_hdr = NULL;
            req[i]->op_type = PASISIO_DRIVE_OP_INVALID;
            PasisFreelist_Put (drive_req_freelist, req[i]);
        }
    }

    count = req_hdr->successful;

    PasisFreelist_Put (drive_req_hdr_freelist, req_hdr);

    PASIS_TIMER_STOP (stats.timers, DRIVE_TIMER_OPERATION + drive_op - 1);

    return count;
}


pasisio_drive_instance_t *
PasisIO_DriveAddInstance (unsigned int  type_index,
                          p_uint32      ip_address,
                          char         *base_path)
{
    pasisio_drive_type_t *type;
    pasisio_drive_instance_t *instance;
    pasis_error_t error;

    /* Get the drive type struct */
    type = PasisIO_DriveFindTypeIndex (type_index);
    if (type == NULL) {
        PASIS_PRINT_ERR ("%s: error: type not found (%d)\n",
                         __FUNCTION__, type_index);
        return NULL;
    }

    /* Try creating an instance */
    error = type->create_instance (ip_address, base_path, &instance);
    if (error) {
        PASIS_PRINT_ERR ("%s: error: create instance failed (%s)\n",
                         __FUNCTION__, pasis_ip_to_hostname (ip_address));
        return NULL;
    }

    return instance;
}


pasisio_drive_instance_t *
PasisIO_DriveFindInstanceIP (unsigned int  type_index,
                             p_uint32      ip_address,
                             char         *base_path)
{
    int i, j, starting_type_index, ending_type_index;

    if (type_index == PASISIO_DRIVE_TYPE_ANY_INDEX) {
        starting_type_index = 0;
        ending_type_index   = PASISIO_DRIVE_TYPE_NUM - 1;
    } else {
        starting_type_index = type_index;
        ending_type_index   = type_index;
    }

    for (i = starting_type_index; i <= ending_type_index; ++i) {
        pasisio_drive_type_t *type = drive_type_array[i];
        pasisio_drive_instance_t **instances = type->select_instances->instances;
        for (j = type->select_instances->count - 1; j >= 0; j--) {
            if (instances[j]->ip_address == ip_address &&
                strcmp (instances[j]->base_path, base_path) == 0) {
                /* If we're in the middle of initializing the connection to this
                   * drive, wait until initialization is finished */
                while (!instances[j]->init_complete) {
                    st_cond_wait (instances[j]->init_cond);
                }
                return instances[j];
            }
        }
    }

    /* Drive not found; try making a new entry for it */
    return PasisIO_DriveAddInstance (type_index, ip_address, base_path);
}


pasisio_drive_instance_t *
PasisIO_DriveFindInstanceName (unsigned int type_index,
                               char *hostname,
                               char *base_path)
{
    struct hostent *hp;

    hp = gethostbyname (hostname);
    
    if (hp == NULL) return NULL;
    PASIS_DASSERT (hp->h_length == 4);

    return PasisIO_DriveFindInstanceIP (type_index,
                                        *(p_uint32 *)hp->h_addr,
                                        base_path);
}


pasisio_drive_type_t *
PasisIO_DriveFindTypeName (char *type_name)
{
    int i;

    for (i = 0; i < PASISIO_DRIVE_TYPE_NUM; i++) {
        if (drive_type_array[i] != NULL && 
            strcmp (drive_type_array[i]->type_name, type_name) == 0) {
            return drive_type_array[i];
        }
    }

    return NULL;
}


pasisio_drive_type_t *
PasisIO_DriveFindTypeIndex (unsigned int type_index)
{
    return drive_type_array[type_index];
}


pasisio_select_t *
PasisIO_DriveGetSelectInstances (unsigned int type_index)
{
    if (type_index >= PASISIO_DRIVE_TYPE_NUM) {
        /* Before, you used to be able to pass in the
         * PASISIO_DRIVE_TYPE_ANY_INDEX constant so you can iterate over all
         * instances.  But now, after we've moved to a more dynamic structure
         * for storing the list of workers, we don't have a single array that
         * holds all instances, and hence it's not possible to get a single list
         * of all instances. -mabdelmalek */
        PASIS_PRINT_ERR ("%s: error:  invalid type_index (%u)\n",
                         __FUNCTION__, type_index);
        return NULL;
    }

    return drive_type_array[type_index]->select_instances;
}


void
PasisIO_DriveDestroyType (pasisio_drive_type_t *type)
{
    free (type->select_instances->instances);
    free (type->select_instances);
    free (type);
}


pasisio_drive_type_t *
PasisIO_DriveCreateType (unsigned int type_index,
                         char *type_name,
                         pasis_error_t (*req_handler) 
                         (pasisio_drive_instance_t *, pasisio_drive_req_t *),
                         pasis_error_t (*create_instance)(p_uint32, char *,
                                                          pasisio_drive_instance_t **),
                         void (*destroy_instance) (pasisio_drive_instance_t *),
                         void (*req_terminate) (pasisio_drive_instance_t *,
                                                st_thread_t))
{
    pasisio_drive_type_t *type;

    if (drive_type_array[type_index] != NULL) {
        return NULL;
    }

    type = (pasisio_drive_type_t *) malloc (sizeof (pasisio_drive_type_t));
    memset (type, 0, sizeof (pasisio_drive_type_t));

    type->type_name = type_name;
    type->type_index = type_index;
    type->req_handler = req_handler;
    type->create_instance = create_instance;
    type->destroy_instance = destroy_instance;
    type->req_terminate = req_terminate;
    type->select_instances = malloc (sizeof (pasisio_select_t));
    type->select_instances->count = 0;
    type->select_instances->instances = malloc (PASISIO_DRIVE_MAXINSTANCES *
                                                sizeof(pasisio_drive_instance_t*));
    PASIS_DASSERT (type->select_instances->instances != NULL);

    /* Link in drive type */
    drive_type_array[type_index] = type;
   
    return type;
}


void
PasisIO_DriveDestroyInstance (pasisio_drive_instance_t *instance)
{
    int i, found = 0;
    pasisio_select_t *drive_instances;

    /* Remove instance from drive type's select_instances array.  I'm doing a
     * memmove here, moving the array elements after this instance up.  This
     * isn't that expensive though, because in almost all cases, this function
     * is called with the instance in the end of the array so there's nothing to
     * move. -mabdelmalek */

    drive_instances = instance->drive->select_instances;
    for (i = drive_instances->count - 1; i >= 0 && !found; --i) {
        if (drive_instances->instances[i] == instance) {
            memmove (drive_instances->instances[i],
                     drive_instances->instances[i + 1],
                     (drive_instances->count - 1 - i) * sizeof(pasisio_select_t*));
            found = 1;
        }
    }
    PASIS_DASSERT (found == 1);

    --drive_instances->count;

    /* Cleanup worker threads */
    instance->dispatch_shutdown = 1;
    for (i = 0; i < instance->thread_num; i++) {
        st_cond_signal (instance->thread_array[i].cond);
        st_thread_join (instance->thread_array[i].thread_id, NULL);
        st_cond_destroy (instance->thread_array[i].cond);
    }
    free (instance->thread_array);

    /* Destroy instance */
    instance->drive->destroy_instance (instance);

    free (instance);
}


static int id=0;
pasisio_drive_instance_t *
PasisIO_DriveCreateInstance (pasisio_drive_type_t *type,
                             p_uint32 ip_address,
                             char *base_path,
                             unsigned int worker_threads)
{
    pasisio_drive_instance_t *instance, *obj;
    int i;

    /* Check that the server_name and base_path's are different to existing
       instances... */
    for (i = 0; i < type->select_instances->count; ++i) {
        obj = type->select_instances->instances[i];

        PASIS_DPRINTF (PASIS_DEBUG_IO_DRIVE, 4, 
                       "Checking: ips[%d][%d], paths[%s][%s]\n",
                       ip_address, obj->ip_address, base_path, obj->base_path);

        if (obj->ip_address == ip_address &&
            strcmp (obj->base_path, base_path) == 0) {
            return NULL;
        }
    }

    instance = (pasisio_drive_instance_t *) malloc (sizeof 
                                                   (pasisio_drive_instance_t));
    memset (instance, 0, sizeof (pasisio_drive_instance_t));

    PASIS_Q_INIT (instance->dispatch_req_queue);
    PASIS_Q_INIT (instance->thread_queue);

    instance->ip_address = ip_address;
    strncpy (instance->base_path, base_path, PASISIO_DRIVE_MAXPATHLEN);
    instance->drive = type;

    instance->status = PASISIO_DRIVE_STATUS_DOWN;

    instance->init_complete = 0;
    instance->init_cond = st_cond_new ();

    /* Set up thread pool */
    instance->thread_num = worker_threads;
    instance->thread_array = malloc (sizeof (pasisio_drive_thread_t) * 
                                     worker_threads);
    for (i = 0; i < worker_threads; i++) {
        pasisio_drive_thread_t *worker;

        worker = &instance->thread_array[i];

        worker->id = id;
        id++;
        worker->cond = st_cond_new ();
        worker->instance = instance;
        worker->thread_id = st_thread_create (PasisIO_DriveThreadDispatch, 
                                              worker, 1, 0);

        PASIS_Q_ADD_TAIL (instance->thread_queue, queue, worker);
    }

    /* Link in to drive type's instance list */
    PASIS_DASSERT (type->select_instances->count < PASISIO_DRIVE_MAXINSTANCES);
    type->select_instances->instances[type->select_instances->count] = instance;
    ++type->select_instances->count;
    
    return instance;
}


void
PasisIO_DriveShutdown (void)
{
    int i, j;

    /* Cleanup drive instances */
    for (i = 0; i < PASISIO_DRIVE_TYPE_NUM; ++i) {
        pasisio_drive_type_t *type = drive_type_array[i];
        for (j = type->select_instances->count - 1; j >= 0; j--) {
            PasisIO_DriveDestroyInstance (type->select_instances->instances[j]);
        }
    }

    /* Shutdown drive types */
    PasisIO_DriveShutdownMD ();

    /* Shtudown rpc layer */
    rpc_auth_client_shutdown ();
    rpc_shutdown ();

    PasisFreelist_Destroy (drive_req_hdr_freelist);
    PasisFreelist_Destroy (drive_req_freelist);    


    PASIS_PRINT_INFO ("\nStatistics:  PasisIO_Drive\n");
    PASIS_PRINT_INFO ("-----------------------------\n");
   
    Pasis_TimerShutdown (stats.timers);

    PASIS_PRINT_INFO ("--- Other Stats ---\n");
    PASIS_PRINT_INFO ("Total Operations     :  %u\n", stats.operations);
    PASIS_PRINT_INFO ("Total Req Enqueues   :  %u\n", stats.enqueues);
    PASIS_PRINT_INFO ("Early Req Termination:  %u\n", stats.reqs_term);
    PASIS_PRINT_INFO ("Failed Early Req Term:  %u\n", stats.reqs_term_failed);
    PASIS_PRINT_INFO ("Enqueue Blocked      :  %u\n", stats.enqueue_blocked);
    PASIS_PRINT_INFO ("Avg. Queued/Instance :  %.3f\n", 
                      stats.enqueue_queued / (double)stats.enqueues);
    PASIS_PRINT_INFO ("Avg. InProg/Instance :  %.3f\n", 
                      stats.threads_in_prog / (double)stats.enqueues);


    PASIS_PRINT_INFO ("-----------------------------\n\n");
}


pasis_error_t
PasisIO_DriveInit (unsigned int max_outstanding)
{
    pasis_error_t error;

    srand (time (NULL));

    /* Create Timers */
    memset (&stats, 0, sizeof (drive_stats_t));
    stats.timers = Pasis_TimerInit ("PasisIO_Drive", DRIVE_TIMER_NUM + 13, 0);
    Pasis_TimerRegister (stats.timers, DRIVE_TIMER_REQUEST, "Request"); 
    
    Pasis_TimerRegister (stats.timers, DRIVE_TIMER_OPERATION, "Create");
    Pasis_TimerRegister (stats.timers, DRIVE_TIMER_OPERATION+1, "Mkdir");
    Pasis_TimerRegister (stats.timers, DRIVE_TIMER_OPERATION+2, "Read");
    Pasis_TimerRegister (stats.timers, DRIVE_TIMER_OPERATION+3, "Write");
    Pasis_TimerRegister (stats.timers, DRIVE_TIMER_OPERATION+4, "OWrite");
    Pasis_TimerRegister (stats.timers, DRIVE_TIMER_OPERATION+5, "Unlink");
    Pasis_TimerRegister (stats.timers, DRIVE_TIMER_OPERATION+6, "Rmdir");
    Pasis_TimerRegister (stats.timers, DRIVE_TIMER_OPERATION+7, "PWrite");
    Pasis_TimerRegister (stats.timers, DRIVE_TIMER_OPERATION+8, "PRead");
    Pasis_TimerRegister (stats.timers, DRIVE_TIMER_OPERATION+9, "PValidate");
    Pasis_TimerRegister (stats.timers, DRIVE_TIMER_OPERATION+10, "PCondWrite");
    Pasis_TimerRegister (stats.timers, DRIVE_TIMER_OPERATION+11, "PGetLTime");
    Pasis_TimerRegister (stats.timers, DRIVE_TIMER_OPERATION+12, "PGC");

 
    /* Create request freelist */
    drive_req_hdr_freelist = PasisFreelist_CreateInitFixed (
                                              sizeof (pasisio_drive_req_hdr_t),
                                                            max_outstanding,
                                                            driveInitReqHdr,
                                                            driveDestReqHdr);

    drive_req_freelist = PasisFreelist_CreateFixed(sizeof(pasisio_drive_req_t),
                                                   max_outstanding * 
                                                   PASIS_MAX_NUM_SERVERS);

    /* Init rpc layer */
    if (rpc_init ()) {
        return PASIS_ERROR;
    }
    
    if (rpc_auth_client_init ()) {
        return PASIS_ERROR;
    }

    /* Init drive types */
    memset (drive_type_array, 0, sizeof (pasisio_drive_type_t) * 
            PASISIO_DRIVE_TYPE_NUM);

    if ((error = PasisIO_DriveInitMD ())) {
        return error;
    }

    return PASIS_SUCCESS;
}

/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */ 
