/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"
#include "pasis_types.h"
#include "pasis_error.h"
#include "pasis_lists.h"
#include "pasis_cache.h"
#include "pasis_timer.h"

#include "pasisio.h"

#include "pasisio_md_protocol.h"
#include "pasisio_md_hist_cache.h"

#include "pasis_md_prot.h"
#include "pasis_md_hist_matrix.h"

#define HIST_CACHE_TIMER_NUM          4
#define HIST_CACHE_TIMER_LOOKUP       0
#define HIST_CACHE_TIMER_UPDATE       1
#define HIST_CACHE_TIMER_HASH         2
#define HIST_CACHE_TIMER_COMPARE      3

typedef struct hist_cache_stats_s {
    pasis_timer_state_t  *timers;
    unsigned int          cache_hits;
    unsigned int          cache_misses;
} hist_cache_stats_t;


static pasis_cache_t      *hist_matrix_cache;
static hist_cache_stats_t  stats;

static int
tag_compare (void *tag1, void *tag2)
{
    int rc;

    PASIS_TIMER_START (); 
    rc = (memcmp (tag1, tag2, sizeof (pasisio_md_hist_tag_t)) == 0);
    PASIS_TIMER_STOP (stats.timers, HIST_CACHE_TIMER_COMPARE);
    return rc;
}


static unsigned int
tag_hash (void *tag, unsigned int buckets)
{
    pasisio_md_hist_tag_t *hist_tag;
    unsigned int           hash, i;
    p_uint32              *fidspace;
    
    PASIS_TIMER_START ();     

    hist_tag = (pasisio_md_hist_tag_t *)tag;
    fidspace = (p_uint32 *)hist_tag->data;
    hash = 0;
    for (i = 0; i < (PASISIO_FILEID_SIZE >> 2); i++) {
        hash ^= fidspace[i];
    }

    PASIS_TIMER_STOP (stats.timers, HIST_CACHE_TIMER_HASH);

    return hash % buckets;
}


pasis_error_t
PasisIO_MDHistCacheInit (unsigned int num_outstanding)
{
    /* Make sure the cache is big enough to hold all history matrices need
     * for inflight requests, if each is also doing repair.
     */
    if (HIST_MATRIX_CACHE_SIZE < num_outstanding*MD_MAX_OBJS_PER_REQ*2) {
        PASIS_PRINT_ERR ("%s: history matrix cache size too small\n",
                         __FUNCTION__);
        PASIS_DASSERT (0);
    }

    memset (&stats, 0, sizeof (hist_cache_stats_t));
    stats.timers = Pasis_TimerInit ("Pasis_IO_Hist_Cache_Timers", 
                                    HIST_CACHE_TIMER_NUM, 0);
    Pasis_TimerRegister (stats.timers, HIST_CACHE_TIMER_HASH, "hash");
    Pasis_TimerRegister (stats.timers, HIST_CACHE_TIMER_COMPARE, "compare");
    Pasis_TimerRegister (stats.timers, HIST_CACHE_TIMER_LOOKUP, "lookup");
    Pasis_TimerRegister (stats.timers, HIST_CACHE_TIMER_UPDATE, "update");

    hist_matrix_cache = PasisCache_Create (HIST_MATRIX_CACHE_SIZE, 1,
                                           sizeof (pasisio_md_hist_tag_t),
                                           sizeof (pasisio_md_hist_data_t),
                                           tag_hash, tag_compare);



    return PASIS_SUCCESS;
}


void
PasisIO_MDHistCacheShutdown (void)
{
    PasisCache_Shutdown (hist_matrix_cache);

    PASIS_PRINT_INFO ("\nPASIS IO Hist Cache Stats\n");
    PASIS_PRINT_INFO ("-------------------------\n");
    PASIS_PRINT_INFO ("cache hits           :   %u\n", stats.cache_hits);
    PASIS_PRINT_INFO ("cache misses         :   %u\n", stats.cache_misses);

    Pasis_TimerShutdown (stats.timers);
}


/* Does a cache lookup based on the tag and fills in the matrix */
pasisio_md_hist_data_t *
PasisIO_MDHistCacheLookup (pasisio_md_hist_tag_t  *tag)
{
    pasisio_md_hist_data_t *data;
    PASIS_TIMER_START ();

    data = PasisCache_Lookup (hist_matrix_cache, tag, 0);
    if (data) {
        stats.cache_hits++;
        PASIS_TIMER_STOP (stats.timers, HIST_CACHE_TIMER_LOOKUP);
        return data;
    }

    stats.cache_misses++;

    PASIS_TIMER_STOP (stats.timers, HIST_CACHE_TIMER_LOOKUP);
    return NULL;
}

pasisio_md_hist_data_t *
PasisIO_MDHistCacheGetEmpty (pasisio_md_hist_tag_t *tag)
{
    pasisio_md_hist_data_t *data;

    PASIS_TIMER_START ();

    data = PasisCache_Lookup (hist_matrix_cache, tag, 0);
    if (data) {
        stats.cache_hits++;
        PASIS_TIMER_STOP (stats.timers, HIST_CACHE_TIMER_LOOKUP);
        return data;
    }

    stats.cache_misses++;

    data = PasisCache_GetEmpty (hist_matrix_cache, tag);
    
    PasisIO_MDHistCacheInvalidateData (data);

    PASIS_TIMER_STOP (stats.timers, HIST_CACHE_TIMER_LOOKUP);

    return data;
}


void
PasisIO_MDHistCachePin (pasisio_md_hist_data_t *data)
{
    PasisCache_PinData (hist_matrix_cache, data);
}
                                

void
PasisIO_MDHistCacheUnpin (pasisio_md_hist_data_t *data)
{
    PasisCache_UnpinData (hist_matrix_cache, data);
}


static inline void
shift_up (pasisio_md_hist_data_t *data,
          unsigned int            start,
          unsigned int            up_by,
          unsigned int            num)
{
     memmove (&data->ts_vector[start - up_by], &data->ts_vector[start],
              num * sizeof (pasis_md_hts_t));

     memmove (&data->ts_bitmap[start - up_by], &data->ts_bitmap[start],
              num * sizeof (p_uint64));
}

static inline void
shift_down (pasisio_md_hist_data_t *data,
            unsigned int            start,
            unsigned int            down_by,
            unsigned int            num)
{
     memmove (&data->ts_vector[start + down_by], &data->ts_vector[start],
              num * sizeof (pasis_md_hts_t));

     memmove (&data->ts_bitmap[start + down_by], &data->ts_bitmap[start],
              num * sizeof (p_uint64));
}


#if PASIS_DEBUG > 0
static void
check_for_duplicate_ts(pasisio_md_hist_data_t *hist_matrix)
{
    p_uint32 i, j;
    for (i = 0; i < hist_matrix->ts_num; ++i) {
        for (j = i + 1; j < hist_matrix->ts_num; ++j) {
            if (!md_hts_compare(&hist_matrix->ts_vector[PASIS_MD_HIST_MAX_VERSIONS + i - hist_matrix->ts_num],
                                &hist_matrix->ts_vector[PASIS_MD_HIST_MAX_VERSIONS + j - hist_matrix->ts_num])) {
                fprintf(stderr, "%s: found duplicate TS in OHS", __FUNCTION__);
                PASIS_DASSERT(0);
            }
        }
    }
}
#endif


void
PasisIO_MDHistCacheUpdate (pasis_md_hist_vector_t *vector,
                           pasisio_md_hist_data_t *data)
{
    unsigned int vec_idx = 0;
    unsigned int mat_idx;
    unsigned int start_idx;
    int          i, rc;
 
    p_uint64     bitmask;

    PASIS_TIMER_START ();

#if PASIS_DEBUG > 0
    check_for_duplicate_ts(data);
#endif

    bitmask = 1ULL << vector->node_id;

    PASIS_DASSERT(data->ts_num < PASIS_MD_HIST_MAX_VERSIONS);

    start_idx = mat_idx = PASIS_MD_HIST_MAX_VERSIONS - data->ts_num;

    while (vec_idx < vector->history_num && 
           mat_idx < PASIS_MD_HIST_MAX_VERSIONS) {
        /*
        PasisIO_MDHistCacheDumpData (data);

        printf ("End op\n\n");
        
        printf ("start_idx=%u, mat_idx=%u, vec_idx=%u\n", 
                start_idx, mat_idx, vec_idx);
        */
        rc = md_hts_compare (&data->ts_vector[mat_idx],
                             &vector->histories[vec_idx]);
        /*
        printf ("hts comparison [%llu]:[%llu], rc=%d\n",
                data->ts_vector[mat_idx].ts.time,
                vector->histories[vec_idx].ts.time, rc);
        */
        if (rc == 0) {
            /* Timestamps equal, set bitmask */
            /*
            printf ("hts equal, setting bitmask (mat_idx=%u, vec_idx=%u)\n",
                    mat_idx, vec_idx);
            */
            data->ts_bitmap[mat_idx] |= bitmask;
            mat_idx++;
            vec_idx++;
            
            continue;
        }

        /* Clear bitmap if it was set */
        data->ts_bitmap[mat_idx] &= ~bitmask;
        if (data->ts_bitmap[mat_idx] == 0) {
            //printf ("Removing element at %u\n", mat_idx);

            /* Shift elements down to cover in hole */
            shift_down (data, start_idx, 1, mat_idx - start_idx);
            data->ts_num--;

            PASIS_DASSERT(data->ts_num < PASIS_MD_HIST_MAX_VERSIONS);

            start_idx++;
            mat_idx++;

            continue;
        }

        if (rc > 0) {
            /* Matrix TS is bigger, move on to next element */
            /*
            printf ("mat hts greater, (mat_idx=%u, vec_idx=%u)\n",
                    mat_idx, vec_idx);
            */
            mat_idx++;

            continue;
        }

        if (rc < 0) {
            /*
            printf ("vec hts greater, (mat_idx=%u, vec_idx=%u)\n",
                    mat_idx, vec_idx);
            */
            /* Vec TS is bigger, shift old entries up */
            shift_up (data, start_idx, 1, mat_idx - start_idx);
            
            memcpy (&data->ts_vector[mat_idx - 1], &vector->histories[vec_idx],
                    sizeof (pasis_md_hts_t));

            data->ts_bitmap[mat_idx - 1] = bitmask;
            
            data->ts_num++;
            start_idx--;
            vec_idx++;

            PASIS_DASSERT(data->ts_num < PASIS_MD_HIST_MAX_VERSIONS);

            continue;
        }
    }


    /* Any entries that are left in vector must be added to end of matrix */
    if (mat_idx == PASIS_MD_HIST_MAX_VERSIONS && 
        vec_idx < vector->history_num) {

        unsigned int vec_num = vector->history_num - vec_idx;
        
        //printf ("no matrix entries remaining (vec_idx=%u)\n", vec_idx);
        
        /* Shift all entries up to make room for vector entries at end */
        shift_up (data, start_idx, vec_num, data->ts_num);
        
        mat_idx -= vec_num;
        start_idx -= vec_num;
        data->ts_num += vec_num;
        PASIS_DASSERT(data->ts_num < PASIS_MD_HIST_MAX_VERSIONS);
        for (i = vec_idx; i < vector->history_num; i++) {
            memcpy(&data->ts_vector[mat_idx], &vector->histories[i],
                   sizeof (pasis_md_hts_t));
            data->ts_bitmap[mat_idx] = bitmask;
            mat_idx++;
        }
    }

    /* Check that all matrix elements were consumed, if not must zero out 
     * any elements that used to belong to this node_id */
    while (mat_idx < PASIS_MD_HIST_MAX_VERSIONS) {
        data->ts_bitmap[mat_idx] &= ~bitmask;
        if (data->ts_bitmap[mat_idx] == 0) {
            /* Shift elements down to cover in hole */
            shift_down (data, start_idx, 1, mat_idx - start_idx);
            start_idx++;
            data->ts_num--;
            PASIS_DASSERT(data->ts_num < PASIS_MD_HIST_MAX_VERSIONS);
        }        
        mat_idx++;            
    }

#if PASIS_DEBUG > 0
    /* Sanity checking.  Check that 1) the elements are in order, and 2) that
     * there are no zero times. */
    PASIS_DASSERT(data->ts_num < PASIS_MD_HIST_MAX_VERSIONS);
    for (i = PASIS_MD_HIST_MAX_VERSIONS - data->ts_num;
         i < PASIS_MD_HIST_MAX_VERSIONS - 1; i++) {
        PASIS_DASSERT (md_hts_compare (&data->ts_vector[i],
                                       &data->ts_vector[i + 1]) > 0);
        //        PASIS_DASSERT (data->ts_vector[i].ts.time != 0);
    }
    //PASIS_DASSERT(data->ts_vector[PASIS_MD_HIST_MAX_VERSIONS - 1].ts.time != 0);

    check_for_duplicate_ts(data);
#endif
    
    /* Copy authenticators */
    data->auth_num[vector->node_id] = vector->auth_num;
    memcpy (data->authenticators[vector->node_id], vector->authenticators,
            vector->auth_num * sizeof (pasis_md_auth_t));

    PASIS_TIMER_STOP (stats.timers, HIST_CACHE_TIMER_UPDATE);
}
                           

void
PasisIO_MDHistCacheSetMatrix (pasisio_md_hist_data_t *data,
                              pasis_md_hist_matrix_t *matrix)
{
    unsigned int i;

#if PASIS_DEBUG > 0
    PASIS_DASSERT(data->ts_num < PASIS_MD_HIST_MAX_VERSIONS);
    check_for_duplicate_ts(data);
#endif

    matrix->ts_num = data->ts_num;

    for (i = 0; i < data->ts_num; i++) {
        matrix->ts_vector[i] = 
            &data->ts_vector[PASIS_MD_HIST_MAX_VERSIONS + i - data->ts_num];
        matrix->ts_bitmap [i] =
            data->ts_bitmap[PASIS_MD_HIST_MAX_VERSIONS + i - data->ts_num];
    }

}



void
PasisIO_MDHistCacheSetAuths (pasis_md_auth_t  auth_to_send[PASIS_MAX_NUM_SERVERS]
                                                          [PASIS_MAX_NUM_SERVERS],
                             unsigned int     auth_to_send_num[PASIS_MAX_NUM_SERVERS],
                             pasis_md_hist_matrix_t *matrix,
                             unsigned int     for_node_id)
{
    // why do we need to change this behaviour if DO_NOT_USE_AUTH > 0?
    //#if DO_NOT_USE_AUTHENTICATORS == 0
    int i;

    /* printf("%s: node = %u, IN\n", __FUNCTION__, for_node_id);
     * pasis_md_hist_matrix_dump (matrix); */
    for (i = 0; i < PASIS_MAX_NUM_SERVERS; ++i) {
        if (auth_to_send_num[i] > 0) {
            PASIS_DASSERT (auth_to_send_num[i] > for_node_id);
            matrix->auth_exists[i] = 1;
            matrix->authenticators[i] = &auth_to_send[i][for_node_id];
            matrix->auth_num++;
            /* printf("\t i = %d, auth = %u\n", i,
            *         *(p_uint32*)matrix->authenticators[i]); */
        } else {
            matrix->auth_exists[i] = 0;
        }
    }
    /* printf("%s: node = %u, OUT\n", __FUNCTION__, for_node_id); */

    /*   printf ("%s: node=%u (matrix->auth_num=%u)\n",
     *           __FUNCTION__, for_node_id, matrix->auth_num); */
    //#endif
}


void
PasisIO_MDHistCacheDumpData (pasisio_md_hist_data_t *data)
{
    int i;
    printf ("\nTS Vector, num=%d\n", data->ts_num);

    for (i = PASIS_MD_HIST_MAX_VERSIONS - data->ts_num; 
         i < PASIS_MD_HIST_MAX_VERSIONS; i++) {
        printf ("[%d]  ts: %llu, bitmap: %#llx\n", 
                i, data->ts_vector[i].time, data->ts_bitmap[i]);        
    }
}
 
/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */  
