/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "limits.h"
#include "pasis_config.h"
#include "pasis_types.h"
#include "pasis_error.h"
#include "pasis_lists.h"
#include "pasis_freelist.h"

#include "pasisio.h"
#include "pasisio_select.h"
#include "pasisio_drives.h"
#include "pasisio_internal.h"
#include "pasisio_drive_md.h"

#include "pasisio_md_protocol.h"
#include "pasisio_md_hist_cache.h"


#include "pasis_md_prot.h"
#include "pasis_md_hist_matrix.h"
#include "pasis_md_dir.h"
#include "pasis_md_xdr.h"


#define USE_EARLY_TERMINATION 1

/* Specifies whether we're stress testing the system or not.  if set to 1, then
 * the code will be more lenient and allow incomplete writes, and not assert. */
int STRESS_TESTING = 0;


/* Specifies whether we should update a client's history cache after the client
 * executes an operation.  Usually, we want to do this, but during testing, we
 * sometimes want this off (e.g., to allow incomplete after a repairable) */
int DO_NOT_UPDATE_HISTORY = 0;



/* Initial wait time (in us) for Q/U protocol whenever we have to backoff */
int INITIAL_BACKOFF_WAIT = 1000;

/* For Mike's Cool Backoff Algorithms, subsequent delay maximum (in us) */
int SUBSEQUENT_MAX_BACKOFF_WAIT = 500;

/* Maximum backoff wait (in us) */
int MAX_BACKOFF_WAIT = 8000;


#define MD_TIMER_NUM             16

#define MD_TIMER_CREATE          0
#define MD_TIMER_LINK            1
#define MD_TIMER_UNLINK          2
#define MD_TIMER_SETATTR         3
#define MD_TIMER_RENAME          4
#define MD_TIMER_GETATTR         5
#define MD_TIMER_LOOKUP          6
#define MD_TIMER_READLINK        7
#define MD_TIMER_READDIR         8
#define MD_TIMER_REPAIR          9
#define MD_TIMER_BARRIER         10
#define MD_TIMER_READ_HISTORY    11
#define MD_TIMER_SYNTH_QUERY     12
#define MD_TIMER_SYNTH_UPDATE    13
#define MD_TIMER_OBJECT_SYNC     14
#define MD_TIMER_GET_STATS       15

#ifndef MAX
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#endif

#ifndef MIN
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#endif

typedef struct md_stats_s {
    unsigned int         read_hist_restart5;
    unsigned int         write_op_hist_inval;
    unsigned int         rereads;
    unsigned int         complete_rereads;
    unsigned int         complete_deductions;
    unsigned int         repairs;
    unsigned int         fast_repairs;
    unsigned int         slow_repairs;
    unsigned int         restarted_repairs;
    pasis_timer_state_t *timers;
} md_stats_t;


static md_stats_t         stats;

static pasis_freelist_t  *proto_freelist;
static pasis_md_hts_t     zero_ts;

static unsigned int       num_outstanding;

static unsigned int       use_random_quorums = 0;

static inline unsigned int
write_res_count_inval_history (pasisio_md_proto_data_t  *proto,
                               pasisio_select_t         *select);

static pasis_error_t
write_op_dispatch (pasis_md_protocol_t      *protocol,
                   pasisio_request_t        *req,
                   pasisio_md_proto_data_t  *proto,
                   pasisio_select_t         *select,
                   unsigned int              fid_num,
                   pasis_md_fid_t           *fids[MD_MAX_OBJS_PER_REQ],
                   unsigned int              op_res_size,
                   void                    **res_out);

static pasis_error_t
write_read_history (pasis_md_protocol_t      *protocol,
                    pasisio_request_t        *req,
                    pasisio_md_proto_data_t  *proto,
                    pasisio_select_t         *select,
                    unsigned int              fid_num,
                    pasis_md_fid_t           *fids[MD_MAX_OBJS_PER_REQ],
                    unsigned int              use_cached_data_flag);


static pasis_error_t
write_classify_history (pasis_md_protocol_t     *protocol,
                        pasisio_request_t       *req,
                        pasisio_md_proto_data_t *proto,
                        pasisio_select_t        *select,
                        unsigned int             fid_num,
                        pasis_md_fid_t          *fids[MD_MAX_OBJS_PER_REQ]);

static void *
read_op_res_count_and_vote (pasisio_select_t       *select,
                            pasis_md_read_op_res_t  read_op_res[PASIS_MAX_NUM_SERVERS],
                            unsigned int            op_size,
                            unsigned int            threshold,
                            pasis_md_hts_t         *result_ts);


void
PasisIO_GetCompleteThreshold (pasis_md_protocol_t *protocol,
                              unsigned int        *complete,
                              unsigned int        *early_term)
{
    switch (protocol->quorum_type) {
    case PASISIO_QUORUM_NONE:
        *complete = protocol->N;
        *early_term = protocol->N - protocol->T;
        break;
    case PASISIO_QUORUM_THRESHOLD_COMP_QC_B_T:
        /* delta already added to Qc in PasisIO_QuorumParamsInit */
        *complete = protocol->Qc + protocol->B + protocol->T;
        *early_term = *complete - protocol->T;
        break;
    case PASISIO_QUORUM_THRESHOLD_COMP_QC_B:
        /* delta already added to Qc in PasisIO_QuorumParamsInit */
        *complete = protocol->Qc + protocol->B;
        *early_term = *complete;
        break;
    case PASISIO_QUORUM_TREE:
        *complete = protocol->Qc + protocol->B;
        *early_term = *complete;
        break;
    default:
        PASIS_DASSERT (0);
    }
}


unsigned int
PasisIO_GetRepairableThreshold (pasis_md_protocol_t *protocol)
{
    switch (protocol->quorum_type) {
    case PASISIO_QUORUM_NONE:
        return protocol->Qc - protocol->T;
    case PASISIO_QUORUM_THRESHOLD_COMP_QC_B_T:
    case PASISIO_QUORUM_THRESHOLD_COMP_QC_B:
        return protocol->Qc - protocol->T - protocol->delta;
    case PASISIO_QUORUM_TREE:
        PASIS_DASSERT (protocol->T == 1 && protocol->B == 1);
        if (protocol->delta == 0) {
            return 3;
        } else if (protocol->delta == 1) {
            return 9; /* but not just ANY 9... */
        } else {
            PASIS_DASSERT (0);
        }
        return 0;
    default:
        PASIS_DASSERT (0);
    }
    return 0;
}


void
PasisIO_UseRandomQuorums (void)
{
    use_random_quorums = 1;
}


void
PasisIO_UsePreferredQuorums (void)
{
    use_random_quorums = 0;
}



/* exponential backoff: 250ms * 2 ^ (num "collisions") */
static inline void
backoff (int num_consecutive_collisions)
{
    float wait_time;

#if 1
    /* exponential backoff */
    /* Limit how much the 1<< num_collisions will be */
    num_consecutive_collisions = MIN(10, num_consecutive_collisions);

    wait_time = INITIAL_BACKOFF_WAIT * (1 << num_consecutive_collisions);
#endif

#if 0
    /* random backoff */
    wait_time = rand() % INITIAL_BACKOFF_WAIT;
#endif

#if 1
    /* Mike's cool backoff algorithm :) 
     * Unfairness results when a single client keeps getting served, and other
     * clients keep getting rejected.  Ordinary exponential backoff or random
     * backoff aren't working as well, because they punish each client the same
     * way.  But we really wanna punish the client doing best, to promote
     * fairness.  One approximation is to have the initial backoff by large, and
     * subsequent backoffs be much smaller.  Once the doing-well client
     * backsoff, it's likely that other clients will have some "breathing" room. */
    if (num_consecutive_collisions <= 1) {
        wait_time = INITIAL_BACKOFF_WAIT;
    } else {
        wait_time = rand() % SUBSEQUENT_MAX_BACKOFF_WAIT;
    }
#endif

    if (wait_time > MAX_BACKOFF_WAIT) {
        wait_time = MAX_BACKOFF_WAIT;
    }
    
#if 1
    /* mostly useful for exponential backoff */
    wait_time *= (1.0 + (rand() % 4)/10.0); // multiply by 1.0, 1.1, 1.2, or 1.3
#endif

    /* printf("<<< backing off for %f microseconds >>>\n", wait_time); */
    st_usleep (wait_time);
}


static inline void
select_reset (int num, pasisio_select_t *select)
{
    /* Call SelectServers, so that we don't select _all_ servers.  we only want
     * to set a quorum of servers, so that the Drives module doesn't look at a
     * bad request structure */
    PasisIO_SelectServers(num, select);
}


static int req_count = 0;


/* The shuffle argument is used for selecting among different preferred
 * quorums.  The caller typically sets this value based on some bits of the
 * FID.
 * The current FID-generation algorithm (see SetCreateFID) sets the FID bits
 * randomly.  If that ever changes, then the caller can't just use any part of
 * the FID bits as the shuffle parameter, if they want to distribute the load
 * evenly.   -mabdelmalek 10/22/2004
 */
static inline void
op_init (pasis_md_protocol_t      *protocol,
         pasis_md_linkage_t       *linkages,
         unsigned int              num_linkages,
         pasisio_md_proto_data_t  *proto,
         pasisio_select_t         *select,
         pasisio_request_t        *req,
         unsigned int              shuffle)
{
    //int i;

    /* Decode metadata object */
    select->count = PASIS_MAX_NUM_SERVERS;

#if 0
    printf ("%s: orig linkages\n", __FUNCTION__);
    for (i = 0; i < num_linkages; i++) {
        printf ("[%d]  %x\n", i, linkages[i].ip_address);
    }
#endif

    switch (protocol->quorum_type) {
    case PASISIO_QUORUM_NONE:
        shuffle = shuffle % protocol->N;
        proto->shuffle = shuffle;
        
        /* We don't use PasisIO_SelectDecodeLinkages because that function
         * shuffles the order of the S-N's.  We want this shuffling for load
         * distribution when performing erasure-coding.  But for the Q/U
         * protocol which uses authenticators, we need to keep the S-N list in
         * order.  And PasisIO_SelectThresholdDecodeLinkages doesn't shuffle. */
        PasisIO_SelectThresholdDecodeLinkages (linkages, protocol->N,
                                               protocol->N,
                                               PASISIO_DRIVE_TYPE_MD_INDEX,
                                               select, shuffle);
        break;
    case PASISIO_QUORUM_THRESHOLD_COMP_QC_B_T:
    case PASISIO_QUORUM_THRESHOLD_COMP_QC_B: {
        unsigned int complete, term;
               
        PasisIO_GetCompleteThreshold (protocol, &complete, &term);

        shuffle = shuffle % protocol->N;

        if (use_random_quorums) {
            shuffle = rand() % protocol->N;
        }

        proto->shuffle = shuffle;

        PasisIO_SelectThresholdDecodeLinkages (linkages, protocol->N,
                                               complete,
                                               PASISIO_DRIVE_TYPE_MD_INDEX,
                                               select, shuffle);
        break;
    }
    case PASISIO_QUORUM_TREE: {
        proto->shuffle = shuffle;

        PasisIO_SelectTreeDecodeLinkages (linkages, protocol,
                                          PASISIO_DRIVE_TYPE_MD_INDEX,
                                          select, shuffle);
        break;
    }
    default:
        PASIS_DASSERT (0);
    }

#if 0
    printf ("\t%s: drive list: ", __FUNCTION__);
    for (i = 0; i < num_linkages; i++) {
        if (select->flags[i] == PASISIO_SELECT_IO_READY) {
            printf ("[%d], ", i);
        }
    }
    printf ("\n");
#endif

    /* Get proto data structure */
    proto->protocol = protocol;
    proto->linkages = (pasis_md_linkage_t *)linkages;

    /* Setup request structure */
    req->drive_select = select;
    req->proto_data = proto;
    req->id = req_count++;

    return;
}


static unsigned int
protocol_issue_reqs (pasisio_request_t      *req,
                     pasisio_drive_op_t      drive_op,
                     unsigned int            issue,
                     unsigned int            min,
                     int                     early_term_threshold)
{
    int completed = 0;
    int selected = 0;
    int early_term = 0;

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "%s: op=%#x, issue=%u, min=%u\n", 
                   __FUNCTION__, drive_op, issue, min);

    do {
        selected = PasisIO_SelectServers (issue-completed, req->drive_select);

        PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "%s: selected=%d, min=%d\n",
                       __FUNCTION__, selected, min);

        /* allow incompletes during testing */
        if (STRESS_TESTING == 0 && selected + completed < min) {
            printf ("%s: no chance for completion, so not even trying; "
                    "selected = %d, completed = %d, min = %d\n",
                    __FUNCTION__, selected, completed, min);
            return completed;
        }

        if (selected > 0) {

            early_term = MAX (early_term_threshold - completed, 0);

            completed += PasisIO_DriveEnqueueShares (drive_op, req, selected,
                                                     early_term);

        }

        PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "%s: completed=%d\n", 
                       __FUNCTION__, completed);

    } while (completed < min && completed < issue && selected > 0);
    
    return completed;
}


static void
validate_hist_matrix (pasis_md_protocol_t     *protocol,
                      pasis_md_hist_matrix_t  *hist_matrix,
                      pasis_md_hts_t         **ts_cond_out,
                      unsigned int            *barrier_out,
                      unsigned int            *repair_out,
                      unsigned int            *fast_repair_out,
                      unsigned int            *multiobject_op)
{
    pasis_md_hts_t  *ts_latest, *ts_classify, *ts_complete, *ts_barrier;
    pasis_error_t    latest_status, classify_status, barrier_status;
    int              latest_is_multiobject_op, classify_is_multiobject_op;

    pasis_md_hist_matrix_classify_node (hist_matrix, protocol, &ts_barrier,   
                                        &barrier_status, &ts_latest,
                                        &latest_status,
                                        &latest_is_multiobject_op,
                                        &ts_classify, 
                                        &classify_status,
                                        &classify_is_multiobject_op,
                                        &ts_complete);

    *ts_cond_out = ts_complete;
    *barrier_out = 0;
    *repair_out = 0;
    *fast_repair_out = 0;

    if (ts_barrier != NULL && barrier_status == PASIS_ERROR_PROTO_COMPLETE) {
        /* If barrier blocking a repairable, just need repair */
        if (classify_status == PASIS_ERROR_PROTO_REPAIR) {
            *ts_cond_out = ts_classify;
            *repair_out = 1;
            *multiobject_op = classify_is_multiobject_op;
        }
        return;
    }

    /* If latest is an incomplete, need a barrier */
    if (latest_status == PASIS_ERROR_PROTO_INCOMPLETE) {
        *barrier_out = 1;

        /* If classification yields a repairable, need to do repair */
        if (classify_status == PASIS_ERROR_PROTO_REPAIR) {
            *ts_cond_out = ts_classify;
            *repair_out = 1;
            *multiobject_op = classify_is_multiobject_op;
        }

    } else if (latest_status == PASIS_ERROR_PROTO_REPAIR) {
        /* If latest is a repairable, need only do repair -- fast repair */
        *ts_cond_out = ts_classify;
        *repair_out = 1;
        *fast_repair_out = 1;
        *multiobject_op = latest_is_multiobject_op;
    }

    if (protocol->quorum_type == PASISIO_QUORUM_TREE &&
        classify_status != PASIS_ERROR_PROTO_COMPLETE) {
        /* have not implemented repair for tree quorums yet */
        PASIS_DASSERT (0);
    }

    return;
}

static inline void
set_hist_matrices (unsigned int              matrices_num,
                   pasis_md_write_op_args_t *write_op_args,
                   pasisio_md_hist_data_t   *cache[MD_MAX_OBJS_PER_REQ])
{                
    int i;

    /* Fill in history matrices from history cache */
    write_op_args->history_matrices_num = matrices_num;
    for (i = 0; i < matrices_num; i++) {
        PasisIO_MDHistCacheSetMatrix (cache[i],
                                      &write_op_args->history_matrices[i]);
    }
}


static int
vote (void         *array[PASIS_MAX_NUM_SERVERS],
      unsigned int  num_array,
      unsigned int  size,
      unsigned int  threshold)
{
    unsigned int  i, count, cmp_idx1, cmp_idx2;
    p_byte        bitmap[PASIS_MAX_NUM_SERVERS];

    if (threshold == 0 || size == 0) {
        return 0;
    }

    memset (bitmap, 1, num_array);

    cmp_idx1 = 0;
    while (cmp_idx1 < num_array) {
        count = 1;
        cmp_idx2 = cmp_idx1 + 1;

        for (i = cmp_idx2; i < num_array; i++) {
            if (bitmap[i]) {

                if (memcmp (array[cmp_idx1], array[i], size) == 0) {
                    bitmap[i] = 0;
                    count++;

                    if (count > threshold) {
                        return cmp_idx1;
                    }
                }
            }
        }

        cmp_idx1++;
    }
    
    return -1;
}


static unsigned int
read_op_count_successes (pasisio_select_t       *select,
                         pasis_md_read_op_res_t  read_op_res[PASIS_MAX_NUM_SERVERS])
{
    unsigned int i, count = 0;
    for (i = 0; i < select->count; ++i) {
        if (select->flags[i] == PASISIO_SELECT_IO_SUCCESS &&
            read_op_res[i].status == PASIS_SUCCESS) {
            ++count;
        }
    }
    return count;
}


static pasis_md_rpc_stat_t
read_op_rpc_op_stat_vote (pasisio_select_t       *select,
                          pasis_md_read_op_res_t  read_op_res[PASIS_MAX_NUM_SERVERS],
                          unsigned int            threshold)
{
    unsigned int i, idx=0, res_num=0;
    void        *res[PASIS_MAX_NUM_SERVERS];
    

    for (i = 0; i < select->count; i++) {
        if (select->flags[i] == PASISIO_SELECT_IO_SUCCESS /* &&
            read_op_res[i].status == PASIS_SUCCESS */) {
            res[res_num] = (void *)&read_op_res[i].op_status;
            res_num++;
        }
    }
    PASIS_DASSERT (res_num > 0);
    
    if (threshold > 0) {
        idx = vote (res, res_num, sizeof (pasis_md_rpc_stat_t), threshold);
        if (idx == -1) return PASIS_ERROR;
    }
 
    return *(pasis_md_rpc_stat_t *)(res[idx]);
}


static unsigned int
write_op_count_successes (pasisio_select_t       *select,
                          pasis_md_write_op_res_t write_op_res[PASIS_MAX_NUM_SERVERS])
{
    unsigned int count = 0, i;
    for (i = 0; i < select->count; ++i) {
        if (select->flags[i] == PASISIO_SELECT_IO_SUCCESS &&
            write_op_res[i].status == PASIS_SUCCESS) {
            ++count;
        }
    }
    return count;
}


static pasis_md_rpc_stat_t
write_op_rpc_op_stat_vote (pasisio_select_t       *select,
                         pasis_md_write_op_res_t write_op_res[PASIS_MAX_NUM_SERVERS],
                         unsigned int            threshold)
{
    unsigned int i, idx=0, res_num=0;
    void        *res[PASIS_MAX_NUM_SERVERS];

    for (i = 0; i < select->count; i++) {
        if (select->flags[i] == PASISIO_SELECT_IO_SUCCESS /*&&
            write_op_res[i].status == PASIS_SUCCESS */) {
            res[res_num] = (void *)&write_op_res[i].op_status;
            res_num++;
        }
    }
    PASIS_DASSERT (res_num > 0);
    
    if (threshold > 0) {
        idx = vote (res, res_num, sizeof (pasis_md_rpc_stat_t), threshold);
        if (idx == -1) return PASIS_ERROR;
    }
 
    return *(pasis_md_rpc_stat_t *)(res[idx]);
}


static void *
write_op_res_vote (pasisio_select_t       *select,
                   pasis_md_write_op_res_t write_op_res[PASIS_MAX_NUM_SERVERS],
                   unsigned int            op_size,
                   unsigned int            threshold)
{
    unsigned int i, idx=0, res_num=0;
    void        *res[PASIS_MAX_NUM_SERVERS];
    

    for (i = 0; i < select->count; i++) {
        if (select->flags[i] == PASISIO_SELECT_IO_SUCCESS &&
            write_op_res[i].status == PASIS_SUCCESS) {

            if (write_op_res[i].op_status != PASIS_SUCCESS) {
                continue; /* Skip this result */
            }
            /* For the synthetic write operation, we need to look at the pointer
             * in the result structure for the actual return data. */
            if (write_op_res[i].op_type == PASIS_MD_SYNTH_UPDATE) {
                res[res_num] = (void *)write_op_res[i].op_res.synth_update.res;
            } else {
                res[res_num] = (void *)&write_op_res[i].op_res;
            }
            res_num++;
        }
    }

    if (res_num == 0) {
        return NULL;
    }

    if (threshold > 0) {
        idx = vote (res, res_num, op_size, threshold);
        if (idx == -1) return NULL;
    }
    
    return res[idx];
}


pasis_error_t
PasisIO_MDInit (unsigned int _num_outstanding)
{
    pasis_error_t error;

    num_outstanding = _num_outstanding;

    memset (&stats, 0, sizeof (md_stats_t));
    memset (&zero_ts, 0, sizeof (pasis_md_hts_t));

    srand (time (NULL));

    stats.timers = Pasis_TimerInit ("PasisMD_Protocol_Timers",
                                    MD_TIMER_NUM, 0);

    Pasis_TimerRegister (stats.timers, MD_TIMER_CREATE, "Create");
    Pasis_TimerRegister (stats.timers, MD_TIMER_LINK, "Link");
    Pasis_TimerRegister (stats.timers, MD_TIMER_UNLINK, "Unlink");
    Pasis_TimerRegister (stats.timers, MD_TIMER_SETATTR, "SetAttr");
    Pasis_TimerRegister (stats.timers, MD_TIMER_RENAME, "Rename");
    Pasis_TimerRegister (stats.timers, MD_TIMER_GETATTR, "GetAttr");
    Pasis_TimerRegister (stats.timers, MD_TIMER_LOOKUP, "Lookup");
    Pasis_TimerRegister (stats.timers, MD_TIMER_READLINK, "ReadLink");
    Pasis_TimerRegister (stats.timers, MD_TIMER_READDIR, "ReadDir");
    Pasis_TimerRegister (stats.timers, MD_TIMER_REPAIR, "Repair");
    Pasis_TimerRegister (stats.timers, MD_TIMER_BARRIER, "Barrier");
    Pasis_TimerRegister (stats.timers, MD_TIMER_READ_HISTORY, "ReadHistory");
    Pasis_TimerRegister (stats.timers, MD_TIMER_SYNTH_QUERY, "Synth_Query");
    Pasis_TimerRegister (stats.timers, MD_TIMER_SYNTH_UPDATE, "Synth_Update");
    Pasis_TimerRegister (stats.timers, MD_TIMER_OBJECT_SYNC, "Object_Sync");
    Pasis_TimerRegister (stats.timers, MD_TIMER_GET_STATS, "GetStats");

    proto_freelist = PasisFreelist_CreateFixed(sizeof (pasisio_md_proto_data_t),
                                               num_outstanding);
                                          
    error = PasisIO_MDHistCacheInit (num_outstanding);

    printf ("Starting up: client id=0x%x\n", rpc_get_host_ip());
    
    return error;
}


void
PasisIO_MDShutdown (void)
{
    PasisFreelist_Destroy (proto_freelist);

    PasisIO_MDHistCacheShutdown ();

    Pasis_TimerShutdown (stats.timers);

    PASIS_PRINT_INFO ("--- Other Stats ---\n");
    PASIS_PRINT_INFO ("Write op history invalid    :  %u\n", 
                      stats.write_op_hist_inval);
    PASIS_PRINT_INFO ("Read history restarts > 5   :  %u\n", 
                      stats.read_hist_restart5); 
    PASIS_PRINT_INFO ("Read op re-reads            :  %u\n",
                      stats.rereads);
    PASIS_PRINT_INFO ("Read op 'complete' re-reads :  %u\n",
                      stats.complete_rereads);
    PASIS_PRINT_INFO ("Repairs                     :  %u\n",
                      stats.repairs);
    PASIS_PRINT_INFO ("Fast repairs                :  %u\n",
                      stats.fast_repairs);
    PASIS_PRINT_INFO ("Slow repairs                :  %u\n",
                      stats.slow_repairs);
    PASIS_PRINT_INFO ("Restarted repairs           :  %u\n\n",
                      stats.restarted_repairs);
    PASIS_PRINT_INFO ("Complete deductions         :  %u\n",
                      stats.complete_deductions);
}



static void
save_authenticators (pasis_md_protocol_t      *protocol,
                     pasisio_md_proto_data_t  *proto,
                     p_uint32                  num_fids,
                     pasisio_md_hist_data_t  **fid_hist_cache)
{
    int i, j;
    for (i = 0; i < num_fids; ++i) {
        memcpy (proto->auth_to_send_num[i],
                fid_hist_cache[i]->auth_num,
                protocol->N * sizeof(unsigned int));
        for (j = 0; j < protocol->N; ++j) {
            memcpy (proto->auth_to_send[i][j],
                    fid_hist_cache[i]->authenticators[j],
                    protocol->N * sizeof(pasis_md_auth_t));
        }
    }
}




static void
save_query_op (pasis_md_protocol_t      *protocol,
               pasisio_md_proto_data_t  *proto,
               pasis_md_rpc_num_t       *saved_type,
               pasis_md_read_flag_t     *saved_read_flags)
{
    int i;

    *saved_type = proto->read_op_args.op_type;
    *saved_read_flags = proto->read_op_args.read_flags;
    memcpy (&proto->read_op_args_backup,
            &proto->read_op_args.op_args,
            sizeof (pasis_md_read_op_args_specific_t));
    for (i = 0; i < protocol->N; ++i) {
        memcpy (&proto->read_op_res_backup[i],
                &proto->read_op_res[i].op_res,
                sizeof (pasis_md_read_op_res_specific_t));
    }
}


static void
save_update_op (pasis_md_protocol_t      *protocol,
                pasisio_md_proto_data_t  *proto,
                pasis_md_rpc_num_t       *saved_type,
                p_uint32                 *saved_write_mtime)
{
    int i;

    PASIS_DASSERT (proto->already_saved_write_op == 0);
    proto->already_saved_write_op = 1;
    *saved_type = proto->write_op_args.op_type;
    saved_write_mtime[0] = proto->write_op_args.mtime[0];
    saved_write_mtime[1] = proto->write_op_args.mtime[1];
    memcpy (&proto->write_op_args_backup,
            &proto->write_op_args.op_args,
            sizeof (pasis_md_write_op_args_specific_t));
    for (i = 0; i < protocol->N; ++i) {
        memcpy (&proto->write_op_res_backup[i],
                &proto->write_op_res[i].op_res,
                sizeof (pasis_md_write_op_res_specific_t));
    }
}


static void
restore_query_op (pasis_md_protocol_t      *protocol,
                  pasisio_md_proto_data_t  *proto,
                  pasis_md_rpc_num_t        saved_type,
                  pasis_md_read_flag_t      saved_read_flags)
{
    int i;

    proto->read_op_args.op_type = saved_type;
    proto->read_op_args.read_flags = saved_read_flags;
    memcpy (&proto->read_op_args.op_args,
            &proto->read_op_args_backup,
            sizeof (pasis_md_read_op_args_specific_t));
    for (i = 0; i < protocol->N; ++i) {
        memcpy (&proto->read_op_res[i].op_res,
                &proto->read_op_res_backup[i],
                sizeof (pasis_md_read_op_res_specific_t));
    }
    
}


static void
restore_update_op (pasis_md_protocol_t      *protocol,
                   pasisio_md_proto_data_t  *proto,
                   pasis_md_rpc_num_t        saved_type,
                   p_uint32                 *saved_write_mtime)
{
    int i;

    PASIS_DASSERT (proto->already_saved_write_op == 1);
    proto->write_op_args.op_type = saved_type;
    proto->write_op_args.mtime[0] = saved_write_mtime[0];
    proto->write_op_args.mtime[1] = saved_write_mtime[1];
    memcpy (&proto->write_op_args.op_args,
            &proto->write_op_args_backup,
            sizeof (pasis_md_write_op_args_specific_t));
    for (i = 0; i < protocol->N; ++i) {
        memcpy (&proto->write_op_res[i].op_res,
                &proto->write_op_res_backup[i],
                sizeof (pasis_md_write_op_res_specific_t));
    }
    proto->already_saved_write_op = 0;
}


/* Repair operation:
 *
 * UPDATE UPDATE UPDATE
 *
 * 1) Read operation, returns the operation and list of FIDs associated with it
 * 2) Perform barrier write, piggyback read history requests for op FIDs
 * 3) Perform repair operation based on operation read in step (1)
 * 4) Perform original operation 
 * At anytime the repair fails, fail with a RETRY error code and retry the
 * op at a higher level.
 */

static pasis_error_t
write_barrier (pasis_md_fid_t           *fid[MD_MAX_OBJS_PER_REQ],
               unsigned int              fid_num,
               pasis_md_protocol_t      *protocol,
               pasisio_request_t        *req,
               pasisio_md_proto_data_t  *proto,
               pasisio_select_t         *select,
               unsigned int              save_write_op)
{
    unsigned int complete, count, num_issue, num_thresh;
    pasis_md_rpc_num_t saved_type = 0;
    p_uint32 saved_write_mtime[2] = {0,0};
    pasis_error_t error;

    PASIS_TIMER_START ();

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, 
                   "IN %s: writing barrier\n", __FUNCTION__);

    /* save original write op args/results*/
    if (save_write_op) {
        save_update_op (protocol, proto, &saved_type, saved_write_mtime);
    }

    /* Set up write for barrier -- assumes history matrices are in place */
    proto->write_op_args.op_type = PASIS_MD_BARRIER;
    proto->write_op_args.repair_flag = 0;
    proto->write_op_args.op_args.barrier.fid_num = fid_num;
    memcpy (proto->write_op_args.op_args.barrier.fid, fid, 
            fid_num * sizeof (pasis_md_fid_t *));
    proto->write_op_args.op_args.barrier.chunk_size = proto->chunk_size;
    proto->write_op_args.op_args.barrier.client_id  = rpc_get_host_ip();

    PasisIO_GetCompleteThreshold (protocol, &num_issue, &num_thresh);

    /* Issue barrier */
    select_reset (num_issue, select);

/*     printf ("%s: writing barrier for %d FID's: ", __FUNCTION__, fid_num); */
/*     for (i = 0; i < fid_num; ++i) { */
/*         printf("%#llx, ", *(p_uint64 *)fid[i]); */
/*     } */
/*     printf("\n"); */

    /* "Backup" the authenticators so requests to servers don't get overwritten
     * with authenticators from new responses  */
    save_authenticators (protocol, proto,
                         proto->write_op_args.history_matrices_num,
                         proto->fid_hist_cache);

#if USE_EARLY_TERMINATION > 0
    complete = protocol_issue_reqs (req, PASISIO_MD_DRIVE_P_WRITE_OP,
                                    num_issue,
                                    num_thresh, /* Min */
                                    num_thresh);  /* Early term */
#else
    complete = protocol_issue_reqs (req, PASISIO_MD_DRIVE_P_WRITE_OP,
                                    protocol->N,
                                    protocol->N, protocol->N);
#endif

    PASIS_DASSERT (complete >= protocol->Qc + protocol->B);

    count = write_op_count_successes (select, proto->write_op_res);

    if (count < protocol->Qc + protocol->B) {
        /* printf ("%s: barrier did not complete\n", __FUNCTION__); */
        error = PASIS_ERROR_PROTO_RETRY;
    } else {
        /* Do this before select_reset, since select_reset resets the drive status*/
        error = write_op_rpc_op_stat_vote (select, proto->write_op_res,
                                           protocol->B); 
    }

    /* Update the history matrix in the write_op */
    set_hist_matrices (fid_num, &proto->write_op_args, proto->fid_hist_cache);

    select_reset (num_issue, select);

    /* restore original write op args/results*/
    if (save_write_op) {
        restore_update_op (protocol, proto, saved_type, saved_write_mtime);
    }

    PASIS_TIMER_STOP (stats.timers, MD_TIMER_BARRIER);

    return error;
}


#if 0 // unused for now, until we get deduction working again

/* A timestamp corresponds to another timestamp if they're part of the same
 * operation.  We can determine if two timestamps are part of the same operation
 * by looking at the operation hash in their timestmaps */
static pasis_md_hts_t*
get_corresponding_ts (pasis_md_hist_matrix_t *hist_matrix,
                      pasis_md_hts_t         *ts,
                      int                    *count)
{
    int i;
    for (i = 0; i < hist_matrix->ts_num; ++i) {
        if (memcmp (hist_matrix->ts_vector[i]->ts.op_hash,
                    ts->op_hash, PASIS_HASH_SIZE) == 0) {
            *count = bitcount (hist_matrix->ts_bitmap[i]);
            return &hist_matrix->ts_vector[i]->ts;
        }
    }
    return NULL;
}


static pasis_error_t
can_deduce_complete (pasis_md_protocol_t      *protocol,
                     int                       is_query_op,
                     pasisio_md_proto_data_t  *proto,
                     p_uint32                  fid_index,/*FID idx in write op*/
                     pasis_md_fid_t           *fid,
                     unsigned int              fid_num,
                     pasis_md_fid_t           *fids[MD_MAX_OBJS_PER_REQ],
                     pasis_md_hts_t           *ts_repair,
                     int                      *restart_op)
{
    int i, j, k, cmp;
    *restart_op = 0;
    for (i = 0; i < fid_num; ++i) {
        pasis_md_hts_t  *ts_latest, *ts_classify, *ts_complete, *ts_barrier;
        pasis_error_t    latest_status, classify_status, barrier_status;
        int              dummy1, dummy2;
        pasis_md_hist_matrix_t* current_hist_matrix =
            &proto->write_op_args.history_matrices[i];

        pasis_md_hist_matrix_classify_node (current_hist_matrix, protocol,
                                            &ts_barrier, &barrier_status,
                                            &ts_latest, &latest_status, &dummy1,
                                            &ts_classify, &classify_status,
                                            &dummy2, &ts_complete);

        if (md_fid_compare (fid, fids[i]) == 0) {
            /* Check if the FID we were repairing is now complete or there is a
             * newer complete version.  If so, then restarting op is the right
             * thing to do in both cases as we may not read the repair. */
            if (md_hts_compare (ts_complete, ts_repair) >= 0) {
                *restart_op = 1;
                return PASIS_ERROR; /* no deduction */
            }
        } else {
            /* For the other FID's in the operation which occured at "repair
             * TS", if there is a complete/repairable version with a later
             * timestamp, then the "repair TS" must be complete, AS LONG as
             * there have been no barriers on the FID that caused the repair.
             * The last condition is needed because it lets us avoid having to
             * trace _through_ the conditioning chain.
             */
            /* First, we need to determine the timestamp for this FID that
             * corresponds to the repair TS.  Keep in mind that the repair TS is
             * the timestamp for the FID that caused the repair.  For the other
             * FID's involved in that operation, their timestamps will differ.*/
            int count;
            pasis_md_hts_t* corresponding_ts =
                get_corresponding_ts (current_hist_matrix, ts_repair, &count);
            if (count <= protocol->B) {
                continue; /* can't trust anything with less than B+1 */
            }

            /* If corresponding_ts is NULL, that means we didn't find a
             * corresponding timestamp in the history matrix.  This implies that
             * the corresponding timestamp has been garbage collected.  This
             * allows us to deduce completeness */
            if (corresponding_ts == NULL ||
                md_hts_compare (ts_classify, corresponding_ts) > 0) {
                pasis_md_hist_matrix_t* fid_hist_matrix;

                /* Ok we found a corresponding timestamp that lets us deduce
                 * that the FID that caused a repair is actually complete.
                 * We need to now check the "barrier condition": there can be no
                 * barrier after this FID's "repairable" timestamp, since a
                 * barrier would imply that this "deducing other timestamp" can
                 * perhaps be conditioned on _another_ chain of timestamps. */
                /* 1) Find the history matrix corresponding to this FID */
                for (j = 0; j < fid_num; ++j) {
                    if (md_fid_compare (fid, fids[j]) == 0) {
                        break;
                    }
                }
                PASIS_DASSERT (j != fid_num);
                fid_hist_matrix = &proto->write_op_args.history_matrices[j];

                /* 2) Go through history matrix, looking for barriers after this
                 * timestamp */
                for (k = 0; k < fid_hist_matrix->ts_num; ++k) {
                    cmp = md_hts_compare (&fid_hist_matrix->ts_vector[k]->ts,
                                          ts_repair);
                    if (cmp > 0 &&
                        fid_hist_matrix->ts_vector[k]->ts.barrier_flag &&
                        bitcount(fid_hist_matrix->ts_bitmap[k]) > protocol->B) {
                        /* Oh no!  We found a barrier, which means the timestamp
                         * that we were going to deduce on doesn't really prove
                         * that this "repairable" is complete.
                         * How do we handle this case???  Can't do copy-over...
                         */
                        fprintf (stderr, "%s: MAJOR ERROR.  Found barrier "
                                 "during deduction\n", __FUNCTION__);
                        PASIS_DASSERT (0);
                    }
                }

                /* 3) If this is an update operation, we need to save this proof
                 * matrix in the write_op_args for the S-N's */
                if (is_query_op == 0) {
                    printf("%s: for Update op, copying proof matrix into spot %d\n",
                           __FUNCTION__, fid_index);
                    PASIS_DASSERT (fid_index < MD_MAX_OBJS_PER_REQ);
                    /* this should only set once; initially UINT_MAX*/
                    PASIS_DASSERT (proto->write_op_args.
                                   proof_fid_indices[fid_index] == UINT_MAX);

                    proto->write_op_args.proof_fid_indices[fid_index] = i;
                    memcpy (&proto->write_op_args.proof_matrices[fid_index],
                            current_hist_matrix, sizeof(pasis_md_hist_matrix_t));
                }

                return PASIS_SUCCESS;
            }
        }
    }
    return PASIS_ERROR;
}

#endif


static pasis_error_t
repair (pasis_md_fid_t           *fid,
        p_uint32                  fid_index, /* index of FID in write op */
        pasis_md_protocol_t      *protocol,
        pasisio_request_t        *req,
        pasisio_md_proto_data_t  *proto,
        pasisio_select_t         *select,
        pasis_md_hts_t           *ts_repair,
        int                       is_query_op,
        int                       repairing_multiobject_op,
        int                      *deduced_complete) /* Out arg */
{
    unsigned int                    i;
    unsigned int                    fid_num = 0, num_issue, num_thresh;
    pasis_error_t                   error;
    pasis_md_fid_t                 *fids[MD_MAX_OBJS_PER_REQ];
    pasisio_md_hist_data_t         *hist_saved[MD_MAX_OBJS_PER_REQ];
#if 0
    unsigned int                    complete, num_issue, num_thresh;
    int                             restart_op;
#endif
    int                             can_do_fast_repair, overwrote_hist_cache = 0;
    unsigned int                    is_multiobject_op;

    int                             saved_query_op_flag = 0;
    int                             saved_update_op_flag = 0;
    pasis_md_rpc_num_t              saved_type;
    pasis_md_read_flag_t            saved_read_flags = 0;
    p_uint32                        saved_write_mtime[2];

    PASIS_TIMER_START ();
    
    /*printf ("%s: FID=%#llx, ts=%llu", __FUNCTION__,
     *       *(p_uint64 *)fid, ts_repair->time); */

/* #if DEBUG_TIMESTAMPS > 0 */
/*     printf(", op desc = %s", ts_repair->op_description); */
/* #endif */
/*     printf("\n"); */

    // haven't thought about how to repair an operation if we're currently
    // executing an update.
    //
    // if we're currently executing a query, fast repair occurs as follows:
    // 1) 1st phase optimistic query failed because a repairable exists
    // 2) our OHS is updated
    // 3) so we go back to the servers with the OHS, and they object sync,
    //    then execute our query

    PASIS_DASSERT (ts_repair->barrier_flag == 0);

    *deduced_complete = 0;

    PasisIO_GetCompleteThreshold (protocol, &num_issue, &num_thresh);

    /* Reset select list */
    select_reset (num_issue, select);

    ++stats.repairs;


    if (!repairing_multiobject_op) {
        fid_num = 1;
        fids[0] = fid;
    } else {

        fprintf(stderr, "%s: need to go read histories for all FIDs involved in "
                "the operation and do the fancy deduced complete stuff.\n",
                __FUNCTION__);
        PASIS_DASSERT (0);

        /* If this is a query op, need to save the query arguments, as we're
         * about to do a read-hist, which will overwrite the args/res pointers
         * of the existing query */
        if (is_query_op) {
            saved_query_op_flag = 1;
            save_query_op (protocol, proto, &saved_type, &saved_read_flags);
        }

#if 0
        /* Store a copy of the history cache elements until repair is done */
        overwrote_hist_cache = 1;
        memcpy (hist_saved, proto->fid_hist_cache, MD_MAX_OBJS_PER_REQ * 
                sizeof (pasisio_md_hist_data_t *));

        /* Set up history cache */
        for (i = 0; i < fid_num; i++) {
            proto->fid_hist_cache[i] = PasisIO_MDHistCacheGetEmpty (fids[i]);
            PasisIO_MDHistCachePin (proto->fid_hist_cache[i]);
        }

        // 1) read which objects are involved in this operation from the workers.
        //    (get their history along)

        // fid_num and fids is read from the servers

        // do deduction by completness

        /* Check if any of the _other_ FID's have a complete/repairable operation.
         * If so, then we can _deduce_ that this version is actually complete.
         * If we've deduced that this version is actually complete:
         * a) for query operations, the caller operation can just use the results
         *    for this version
         * b) for update operations, the caller operation can give this "deducing
         *    matrix to the S-N's to cause object syncing _just_ for the FID's that
         *    are out of sync
         * Jay thinks we have to follow the condition-on chain to deduce that this
         * version is complete.  Mike thinks that as long as there are no barriers
         * for this FID, then this version must be complete if there are future
         * versions which are repairable/complete and depend on this one.
         * We're implementing Mike's reasoning since it's simpler :) (Jay's requires
         * that we have all the past versions available to us.)
         * -mabdelmalek 10/26/2004. */
        /* Update: Jay and I agree that he's right.  Will need some tricks with
         * GC to ensure this history is available... -mabdelmalek 8/1/2005 */
        if (can_deduce_complete (protocol, is_query_op, proto, fid_index, fid, fid_num,
                                 fids, ts_repair, &restart_op) == PASIS_SUCCESS) {
            if (is_query_op) {
                /* return and signal caller that they can use the value from
                 * repair_ts */
                printf("%s: Q op deducing that the repairable FID %llx is actually complete.\n",
                       __FUNCTION__, *(p_uint64*)fid);
                *deduced_complete = 1;
                ++stats.complete_deductions;
                error = PASIS_SUCCESS;
                goto exit;
            } else {
                /* can_deduce_complete has copied the deductive matrix to the
                 * write_op_args.  The S-N's will need this for their valdiation.
                 * Now, we return, signalling the caller that they can condition on
                 * the repair-ts */
                printf("%s: U op deducing that the repairable FID %llx is actually complete.\n",
                       __FUNCTION__, *(p_uint64*)fid);
                *deduced_complete = 1;
                ++stats.complete_deductions;
                error = PASIS_SUCCESS;
                goto exit;
            }
        }

        /* The above function also performed classification again, with the new
         * history that we've read.  (The classification occured for free as part of
         * the above logic.)  If the reclassification found 1) that the "repairable"
         * timestamp is now complete, or 2) that there is a new complete/repairable
         * candidate that's newer than the TS we were repairing, then we restart op*/
        if (restart_op) {
            printf("%s: restarting op because read-op call in repair found that the "
                   "1) repairable TS is now complete, or 2) there is a newer candidate "
                   "than the one that we were trying to repair\n", __FUNCTION__);
            error = PASIS_SUCCESS;
            ++stats.restarted_repairs;
            goto exit;
        }
#endif
    }

    /* Go through read history matrices and see if we can perform fast repair */
    can_do_fast_repair = 1;
    for (i = 0; i < fid_num; i++) {
        unsigned int       barrier_flag, repair_flag, fast_repair_flag;
        pasis_md_hts_t    *ts_tmp;

        validate_hist_matrix (protocol, 
                              &proto->write_op_args.history_matrices[i],
                              &ts_tmp, &barrier_flag, &repair_flag,
                              &fast_repair_flag, &is_multiobject_op);

        PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, 
                       "%s: classification: ts=%llu, barrier=%u, repair=%u\n",
                       __FUNCTION__, ts_tmp->time, barrier_flag, repair_flag);

        if (fast_repair_flag == 0) {
            can_do_fast_repair = 0;
            break;
        }
    }

    if (can_do_fast_repair) {
        ++stats.fast_repairs;
        //printf ("%s: can do fast repair!\n", __FUNCTION__);
        if (STRESS_TESTING) {
            printf("%s: matrix:\n", __FUNCTION__);
            pasis_md_hist_matrix_dump (&proto->write_op_args.history_matrices[0]);
        }
        error = PASIS_SUCCESS;
    } else {
        ++stats.slow_repairs;

        /* If this is an update op, need to save the update arguments, as we're
         * about to do a write-barrier, which will overwrite the args/res pointers
         * of the existing update */
        if (!is_query_op) {
            saved_update_op_flag = 1;
            save_update_op (protocol, proto, &saved_type, saved_write_mtime);
        }

#if 1
        /* I don't know if we're correctly saving/restoring
         * history. -mabdelmalek 8/1/2005 */
#else
        /* Set up history cache */
        overwrote_hist_cache = 1;
        memcpy (hist_saved, proto->fid_hist_cache, MD_MAX_OBJS_PER_REQ * 
                sizeof (pasisio_md_hist_data_t *));
        for (i = 0; i < fid_num; i++) {
            proto->fid_hist_cache[i] = PasisIO_MDHistCacheGetEmpty (fids[i]);
            PasisIO_MDHistCachePin (proto->fid_hist_cache[i]);
        }
#endif

        //printf("%s: doing slow repair; writing barrier\n", __FUNCTION__);

        /* Write barrier */
        /* The resultant history vectors (returned from the barrier write) are
         * merged with the current history matrix -- in the cache, and can thus be
         * used in the repair operation.  This is necessary to ensure that the
         * history matrix that shows both the repairable and the barrier, even
         * though a complete barrier has been written (and a repairable exists). */
        error = write_barrier (fids, fid_num, protocol, req, proto, select, 0);

        /* restore write op arguments/results right away */
        if (saved_update_op_flag) {
            restore_update_op (protocol, proto, saved_type, saved_write_mtime);
            saved_update_op_flag = 0;
        }

        if (error) {
            /* Barrier write did not complete, probably due to concurrency. 
             * Retry op.  Caller will backoff, if necessary (for queries, no
             * need to backoff). */
            /* printf ("%s: barrier did not complete\n", __FUNCTION__); */
            error = PASIS_ERROR_PROTO_RETRY;
            goto exit;
        }

        /* Go through read history matrices, perform classification, make sure
         * all FIDs in the repairable operation are still repairable w/the same
         * timestamp.
         */
        for (i = 0; i < fid_num; i++) {
            unsigned int       barrier_flag, repair_flag, fast_repair_flag, dum;
            pasis_md_hts_t    *ts_tmp;

            /* Assumes set_hist_matrices has been previously called */
            validate_hist_matrix (protocol, 
                                  &proto->write_op_args.history_matrices[i],
                                  &ts_tmp, &barrier_flag, &repair_flag,
                                  &fast_repair_flag, &dum);

            PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, 
                           "%s: classification: ts=%llu, barrier=%u, repair=%u\n",
                           __FUNCTION__, ts_tmp->time, barrier_flag, repair_flag);

            if (barrier_flag) {
                /* Still need a barrier???  Probably need to retry op */
                /* printf ("%s: barrier flag is now set\n", __FUNCTION__); */
                error = PASIS_ERROR_PROTO_RETRY;
                goto exit;
            }

            if (repair_flag == 0 && (md_hts_compare (ts_tmp, &zero_ts) != 0)) {
                /* Some part of the op is not repairable, it is either incomplete
                 * or complete --
                 *   1) if complete, it may have originally been complete but
                 *   just viewed as repairable, or another client has repaired
                 *   this object.
                 *   2) if incomplete, or it may have been overwritten with a
                 *   new value, meaning it was really previously incomplete.
                 * Either way, we can now retry the op with this history matrix.
                 */
                /* XXX return something suitable -- this is tricky... */
                /* printf ("%s: not classified repairable any longer (new_ts=%llu)\n",
                 *       __FUNCTION__, ts_tmp->time); */
                error = PASIS_ERROR_PROTO_RETRY;
                goto exit;
            }

            /* Pertains only to original ts */
            if ((memcmp (fids[i], fid, PASISIO_FILEID_SIZE) == 0) &&
                (md_hts_compare (ts_tmp, ts_repair) != 0)) {
                /* Something changed since last time -- new repairable??? Retry */
#if 0
                printf ("%s: classification ts changed (new_ts=%llu)\n", 
                        __FUNCTION__, ts_tmp->time);
                printf("%s: matrix:\n", __FUNCTION__);
                pasis_md_hist_matrix_dump (&proto->write_op_args.history_matrices[i]);
#endif
                error = PASIS_ERROR_PROTO_RETRY;
                goto exit;
            }
        }
    }

    /* for query operations, the caller will reissue the query with the OHS,
     * which will cause the servers to perform repair.
     * for updates, we issue those operations here.  the servers won't actually
     * perform the update -- they'll just do whatever is necessary for repair.
     * -mabdelmalek 22/3/2005. */
    if (!is_query_op) {
        select_reset (num_issue, select);
        /* "Backup" the authenticators so requests to servers don't get overwritten
         * with authenticators from new responses  */
        save_authenticators (protocol, proto,
                             proto->write_op_args.history_matrices_num,
                             proto->fid_hist_cache);

        //printf ("%s: issuing update request so servers can do copy-forward\n", 
        //       __FUNCTION__);
        /* Issue request */
#if USE_EARLY_TERMINATION > 0
        protocol_issue_reqs (req, PASISIO_MD_DRIVE_P_WRITE_OP,
                             num_issue,
                             num_thresh,  /* Min */
                             num_thresh); /* Early term */
#else
        protocol_issue_reqs (req, PASISIO_MD_DRIVE_P_WRITE_OP,
                             protocol->N,
                             protocol->N, protocol->N);
#endif

        /* Update the history matrix in the write_op */
        set_hist_matrices (fid_num, &proto->write_op_args, proto->fid_hist_cache);

        //pasis_md_hist_matrix_dump (&proto->write_op_args.history_matrices[0]);

        /* Caller will reclassify and see the status */
    }


    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, 
                   "%s: issuing repair of op\n", __FUNCTION__);

 exit:
    select_reset (num_issue, select);

    /* Unpin history cache objects */
    if (overwrote_hist_cache) {
        for (i = 0; i < fid_num; i++) {
            PasisIO_MDHistCacheUnpin (proto->fid_hist_cache[i]);
        }

        /* Restore history cache objects that were saved */
        memcpy (proto->fid_hist_cache, hist_saved, MD_MAX_OBJS_PER_REQ *
                sizeof (pasisio_md_hist_data_t *));
    }

    /* Restore arguments/results */
    if (saved_query_op_flag) {
        restore_query_op (protocol, proto, saved_type, saved_read_flags);
    }

    if (is_query_op) {
        // get servers to object-sync (during repair) by giving them an OHS

        PASIS_DASSERT (repairing_multiobject_op == 0); // XXX, for now

        PASIS_DASSERT (fid_num == 1);
        proto->read_op_args.history_matrices_num = 1;
        PasisIO_MDHistCacheSetMatrix (proto->fid_hist_cache[0],
                                      proto->read_op_args.history_matrices);
        /* Set/"backup" the authenticators */
        save_authenticators (protocol, proto, 1, proto->fid_hist_cache);

        //printf("%s: at end of repair, history matrix is:\n", __FUNCTION__);
        //pasis_md_hist_matrix_dump (proto->read_op_args.history_matrices);
    }

    PASIS_TIMER_STOP (stats.timers, MD_TIMER_REPAIR);

    //printf ("%s: repair done (error=%u)\n", __FUNCTION__, error);
    return error;
}


/* Write operations:
 * 1) Read timestamp histories.
 * 2) Perform read classification -- need only check for repairables
 * 3) If repairable, perform repair
 * 4) If complete, issue operation
 * 5) Count successes to see if operation completes
 */

static pasis_error_t
read_histories (pasis_md_protocol_t      *protocol,
                pasisio_request_t        *req,
                pasisio_md_proto_data_t  *proto,
                pasisio_select_t         *select,
                unsigned int              fid_num,
                pasis_md_fid_t           *fids[MD_MAX_OBJS_PER_REQ],
                unsigned int              use_cached_data)
{
    pasis_error_t    error = PASIS_SUCCESS;
    unsigned int     complete, cache_hits, i, num_issue, num_thresh;

    PASIS_TIMER_START ();

    /* First check hist cache objects -- if use_cached_data is set */
    if (use_cached_data) {
        cache_hits = 0;
        for (i = 0; i < fid_num; i++) {
            PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "%s: ts_num=%u\n", 
                           __FUNCTION__, proto->fid_hist_cache[i]->ts_num);
            if (proto->fid_hist_cache[i]->ts_num > 0) {
                cache_hits++;
            }
        }
        
        PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "%s: hits=%u, fids=%u\n", 
                       __FUNCTION__, cache_hits, fid_num);
        
        if (cache_hits == fid_num) {
            PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "%s: history cache hit\n",
                           __FUNCTION__);
            PASIS_TIMER_STOP (stats.timers, MD_TIMER_READ_HISTORY);
            return PASIS_SUCCESS;
        }
    }

    /* Setup read operation -- Read History */
    proto->read_op_args.op_type = PASIS_MD_READ_HISTORY;
    proto->read_op_args.read_flags = 0;
    proto->read_op_args.op_args.read_history.fid_num = fid_num;
    proto->read_op_args.history_matrices_num = 0;

    for (i = 0; i < fid_num; i++) {
        proto->read_op_args.op_args.read_history.fid[i] = fids[i];
    }

    PasisIO_GetCompleteThreshold (protocol, &num_issue, &num_thresh);

    select_reset (num_issue, select);

#if USE_EARLY_TERMINATION > 0
    complete = protocol_issue_reqs (req, PASISIO_MD_DRIVE_P_READ_OP,
                                    num_issue,
                                    num_thresh,  /* Min */
                                    num_thresh); /* Early term */
#else
    complete = protocol_issue_reqs (req, PASISIO_MD_DRIVE_P_READ_OP,
                                    protocol->N,
                                    protocol->N, protocol->N);
#endif

    /* Check complete */
    if (complete < protocol->Qc + protocol->B) {
        PASIS_PRINT_ERR ("\n%s: failed (%d)\n", __FUNCTION__, error);
        if (STRESS_TESTING == 0) { /* we want to allow incompletes during testing*/
            PASIS_DASSERT (0);
        }
    }

    select_reset (num_issue, select);

    PASIS_TIMER_STOP (stats.timers, MD_TIMER_READ_HISTORY);

    PASIS_DASSERT (error == 0);

    return error;
}

static inline unsigned int
write_res_count_inval_history (pasisio_md_proto_data_t  *proto,
                               pasisio_select_t         *select)
{
    unsigned int i, count=0;

    /* Check error return codes, count number of invalid history responses */
    for (i = 0; i < select->count; i++) {
        if (select->flags[i] == PASISIO_SELECT_IO_SUCCESS && 
            (proto->write_op_res[i].status == PASIS_MD_ERR_HIST_NOT_LATEST ||
             proto->write_op_res[i].status == PASIS_MD_ERR_OBJSYNC_NOT_EXIST ||
             proto->write_op_res[i].status == PASIS_MD_ERR_HIST_BARRIER)) {
            /* No need to do more, history cache is updated in md drive */
            count++;
        }
    }

    return count;
}


static pasis_error_t
write_classify_history (pasis_md_protocol_t      *protocol,
                        pasisio_request_t        *req,
                        pasisio_md_proto_data_t  *proto,
                        pasisio_select_t         *select,
                        unsigned int              fid_num,
                        pasis_md_fid_t           *fids[MD_MAX_OBJS_PER_REQ])
{
    pasis_error_t    error;
    unsigned int     barrier_flag, repair_flag, fast_repair_flag;
    unsigned int     is_multiobject_op = 0;
    int              i, backoff_flag, restart_count;
    int              fid_indices_to_ignore[MD_MAX_OBJS_PER_REQ];

    memset (fid_indices_to_ignore, 0, sizeof(int)*MD_MAX_OBJS_PER_REQ);

    backoff_flag = 0;
    restart_count = 0;

 restart:

    if (backoff_flag) {
        /* PASIS_DASSERT (restart_count < 50); */
        backoff (restart_count);
    }

    /* Fill in history matrices from history cache */
    set_hist_matrices (fid_num, &proto->write_op_args, proto->fid_hist_cache);

    for (i = 0; i < fid_num; i++) {
/*         printf("%s: history matrix:\n", __FUNCTION__); */
/*         pasis_md_hist_matrix_dump (&proto->write_op_args.history_matrices[i]); */

        if (fid_indices_to_ignore[i]) {
            continue;
        }

        validate_hist_matrix (protocol, 
                              &proto->write_op_args.history_matrices[i],
                              &proto->ts_classify, &barrier_flag, &repair_flag,
                              &fast_repair_flag, &is_multiobject_op);
        
        /* 3 Cases at this stage:
         * 1) complete -- new write can proceed
         * 2) barrier needed -- may or may not require repair
         * 3) repair needed w/o barrier
         */
        

        /* Do repair -- will write a barrier if necessary */
        if (repair_flag) {
            int deduced_complete;

            //printf ("%s: doing repair for FID=%#llx (index %d, "
            //        "latest=%llu, classify=%llu)\n", __FUNCTION__,
            //        *(p_uint64 *)fids[i], i,
            //        proto->write_op_args.history_matrices[i].
            //        ts_vector[0]->time, proto->ts_classify->time);
/*             pasis_md_hist_matrix_dump (&proto->write_op_args.history_matrices[i]); */

            error = repair (fids[i], i, protocol, req, proto, select, 
                            proto->ts_classify, 0, is_multiobject_op,
                            &deduced_complete);

            if (deduced_complete) {
                /* We've deduced that this "repairable" timestamp is in fact
                 * complete.  repair() has put the "proof" matrix in the right
                 * spot in the write op args. Now, continue validation */
                // XXX: jump to restart so we refill the history matrix
                printf("%s: deduced that FID %llx is complete\n", __FUNCTION__,
                       *(p_uint64*)fids[i]);
                fid_indices_to_ignore[i] = 1;
                continue;
            } else if (error == PASIS_ERROR_PROTO_RETRY) {
                /* printf("%s: repair failed; backing off\n", __FUNCTION__); */
                restart_count++;
                backoff_flag = 1;
            } else {
                PASIS_DASSERT (error == PASIS_SUCCESS);
                // if error = success, we won't perform backup at restart:
                backoff_flag = 0;
            }
            
            /* After repair is done, reclassify objects and see if repair or
             * barriers are needed again.  If not, we'll return with new
             * history. */
            /* printf("%s: before second goto restart: %u\n", __FUNCTION__,
               restart_count); */
            goto restart;

        } else if (barrier_flag) {  /* Need a barrier */
            /* Could optimize to issue one barrier request for all fids... */
            pasisio_md_hist_data_t  *fid_cache_state[MD_MAX_OBJS_PER_REQ];

/*             printf ("%s: doing barrier for FID=%#llx (matrix %d, " */
/*                     "latest_ts=%llu, classify_ts=%llu)\n", __FUNCTION__, */
/*                     *(p_uint64 *)fids[i], i, */
/*                     proto->write_op_args.history_matrices[i]. */
/*                     ts_vector[0]->ts.time, proto->ts_classify->time); */
/*             printf ("%s: matrix %d:\n", __FUNCTION__, i); */
/*             pasis_md_hist_matrix_dump (&proto->write_op_args.history_matrices[i]); */
/*             printf ("\n"); */
            

            /* Save history cache pointers */
            memcpy (fid_cache_state, proto->fid_hist_cache, 
                    fid_num * sizeof (pasisio_md_hist_data_t *));
            
            proto->fid_hist_cache[0] = proto->fid_hist_cache[i];

            proto->write_op_args.history_matrices_num = 1;
            PasisIO_MDHistCacheSetMatrix (proto->fid_hist_cache[0],
                                          &proto->write_op_args.
                                          history_matrices[0]);
            
            /* Write a barrier */
            error = write_barrier (&fids[i], 1, protocol, req, proto, select, 1);

            if (error == PASIS_ERROR_PROTO_RETRY) {
                /* printf("%s: barrier write failed; backing off\n",
                   __FUNCTION__); */
                restart_count++;
                backoff_flag = 1;
            } else {
                PASIS_DASSERT (error == PASIS_SUCCESS);
                // if error = success, we won't perform backup at restart:
                backoff_flag = 0;
            }

            /* Restore history cache pointers */
            memcpy (proto->fid_hist_cache, fid_cache_state, 
                    fid_num * sizeof (pasisio_md_hist_data_t *));
            
            /* After barrier is done, reclassify objects and see if repair or
             * barriers are needed again.  If not, we'll return with new
             * history. */
            /* printf("%s: wrote barrier: error = %u; restart count = %u\n",
               __FUNCTION__, error, restart_count); */
            goto restart;
        }
    }

    return PASIS_SUCCESS;
}


static pasis_error_t
write_read_history (pasis_md_protocol_t      *protocol,
                    pasisio_request_t        *req,
                    pasisio_md_proto_data_t  *proto,
                    pasisio_select_t         *select,
                    unsigned int              fid_num,
                    pasis_md_fid_t           *fids[MD_MAX_OBJS_PER_REQ],
                    unsigned int              use_cached_data_flag)
{
    pasis_error_t    error;

    error = read_histories (protocol, req, proto, select, fid_num, fids, 
                            use_cached_data_flag);
    if (error) return error;
    
    /* Classify history. Perform repair/barrier-write if necessary. This
     * function will keep retrying in order to make sure that the candidate we
     * condition on is complete. */
    return write_classify_history (protocol, req, proto, select, fid_num, fids);
}


static pasis_error_t
write_op_dispatch (pasis_md_protocol_t      *protocol,
                   pasisio_request_t        *req,
                   pasisio_md_proto_data_t  *proto,
                   pasisio_select_t         *select,
                   unsigned int              fid_num,
                   pasis_md_fid_t           *fids[MD_MAX_OBJS_PER_REQ],
                   unsigned int              op_res_size,
                   void                    **res_out)
{
    unsigned int complete, count, restart_count, num_issue, num_thresh, first_time;
    pasis_error_t error;

    PasisIO_GetCompleteThreshold (protocol, &num_issue, &num_thresh);

    first_time = 1;
    restart_count = 0;

 restart:

    if (first_time) {
        first_time = 0;
    } else {
        /* PASIS_DASSERT (restart_count < 50); */
        /* I commented out the backoff() call here. Here's why: imagine two
         * clients updating the same object. Now, say client A does an update,
         * and the servers respond with an error because client A's OHS is out
         * of date.  The servers will respond with the current replica
         * histories.  But if A backs-off and retries the operation, then surely
         * his OHS will be out of date, because the second client is performing
         * the updates.  So we have to be careful in where we do _not_ back off.
         * We'll still leave repair() doing back-off, if repair or
         * barrier-write fail.  -mabdelmalek 8/2/2005. */
        /* backoff (restart_count); */
    }

    select_reset (num_issue, select);

    /* "Backup" the authenticators so requests to servers don't get overwritten
     * with authenticators from new responses  */
    save_authenticators (protocol, proto,
                         proto->write_op_args.history_matrices_num,
                         proto->fid_hist_cache);

    /* Issue request */
#if USE_EARLY_TERMINATION > 0
    complete = protocol_issue_reqs (req, PASISIO_MD_DRIVE_P_WRITE_OP,
                                    num_issue,
                                    num_thresh,  /* Min */
                                    num_thresh); /* Early term */
#else
    complete = protocol_issue_reqs (req, PASISIO_MD_DRIVE_P_WRITE_OP,
                                    protocol->N,
                                    protocol->N, protocol->N);
#endif

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, 
                   "%s: num_issue=%u, num_thresh=%u, complete=%u\n",
                   __FUNCTION__, num_issue, num_thresh, complete);

    if (STRESS_TESTING == 0) {
        PASIS_DASSERT (complete >= protocol->Qc + protocol->B);
    }

    count = write_op_count_successes (select, proto->write_op_res);

    if (count < protocol->Qc + protocol->B) {
        unsigned int inval_count;

        if (protocol->quorum_type == PASISIO_QUORUM_TREE) {
            /* have not implemented repair for tree quorums */
            /* printf("%s: count = %d, protocol->Qc = %d, protocol->B = %d\n",
               __FUNCTION__, count, protocol->Qc, protocol->B); */
            PASIS_DASSERT (0);
        }

        /* Check why this failed, if due to invalid history and/or repair is
         * necessary, restart op.  New history vectors accompany responses. */
        inval_count = write_res_count_inval_history (proto, select);
        if (inval_count + count >= protocol->Qc + protocol->B) {

            /* Classify history.  Perform repair/barrier-write if necessary.
             * This function will keep retrying in order to make sure that the
             * candidate we condition on is complete.
             * This function will also perform back-off. */
            write_classify_history (protocol, req, proto, select, fid_num, fids);


            /* PASIS_PRINT_INFO ("%s: restarting op due to 1) out-of-date history, "
             *                 "or 2) repair necessary; restart count =  %d\n",
             *                 __FUNCTION__, restart_count); */

            stats.write_op_hist_inval++;

            ++restart_count;

            /* PASIS_PRINT_INFO (" - op_type=%u, latest ts=%llu\n",
             *                 proto->write_op_args.op_type,
             *                 proto->write_op_args.history_matrices[0].
             *                 ts_vector[0]->ts.time); */

            /* PASIS_DASSERT (restart_count < 100);  */

            goto restart;
        }

#if 1
        // I don't know why the (usually) 5th error isn't INVAL_HISTORY; oh
        // well, just retry again. -mabdelmamlek 8/10/2005
        write_classify_history (protocol, req, proto, select, fid_num, fids);
        stats.write_op_hist_inval++;
        ++restart_count;
        goto restart;
#else
        if (STRESS_TESTING == 0) {
            PASIS_PRINT_ERR ("\n%s: failed (count=%d, inval_count=%d)\n", 
                             __FUNCTION__, count, inval_count);
            PASIS_DASSERT (0);
        }
#endif
        return PASIS_ERROR;
    }

    /* Vote across the results, looking for B + 1 matching */
    *res_out = write_op_res_vote (select, proto->write_op_res, op_res_size,
                                  protocol->B);
    if (*res_out == NULL) {
        PASIS_DASSERT (0);
    }

    /* Do this before select_reset, since select_reset resets drive statuses */
    error = write_op_rpc_op_stat_vote (select, proto->write_op_res,
                                       protocol->B); 

    select_reset (num_issue, select);
    PASIS_DASSERT (error==0);
    return error;
}


/******************* Write operations start here ********************/


pasis_error_t
PasisIO_MDCreate (pasis_md_protocol_t *protocol,
                  pasis_md_linkage_t  *linkages,
                  unsigned int         num_linkages,
                  pasis_md_fid_t      *parent_fid,
                  pasis_md_fid_t      *name_fid,
                  char                *name,
                  char                *symlink_name,
                  pasis_md_attr_t     *name_attr_in,
                  pasis_md_protocol_t *name_protocol,
                  pasis_md_linkage_t  *name_linkages,
                  unsigned int         num_name_linkages,
                  pasis_md_attr_t     *parent_attr_out,    /* Out args */
                  pasis_md_attr_t     *name_attr_out)      /* Out args */
{
    pasis_error_t              error;
    pasisio_request_t          req;
    pasisio_select_t           select_out;
    pasisio_md_proto_data_t   *proto;
    pasisio_drive_instance_t  *out_list[PASIS_MAX_NUM_SERVERS];
    pasis_md_create_res_t     *res;
    pasis_md_fid_t            *fids[MD_MAX_OBJS_PER_REQ];
    unsigned int               shuffle;

    PASIS_TIMER_START ();

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "\n%s: \n", __FUNCTION__); 

    // For now, make sure we write- and read-all, to ensure repair doesn't
    // happen. -mabdelmalek 8/10/2005
    PASIS_DASSERT(protocol->quorum_type == PASISIO_QUORUM_NONE);

    select_out.instances = out_list;
    proto = PasisFreelist_Get (proto_freelist);
    proto->already_saved_write_op = 0;
    proto->chunk_size = 0;
    memset (proto->write_op_args.proof_fid_indices, UCHAR_MAX,
            sizeof(p_uint32) * MD_MAX_OBJS_PER_REQ);

    shuffle = *(unsigned int*) name_fid;
    op_init (protocol, linkages, num_linkages, proto, &select_out, &req,
             shuffle);

    fids[0] = parent_fid;
    fids[1] = name_fid;

    /* Init history cache objects */
    /* Do a lookup, or get empty if lookup fails, followed by a pin */
    proto->fid_hist_cache[0] = PasisIO_MDHistCacheGetEmpty (fids[0]);
    PasisIO_MDHistCachePin (proto->fid_hist_cache[0]);
    proto->fid_hist_cache[1] = PasisIO_MDHistCacheGetEmpty (fids[1]);
    PasisIO_MDHistCachePin (proto->fid_hist_cache[1]);

    /* Write operation -- Create */
    proto->write_op_args.op_type = PASIS_MD_CREATE;
    proto->write_op_args.repair_flag = 0;

    pasis_md_attr_set_time (proto->write_op_args.mtime);
    proto->write_op_args.op_args.create.parent_dir_fid = parent_fid;
    proto->write_op_args.op_args.create.create_fid = name_fid;
    proto->write_op_args.op_args.create.create_name = name;
    proto->write_op_args.op_args.create.symlink_name = symlink_name;
    memcpy (&proto->write_op_args.op_args.create.create_attr, 
            name_attr_in, sizeof (pasis_md_attr_t));
    proto->write_op_args.op_args.create.client_id = rpc_get_host_ip();

    /* Fill in distribution */
    memcpy(&proto->write_op_args.op_args.create.distribution.protocol,
           name_protocol, sizeof(pasis_md_protocol_t));

    /* For now -- copy the first N linkages */
    /* If we want to have quorum support for data, we'd modify the following
     * step to do quorum-specific linkage-selection, just like op_init does. */
    memcpy (&proto->write_op_args.op_args.create.distribution.linkages,
            name_linkages, name_protocol->N * sizeof (pasis_md_linkage_t));

    error = write_read_history (protocol, &req, proto, &select_out, 
                                2, fids, 1);
    if (error) goto exit;

    /* Dispatch write operation */
    error = write_op_dispatch (protocol, &req, proto, &select_out, 2, fids,
                               sizeof (pasis_md_create_res_t), (void **)&res);
    if (error) goto exit;

    /* Copy results */
    memcpy (parent_attr_out, &res->parent_attr, sizeof (pasis_md_attr_t));
    memcpy (name_attr_out, &res->create_attr, sizeof (pasis_md_attr_t));

 exit:
    PasisIO_MDHistCacheUnpin (proto->fid_hist_cache[0]);
    PasisIO_MDHistCacheUnpin (proto->fid_hist_cache[1]);
    PasisFreelist_Put (proto_freelist, proto);

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "OUT %s: error=%u", 
                   __FUNCTION__, error);

    PASIS_TIMER_STOP (stats.timers, MD_TIMER_CREATE);

    return error;
}

pasis_error_t
PasisIO_MDLink (pasis_md_protocol_t *protocol,
                pasis_md_linkage_t  *linkages,
                unsigned int         num_linkages,
                pasis_md_fid_t      *from_fid,
                pasis_md_fid_t      *to_dir_fid,
                char                *to_name,
                pasis_md_attr_t     *from_attr_out,       /* Out args */
                pasis_md_attr_t     *to_dir_attr_out)     /* Out args */
{
    pasis_error_t              error;
    pasisio_request_t          req;
    pasisio_select_t           select_out;
    pasisio_md_proto_data_t   *proto;
    pasisio_drive_instance_t  *out_list[PASIS_MAX_NUM_SERVERS];
    pasis_md_link_res_t       *res;
    pasis_md_fid_t            *fids[MD_MAX_OBJS_PER_REQ];
    int                        i;
    unsigned int               shuffle;

    PASIS_TIMER_START ();

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "\n%s: \n", __FUNCTION__); 

    select_out.instances = out_list;
    proto = PasisFreelist_Get (proto_freelist);
    proto->already_saved_write_op = 0;
    proto->chunk_size = 0;
    memset (proto->write_op_args.proof_fid_indices, UCHAR_MAX,
            sizeof(p_uint32) * MD_MAX_OBJS_PER_REQ);

    shuffle = *(unsigned int*) from_fid;
    op_init (protocol, linkages, num_linkages, proto, &select_out, &req,
             shuffle);

    fids[0] = from_fid;
    fids[1] = to_dir_fid;

    /* Init history cache objects */
    for (i = 0; i < 2; i++) {
        /* Do a lookup, or get empty if lookup fails, followed by a pin */
        proto->fid_hist_cache[i] = PasisIO_MDHistCacheGetEmpty (fids[i]);
        PasisIO_MDHistCachePin (proto->fid_hist_cache[i]);
    }

    /* Write operation -- Link */
    proto->write_op_args.op_type = PASIS_MD_LINK;
    proto->write_op_args.repair_flag = 0;

    pasis_md_attr_set_time (proto->write_op_args.mtime);
    proto->write_op_args.op_args.link.from_file_fid = from_fid;
    proto->write_op_args.op_args.link.to_dir_fid = to_dir_fid;
    proto->write_op_args.op_args.link.to_name = to_name;
    proto->write_op_args.op_args.link.client_id = rpc_get_host_ip();

    error = write_read_history (protocol, &req, proto, &select_out, 
                                2, fids, 1);
    if (error) goto exit;

    /* Dispatch write operation */
    error = write_op_dispatch (protocol, &req, proto, &select_out, 2, fids,
                               sizeof (pasis_md_link_res_t), (void **)&res);
    if (error) goto exit;

    /* Copy results */
    memcpy (from_attr_out, &res->from_file_attr, sizeof (pasis_md_attr_t));
    memcpy (to_dir_attr_out, &res->to_dir_attr, sizeof (pasis_md_attr_t));

 exit:

    for (i = 0; i < 2; i++) {
        PasisIO_MDHistCacheUnpin (proto->fid_hist_cache[i]);
    }

    PasisFreelist_Put (proto_freelist, proto);

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "OUT %s: error=%u", 
                   __FUNCTION__, error);

    PASIS_TIMER_STOP (stats.timers, MD_TIMER_LINK);

    return error;
}


pasis_error_t
PasisIO_MDUnlink (pasis_md_protocol_t *protocol,
                  pasis_md_linkage_t  *linkages,
                  unsigned int         num_linkages,
                  pasis_md_fid_t      *parent_fid,
                  pasis_md_fid_t      *from_fid,
                  char                *from_name,
                  pasis_md_attr_t     *parent_attr_out)   /* Out args */
{
    pasis_error_t              error;
    pasisio_request_t          req;
    pasisio_select_t           select_out;
    pasisio_md_proto_data_t   *proto;
    pasisio_drive_instance_t  *out_list[PASIS_MAX_NUM_SERVERS];
    pasis_md_unlink_res_t     *res;
    pasis_md_fid_t            *fids[MD_MAX_OBJS_PER_REQ];
    int                        i;
    unsigned int               shuffle;

    PASIS_TIMER_START ();

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "\n%s: \n", __FUNCTION__); 

    select_out.instances = out_list;
    proto = PasisFreelist_Get (proto_freelist);
    proto->already_saved_write_op = 0;
    proto->chunk_size = 0;
    memset (proto->write_op_args.proof_fid_indices, UCHAR_MAX,
            sizeof(p_uint32) * MD_MAX_OBJS_PER_REQ);
    
    shuffle = *(unsigned int*) parent_fid;
    op_init (protocol, linkages, num_linkages, proto, &select_out, &req,
             shuffle);

    fids[0] = parent_fid;
    fids[1] = from_fid;

    /* Init history cache objects */
    for (i = 0; i < 2; i++) {
        /* Do a lookup, or get empty if lookup fails, followed by a pin */
        proto->fid_hist_cache[i] = PasisIO_MDHistCacheGetEmpty (fids[i]);
        PasisIO_MDHistCachePin (proto->fid_hist_cache[i]);
    }

    /* Write operation -- Unlink */
    proto->write_op_args.op_type = PASIS_MD_UNLINK;
    proto->write_op_args.repair_flag = 0;

    pasis_md_attr_set_time (proto->write_op_args.mtime);
    proto->write_op_args.op_args.unlink.parent_dir_fid = parent_fid;
    proto->write_op_args.op_args.unlink.name_fid = from_fid;
    proto->write_op_args.op_args.unlink.name = from_name;
    proto->write_op_args.op_args.unlink.client_id = rpc_get_host_ip();

    error = write_read_history (protocol, &req, proto, &select_out, 
                                2, fids, 1);
    if (error) goto exit;

    /* Dispatch write operation */
    error = write_op_dispatch (protocol, &req, proto, &select_out, 2, fids, 
                               sizeof (pasis_md_unlink_res_t), (void **)&res);
    if (error) goto exit;

    /* Copy results */
    memcpy (parent_attr_out, &res->parent_dir_attr, sizeof (pasis_md_attr_t));


 exit:
    for (i = 0; i < 2; i++) {
        PasisIO_MDHistCacheUnpin (proto->fid_hist_cache[i]);
    }

    PasisFreelist_Put (proto_freelist, proto);

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "OUT %s: error=%u", 
                   __FUNCTION__, error);

    PASIS_TIMER_STOP (stats.timers, MD_TIMER_UNLINK);

    return error;
}


pasis_error_t
PasisIO_MDSetattr (pasis_md_protocol_t *protocol,
                   pasis_md_linkage_t  *linkages,
                   unsigned int         num_linkages,
                   pasis_md_fid_t      *fid,
                   pasis_md_sattr_t    *sattr,
                   p_uint32             guard_ctime[2],
                   pasis_md_attr_t     *attr_out)         /* Out args */
{
    pasis_error_t              error;
    pasisio_request_t          req;
    pasisio_select_t           select_out;
    pasisio_md_proto_data_t   *proto;
    pasisio_drive_instance_t  *out_list[PASIS_MAX_NUM_SERVERS];
    pasis_md_setattr_res_t    *res;
    pasis_md_fid_t            *fids[MD_MAX_OBJS_PER_REQ];
    unsigned int               shuffle;

    PASIS_TIMER_START ();

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "\n%s: \n", __FUNCTION__); 

    select_out.instances = out_list;
    proto = PasisFreelist_Get (proto_freelist);
    proto->already_saved_write_op = 0;
    proto->chunk_size = 0;
    memset (proto->write_op_args.proof_fid_indices, UCHAR_MAX,
            sizeof(p_uint32) * MD_MAX_OBJS_PER_REQ);

    shuffle = *(unsigned int*) fid;
    op_init (protocol, linkages, num_linkages, proto, &select_out, &req,
             shuffle);

    fids[0] = fid;

    /* Init history cache objects */
    /* Do a lookup, or get empty if lookup fails, followed by a pin */
    proto->fid_hist_cache[0] = PasisIO_MDHistCacheGetEmpty (fids[0]);
    PasisIO_MDHistCachePin (proto->fid_hist_cache[0]);

    /* Write operation -- Setattr */
    proto->write_op_args.op_type = PASIS_MD_SETATTR;
    proto->write_op_args.repair_flag = 0;

    proto->write_op_args.op_args.setattr.fid = fid;
    memcpy (&proto->write_op_args.op_args.setattr.attr, sattr, 
            sizeof (pasis_md_sattr_t));
    proto->write_op_args.op_args.setattr.guard_ctime[0] = guard_ctime[0];
    proto->write_op_args.op_args.setattr.guard_ctime[1] = guard_ctime[1];
    proto->write_op_args.op_args.setattr.client_id  = rpc_get_host_ip();

    error = write_read_history (protocol, &req, proto, &select_out, 
                                1, fids, 1);
    if (error) goto exit;
   
    /* Dispatch write operation */
    error = write_op_dispatch (protocol, &req, proto, &select_out, 1, fids,
                               sizeof (pasis_md_setattr_res_t), (void **)&res);
    if (error) goto exit;

    /* Copy results */
    memcpy (attr_out, &res->attr, sizeof (pasis_md_attr_t));

 exit:
    PasisIO_MDHistCacheUnpin (proto->fid_hist_cache[0]);
    PasisFreelist_Put (proto_freelist, proto);

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "OUT %s: error=%u", 
                   __FUNCTION__, error);

    PASIS_TIMER_STOP (stats.timers, MD_TIMER_SETATTR);

    return error;
}


pasis_error_t
PasisIO_MDRename (pasis_md_protocol_t   *protocol,
                  pasis_md_linkage_t    *linkages,
                  unsigned int           num_linkages,
                  pasis_md_fid_t        *from_dir,
                  pasis_md_fid_t        *to_dir,
                  pasis_md_fid_t        *from_fid,
                  pasis_md_fid_t        *to_fid,
                  char                  *from_name,
                  char                  *to_name,
                  int                    to_name_valid,
                  pasis_md_attr_t       *from_dir_attr_out, /* Out args */
                  pasis_md_attr_t       *to_dir_attr_out)   /* Out args */
{
    pasis_error_t              error;
    pasisio_request_t          req;
    pasisio_select_t           select_out;
    pasisio_md_proto_data_t   *proto;
    pasisio_drive_instance_t  *out_list[PASIS_MAX_NUM_SERVERS];
    pasis_md_rename_res_t     *res;
    pasis_md_fid_t            *fids[MD_MAX_OBJS_PER_REQ];
    int                        i, to_dir_valid=1, num_fids=0;
    unsigned int               shuffle;

    PASIS_TIMER_START ();

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "\n%s: \n", __FUNCTION__); 

    select_out.instances = out_list;
    proto = PasisFreelist_Get (proto_freelist);
    proto->already_saved_write_op = 0;
    proto->chunk_size = 0;
    memset (proto->write_op_args.proof_fid_indices, UCHAR_MAX,
            sizeof(p_uint32) * MD_MAX_OBJS_PER_REQ);

    shuffle = *(unsigned int*) from_dir;
    op_init (protocol, linkages, num_linkages, proto, &select_out, &req,
             shuffle);

    /* Check if the from_dir is the same as to_dir, if so ignore the to_dir */
    if (md_fid_compare (from_dir, to_dir) == 0) {
        to_dir_valid = 0;
    }

    fids[num_fids] = from_dir;
    num_fids++;
    if (to_dir_valid != 0) {
        fids[num_fids] = to_dir;
        num_fids++;
    }
    fids[num_fids] = from_fid;
    num_fids++;
    if (to_name_valid != 0) {
        fids[num_fids] = to_fid;
        num_fids++;
    }

    /* Init history cache objects */
    for (i = 0; i < num_fids; i++) {
        /* Do a lookup, or get empty if lookup fails, followed by a pin */
        proto->fid_hist_cache[i] = PasisIO_MDHistCacheGetEmpty (fids[i]);
        PasisIO_MDHistCachePin (proto->fid_hist_cache[i]);
    }

    /* Write operation -- Rename */
    proto->write_op_args.op_type = PASIS_MD_RENAME;
    proto->write_op_args.repair_flag = 0;

    pasis_md_attr_set_time (proto->write_op_args.mtime);
    proto->write_op_args.op_args.rename.from_dir_fid = from_dir;
    proto->write_op_args.op_args.rename.to_dir_valid = to_dir_valid;
    proto->write_op_args.op_args.rename.to_dir_fid = to_dir;
    proto->write_op_args.op_args.rename.from_name_fid = from_fid;
    proto->write_op_args.op_args.rename.to_name_valid = to_name_valid;
    proto->write_op_args.op_args.rename.to_name_fid = to_fid;
    proto->write_op_args.op_args.rename.from_name = from_name;
    proto->write_op_args.op_args.rename.to_name = to_name;
    proto->write_op_args.op_args.rename.client_id = rpc_get_host_ip();

    error = write_read_history (protocol, &req, proto, &select_out, 
                                num_fids, fids, 1);
    if (error) goto exit;


    /* Dispatch write operation */
    error = write_op_dispatch (protocol, &req, proto, &select_out, num_fids,
                               fids, sizeof (pasis_md_rename_res_t),
                               (void **)&res);
                               
    if (error) goto exit;

    /* Copy results */
    memcpy (from_dir_attr_out, &res->from_dir_attr, sizeof (pasis_md_attr_t));
    memcpy (to_dir_attr_out, &res->to_dir_attr, sizeof (pasis_md_attr_t));

 exit:

    for (i = 0; i < num_fids; i++) {
        PasisIO_MDHistCacheUnpin (proto->fid_hist_cache[i]);
    }

    PasisFreelist_Put (proto_freelist, proto);

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "OUT %s: error=%u", 
                   __FUNCTION__, error);

    PASIS_TIMER_STOP (stats.timers, MD_TIMER_RENAME);

    return error;
}



/* Read-Only operations 
 *
 * UPDATE UPDATE UPDATE
*
 * 1) Issue request optimistically, obtain set of timestamp vectors 
 * 2) Classify candidate -- either complete or repairable 
 * 3) If candidate is complete, count number of values, done if we have b+1 
 *    if not, request additional entries
 * 4) If candidate is repairable, perform repair
 * 5) Return value
 */    


static unsigned int
read_op_ts_count (pasisio_select_t       *select,
                  unsigned int            vec_set,
                  pasis_md_read_op_res_t  read_op_res[PASIS_MAX_NUM_SERVERS],
                  pasis_md_hts_t         *ts)
{
    unsigned int i, count=0;
    
    /* Go through list of S-Ns whose data corresponds to timestamp 
     * observed as being the latest write */
    for (i = 0; i < select->count; i++) {
        if (select->flags[i] == PASISIO_SELECT_IO_SUCCESS &&
            read_op_res[i].status == PASIS_SUCCESS) {
            if (md_hts_compare (ts, &read_op_res[i].ts) == 0) {
                count++;
            }
        } else {
            select->flags[i] = PASISIO_SELECT_IO_IGNORE;
        }
    }

    return count;
}


static void *
read_op_res_count_and_vote (pasisio_select_t       *select,
                            pasis_md_read_op_res_t  read_op_res[PASIS_MAX_NUM_SERVERS],
                            unsigned int            op_size,
                            unsigned int            threshold,
                            pasis_md_hts_t         *result_ts)
{
    unsigned int i, idx=0, res_num=0;
    void        *to_vote[PASIS_MAX_NUM_SERVERS];
    void        *to_return[PASIS_MAX_NUM_SERVERS];

    for (i = 0; i < select->count; i++) {
        if (select->flags[i] == PASISIO_SELECT_IO_SUCCESS &&
            read_op_res[i].status == PASIS_SUCCESS) {

#if 0
            /* I don't think this is necessary.  And I don't like the idea of
             * destroying soem result because it belongs to a different
             * version. -mabdelmalek 8/1/2005 */
            if (read_op_res[i].op_status != PASIS_SUCCESS ||
                md_hts_compare (&read_op_res[i].ts, result_ts) != 0) {
                memset (&read_op_res[i].op_res, 0, op_size);
            }
#endif

            /* For the synthetic query operation, we need to look at the pointer
             * in the result structure for the actual return data for voting. */
            if (read_op_res[i].op_type == PASIS_MD_SYNTH_QUERY) {
                to_vote[res_num] = (void *)read_op_res[i].op_res.synth_query.res;
            } else {
                to_vote[res_num] = (void *)&read_op_res[i].op_res;
            }
            to_return[res_num] = (void *)&read_op_res[i].op_res;
            res_num++;
        }
    }

    if (res_num == 0) {
        return NULL;
    }
    
    if (threshold > 0) {
        idx = vote (to_vote, res_num, op_size, threshold);
        if (idx == -1) return NULL;
    }

    return to_return[idx];
}


static pasis_error_t
read_op_dispatch (pasis_md_protocol_t      *protocol,
                  pasisio_request_t        *req,
                  pasisio_md_proto_data_t  *proto,
                  pasisio_select_t         *select,
                  pasis_md_fid_t           *fid,
                  unsigned int              fid_num,
                  unsigned int              op_res_size,
                  void                    **res_out)
{
    unsigned int             complete, num_issue, num_thresh, count;
    pasis_error_t            error;
    pasisio_select_flags_t   pre_repair_flags[PASIS_MAX_NUM_SERVERS];
    unsigned int             barrier_flag, repair_flag, fast_repair_flag;
    unsigned int             is_multiobject_op = 0;

    /* First phase is optimistic: we read the latest version of the data,
     * expecting that all the servers will have the same latest version.
     * We also read the replica histories to verify this assumption.  If our
     * optimism was uncalled for, we go back and perform repair  (if necessary)
     * and then read at the correct version */
    proto->read_op_args.read_flags = PASIS_MD_READ_LATEST_DATA;

    /* We only send an OHS with a query on second+subsequent phases, when we see
     * a repair is necessary.  Here, in the first (optimistic) phase of a query,
     * we don't send any OHS. -mabdelmalek 19/3/2005 */
    proto->read_op_args.history_matrices_num = 0;

    PasisIO_GetCompleteThreshold (protocol, &num_issue, &num_thresh);

 restart:
    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 5, "IN %s\n", __FUNCTION__);

#if USE_EARLY_TERMINATION > 0
    complete = protocol_issue_reqs (req, PASISIO_MD_DRIVE_P_READ_OP,
                                    num_issue,
                                    num_thresh,  /* Min */
                                    num_thresh); /* Early term */
#else
    complete = protocol_issue_reqs (req, PASISIO_MD_DRIVE_P_READ_OP,
                                    protocol->N,
                                    protocol->N, protocol->N);
#endif

    /* Check complete */
    if (STRESS_TESTING == 0) {
        if (complete < protocol->Qc + protocol->B) {
            PASIS_DASSERT (0);
        }
    }

    /* Fill in history matrices from history cache */
    set_hist_matrices (fid_num, &proto->write_op_args,
                       proto->fid_hist_cache);

    /* Perform read classification over matrix */
    validate_hist_matrix (protocol,
                          &proto->write_op_args.history_matrices[0],
                          &proto->ts_classify, &barrier_flag, 
                          &repair_flag, &fast_repair_flag,
                          &is_multiobject_op);

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, 
                   "%s: validation: barrier=%u, repair=%u, time=%llu\n",
                   __FUNCTION__, barrier_flag, repair_flag, 
                   proto->ts_classify->time);

    /* Perform repair if repair flag is set, don't care about barriers */
    if (repair_flag) {
        int deduced_complete;

        /* Save copy of current drive flags */
        memcpy (pre_repair_flags, select->flags,
                protocol->N * sizeof(pasisio_select_flags_t));
        
        /* Perform repair */
        error = repair (fid, UINT_MAX, protocol, req, proto, select, 
                        proto->ts_classify, 1, is_multiobject_op,
                        &deduced_complete);

        if (deduced_complete) {
            /* We've deduced that this "repairable" timestamp is in fact
             * complete.  So there's no need to restart the operation --
             * we'll just use the results */
            memcpy (select->flags, pre_repair_flags,
                    protocol->N * sizeof(pasisio_select_flags_t));
        } else if (error) {
            /* If the repair failed (due to concurrency), then we want to
             * restart the whole query operation from beginning (by reading
             * history and latest data, etc...
             * No need to backoff, since this is a query. */
            stats.complete_rereads++;
            proto->read_op_args.read_flags = PASIS_MD_READ_LATEST_DATA;
            // reset history_matrices_num????
            goto restart;
        } else {
            /* When doing repair for query operations, we just send an OHS
             * with the query op, and the server can validate that object
             * syncing is necessary to repair a repairable operation.
             * Then, our query is executed by the server after the object
             * syncing is done (if necessary).
             * For query ops, repair() will write a barrier if necessary.
             * Then, it will set the history_matrices and authenticators for
             * the query op, so that when we resend the query op, the server
             * can do the object syncing.
             * So now, we just have to resend the query op (along with the
             * OHS).
             * -mabdelmalek 20/3/2005. */
            proto->read_op_args.read_flags = PASIS_MD_READ_HTS_DATA;
            md_hts_copy (&proto->read_op_args.ts, proto->ts_classify);
            goto restart;
        }
    } else if (barrier_flag) {
        /* Because this is a read operation, we don't have to actually write
         * a barrier if the latest version is incomplete. */
    }

    *res_out = NULL;

    /* Now, look at the query responses */
    if (proto->read_op_args.read_flags == PASIS_MD_READ_HTS_DATA) {
        /* Count how many query responses were at the timestamp we requested.
         * Also, we need at least B+1 matching results. */
        count = read_op_ts_count (select, 0, proto->read_op_res, 
                                  &proto->read_op_args.ts);
        if (count < protocol->B + 1) {
            /* Hmm, didn't get enough responses at the timestamp we asked for.
             * That could happen because repair occured and our query couldn't
             * be handled at the timestamp we requested.
             * In some cases, the servers will be optimistic and give us the
             * response for the latest version, if they don't have the version
             * we asked for.  So look at the responses again, but this time with
             * the latest complete timestamp. */
            count = read_op_ts_count (select, 0, proto->read_op_res, 
                                      proto->ts_classify);
            if (count < protocol->B + 1) {
                /* Restart the op. */
            } else {
                /* Use this data */
                *res_out = read_op_res_count_and_vote (select, proto->read_op_res,
                                                       op_res_size, protocol->B,
                                                       proto->ts_classify);
            }
        } else {
            /* Use this data */
            *res_out = read_op_res_count_and_vote (select, proto->read_op_res,
                                                   op_res_size, protocol->B,
                                                   &proto->read_op_args.ts);
        }
    } else {
        /* Count how many query responses were at the timestamp we requested.
         * Also, we need at least B+1 matching results. */
        count = read_op_ts_count (select, 0, proto->read_op_res, 
                                  proto->ts_classify);
        if (count < protocol->B + 1) {
            /* Restart the op. */
        } else {
            /* Use this data */
            *res_out = read_op_res_count_and_vote (select, proto->read_op_res,
                                                   op_res_size, protocol->B,
                                                   proto->ts_classify);
        }
    }

    if (*res_out == NULL) {
        /* Restart the op. How do we get here??? */
        PASIS_PRINT_INFO ("%s: completely restarting query operation (%d)\n",
                          __FUNCTION__, proto->read_op_args.op_type);

        /* Reset read flags and re-issue */
        select_reset (num_issue, select);
        proto->read_op_args.read_flags = PASIS_MD_READ_LATEST_DATA;
        stats.complete_rereads++;

        goto restart;
    }

    /* Vote on return code and return */
    return read_op_rpc_op_stat_vote (select, proto->read_op_res, protocol->B);
}


pasis_error_t
PasisIO_MDGetattr (pasis_md_protocol_t *protocol,
                   pasis_md_linkage_t  *linkages,
                   unsigned int         num_linkages,
                   pasis_md_fid_t      *fid,
                   pasis_md_attr_t     *attr_out,         /* Out args */
                   pasis_md_loc_t      *distribution_out) /* Out args */
{
    pasis_error_t              error;
    pasisio_request_t          req;
    pasisio_select_t           select_out;
    pasisio_md_proto_data_t   *proto;
    pasisio_drive_instance_t  *out_list[PASIS_MAX_NUM_SERVERS];
    pasis_md_getattr_res_t    *res;
    unsigned int               shuffle;

    PASIS_TIMER_START ();

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "\n%s: \n", __FUNCTION__); 

    select_out.instances = out_list;
    proto = PasisFreelist_Get (proto_freelist);
    proto->already_saved_write_op = 0;
    proto->chunk_size = 0;
    memset (proto->write_op_args.proof_fid_indices, UCHAR_MAX,
            sizeof(p_uint32) * MD_MAX_OBJS_PER_REQ);

    shuffle = *(unsigned int*) fid;
    op_init (protocol, linkages, num_linkages, proto, &select_out, &req,
             shuffle);

    /* Init history cache objects */
    /* Do a lookup, or get empty if lookup fails, followed by a pin */
    proto->fid_hist_cache[0] = PasisIO_MDHistCacheGetEmpty (fid);
    PasisIO_MDHistCachePin (proto->fid_hist_cache[0]);

    /* Setup read operation -- Getattr */
    proto->read_op_args.op_type = PASIS_MD_GETATTR;
    proto->read_op_args.op_args.getattr.fid = fid;


    error = read_op_dispatch (protocol, &req, proto, &select_out, fid, 1,
                              sizeof (pasis_md_getattr_res_t), (void **)&res);
    if (error) goto exit;

    /* Copy output results */
    memcpy (attr_out, &res->attr, sizeof (pasis_md_attr_t));
    if (distribution_out != NULL) {
        memcpy (distribution_out, &res->distribution, sizeof (pasis_md_loc_t));
    }

 exit:

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "OUT %s: error=%u",
                   __FUNCTION__, error);

    PasisIO_MDHistCacheUnpin (proto->fid_hist_cache[0]);
    PasisFreelist_Put (proto_freelist, proto);

    PASIS_TIMER_STOP (stats.timers, MD_TIMER_GETATTR);

    return error;
}


pasis_error_t
PasisIO_MDLookup (pasis_md_protocol_t *protocol,
                  pasis_md_linkage_t  *linkages,
                  unsigned int         num_linkages,
                  pasis_md_fid_t      *parent_fid,
                  char                *name,
                  pasis_md_fid_t      *fid_out,           /* Out args */
                  pasis_md_attr_t     *dir_attr_out,      /* Out args */
                  pasis_md_loc_t      *distribution_out)  /* Out args */
{
    pasis_error_t              error;
    pasisio_request_t          req;
    pasisio_select_t           select_out;
    pasisio_md_proto_data_t   *proto;
    pasisio_drive_instance_t  *out_list[PASIS_MAX_NUM_SERVERS];
    pasis_md_lookup_res_t     *res;
    unsigned int               shuffle;

    PASIS_TIMER_START ();

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "\n%s: \n", __FUNCTION__); 

    select_out.instances = out_list;
    proto = PasisFreelist_Get (proto_freelist);
    proto->already_saved_write_op = 0;
    proto->chunk_size = 0;
    memset (proto->write_op_args.proof_fid_indices, UCHAR_MAX,
            sizeof(p_uint32) * MD_MAX_OBJS_PER_REQ);

    shuffle = *(unsigned int*) parent_fid;
    op_init (protocol, linkages, num_linkages, proto, &select_out, &req,
             shuffle);

    /* Init history cache objects */
    /* Do a lookup, or get empty if lookup fails, followed by a pin */
    proto->fid_hist_cache[0] = PasisIO_MDHistCacheGetEmpty (parent_fid);
    PasisIO_MDHistCachePin (proto->fid_hist_cache[0]);


    /* Setup read operation -- lookup */
    proto->read_op_args.op_type = PASIS_MD_LOOKUP;
    proto->read_op_args.op_args.lookup.parent_dir_fid = parent_fid;
    proto->read_op_args.op_args.lookup.name = name;

    /* Dispatch read operation */
    error = read_op_dispatch (protocol, &req, proto, &select_out, parent_fid,
                              1, sizeof (pasis_md_lookup_res_t), 
                              (void **)&res);
    if (error) goto exit;

    /* Copy output results */
    memcpy (fid_out, &res->fid, sizeof (pasis_md_fid_t));
    memcpy (dir_attr_out, &res->dir_attr, sizeof (pasis_md_attr_t));
    if (distribution_out != NULL) {
        memcpy (distribution_out, &res->distribution, sizeof (pasis_md_loc_t));
    }

 exit:

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "OUT %s: error=%u", 
                   __FUNCTION__, error);

    PasisIO_MDHistCacheUnpin (proto->fid_hist_cache[0]);
    PasisFreelist_Put (proto_freelist, proto);

    PASIS_TIMER_STOP (stats.timers, MD_TIMER_LOOKUP);

    return error;
}


pasis_error_t
PasisIO_MDReadlink (pasis_md_protocol_t *protocol,
                    pasis_md_linkage_t  *linkages,
                    unsigned int         num_linkages,
                    pasis_md_fid_t      *fid,
                    char                *path_out,        /* Out args */
                    pasis_md_attr_t     *fid_attr_out)    /* Out args */
{
    pasis_error_t              error;
    pasisio_request_t          req;
    pasisio_select_t           select_out;
    pasisio_md_proto_data_t   *proto;
    pasisio_drive_instance_t  *out_list[PASIS_MAX_NUM_SERVERS];
    pasis_md_readlink_res_t   *res;
    unsigned int               shuffle;

    PASIS_TIMER_START ();

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "\n%s: \n", __FUNCTION__); 

    select_out.instances = out_list;
    proto = PasisFreelist_Get (proto_freelist);
    proto->already_saved_write_op = 0;
    proto->chunk_size = 0;
    memset (proto->write_op_args.proof_fid_indices, UCHAR_MAX,
            sizeof(p_uint32) * MD_MAX_OBJS_PER_REQ);

    shuffle = *(unsigned int*) fid;
    op_init (protocol, linkages, num_linkages, proto, &select_out, &req,
             shuffle);

    /* Init history cache objects */
    /* Do a lookup, or get empty if lookup fails, followed by a pin */
    proto->fid_hist_cache[0] = PasisIO_MDHistCacheGetEmpty (fid);
    PasisIO_MDHistCachePin (proto->fid_hist_cache[0]);

    /* Setup read operation -- lookup */
    proto->read_op_args.op_type = PASIS_MD_READLINK;
    proto->read_op_args.op_args.readlink.fid = fid;

    /* Dispatch read operation */
    error = read_op_dispatch (protocol, &req, proto, &select_out, fid, 1,
                              sizeof (pasis_md_readlink_res_t), (void **)&res);
    if (error) goto exit;

    /* Copy output results */
    strncpy (path_out, res->name, PASIS_MD_MAX_PATHLEN);
    memcpy (fid_attr_out, &res->fid_attr, sizeof (pasis_md_attr_t));

 exit:

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "OUT %s: error=%u", 
                   __FUNCTION__, error);

    PasisIO_MDHistCacheUnpin (proto->fid_hist_cache[0]);
    PasisFreelist_Put (proto_freelist, proto);

    PASIS_TIMER_STOP (stats.timers, MD_TIMER_READLINK);

    return error;
}


pasis_error_t
PasisIO_MDReaddir (pasis_md_protocol_t *protocol,
                   pasis_md_linkage_t  *linkages,
                   unsigned int         num_linkages,
                   pasis_md_fid_t      *fid,
                   p_uint64             blkno,
                   p_byte              *buffer_out,
                   pasis_md_attr_t     *dir_attr_out)    /* Out args */
{
    pasis_error_t              error;
    pasisio_request_t          req;
    pasisio_select_t           select_out;
    pasisio_md_proto_data_t   *proto;
    pasisio_drive_instance_t  *out_list[PASIS_MAX_NUM_SERVERS];
    pasis_md_readdir_res_t    *res;
    pasis_md_fid_t            *fids[MD_MAX_OBJS_PER_REQ];
    unsigned int               i;
    p_byte                    *tmp_bufs;
    unsigned int               shuffle;
    unsigned int               num_issue, num_thresh;


    PASIS_TIMER_START ();

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "\n%s: \n", __FUNCTION__); 

    tmp_bufs = malloc (PASIS_MD_DIR_BLK_SIZE * num_linkages);

    /* BUG BUG BUG BUG
     * Any replies we hear back that occur _after_ this function has returned
     * will be writing their data to bad memory, since we'd free the tmp_bufs
     * array on exit.
     */

    PasisIO_GetCompleteThreshold (protocol, &num_issue, &num_thresh);

    select_out.instances = out_list;
    proto = PasisFreelist_Get (proto_freelist);
    proto->already_saved_write_op = 0;
    proto->chunk_size = 0;
    memset (proto->write_op_args.proof_fid_indices, UCHAR_MAX,
            sizeof(p_uint32) * MD_MAX_OBJS_PER_REQ);

    shuffle = *(unsigned int*) fid;
    op_init (protocol, linkages, num_linkages, proto, &select_out, &req,
             shuffle);

    fids[0] = fid;
    
    proto->fid_hist_cache[0] = PasisIO_MDHistCacheGetEmpty (fids[0]);
    PasisIO_MDHistCachePin (proto->fid_hist_cache[0]);

    /* Force refresh of history data if reading the first block of the dir */
    if (blkno == 0) {
        PasisIO_MDHistCacheInvalidateData (proto->fid_hist_cache[0]);
    }

    /* Setup read operation -- readdir */
    proto->read_op_args.op_type = PASIS_MD_READDIR;

    proto->read_op_args.op_args.readdir.fid = fid;
    proto->read_op_args.op_args.readdir.block_num = blkno;
    
    /* Setup readdir results */
    for (i = 0; i < select_out.count; i++) {
        /* Set pointers for data buffers */
        proto->read_op_res[i].op_res.readdir.data = tmp_bufs + 
            (i * PASIS_MD_DIR_BLK_SIZE);
        proto->read_op_res[i].op_res.readdir.data_size = PASIS_MD_DIR_BLK_SIZE;
    }

    /* Reset select flags */
    select_reset (num_issue, &select_out);
    

    /* Dispatch read operation -- subtract of buffer pointer for res voting */
    error = read_op_dispatch (protocol, &req, proto, &select_out, fid, 1,
                              sizeof (pasis_md_readdir_res_t) - 
                              sizeof (p_byte *), (void **)&res);
    if (error) goto exit;

    /* Copy output results */
    memcpy (buffer_out, res->data, res->data_size);
    memcpy (dir_attr_out, &res->dir_attr, sizeof (pasis_md_attr_t));

 exit:
    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "OUT %s: error=%u", 
                   __FUNCTION__, error);

    PasisIO_MDHistCacheUnpin (proto->fid_hist_cache[0]);

    /* Reset pointers to the buffer we're about to free.  This way, due to early
     * termination, if we get a readdir response and we've already freed the
     * buffer, the xdr_decode can tell this by finding a null pointer and they
     * would ignore the incoming data and not try to write it to an
     * already-freed buffer. */
    for (i = 0; i < select_out.count; i++) {
        proto->read_op_res[i].op_res.readdir.data = NULL;
    }


    PasisFreelist_Put (proto_freelist, proto);
    free (tmp_bufs);

    PASIS_TIMER_STOP (stats.timers, MD_TIMER_READDIR);

    return error;
}


pasis_error_t
PasisIO_MDSynthQuery (pasis_md_protocol_t   *protocol,
                      pasis_md_fid_t       **fids,
                      p_uint32               num_fids,
                      pasis_md_linkage_t    *linkages,
                      unsigned int           num_linkages,
                      p_uint32               arg_size,
                      p_uint32               res_size,
                      p_uint32               chunk_size,
                      p_uint32              *read_chunks, /* bitmaps */
                      p_uint32               compute_time,
                      unsigned int           use_cached_hist_flag,
                      p_uint32               server_override_bitmap,
                      /* Only pass data if modified_chunks has just one
                       * element, and chunk_size = arg_size */
                      p_byte                *data)

{
    pasis_error_t               error;
    pasisio_request_t           req;
    pasisio_select_t            select_out;
    pasisio_md_proto_data_t    *proto;
    pasisio_drive_instance_t   *out_list[PASIS_MAX_NUM_SERVERS];
    pasis_md_synth_query_res_t *res;
    unsigned int                shuffle;
    p_byte                     *arg_array, *res_bufs;
    int                         i;

    PASIS_TIMER_START ();

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "\n%s: \n", __FUNCTION__); 

    /* For now, we only support single-object query operations */
    PASIS_DASSERT (num_fids == 1);

    select_out.instances = out_list;
    proto = PasisFreelist_Get (proto_freelist);
    proto->already_saved_write_op = 0;
    proto->chunk_size = chunk_size;
    memset (proto->write_op_args.proof_fid_indices, UCHAR_MAX,
            sizeof(p_uint32) * MD_MAX_OBJS_PER_REQ);

    shuffle = *(unsigned int*) fids[0];
    op_init (protocol, linkages, num_linkages, proto, &select_out, &req,
             shuffle);

    if (server_override_bitmap) {
        PasisIO_SelectThresholdDecodeLinkages (linkages, protocol->N,
                                               protocol->N,
                                               PASISIO_DRIVE_TYPE_MD_INDEX,
                                               &select_out, 0);
        for (i = 0; i < select_out.count; i++) {
            if ((server_override_bitmap & (1 << i)) == 0) {
                select_out.instances[i] = NULL;
                select_out.flags[i] = 0xff; /* so select_reset does nothing */
            }
        }
    }
#if 0
    printf ("\t%s: drive list: ", __FUNCTION__);
    for (i = 0; i < num_linkages; i++) {
        if (select_out.flags[i] == PASISIO_SELECT_IO_READY) {
            printf ("[%d], ", i);
        }
    }
    printf ("\n");
#endif

    arg_array = malloc (arg_size);
    res_bufs = malloc (res_size * num_linkages);

    /* Init history cache objects */
    for (i = 0; i < num_fids; i++) {
        /* Do a lookup, or get empty if lookup fails, followed by a pin */
        proto->fid_hist_cache[i] = PasisIO_MDHistCacheGetEmpty (fids[i]);
        PasisIO_MDHistCachePin (proto->fid_hist_cache[i]);
        proto->read_op_args.op_args.synth_query.fids[i] = fids[i];
        proto->read_op_args.op_args.synth_query.read_chunks[i] = read_chunks[i];

        if (use_cached_hist_flag == 0) {
            if (num_outstanding > 1) {
                PASIS_PRINT_ERR ("%s: forcing cache invalidation with multiple "
                                 "outstanding requests is asking for trouble.",
                                 __FUNCTION__);
                PASIS_DASSERT (0);
            }
            PasisIO_MDHistCacheInvalidateData (proto->fid_hist_cache[i]);
        }
    }

    /* Setup read operation */
    proto->read_op_args.op_type = PASIS_MD_SYNTH_QUERY;
    proto->read_op_args.op_args.synth_query.num_fids     = num_fids;
    proto->read_op_args.op_args.synth_query.arg_size     = arg_size;
    proto->read_op_args.op_args.synth_query.res_size     = res_size;
    proto->read_op_args.op_args.synth_query.chunk_size   = chunk_size;
    proto->read_op_args.op_args.synth_query.compute_time = compute_time;
    proto->read_op_args.op_args.synth_query.args         = arg_array;

    /* Setup result pointers */
    for (i = 0; i < select_out.count; i++) {
        /* Set pointers for data buffers */
        proto->read_op_res[i].op_res.synth_query.res = res_bufs + 
            (i * res_size);
        proto->read_op_res[i].op_res.synth_query.res_size = res_size;
    }

    error = read_op_dispatch (protocol, &req, proto, &select_out, fids[0], num_fids,
                              res_size, (void **)&res);
    if (error) goto exit;

    if (data) {
        PASIS_DASSERT(res->res_size == chunk_size);
        PASIS_DASSERT(chunk_size == res_size);
        memmove(data, res->res, res->res_size);
    }

#if 0
    printf("\t%s: returned value at TS %lld for req %u\n", __FUNCTION__,
           ((pasis_md_hts_t*)res)->time, req.id);
#endif
 exit:

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "OUT %s: error=%u",
                   __FUNCTION__, error);

    for (i = 0; i < num_fids; i++) {
        PasisIO_MDHistCacheUnpin (proto->fid_hist_cache[i]);
    }
    PasisFreelist_Put (proto_freelist, proto);

    free (arg_array);
    free (res_bufs);

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "OUT %s: error=%u", 
                   __FUNCTION__, error);

    PASIS_TIMER_STOP (stats.timers, MD_TIMER_SYNTH_QUERY);

    return error;
}



pasis_error_t
PasisIO_MDSynthUpdate (pasis_md_protocol_t *protocol,
                       pasis_md_fid_t     **fids,
                       p_uint32             num_fids,
                       pasis_md_linkage_t  *linkages,
                       unsigned int         num_linkages,
                       p_uint32             arg_size,
                       p_uint32             res_size,
                       p_uint32             chunk_size, /* set on create */
                       p_uint32            *modified_chunks, /* bitmaps */
                       p_uint32             compute_time,
                       /* For testing, useful for client to force server to
                        * increment logical TS by more than 1 */
                       p_uint32             ts_increment, /* Usually 1 */
                       unsigned int         use_cached_hist_flag,
                       p_uint32             server_override_bitmap,
                       /* Only pass data if modified_chunks has just one
                        * element, and chunk_size = arg_size */
                       p_byte              *data)
{
    pasis_error_t                error;
    pasisio_request_t            req;
    pasisio_select_t             select_out;
    pasisio_md_proto_data_t     *proto;
    pasisio_drive_instance_t    *out_list[PASIS_MAX_NUM_SERVERS];
    pasis_md_synth_update_res_t *res;
    unsigned int                 shuffle;
    int                          i;
    p_byte                      *arg_array, *res_bufs;

    PASIS_TIMER_START ();

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "\n%s: \n", __FUNCTION__); 

    select_out.instances = out_list;
    proto = PasisFreelist_Get (proto_freelist);
    proto->already_saved_write_op = 0;
    proto->chunk_size = chunk_size;
    memset (proto->write_op_args.proof_fid_indices, UCHAR_MAX,
            sizeof(p_uint32) * MD_MAX_OBJS_PER_REQ);

    shuffle = *(unsigned int*) fids[0];
    op_init (protocol, linkages, num_linkages, proto, &select_out, &req,
             shuffle);
    
    if (data) {
        PASIS_DASSERT(arg_size == chunk_size);
        arg_array = data;
    } else {
        arg_array = malloc (arg_size);
    }
    res_bufs = malloc (res_size * num_linkages);

    /* Init history cache objects */
    for (i = 0; i < num_fids; i++) {
        /* Do a lookup, or get empty if lookup fails, followed by a pin */
        proto->fid_hist_cache[i] = PasisIO_MDHistCacheGetEmpty (fids[i]);
        PasisIO_MDHistCachePin (proto->fid_hist_cache[i]);
        proto->write_op_args.op_args.synth_update.fids[i] = fids[i];
        proto->write_op_args.op_args.synth_update.modified_chunks[i] = modified_chunks[i];
    }

    /* Write operation -- synth update */
    proto->write_op_args.op_type = PASIS_MD_SYNTH_UPDATE;
    proto->write_op_args.repair_flag = 0;

    proto->write_op_args.op_args.synth_update.num_fids   = num_fids;
    proto->write_op_args.op_args.synth_update.arg_size   = arg_size;
    proto->write_op_args.op_args.synth_update.res_size   = res_size;
    proto->write_op_args.op_args.synth_update.chunk_size = chunk_size;
    proto->write_op_args.op_args.synth_update.compute_time = compute_time;
    proto->write_op_args.op_args.synth_update.ts_increment = ts_increment;
    proto->write_op_args.op_args.synth_update.client_id  = rpc_get_host_ip();

    proto->write_op_args.op_args.synth_update.args       = arg_array;

    if (server_override_bitmap) {
        PasisIO_SelectThresholdDecodeLinkages (linkages, protocol->N,
                                               protocol->N,
                                               PASISIO_DRIVE_TYPE_MD_INDEX,
                                               &select_out, 0);
        for (i = 0; i < select_out.count; i++) {
            if ((server_override_bitmap & (1 << i)) == 0) {
                select_out.instances[i] = NULL;
                select_out.flags[i] = 0xff; /* so select_reset does nothing */
            }
        }
    }
#if 0
    printf ("\t%s: drive list: ", __FUNCTION__);
    for (i = 0; i < num_linkages; i++) {
        if (select_out.flags[i] == PASISIO_SELECT_IO_READY) {
            printf ("[%d], ", i);
        }
    }
    printf ("\n");
#endif

    /* Setup result pointers */
    for (i = 0; i < select_out.count; i++) {
        /* Set pointers for data buffers */
        proto->write_op_res[i].op_res.synth_update.res = res_bufs + 
            (i * res_size);
        proto->write_op_res[i].op_res.synth_update.res_size = res_size;
    }

    /* Call write_read_history after the above write_op_args members have been
     * set.  This is important, because the way we no do repair, the write op
     * will be sent and will cause the repair to occur "alongside".  So it's
     * vital the write_op_args struct is already filled.  -mabdelmalek 8/1/2005 */
    error = write_read_history (protocol, &req, proto, &select_out, 
                                num_fids, fids, use_cached_hist_flag);
    if (error) goto exit;
   

    /* Dispatch write operation */
    error = write_op_dispatch (protocol, &req, proto, &select_out, num_fids,
                               fids, res_size, (void **)&res);
    if (error) goto exit;

 exit:
    /* Reset pointers to the buffer we're about to free.  This way, due to early
     * termination, if we get a response and we've already freed the buffer, the
     * xdr_decode can tell this by finding a null pointer and they would ignore
     * the incoming data and not try to write it to an already-freed buffer. */
    for (i = 0; i < select_out.count; i++) {
        proto->write_op_res[i].op_res.synth_update.res = NULL;
    }

    for (i = 0; i < num_fids; i++) {
        PasisIO_MDHistCacheUnpin (proto->fid_hist_cache[i]);
    }
    PasisFreelist_Put (proto_freelist, proto);

    if (data) {
    } else {
        free (arg_array);
    }
    free (res_bufs);

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "OUT %s: error=%u", 
                   __FUNCTION__, error);

    PASIS_TIMER_STOP (stats.timers, MD_TIMER_SYNTH_UPDATE);

    return error;
}


pasis_error_t
PasisIO_MDObjectSync (pasis_md_protocol_t   *protocol,
                      pasis_md_fid_t        *fid,
                      pasis_md_linkage_t    *linkages,  /* B+1 servers */
                      pasis_md_hts_t        *sync_ts,
                      p_uint32               block_size,
              /* buffers for each server's blocks' response; prevents memcpy */
                      p_byte                *buffer,
                      p_uint32               starting_block_num,
                      p_uint32               num_blocks_to_ret,
                      p_uint32              *num_total_blocks, /* Result */
                      p_uint32              *num_returned_blocks) /* Result */
{
    pasis_error_t               error;
    pasisio_request_t           req;
    pasisio_select_t            select_out;
    pasisio_md_proto_data_t    *proto;
    pasisio_drive_instance_t   *out_list[PASIS_MAX_NUM_SERVERS];
    void                       *block_buffers[PASIS_MAX_NUM_SERVERS];
    void                       *returned_args[PASIS_MAX_NUM_SERVERS];
    pasis_md_synth_query_res_t *res;
    int                         i, complete, idx;

    PASIS_TIMER_START ();

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "\n%s: \n", __FUNCTION__); 

    select_out.instances = out_list;
    proto = PasisFreelist_Get (proto_freelist);
    PASIS_DASSERT(proto != NULL);
    proto->already_saved_write_op = 0;
    proto->chunk_size = block_size;
    memset (proto->write_op_args.proof_fid_indices, UCHAR_MAX,
            sizeof(p_uint32) * MD_MAX_OBJS_PER_REQ);

    /* Do the setup ourselves, since we want all B + 1 servers */
    select_out.count = protocol->B + 1;
    proto->shuffle = 0;
    proto->protocol = protocol;
    proto->linkages = (pasis_md_linkage_t *)linkages;
    req.drive_select = &select_out;
    req.proto_data = proto;
    req.id = req_count++;
    PasisIO_SelectDecodeLinkages(linkages, protocol->B + 1,
                                 PASISIO_DRIVE_TYPE_MD_INDEX, &select_out, 0);

    /* Setup read operation */
    proto->read_op_args.op_type = PASIS_MD_OBJECT_SYNC;
    proto->read_op_args.op_args.object_sync.fid         = fid;
    proto->read_op_args.op_args.object_sync.return_attrs = 0; // for now
    proto->read_op_args.op_args.object_sync.start_block_num =
        starting_block_num;
    proto->read_op_args.op_args.object_sync.num_blocks_to_ret =
        num_blocks_to_ret;

    /* Setup result pointers */
    for (i = 0; i < protocol->B + 1; i++) {
        /* Set pointers for data buffers */
        block_buffers[i] = buffer + (i * num_blocks_to_ret * block_size);
        proto->read_op_res[i].op_res.object_sync.data.client = block_buffers[i];
        returned_args[i] = &proto->read_op_res[i].op_res;
    }

    proto->read_op_args.read_flags = PASIS_MD_READ_HTS_DATA;
    md_hts_copy(&proto->read_op_args.ts, sync_ts);
    proto->read_op_args.history_matrices_num = 0;
    complete = protocol_issue_reqs (&req, PASISIO_MD_DRIVE_P_READ_OP,
                                    protocol->B + 1, protocol->B + 1,
                                    protocol->B + 1);
    if (complete < protocol->B + 1) {
        printf("%s: not enough responses from servers (%u, %u)\n",
               __FUNCTION__, complete, protocol->B + 1);
        error = PASIS_ERROR;
        goto exit;
    }

    /* Vote on the returned error code from the clients */
    error = read_op_rpc_op_stat_vote (&select_out, proto->read_op_res, protocol->B);
    if (error) {
        /* printf("%s: returned error codes indicate error %u\n", __FUNCTION__,
         *      error); */
        goto exit;
    }

    /* Vote on the object sync result return types (attrs/num total blocks/num
     * returned blocks) */
    idx = vote (returned_args, protocol->B + 1,
                sizeof (pasis_md_object_sync_res_t) - sizeof (pasis_md_data_t),
                protocol->B);
    if (idx == -1) {
        printf("%s: results don't agree (1)\n", __FUNCTION__);
        error = PASIS_ERROR;
        goto exit;
    }

    *num_total_blocks = proto->read_op_res[0].op_res.object_sync.num_total_blocks;
    *num_returned_blocks = proto->read_op_res[0].op_res.object_sync.num_returned_blocks;

    if (*num_returned_blocks > 0) {
        /* Check that the caller had expected the right buffer size; otherwise,
         * there will be a buffer overrun */
        PASIS_DASSERT(block_size ==
                      proto->read_op_res[0].op_res.object_sync.block_size);
    }

    /* Vote on the block data returned */
    idx = vote (block_buffers, protocol->B + 1,
                *num_returned_blocks * block_size, protocol->B);
    if (idx == -1) {
        printf("%s: results don't agree (2)\n", __FUNCTION__);
        error = PASIS_ERROR;
        goto exit;
    }

 exit:

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "OUT %s: error=%u",
                   __FUNCTION__, error);
    if (error) {
        /* printf("%s: returning error %u\n", __FUNCTION__, error); */
    }

    PasisFreelist_Put (proto_freelist, proto);

    PASIS_TIMER_STOP (stats.timers, MD_TIMER_OBJECT_SYNC);

    return error;
}


/* Get statistics from servers about how many requests they're handling */
pasis_error_t
PasisIO_MDGetStats (pasis_md_protocol_t   *protocol,
                    pasis_md_linkage_t    *linkages,
                    unsigned int           num_linkages,
                    p_uint32             **stats_,    /* result */
                    p_uint32               element_size,
                    p_byte                *flags,
                    p_uint32              *num_buckets, /* result */
                    p_uint32              *bucket_size) /* result */
{
    pasis_error_t              error = PASIS_SUCCESS;
    pasisio_request_t          req;
    pasisio_select_t           select_out;
    pasisio_md_proto_data_t   *proto;
    pasisio_drive_instance_t  *out_list[PASIS_MAX_NUM_SERVERS];
    pasis_md_getattr_res_t    *res;
    unsigned int               i, complete, num_issue, num_thresh, first_time;

    PASIS_TIMER_START ();

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "\n%s: \n", __FUNCTION__); 

    select_out.instances = out_list;
    proto = PasisFreelist_Get (proto_freelist);
    proto->already_saved_write_op = 0;
    proto->chunk_size = 0;
    proto->fid_hist_cache[0] = NULL;
    proto->read_op_args.op_type = PASIS_MD_GET_STATS;
    proto->read_op_args.history_matrices_num = 0;
    proto->read_op_args.op_args.get_stats.res_max_num_buckets = element_size;
    memset (proto->write_op_args.proof_fid_indices, UCHAR_MAX,
            sizeof(p_uint32) * MD_MAX_OBJS_PER_REQ);

    op_init (protocol, linkages, num_linkages, proto, &select_out, &req, 0);

    PasisIO_GetCompleteThreshold (protocol, &num_issue, &num_thresh);

    /* Setup arguments and result pointers */
    for (i = 0; i < protocol->N; i++) {
        proto->read_op_res[i].op_res.get_stats.res = stats_[i];
    }

    /* Contact servers */
#if USE_EARLY_TERMINATION > 0
    complete = protocol_issue_reqs (&req, PASISIO_MD_DRIVE_P_READ_OP,
                                    num_issue,
                                    num_thresh,  /* Min */
                                    num_thresh); /* Early term */
#else
    complete = protocol_issue_reqs (&req, PASISIO_MD_DRIVE_P_READ_OP,
                                    protocol->N,
                                    protocol->N, protocol->N);
#endif

    /* Check complete */
    if (complete < protocol->Qc + protocol->B) {
        /* If the other servers aren't all up, the above function will fail. */
        error = PASIS_ERROR;
        goto exit;
    }

    /* Go through results and check that num_buckets/bucket_size agree, and set
     * "flags" apropriately */
    first_time = 1;
    for (i = 0; i < select_out.count; i++) {
        if (select_out.flags[i] == PASISIO_SELECT_IO_SUCCESS &&
            proto->read_op_res[i].status == PASIS_SUCCESS &&
            proto->read_op_res[i].op_status == PASIS_SUCCESS) {

            pasis_md_get_stats_res_t *r = &proto->read_op_res[i].op_res.get_stats;

            if (first_time) {
                *num_buckets = r->num_buckets;
                *bucket_size = r->bucket_size;
                first_time = 0;
            } else {
                if (*num_buckets != r->num_buckets ||
                    *bucket_size != r->bucket_size) {
                    break;
                }
            }

            flags[i] = 1;
        } else {
            select_out.flags[i] = PASISIO_SELECT_IO_IGNORE;
            flags[i] = 0;
        }
    }

    if (i != select_out.count) {
        printf("%s: some servers returned a different num_buckets and/or "
               "bucket_size from others\n", __FUNCTION__);
        error = PASIS_ERROR;
        goto exit;
    }


 exit:

    PASIS_DPRINTF (PASIS_DEBUG_IO_MD, 4, "OUT %s: error=%u",
                   __FUNCTION__, error);

    PasisFreelist_Put (proto_freelist, proto);

    PASIS_TIMER_STOP (stats.timers, MD_TIMER_GET_STATS);

    return error;
}


/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */ 
