/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"
#include "pasis_debug.h"
#include "pasis_types.h"

#include "pasis_md_prot.h"
#include "pasis_md_hist_matrix.h"
#include "pasisio_md_protocol.h"
#include "pasis_md_log.h"


void
pasis_md_hist_matrix_dump (pasis_md_hist_matrix_t *matrix)
{
    int i;
    printf ("OHS:\n");

    for (i = 0; i < matrix->ts_num; i++) {
        printf ("[%2d]  ts: %llu, bitmap: %#2llx, total: %d, object hash: %x, "
                "op hash: %x, client_id: %x", i, matrix->ts_vector[i]->time,
                matrix->ts_bitmap[i], bitcount (matrix->ts_bitmap[i]),
#if USE_OBJECT_HASHES > 0
                *(p_uint32*)matrix->ts_vector[i]->object_hash,
#else
                0,
#endif
                *(p_uint32*)matrix->ts_vector[i]->op_hash,
                matrix->ts_vector[i]->client_id);

#if DEBUG_TIMESTAMPS > 0
        printf (", %s", matrix->ts_vector[i]->op_description);
#endif

        printf ("\n");
    }
}

void
pasis_md_hist_matrix_log_dump (pasis_md_hist_matrix_t *matrix)
{
    int i;
    pasis_md_log_msg ("OHS:\n");

    for (i = 0; i < matrix->ts_num; i++) {
        pasis_md_log_msg ("[%2d]  ts: %llu, bitmap: %#2llx, total: %d, object "
                          "hash: %llx", i, matrix->ts_vector[i]->time,
                          matrix->ts_bitmap[i], bitcount (matrix->ts_bitmap[i]),
#if USE_OBJECT_HASHES > 0
                          *(p_uint64*)matrix->ts_vector[i]->object_hash);
#else
                          0);
#endif

#if DEBUG_TIMESTAMPS > 0
        pasis_md_log_msg (", %s", matrix->ts_vector[i]->op_description);
#endif

        pasis_md_log_msg ("\n");
    }
}


/* Initialize and empty history matrix */
void
pasis_md_hist_matrix_init (pasis_md_hist_matrix_t *matrix)
{
    matrix->ts_num    = 0;
    memset (matrix->ts_bitmap, 0, sizeof (p_uint64) * 
            PASIS_MD_HIST_MAX_VERSIONS);
    matrix->auth_num  = 0;
}

/* Add a history vector to a history matrix -- keeps pointers to the vector */
void
pasis_md_hist_matrix_add_vec (pasis_md_hist_vector_t  *vector,
                              pasis_md_hist_matrix_t  *matrix)
{
    unsigned int mat_idx=0;
    unsigned int vec_idx=0;
    int          i, rc;

    p_uint64     bitmask;

    bitmask = 1ULL << vector->node_id;

    while (vec_idx < vector->history_num) {

        if (mat_idx == matrix->ts_num) {  /* Consumed all element in matrix */
            for (i = vec_idx; i < vector->history_num; i++) {
                PASIS_DASSERT (matrix->ts_num < PASIS_MD_HIST_MAX_VERSIONS);
                matrix->ts_vector[mat_idx] = &vector->histories[i];
                matrix->ts_bitmap[mat_idx] = bitmask; 
                matrix->ts_num++;
                mat_idx++;
            } 
            break;
        }

        rc = md_hts_compare (matrix->ts_vector[mat_idx],
                             &vector->histories[vec_idx]);
        if (rc == 0) {
            matrix->ts_bitmap[mat_idx] |= bitmask;
            mat_idx++;
            vec_idx++;
        } else if (rc > 0) {
            /* Matrix TS is bigger, incr mat_idx */
            mat_idx++;
        } else {
            /* Vec TS is bigger, insert here and shift */
            PASIS_DASSERT (matrix->ts_num < PASIS_MD_HIST_MAX_VERSIONS);

            memmove (&matrix->ts_vector[mat_idx+1], 
                     &matrix->ts_vector[mat_idx], 
                     (matrix->ts_num - mat_idx) * 
                     sizeof (pasis_md_hts_t *));

            memmove (&matrix->ts_bitmap[mat_idx+1],
                     &matrix->ts_bitmap[mat_idx],
                     (matrix->ts_num - mat_idx) * sizeof (p_uint64));

            matrix->ts_vector[mat_idx] = &vector->histories[vec_idx];
            matrix->ts_bitmap[mat_idx] = bitmask; 
            
            matrix->ts_num++;
            vec_idx++;
        }
    }

    /* pasis_md_hist_matrix_dump (matrix); */
}


/* Copies history matrix */
void
pasis_md_hist_matrix_copy (pasis_md_hist_matrix_t *dest,
                           pasis_md_hist_matrix_t *src)
{
    int i;

    dest->ts_num = src->ts_num;    
    for (i = 0; i < src->ts_num; i++) {
        dest->ts_vector[i] = src->ts_vector[i];
        dest->ts_bitmap[i] = src->ts_bitmap[i];
    }

    dest->auth_num = src->auth_num;
    for (i = 0; i < PASIS_MAX_NUM_SERVERS; i++) {
        dest->auth_exists[i] = src->auth_exists[i];
    }
    for (i = 0; i < src->auth_num; i++) {
        dest->authenticators[i] = src->authenticators[i];
    }
}


#if 0
/* Adds authenticators specific to a specific node_id to a list of auths */
/* auth_list: contains a 2-D array of NxN authenticators  */
void
pasis_md_hist_matrix_set_auths (p_uint32                for_node_id,
                                unsigned int            auth_num,
                                pasis_md_hist_matrix_t *matrix,              
                                pasis_md_auth_t        *auths[PASIS_MAX_NUM_SERVERS]
                                                             [PASIS_MAX_NUM_SERVERS])
{
    memcpy (matrix->authenticators, auths[for_node_id],
            auth_num * sizeof(pasis_md_auth_t*));
    matrix->auth_num = auth_num;
}


/* Generates a NxN matrix of authenticators.  The first index specifies the
 * storage-node the authenticator is for.  The second index specifies the 
 * storage-node the authenticator is from.
 */
void
pasis_md_hist_matrix_add_auths (pasis_md_hist_vector_t *vec,
                                pasis_md_auth_t    *auths_out[PASIS_MAX_NUM_SERVERS]
                                                             [PASIS_MAX_NUM_SERVERS])
{
    int i;
    for (i = 0; i < vec->auth_num; i++) {
        auths_out[i][vec->node_id] = &vec->authenticators[i];
    }
}
#endif



void
pasis_md_hist_matrix_get_vec (p_uint32 node_id,
                              pasis_md_hist_matrix_t *matrix,
                              unsigned int *ts_num,
                              pasis_md_auth_t **auth_out,
                              pasis_md_hts_t  **recs_out)

{
    p_uint64 bitmask = (1ULL << node_id);
    int      i, out_idx=0;

    for (i = 0; i < matrix->ts_num; i++) {
        if (bitmask & matrix->ts_bitmap[i]) {
            recs_out[out_idx] = matrix->ts_vector[i];
            out_idx++;
        }
    }
    
    *ts_num = out_idx;
    if (auth_out && out_idx) {
        if (DO_NOT_TRANSFER_AUTHENTICATORS == 0) {
            PASIS_DASSERT (matrix->auth_exists[node_id]);
        } else {
            /* If we're not transferring authenticators, this is OK */
        }
        *auth_out = matrix->authenticators[node_id];
    } else {
        *auth_out = NULL; /* no timestamps/auth for this node */
    }
}


p_uint32
pasis_md_hist_matrix_node_has_ts (p_uint32                node_id,
                                  pasis_md_hist_matrix_t *matrix,
                                  pasis_md_hts_t         *ts)
{
    p_uint64 bitmask = (1ULL << node_id);
    int i, cmp;

    for (i = 0; i < matrix->ts_num; i++) {
        cmp = md_hts_compare (matrix->ts_vector[i], ts);

        if (cmp == 0) {
            return matrix->ts_bitmap[i] & bitmask;
        } else if (cmp < 0) {
            /* the specified timestamp isn't in the history matrix */
            return 0;
        }
    }
    
    return 0;
}


static inline pasis_error_t
hist_classify (pasis_md_protocol_t *protocol,
               unsigned int         count)
{
    unsigned int complete_threshold, early_ter_threshold, repairable_threshold;

    PasisIO_GetCompleteThreshold (protocol, &complete_threshold,
                                  &early_ter_threshold);

    if (count >= early_ter_threshold) {
        /* Complete */
        return PASIS_ERROR_PROTO_COMPLETE;
    }

    repairable_threshold = PasisIO_GetRepairableThreshold (protocol);

    if (count >= repairable_threshold) {
        /* Repairable */
        return PASIS_ERROR_PROTO_REPAIR;
    }

    /* Incomplete */
    return PASIS_ERROR_PROTO_INCOMPLETE;
}


pasis_error_t
pasis_md_hist_matrix_classify_client (pasis_md_hist_matrix_t *matrix,
                                      pasis_md_protocol_t    *protocol,
                                      pasis_md_hts_t        **ts_classify_out,
                                      pasis_md_hts_t        **ts_latest_out)

{
    pasis_error_t classification;
    int idx=0, count;

    classification = PASIS_ERROR_PROTO_INCOMPLETE;

    if (ts_latest_out) {
        *ts_latest_out = matrix->ts_vector[0];
    }

    /* XXX: need to distinguish whether a barrier is required for a write */
    while (classification == PASIS_ERROR_PROTO_INCOMPLETE &&
           idx < matrix->ts_num) {

        /* Skip over barriers */
        if (matrix->ts_vector[idx]->barrier_flag) {
            idx++;
            continue;
        }

        count = bitcount (matrix->ts_bitmap[idx]);
        //PASIS_DASSERT (count == protocol->N);
        
        classification = hist_classify (protocol, count);
        /* Skip over incompletes */
        if (classification != PASIS_ERROR_PROTO_INCOMPLETE) {
            *ts_classify_out = matrix->ts_vector[idx];
            return classification;
        }

        idx++;
    }

    return PASIS_ERROR_PROTO_NOWRITE;
}


void
pasis_md_hist_matrix_classify_node (pasis_md_hist_matrix_t *matrix,
                                    pasis_md_protocol_t    *protocol,
                                    pasis_md_hts_t        **ts_barrier_out,
                                    pasis_error_t          *barrier_status_out,
                                    pasis_md_hts_t        **ts_latest_out,
                                    pasis_error_t          *latest_status_out,
                                    int             *latest_is_multiobject_op,
                                    pasis_md_hts_t        **ts_classify_out,
                                    pasis_error_t         *classify_status_out,
                                    int             *classify_is_multiobject_op,
                                    pasis_md_hts_t        **ts_complete_out)

{
    pasis_error_t   classification;
    pasis_md_hts_t *last_repairable=NULL;
    int idx=0, barrier_count=0, latest_out=0, classify_out=0, count;

    *latest_is_multiobject_op = *classify_is_multiobject_op = 0;

    /* Set barrier if the latest timestamp is in fact a barrier */
    *ts_barrier_out = NULL;
    *barrier_status_out = PASIS_ERROR_PROTO_INCOMPLETE;
    if (matrix->ts_vector[idx]->barrier_flag) {
        count = bitcount (matrix->ts_bitmap[0]);
        classification = hist_classify (protocol, count);

        *ts_barrier_out = matrix->ts_vector[0];
        *barrier_status_out = classification;
        barrier_count = count;
    }
    classification = PASIS_ERROR_PROTO_INCOMPLETE;

    while (classification != PASIS_ERROR_PROTO_COMPLETE &&
           idx < matrix->ts_num) {

        count = bitcount (matrix->ts_bitmap[idx]);

        /* If this is a barrier and is either the first timestamp or a
           later timestamp with no intermediate timestamps that have
           not been barriers, update barrier count and reclassify.
           Otherwise skip over the barrier. */
        if (matrix->ts_vector[idx]->barrier_flag) {
            if (idx > 0 && barrier_count > 0) {           
                barrier_count += count;
                *barrier_status_out = hist_classify (protocol, barrier_count);
            }
            idx++;
            continue;
        }

        barrier_count = 0;
        
        /* Classify current ts -- it is not a barrier */
        classification = hist_classify (protocol, count);

        /* Latest holds the latest non-barrier and its classification */
        if (latest_out == 0) {
            *ts_latest_out = matrix->ts_vector[idx];
            *latest_status_out = classification;
            *latest_is_multiobject_op = matrix->ts_vector[idx]->multiobject_op;
            latest_out = 1;
        }

        /* Classify holds the latest non-incomplete and its classification */
        if (classify_out == 0 && 
            classification != PASIS_ERROR_PROTO_INCOMPLETE) {
            *ts_classify_out = matrix->ts_vector[idx];
            *classify_status_out = classification;
            *classify_is_multiobject_op = matrix->ts_vector[idx]->multiobject_op;
            classify_out = 1;
        }

        /* Complete holds the latest complete timestamp */
        if (classification == PASIS_ERROR_PROTO_COMPLETE) {
            *ts_complete_out = matrix->ts_vector[idx];
            return;
        }

        /* Keeps track of oldest repairable -- it may actually be complete */
        if (classification == PASIS_ERROR_PROTO_REPAIR) {
            last_repairable = matrix->ts_vector[idx];
        }

        idx++;
    }

    /* Since good storage-nodes prune their history up to the latest complete,
     * if a repairable exists and no other complete is found then that
     * repairable must actually be complete (by deduction) 
     */
    if (last_repairable != NULL) {
        *ts_complete_out = last_repairable;
        return;
    }
    
    *ts_complete_out = NULL;
    *ts_classify_out = NULL;
    *classify_status_out = PASIS_ERROR_PROTO_NOWRITE;
    *classify_is_multiobject_op = 0;

    /* For now not sure what to do if no complete write exists */
    PASIS_DASSERT (0);

    return;
}

                               
/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
