/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"
#include "pasis_types.h"
#include "pasis_error.h"


#include "pasis_md_log.h"

int              log_key;        /* Must be global */
static FILE*     log_file=NULL;


#define USE_LOG_FILE 0


pasis_error_t
pasis_md_log_init (p_uint32 node_id)
{
#if USE_LOG_FILE > 0
    char log_filename[64];

    sprintf (log_filename, "log-%d", node_id);
    log_file = fopen (log_filename, "w");
    
    if (log_file == NULL) {
        PASIS_PRINT_ERR ("Error: creating log file\n");
        return errno;
    }
#else
    log_file = NULL;
#endif

    /* it simplifies world when we always create the log_key */
    st_key_create(&log_key, NULL);

    return PASIS_SUCCESS;
}


void
pasis_md_log_msg (char *fmt, ...) 
{
#if USE_LOG_FILE > 0
    pasis_md_log_t *log_info;
    va_list ap;

    log_info = (pasis_md_log_t *)st_thread_getspecific(log_key);

    if (log_info == NULL || log_file == NULL) return;

    fprintf (log_file, "%u: ", log_info->op_id);
    va_start(ap, fmt);
    vfprintf (log_file, fmt, ap);
    va_end (ap);
    fprintf (log_file, "\n");

    fflush (log_file);
#endif
}


void
pasis_md_log_matrices (pasis_md_hist_matrix_t *matrix, 
                       unsigned int num)
{
#if USE_LOG_FILE > 0
    int k, i;
    
    pasis_md_log_msg ("History matrices [%u]", num);
    for (k = 0; k < num; k++) {
        pasis_md_log_msg ("  Matrix %d: %u entries", k, matrix[k].ts_num);
        for (i = 0; i < matrix[k].ts_num; i++) {
            pasis_md_log_msg ("  [%d] :  %llu  (bitmap=%#llx) %s",
                              i, matrix[k].ts_vector[i]->time,
                              matrix[k].ts_bitmap[i], 
                              matrix[k].ts_vector[i]->barrier_flag ? 
                              "[BARRIER]" : "");
        }
    }
#endif
}


void
pasis_md_log_vectors (pasis_md_hist_vector_t *vec,
                      unsigned int num)
{
#if USE_LOG_FILE > 0
    int k, i;
    
    pasis_md_log_msg ("History vector [%u]", num);
    for (k = 0; k < num; k++) {
        pasis_md_log_msg ("  Vector %d: %u entries", k, vec[k].history_num);
        for (i = 0; i < vec[k].history_num; i++) {
            pasis_md_log_msg ("  [%d] : %llu %s", i, 
                              vec[k].histories[i].time,
                              vec[k].histories[i].barrier_flag ? 
                              "[BARRIER]" : "");
        }
    }
#endif
}



void
pasis_md_log_shutdown (void)
{
#if USE_LOG_FILE > 0
    fclose (log_file);
#endif
}
