/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"
#include "pasis_types.h"
#include "pasis_error.h"
#include "pasis_timer.h"
#include "pasis_md_stats.h"
#include "rpc_common.h"


#define PASIS_MD_STATS_TIMER_NUM 2
#define PASIS_MD_STATS_INCREMENT 0
#define PASIS_MD_STATS_GET       1

typedef struct pasis_md_stats_stats_s {
    pasis_timer_state_t  *timers;
    /* If you add any counters, reset them in pasis_fe_stats_dump */
} pasis_md_stats_stats_t;

static pasis_md_stats_stats_t meta_stats;




typedef struct pasis_md_stats_s {
    p_int32   num_buckets;
    p_uint32  bucket_size; /* in us */
    p_int32   current_bucket;
    p_uint32 *buckets;
} pasis_md_stats_t;


static pasis_md_stats_t stats;
static st_utime_t current_bucket_start_time;


pasis_error_t
pasis_md_stats_init(p_uint32 num_buckets,
                    p_uint32 bucket_size) /* in ms */
{
    if (num_buckets * sizeof(p_uint32) > RPC_MAX_PAYLOAD) {
        fprintf(stderr, "%s: buffer overflow would occur in get()\n",
                __FUNCTION__);
        exit(-1);
    }

    stats.num_buckets = num_buckets;
    stats.bucket_size = bucket_size * 1000;
    stats.current_bucket = 0;
    stats.buckets = malloc(sizeof(p_uint32) * num_buckets);
    if (stats.buckets == NULL) {
        fprintf(stderr, "%s: couldn't allocate stats.bucket\n",
                __FUNCTION__);
        exit(-1);
    }
    memset(stats.buckets, 0, sizeof(p_uint32) * num_buckets);
    
    current_bucket_start_time = st_utime();

    meta_stats.timers = Pasis_TimerInit ("Pasis_MD_Statistics_Timers",
                                         PASIS_MD_STATS_TIMER_NUM, 0);
    Pasis_TimerRegister (meta_stats.timers, PASIS_MD_STATS_INCREMENT,
                         "Increment"); 
    Pasis_TimerRegister (meta_stats.timers, PASIS_MD_STATS_GET, "Get");

    return PASIS_SUCCESS;
}


static void print_buckets(void)
{
    p_int32 i;

    printf("Number of requests (%u buckets, %u ms each)\n[",
           stats.num_buckets, stats.bucket_size / 1000);
    for (i = 0; i < stats.num_buckets; ++i) {
        /* Add stats.num_buckets to ensure we have a positive index*/
        printf("%u, ",
               stats.buckets[(stats.current_bucket + stats.num_buckets - i) %
                             stats.num_buckets]);
    }
    printf("]\n");
}


void
pasis_md_stats_shutdown(void)
{
    Pasis_TimerShutdown(meta_stats.timers);

    print_buckets();
}


void pasis_md_stats_callback(void)
{
    st_utime_t now = st_utime();
    st_utime_t diff = now - current_bucket_start_time;
    int switched = 0;

    PASIS_TIMER_START ();

    /* XXX: if we had a lot of buckets, the following loop could be slow. But I
     * don't think one will use many buckets. */
    while (diff >= stats.bucket_size) {
        /* Switch to next bucket. In case there has been a "quiet" time with no
         * requests, we may skip more than one buckets. */
        
        stats.current_bucket = (stats.current_bucket + 1 ) %
            stats.num_buckets;

        /* Reset bucket count */
        stats.buckets[stats.current_bucket] = 0;

        current_bucket_start_time += stats.bucket_size;
        diff -= stats.bucket_size;

        ++switched;
    }

    stats.buckets[stats.current_bucket]++;

#if 0
    if (switched) {
        printf("Switched over by %u buckets; current bucket = %u\n",
               switched, stats.current_bucket);
        print_buckets();
    }
#endif

    PASIS_TIMER_STOP (meta_stats.timers, PASIS_MD_STATS_INCREMENT);
}


void pasis_md_stats_get(pasis_md_read_op_args_t *args,
                        pasis_md_read_op_res_t  *res)

{
    pasis_md_get_stats_args_t *get_stats_args;
    pasis_md_get_stats_res_t  *get_stats_res;
    pasis_error_t              error;
    p_uint32                   real_node_id;
    p_int32 num_elements_to_return, i;

    PASIS_TIMER_START ();

    get_stats_args = &args->op_args.get_stats;
    get_stats_res  = &res->op_res.get_stats;

    /* Setup crucial return fields right away */
    res->op_type = PASIS_MD_GET_STATS;
    res->history_vec_num = 0;
    memset(&res->ts, 0, sizeof (pasis_md_hts_t));
    res->op_status = (pasis_md_rpc_stat_t)PASIS_SUCCESS;
    res->status = (pasis_md_rpc_stat_t)PASIS_SUCCESS;

    if (get_stats_args->res_max_num_buckets * 4 > RPC_MAX_PAYLOAD) {
        /* Avoid buffer overflow since we're using worker thread's "data" buffer
         * which has RPC_MAX_PAYLOAD size. */
        get_stats_args->res_max_num_buckets = RPC_MAX_PAYLOAD / 4;
    }

    num_elements_to_return = get_stats_args->res_max_num_buckets;
    if (num_elements_to_return > stats.num_buckets) {
        num_elements_to_return = stats.num_buckets;
    }

    get_stats_res->num_buckets = num_elements_to_return;
    get_stats_res->bucket_size = stats.bucket_size / 1000;

    /* We reverse the order of elements when giving it to the user: the most
     * recent element is the first one in the returned list. */
    for (i = 0; i < num_elements_to_return; ++i) {
        /* Add stats.num_buckets to ensure we have a positive index*/
        *(get_stats_res->res + i) =
            stats.buckets[(stats.current_bucket + stats.num_buckets - i)
                          % stats.num_buckets];
    }

    PASIS_TIMER_STOP (meta_stats.timers, PASIS_MD_STATS_GET);
}
