/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"
#include "pasis_lists.h"
#include "pasis_types.h"
#include "pasis_timer.h"

#include "pasisio.h"

#include "pasis_md_prot.h"

#include "pasis_md_s4.h"
#include "pasis_md_auth.h"
#include "pasis_md_dir.h"
#include "pasis_md_fid_lock.h"
#include "pasis_md_object_sync.h"
#include "pasis_md_ops.h"
#include "pasis_md_validation.h"
#include "pasis_md_work.h"
#include "pasis_md_log.h"
#include "pasis_md_delay.h"
#include "pasis_md_hist_matrix.h"
#include "pasis_md_memory_versioning.h"

#include "pasis_md_err.h"


#define PASIS_MD_WORK_TIMER_NUM            5

#define PASIS_MD_WORK_SYNTH_QUERY          0
#define PASIS_MD_WORK_GET_HIST_GEN_AUTH    1
#define PASIS_MD_WORK_LOOKUP_AND_LOCK      2
#define PASIS_MD_WORK_VALIDATE_UPDATE      3
#define PASIS_MD_WORK_EXECUTE_SYNTH_UPDATE 4


typedef struct pasis_md_work_stats_s {
    pasis_timer_state_t  *timers;
    /* If you add any counters, reset them in pasis_fe_work_stats_dump */
} pasis_md_work_stats_t;

static pasis_md_work_stats_t stats;

static pasis_timespec_t gethistgenauth_1_ts, gethistgenauth_2_ts;
static unsigned int gethistgenauth_num=0;



/* Protocol/linkage records; helps us locate other metadata S-N's and know the
 * protocol membership.  (XXX: When we want R-W versatility for Q/U, we'll have
 * to get rid of this global variable and make this part of a client's
 * arguments.  There is a bit of a bootstrap problem of how the client knows
 * this, but for the MD, it has to do lookup() first (has to, to get OID
 * anyways). */ 
static pasis_md_loc_t *distribution = NULL;


inline static void
calculate_op_hash (u_int32_t     op_size,
                   u_int32_t    *encoded_op,
                   p_byte       *hash) /* Out arg */
{
    pasis_md_auth_hash (op_size * 4, (p_byte*)encoded_op, hash);
}


inline static void
calculate_matrix_hashes (pasis_md_hist_matrix_t matrix[MD_MAX_OBJS_PER_REQ],
                         unsigned int            num_matrices,
                         p_byte                 *hash) /* Out arg */
{
    if (num_matrices == 1) {
        pasis_md_auth_hash_matrix (matrix, hash);
    } else {
        p_byte   intermediate_hashes[MD_MAX_OBJS_PER_REQ][PASIS_HASH_SIZE];

        for (p_uint32 i = 0; i < num_matrices; i++) {
            pasis_md_auth_hash_matrix (&matrix[i],
                                       (p_byte *)&intermediate_hashes[i]);
        }

        pasis_md_auth_hash (PASIS_HASH_SIZE * num_matrices, 
                            (p_byte *)intermediate_hashes, hash);
    }
}


static void
set_history_vector(p_uint32                 node_id,
                   MDVersionedObject       *object,
                   pasis_md_hist_vector_t  *history_vector)
{
    pasis_timer_t    timer;
    pasis_timespec_t ts;

    PASIS_TIMER_START ();
    PASIS_TIMER_START_TM (&timer);

    ++gethistgenauth_num;

    history_vector->node_id  = node_id;
    history_vector->auth_num = object->object_attrs.distribution.protocol.N;

    /* Get history entries */
    MDVersion latest;
    latest.setLatest();
    object->get_history(latest,
                        PASIS_MD_HIST_MAX_VERSIONS,
                        history_vector->histories,
                        &history_vector->history_num,
                        history_vector->authenticators);
    PASIS_DASSERT (history_vector->history_num <= PASIS_MD_HIST_MAX_VERSIONS);

    PASIS_TIMER_STOP_TM (&timer);
    PASIS_TIMER_ELAPSED_TS (&timer, &ts);
    PASIS_TIMER_ADD_TS (&ts, &gethistgenauth_1_ts, &gethistgenauth_1_ts);

    PASIS_TIMER_START_TM (&timer);

#if STORE_AUTHENTICATORS_IN_HIST_TABLE > 0
    // Authenticators already returned in get_history
#else
    /* Generate authenticators */
    pasis_md_auth_generate (node_id,
                            history_vector->history_num,
                            history_vector->histories,
                            history_vector->auth_num,
                            history_vector->authenticators);
#endif

    PASIS_TIMER_STOP_TM (&timer);
    PASIS_TIMER_ELAPSED_TS (&timer, &ts);
    PASIS_TIMER_ADD_TS (&ts, &gethistgenauth_2_ts, &gethistgenauth_2_ts);

    PASIS_TIMER_STOP (stats.timers, PASIS_MD_WORK_GET_HIST_GEN_AUTH);
}


static inline void
set_history_vectors(p_uint32                node_id,
                    unsigned int            num_fids,
                    MDVersionedObject      *objects[MD_MAX_OBJS_PER_REQ],
                    pasis_md_hist_vector_t  vectors[MD_MAX_OBJS_PER_REQ])
{
    for (p_uint32 i = 0; i < num_fids; i++) {
        set_history_vector(node_id, objects[i], &vectors[i]);
    }
}
                      

static void
slow_repair_generate_ts(pasis_md_hts_t  *new_hts,
                        pasis_md_hts_t  *repairable_hts,
                        pasis_md_hts_t  *latest_hts)
{
    /* Since this is a copy-forward, the object/operation/matrix hashes are the
     * same as the ones in the timestamp of the operation we're repairing. */
    md_hts_copy(new_hts, repairable_hts);

    /* Timestamp is 1 + the latest time in the OHS */
    new_hts->time = latest_hts->time + 1;

    /* do i copy-forward barriers? don't think so...*/
    PASIS_DASSERT(repairable_hts->barrier_flag == 0);
}


/* My apologies to anyone who has to look at this function (besides Garth and I)
 * -mabdelmalek, 6/22/2004 */
static pasis_error_t
validate_write_op (p_uint32                node_id,
                   p_uint32                num_objects,
                   MDVersionedObject      *objects[MD_MAX_OBJS_PER_REQ],
                   p_byte                 *op_hash,
                   pasis_md_hist_matrix_t  history_matrices[MD_MAX_OBJS_PER_REQ],
                   p_uint32                proof_fid_indices[MD_MAX_OBJS_PER_REQ],
                   pasis_md_hist_matrix_t  proof_matrices[MD_MAX_OBJS_PER_REQ],
                   pasis_md_hts_t         *lcwts_hts[MD_MAX_OBJS_PER_REQ],
                   pasis_md_hts_t         *latest_hts[MD_MAX_OBJS_PER_REQ],
                   pasis_md_hts_t         *candidate_hts[MD_MAX_OBJS_PER_REQ],
                   int                    *already_executed,
                   int                    *doing_repair,
                   int                    *doing_fast_repair,
                   p_uint32                chunk_size)
{
    PASIS_TIMER_START ();

    for (p_uint32 i = 0; i < num_objects; i++) {
        pasis_error_t error;
        pasis_md_hts_t new_hts, *sync_ts = NULL;

        /* OHS classification, authenticators check, and stale RH check */
        error = pasis_md_validate_hist_matrix (node_id, objects[i],
                                               &distribution->protocol,
                                               &history_matrices[i],
                                               proof_fid_indices[i],
                                               &proof_matrices[i], op_hash,
                                               0, &lcwts_hts[i],
                                               &latest_hts[i], &candidate_hts[i],
                                               already_executed, doing_repair,
                                               doing_fast_repair);
        if (error) {
            PASIS_TIMER_STOP (stats.timers, PASIS_MD_WORK_VALIDATE_UPDATE);
            return error;
        }

        /* Handle repair case */
        if (*doing_repair) {
            if (*already_executed) {
                if (*doing_fast_repair) {
                    /* nothing to do for fast repairs if we've executed op
                     * already */
                    /* printf("%s: repair and already executed; fast repair\n",
                     *      __FUNCTION__); */
                } else {
                    /* copy-forward + new TS for slow repairs */

                    /* Ok, here's where we're at: we received an update that
                     * signalled we should do slow repair.  We've already
                     * executed the repairable operation and it's still in our
                     * history, so we just copy-forward.  */
                    pasis_md_hts_t new_hts;

                    //printf("%s: repair and already executed; slow repair; "
                    //       "doing copy-forward\n", __FUNCTION__);

                    /* Generate the new timestamp */
                    slow_repair_generate_ts(&new_hts, candidate_hts[i],
                                            latest_hts[i]);

                    /* Before inserting the new history entry need to copy forward */
                    MDVersion new_version(&new_hts),
                        candidate_version(candidate_hts[i]),
                        lcwts_version(lcwts_hts[i]);
                    objects[i]->copy_forward(new_version, candidate_version,
                                             lcwts_version);

                    //printf("%s: the versions we have are: %s\n", __FUNCTION__,
                    //       objects[0]->dump_history());
                }
            } else { /* !already_executed */
                /* This is a repair which needs us to retrieve a version. Object
                 * sync below. */
                sync_ts = candidate_hts[i];
                //printf("%s: repair and we need version %lld; will object sync\n",
                //       __FUNCTION__, candidate_hts[i]->time);
                if (*doing_fast_repair) {
                    /* for fast repairs, the timestamp of the version we're
                     * object syncing is the same as the version we want */
                    md_hts_copy (&new_hts, candidate_hts[i]);
                } else {
                    /* for slow repairs, the timestamp of the new version is
                     * equal to the new version timestamp, which is higher than
                     * the barrier */
                    slow_repair_generate_ts (&new_hts, candidate_hts[i],
                                             latest_hts[i]);
                }
            }
        } else {
            /* in case we object sync because we don't have the conditioned-on
             * version, setup new_hts and sync_ts */
            md_hts_copy (&new_hts, lcwts_hts[i]);
            sync_ts = lcwts_hts[i];
        }

        /* Now we check if object syncing is needed.  It is required if 1) the
         * conditioned-on timestamp is not in our replica history, or 2) we're
         * performing repair, and we're missing the repairable version.  For the
         * latter case, we've already done the test of whether we have the
         * necessary version or not. For the former case, we need to check our
         * RH. */
        bool do_object_sync = false;
        MDVersion sync_version(lcwts_hts[i]);
        if (*doing_repair) {
            /* no need to do a has_version, since we've already checked if we
             * have the necessary version or not (in already_executed) */
        } else {
            /* Normal-case operation; check if we have the conditioned-on TS */
            do_object_sync = objects[i]->has_version(sync_version) == false;
        }

        /* Do object syncing if 1) we don't have cond_ts, or 2) we're doing
         * repair and we need to retrieve a version */
        if (do_object_sync || (*doing_repair && !*already_executed)) {
            error = pasis_md_object_sync_client (node_id, objects[i],
                                                 sync_ts, lcwts_hts[i],
                                                 &new_hts, &history_matrices[i]);

            if (error) {
#if 1
                PASIS_DPRINTF (PASIS_DEBUG_S4_MD_OBJECT_SYNC, 3,
#else
                printf(
#endif
                       "%s: object sync failed (%u)\n", __FUNCTION__, error);
                PASIS_TIMER_STOP (stats.timers, PASIS_MD_WORK_VALIDATE_UPDATE);
                return error;
            }

            /* Try to check if we have the conditioned-on TS again */
            if (objects[i]->has_version(sync_version) == false) {
                PASIS_DPRINTF (PASIS_DEBUG_S4_MD_OBJECT_SYNC, 3,
                               "%s: object sync didn't help\n", __FUNCTION__);
                PASIS_TIMER_STOP (stats.timers, PASIS_MD_WORK_VALIDATE_UPDATE);
                // could happen because of concurrency, the conditioend-on
                // version is not available anymore
                return PASIS_MD_ERR_OBJSYNC_NOT_EXIST;
            }
        }
    }
    PASIS_TIMER_STOP (stats.timers, PASIS_MD_WORK_VALIDATE_UPDATE);
    return PASIS_SUCCESS;
}


inline static void
lookup_and_lock_objects (p_uint32            node_id,
                         p_uint32            n,
                         pasis_md_fid_t    **fids,
                         p_uint32            chunk_size, /* for synth object */
                         MDVersionedObject **objects)
{
    PASIS_TIMER_START ();

    // 1) First, lookup (and create, if object DNE) all objects.  It's OK to
    // create objects, even if the operation doesn't succeed, since create only
    // happens in-memory (until eviction).  For Byzantine fault-tolerance, we
    // should probably delete them if the operation fails, but this is a small
    // performance hit (of having to unnecessarily create objects in back-end on
    // eviction).  The upside of aggressive object creation is simplified
    // locking, since we always have an MDVersionedObject, even for new
    // objects. -mabdelmalek 7/3/2005
    for (p_uint32 i = 0; i < n; ++i) {
        MDObjectID oid(node_id, fids[i]);
        objects[i] = md_universe->lookup(oid);
        if (objects[i] == NULL) {
            // XXX: use global distribution; not concerned with versatility
            objects[i] = md_universe->create(oid, chunk_size, distribution);
            PASIS_DASSERT(objects[i] != NULL);
        }
        // set because if object was created in read_history or write barrier,
        // chunk_size wasn't set
        if (chunk_size > 0) {
            if (objects[i]->object_attrs.block_size > 0) {
                PASIS_DASSERT(objects[i]->object_attrs.block_size == chunk_size);
            } else {
                objects[i]->object_attrs.block_size = chunk_size;
            }
        }
    }
    // 2) Lock all objects (in order, to avoid deadlocks), and do some fancy
    // operation serialization
    pasis_md_fid_lock (n, objects);

    PASIS_TIMER_STOP (stats.timers, PASIS_MD_WORK_LOOKUP_AND_LOCK);
}


inline static void
generate_new_timestamps (p_uint32                client_id,
                         pasis_md_rpc_num_t      op_type,
                         char                   *op_description,
                         p_byte                  op_hash[PASIS_HASH_SIZE],
                         unsigned int            num_fids,
                         pasis_md_hist_matrix_t  matrices[MD_MAX_OBJS_PER_REQ],
                         pasis_md_hts_t         *latest_hts[MD_MAX_OBJS_PER_REQ],
                         /* For testing, useful for client to force server to
                          * increment logical TS by more than 1 */
                         p_uint32                ts_increment, /* Usually 1 */
                         pasis_md_hts_t          new_hts[MD_MAX_OBJS_PER_REQ])
{
    bool   is_multiobject_op = num_fids > 1;
    p_byte matrices_hash[PASIS_HASH_SIZE];

    /* calculate the hash of the _matrices_ hashes */
    calculate_matrix_hashes (matrices, num_fids, matrices_hash);

    for (p_uint32 i = 0; i < num_fids; i++) {
        memcpy (new_hts[i].op_hash, op_hash, PASIS_HASH_SIZE);
        memcpy (new_hts[i].matrix_hash, matrices_hash, PASIS_HASH_SIZE);
        new_hts[i].client_id = client_id;
        // XXX: if barrier ops call this function, will need to accept
        // barrier_flag as an argument to this function
        new_hts[i].barrier_flag = 0;
        new_hts[i].multiobject_op = is_multiobject_op;
        new_hts[i].op_type = ((unsigned)op_type & 0x3fffffff);

#if DEBUG_TIMESTAMPS > 0
        memcpy (new_hts[i].op_description, op_description, DEBUG_OP_DESC_SIZE);
#endif

#if USE_OBJECT_HASHES > 0
        // object hash will be set by MDVersionedObject, if we're using it
        memset (new_hts[i].object_hash, 0, PASIS_HASH_SIZE);
#endif

        /* Increment logical time */
        new_hts[i].time = latest_hts[i]->time + ts_increment;

#if 0
        printf("%s: new TS for FID %llx\n", __FUNCTION__, *(p_uint64*)fid_info[i].fid);
        printf("%s: TS = %lld, object hash = %llx, op hash = %llx, matrix hash = %llx\n",
               __FUNCTION__, new_hts.time,
#if USE_OBJECT_HASHES > 0
               *(p_uint64*)new_hts.object_hash,
#else
               0,
#endif
               *(p_uint64*)new_hts.op_hash, *(p_uint64*)new_hts.matrix_hash);
#endif
    }
}


pasis_error_t
query_do_fast_repair (p_uint32                 real_node_id,
                      pasis_md_read_op_args_t *args,
                      MDVersionedObject       *object,
                      int                      already_executed,
                      int                      out_of_date_ohs,
                      pasis_md_hts_t          *repairable_hts,
                      pasis_md_hts_t          *lcwts_hts,
                      pasis_md_hist_matrix_t  *history_matrix)
{
    /* If we don't host the repairable version, then object sync to get it */

    if (already_executed) {
        /* No need to object sync, since we have this version already */

        /* If the client's RH is out of date, then let's execute the query at
         * the latest timestamp, NOT at the timestamp they specified.  This is
         * because they're likely in the middle of read-write concurrency and
         * will have to reissue the query anyway, so this is saving them a
         * phase. -mabdelmalek 20/3/2005 */
        if (out_of_date_ohs) {
            args->read_flags = PASIS_MD_READ_LATEST_DATA;
        }
        return PASIS_SUCCESS;
    } else {
        /* Either we haven't executed this operation (since it's not in OHS) or
         * we have executed it after the OHS was constructed, but GC has removed
         * that version.
         *
         * In either case, only attempt to object sync if our RH in the OHS is
         * current; otherwise, execute the query at the latest time */

        if (out_of_date_ohs) {
            /* printf("%s: 222 OHS out of date; issuing query at latest time\n", __FUNCTION__); */
            args->read_flags = PASIS_MD_READ_LATEST_DATA;
            return PASIS_SUCCESS;
        } else {
            pasis_error_t error;

            printf("%s: doing object syncing\n", __FUNCTION__);

            /* Do the object syncing.
             * Object syncing will create a new entry in the replica history */
            error =
                pasis_md_object_sync_client (real_node_id, object,
                                             repairable_hts, lcwts_hts,
                                             repairable_hts, history_matrix);

            /* If the other server doesn't have the version we want, then
             * another client must have updated the object and GC has removed
             * the version we wanted to sync to.  We could just return an error
             * to the client.  The client would then retry the query.  We can be
             * optimistic by performing the query at the latest version and
             * giving the client the updated history.  The client code always
             * classifies over the history and so things will work fine...
             * -mabdelmalek 20/3/2005
             */
            if (error == PASIS_MD_ERR_OBJSYNC_NOT_EXIST) {
                printf("MIKE MIKE MIKE MIKE MIKE MIKE MIKE MIKE MIKE MIKE \n");
                args->read_flags = PASIS_MD_READ_LATEST_DATA;
                error = PASIS_SUCCESS;
            }

            return error;
        }
    }
}


pasis_error_t
query_do_slow_repair (p_uint32                 real_node_id,
                      pasis_md_read_op_args_t *args,
                      MDVersionedObject       *object,
                      int                      already_executed,
                      int                      out_of_date_ohs,
                      pasis_md_hts_t          *repairable_hts,
                      pasis_md_hts_t          *lcwts_hts,
                      pasis_md_hts_t          *latest_hts,
                      pasis_md_hist_matrix_t  *history_matrix)
{
    /* Need to copy-forward the operation.  If we have already executed the
     * operation, just do a copy-forward.  If we haven't executed the operation,
     * then do object syncing and then copy-forward. */

    if (already_executed) {
        /* No need to object sync, since we have this version already */

        if (out_of_date_ohs) {
            /* If the client's RH is out of date, we don't do a copy-forward.
             * Optimistically, we'll execute their query at the latest
             * timestamp. */
            args->read_flags = PASIS_MD_READ_LATEST_DATA;
        } else {
            /* Ok, here's where we're at: we received a query that signalled
             * we should do slow repair.  We've already executed the repairable
             * operation and it's still in our history.   */
            pasis_md_hts_t new_hts;
            pasis_error_t error;

            /* Generate the new timestamp */
            slow_repair_generate_ts (&new_hts, repairable_hts, latest_hts);

            /* Before inserting the new history entry need to copy forward */
            MDVersion new_version(&new_hts), repairable_version(repairable_hts),
                lcwts_version(lcwts_hts);
            object->copy_forward(new_version, repairable_version, lcwts_version);

            //printf("%s: node %u: already executed, copied forward %llu to %llu . "
            //       "now, we have versions: %s\n", __FUNCTION__, real_node_id,
            //       repairable_hts->time, new_hts.time, object->dump_history());
            
            /* Execute the query on the latest timestamp, since the timestamp
             * the client gave us was for the "original" repairable, and we've
             * done a copy forward now which has a new timestamp */
            args->read_flags = PASIS_MD_READ_LATEST_DATA;
        }
        return PASIS_SUCCESS;
    } else { /* !already_executed */
        pasis_error_t error;

        /* Either we haven't executed this operation (since it's not in OHS) or
         * we have executed it after the OHS was constructed, but GC has removed
         * that version.

         * In either case, if the OHS is not current, then we won't try to
         * object sync. */

        if (out_of_date_ohs) {
            /* printf("%s: 222 OHS out of date; not doing slow repair\n", __FUNCTION__); */
            /* Optimistically execute query on latest version */
            args->read_flags = PASIS_MD_READ_LATEST_DATA;
            return PASIS_SUCCESS;
        } else {
            pasis_md_hts_t new_hts;

            /* Generate the new timestamp */
            slow_repair_generate_ts (&new_hts, repairable_hts, latest_hts);

            //printf("%s: node %u:  object syncing version %llu, making it version %llu\n",
            //       __FUNCTION__, real_node_id, repairable_hts->time, new_hts.time);

            /* Do the object syncing */
            error =
                pasis_md_object_sync_client (real_node_id, object,
                                             repairable_hts, lcwts_hts, &new_hts,
                                             history_matrix);

            /* If the other server doesn't have the version we want, then
             * another client must have updated the object and GC has removed
             * the version we wanted to sync to.  We could just return an error
             * to the client.  The client would then retry the query.  We can be
             * optimistic by performing the query at the latest version and
             * giving the client the updated history.  The client code always
             * classifies over the history and so things will work fine...
             * -mabdelmalek 21/3/2005
             */
            if (error == PASIS_MD_ERR_OBJSYNC_NOT_EXIST) {
                printf("MIKE MIKE MIKE MIKE MIKE MIKE MIKE MIKE MIKE MIKE \n");
                args->read_flags = PASIS_MD_READ_LATEST_DATA;
                error = PASIS_SUCCESS;
            }

            //printf("%s: node %u: after object syncing, here are our versions: %s\n",
            //       __FUNCTION__, real_node_id, object->dump_history());

            /* Execute the query on the latest timestamp, since the tiemstamp
             * the client gave us was for the "original" repairable, and we've
             * done a copy forward now which has a new timestamp */
            args->read_flags = PASIS_MD_READ_LATEST_DATA;

            return error;
        }
    }
}


pasis_error_t
query_do_repair (p_uint32                 node_id,
                 pasis_md_read_op_args_t *args,
                 /* repairing synthetic update operation? */
                 int                      repairing_synth,
                 p_uint32                 chunk_size,
                 MDVersionedObject       *object,
                 pasis_md_hist_matrix_t  *history_matrix)
{
    pasis_md_hts_t *lcwts_hts, *latest_hts, *candidate_hts;
    int already_executed, doing_repair, doing_fast_repair, out_of_date_ohs;
    pasis_error_t error;

    if (history_matrix->ts_num == 0) {
        fprintf(stderr, "%s: GETTING AN EMPTY HISTORY MATRIX!\n", __FUNCTION__);
        PASIS_DASSERT (0);
    }

    /* Validate OHS (authenticators), classify OHS, check if we've already
     * executed this operation,  and whether the client's OHS is out of date */
    error = pasis_md_validate_hist_matrix (node_id, object,
                                           &distribution->protocol,
                                           history_matrix,
                                           UINT_MAX, /* proof matrix index */
                                           NULL,     /* proof matrix */
                                           NULL,     /* op hash */
                                           0, &lcwts_hts, &latest_hts,
                                           &candidate_hts, &already_executed,
                                           &doing_repair, &doing_fast_repair);
    PASIS_DASSERT (error == PASIS_SUCCESS ||
                   error == PASIS_MD_ERR_HIST_NOT_LATEST);

    out_of_date_ohs = (error == PASIS_MD_ERR_HIST_NOT_LATEST);

    if (doing_fast_repair) {
        /* For fast repairs, perform object syncing if necessary.  Then, caller
         * will execute the query  */
        return query_do_fast_repair (node_id, args, object, already_executed,
                                     out_of_date_ohs, candidate_hts, lcwts_hts,
                                     history_matrix);
    } else if (doing_repair) {
        /* For slow repairs,  */
        return query_do_slow_repair (node_id, args, object, already_executed,
                                     out_of_date_ohs, candidate_hts, lcwts_hts,
                                     latest_hts, history_matrix);
    } else {
        // client thought they needed to do repair, but now the state of the
        // world has changed, and repair is not needed anymore.  

        if (out_of_date_ohs) {
            /* Optimistically execute query on latest version */
            args->read_flags = PASIS_MD_READ_LATEST_DATA;
        }

        return PASIS_SUCCESS;
    }
}


static void
init_return_values(pasis_md_write_op_res_t  *res,
                   pasis_md_rpc_num_t        op_type)
{
    res->op_type               = op_type;
    res->history_vec_num       = 0;
    res->op_status = (pasis_md_rpc_stat_t)PASIS_ERROR;
}


void
pasis_md_work_read_history(pasis_fe_s4_ipc_t       *ipc_op,
                           pasis_md_read_op_args_t *args,
                           pasis_md_read_op_res_t  *res)
{
    pasis_md_read_history_args_t *read_args;
    p_uint32                      i, real_node_id;
    MDVersionedObject            *objects[MD_MAX_OBJS_PER_REQ];

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4, "\nIN %s\n", __FUNCTION__);

    /* Setup crucial return fields right away */
    res->op_type = PASIS_MD_READ_HISTORY;
    res->history_vec_num = 0;
    memset (&res->ts, 0, sizeof (pasis_md_hts_t));
    res->op_res.read_history.result_num = 0;

    // Used when a single server is emulating many servers, by using the
    // incoming partition number as the node ID
    real_node_id = (node_id == PASIS_MAX_NUM_SERVERS ? args->part - 1: node_id);

    read_args = &args->op_args.read_history;

    /* Lookup objects and lock them; will create objects (in memory) if they
     * DNE. (This simplifies locking and history-retrieval.) */
    lookup_and_lock_objects(real_node_id, read_args->fid_num, read_args->fid, 0,
                            objects);

    set_history_vectors(real_node_id, read_args->fid_num, objects,
                        res->history_vectors);

    for (i = 0; i < read_args->fid_num; ++i) {
        res->op_res.read_history.statuses[i] = (pasis_md_rpc_stat_t)PASIS_SUCCESS;
    }

    pasis_md_fid_unlock(read_args->fid_num, objects);

    res->history_vec_num = read_args->fid_num;
    res->op_res.read_history.result_num = read_args->fid_num;

    res->status = res->op_status = (pasis_md_rpc_stat_t)PASIS_SUCCESS;
}


void
pasis_md_work_barrier(pasis_fe_s4_ipc_t        *ipc_op,
                      pasis_md_write_op_args_t *args,
                      pasis_md_write_op_res_t  *res)
{
    pasis_md_barrier_args_t  *barrier_args;
    pasis_md_barrier_res_t   *barrier_res;
    pasis_md_fid_data_t       fid_data[MD_MAX_OBJS_PER_REQ];
    pasis_error_t             error = PASIS_SUCCESS;
    pasis_md_hts_t           *lcwts_hts[MD_MAX_OBJS_PER_REQ];
    pasis_md_hts_t           *latest_hts[MD_MAX_OBJS_PER_REQ];
    pasis_md_hts_t           *candidate_hts[MD_MAX_OBJS_PER_REQ];
    int                       need_barrier[MD_MAX_OBJS_PER_REQ];
    pasis_md_hts_t            new_hts;
    p_uint32                  i, real_node_id;
    int                       is_multiobject_op;
    int                       already_executed, doing_repair, doing_fast_repair;
    MDVersionedObject        *objects[MD_MAX_OBJS_PER_REQ];

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4, "\nIN %s\n", __FUNCTION__);

    barrier_args = &args->op_args.barrier;
    barrier_res = &res->op_res.barrier;

    /* Setup crucial return fields right away */
    res->history_vec_num = 0;
    res->status = (pasis_md_rpc_stat_t)PASIS_SUCCESS;
    res->op_type = PASIS_MD_BARRIER;
    memset (objects, 0, sizeof(MDVersionedObject*) * MD_MAX_OBJS_PER_REQ);

    // Used when a single server is emulating many servers, by using the
    // incoming partition number as the node ID
    real_node_id = (node_id == PASIS_MAX_NUM_SERVERS ? args->part - 1 : node_id);

    if (args->history_matrices_num < 1 || 
        args->history_matrices_num > MD_MAX_OBJS_PER_REQ) {
        res->status = PASIS_MD_ERR_HIST_MATRIX_NUM;
        return;
    }

    /* Lookup objects and lock them; will create objects (in memory) if they
     * DNE. (This simplifies locking and history retrieval.) */
    lookup_and_lock_objects(real_node_id, barrier_args->fid_num,
                            barrier_args->fid, 0, objects);

    /* Validate that all FIDs in the barrier set are either all complete or
     * repairable, and that they all share the same timestamp */
    for (i = 0; i < barrier_args->fid_num; i++) {
        /* Validation is slightly different for barriers */
        error = pasis_md_validate_hist_matrix (real_node_id, objects[i],
                                               &distribution->protocol,
                                               &args->history_matrices[i],
                                               args->proof_fid_indices[i],
                                               &args->proof_matrices[i],
                                               NULL, 
                                               1, &lcwts_hts[i],
                                               &latest_hts[i],
                                               &candidate_hts[i],
                                               &already_executed, &doing_repair,
                                               &doing_fast_repair);

        if (error != PASIS_SUCCESS &&
            error != PASIS_MD_ERR_HIST_BARRIER &&
            error != PASIS_MD_ERR_HIST_NOBARRIER) {
            /* printf ("%s (%d): error=%u\n", __FUNCTION__, __LINE__, error); */
            goto exit;
        }
        
        // Why do we have this need_barrier array?  I think Garth put it in so
        // that if some objects have a barrier, from another client, while other
        // objects still need a barrier, the operation can succeed.
        // -mabdelmalek 7/31/2005 
        if (error == PASIS_MD_ERR_HIST_NOBARRIER) {
            need_barrier[i] = 0;
        } else {
            need_barrier[i] = 1;
        }

        /* These flags should never be wet when we're writing a barrier */
        PASIS_DASSERT (already_executed == 0);
        PASIS_DASSERT (doing_repair == 0);
        PASIS_DASSERT (doing_fast_repair == 0);

        /* XXX: need to handle case when only some of the FIDs are repairable.
         * This may be because some other client observed the repairable as
         * incomplete, and thus performed an op that included at least one
         * of the FIDs in this op */
        /* I don't understand the above comment (by me). -mabdelmalek, 7/31/2005 */
    }

    is_multiobject_op = barrier_args->fid_num > 1;

    /* Setup barrier timestamp */
    memset (&new_hts, 0, sizeof (pasis_md_hts_t));
    new_hts.client_id = barrier_args->client_id;
    new_hts.barrier_flag = 1;
    new_hts.multiobject_op = is_multiobject_op;
    new_hts.op_type = PASIS_MD_BARRIER;

#if DEBUG_TIMESTAMPS > 0
    /* Copy the operation description */
    snprintf (new_hts.op_description, DEBUG_OP_DESC_SIZE, "B %d %x",
              barrier_args->fid_num, *(p_uint32*)barrier_args->fid[0]);
#endif

    /* Go through all FIDs writing the barriers */
    for (i = 0; i < barrier_args->fid_num; i++) {
        
        if (need_barrier[i] == 0) {
            error = PASIS_SUCCESS;
            continue;
        }
        
        /* Set rest of timestamp */
        new_hts.time = latest_hts[i]->time + 1;

#if DEBUG_TIMESTAMPS > 0
        printf("%s: op_description: %s\n", __FUNCTION__, new_hts.op_description);
#endif

        /* Write barrier, conditioning on lcwts */
        MDVersion new_version(&new_hts), cond_version(lcwts_hts[i]);
        objects[i]->write_barrier(new_version, cond_version);

        //printf("%s: node %u: wrote a barrier; the versions we have are: %s\n",
        //       __FUNCTION__, real_node_id, objects[i]->dump_history());
    }

 exit:
    /* Always return the current history, even on operation error */
    set_history_vectors(real_node_id, barrier_args->fid_num, objects,
                        res->history_vectors);
    res->history_vec_num = barrier_args->fid_num;

    pasis_md_fid_unlock (barrier_args->fid_num, objects);

    res->status = res->op_status = (pasis_md_rpc_stat_t)error;

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4, "\nOUT %s, error=%u\n",
                   __FUNCTION__, error);
}


void
pasis_md_work_create(pasis_md_write_op_args_t *args,
                     u_int32_t                 op_size,
                     u_int32_t                *encoded_op,
                     pasis_md_write_op_res_t  *res)
{
    pasis_md_fid_t         *fid_array[2];
    pasis_md_hts_t         *lcwts_hts[2];
    pasis_md_hts_t         *latest_hts[2];
    pasis_md_hts_t         *candidate_hts[2];
    p_byte                  op_hash[PASIS_HASH_SIZE];
    pasis_error_t           error;
    pasis_md_create_args_t *create_args = &args->op_args.create;
    MDVersionedObject      *objects[2];
    p_uint32                real_node_id;
    int                     already_executed, already_executed2;
    int                     doing_repair, doing_fast_repair;

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4, "\nIN %s\n", __FUNCTION__);

    /* Setup crucial return fields right away */
    init_return_values(res, PASIS_MD_CREATE);
    memset (objects, 0, sizeof(MDVersionedObject*) * 2);

    real_node_id = (node_id == PASIS_MAX_NUM_SERVERS ? args->part - 1: node_id);

    if (args->history_matrices_num != 2) {
        PASIS_DASSERT (0);
        res->status = PASIS_MD_ERR_HIST_MATRIX_NUM;
        return;
    }

    /* Lookup objects and lock them; will create objects (in memory) if they
     * DNE. (This simplifies locking and history retrieval.) */
    fid_array[0] = create_args->parent_dir_fid;
    fid_array[1] = create_args->create_fid;
    lookup_and_lock_objects(real_node_id, 2, fid_array, PASIS_MD_DIR_BLK_SIZE,
                            objects);

    /* Calculate the operation hash early on, so the following code can test
     * whether this operation has already been executed */
    calculate_op_hash(op_size, encoded_op, op_hash);


    /* Validate the parent directory */
    /* This call does too much and needs to be split up.  Here is where we
     * classify the OHS, check the authenticators, check for stale RH, and if
     * we're doing repair we do copy-forward, and finally we do object syncing
     * if necessary.  */
    error = validate_write_op(real_node_id, 1, objects, op_hash, args->history_matrices,
                              args->proof_fid_indices, args->proof_matrices,
                              lcwts_hts, latest_hts, candidate_hts,
                              &already_executed, &doing_repair,
                              &doing_fast_repair, PASIS_MD_DIR_BLK_SIZE);
    if (error) {
        printf ("%s: validating parent dir failed (error=%d)\n",
                __FUNCTION__, error);
        goto exit;
    }

    /* Validate create FID's history matrix */
    error = pasis_md_validate_hist_matrix (real_node_id, objects[1],
                                           &distribution->protocol,
                                           &args->history_matrices[1],
                                           args->proof_fid_indices[1],
                                           &args->proof_matrices[1],
                                           op_hash, 0, &lcwts_hts[1],
                                           &latest_hts[1],
                                           &candidate_hts[1],
                                           &already_executed2,
                                           &doing_repair, &doing_fast_repair);
    if (error) {
        printf ("%s: validate_hist_matrix error=%u\n", __FUNCTION__, error);
        goto exit;
    }

    /* Some sanity checking */
    if (already_executed)  PASIS_DASSERT (already_executed2);
    if (already_executed2) PASIS_DASSERT (already_executed);

#if 0
    /* The code below doesn't seem correct.  -mabdelmalek 10/25/2004 */

    /* Check to see if this is an allowable create -- the FID should not be
     * complete.  If it is, see if this is a repair of the completed create.
     * This can be checked by comparing the op_hashes. */

    pasis_md_hts_t          zero_hts;
    memset (&zero_hts, 0, sizeof (pasis_md_hts_t));

    if (md_hts_compare (lcwts_hts[1], &zero_hts) != 0) {
        printf ("%s: create not conditioning on zero ts (%llu)\n",
                __FUNCTION__, lcwts_hts[1]->time);

        if (memcmp (op_hash, lcwts_hts[1]->op_hash, 
                    PASIS_HASH_SIZE) == 0) {
            printf ("%s: op hash validates, return success\n", __FUNCTION__);
            res->op_status = error = PASIS_SUCCESS;
        } else {
            PASIS_DASSERT (0);
            error = PASIS_ERROR;
        }

        goto exit;
    }
#endif

    // If we're performing repair, then we've already done object syncing if we
    // needed to and we don't execute the operation.  Also, if we've already
    // executed the operation, then we obviously don't execute it again.
    if (doing_repair || already_executed == 1) {
        printf ("%s: NOT executing, since this is a repair or we already executed "
                "this operation\n", __FUNCTION__);
        res->op_status = (pasis_md_rpc_stat_t)OP_OK;
    } else {
        pasis_md_hts_t new_hts[2];
        char           op_description[DEBUG_OP_DESC_SIZE];

        pasis_md_log_msg ("create name=%s, lcwts=%llu",
                          create_args->create_name, lcwts_hts[0]->time);

#if DEBUG_TIMESTAMPS > 0
        memset (op_description, 0, DEBUG_OP_DESC_SIZE);
        snprintf(op_description, DEBUG_OP_DESC_SIZE, "C %x %x %s",
                 *(p_uint32*)create_args->parent_dir_fid,
                 *(p_uint32*)create_args->create_fid, create_args->create_name);
#endif

        generate_new_timestamps(create_args->client_id, args->op_type,
                                op_description, op_hash, 2,
                                args->history_matrices, latest_hts, 1, new_hts);

        /* Execute create operation */
        res->op_status = (pasis_md_rpc_stat_t)
            pasis_md_op_create (args->mtime, objects[0], objects[1], new_hts,
                                lcwts_hts, create_args->create_name,
                                &create_args->create_attr,
                                create_args->symlink_name,
                                &res->op_res.create.create_attr,
                                &res->op_res.create.parent_attr);
        if (res->op_status) {
            error = PASIS_SUCCESS; /* redundant */
            goto exit;
        }
    }

#if 0
    printf("%s: successful CREATE for fid %x, filename %s\n", __FUNCTION__,
           *(unsigned int*)create_args->create_fid, create_args->create_name);
#endif

 exit:
    /* Always return the current history, even on operation error */
    set_history_vectors(real_node_id, 2, objects, res->history_vectors);
    res->history_vec_num = 2;
    
    pasis_md_fid_unlock(2, objects);

    res->status = (pasis_md_rpc_stat_t)error;

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4, "\nOUT %s, error=%u\n",
                   __FUNCTION__, error);
}


void
pasis_md_work_link (pasis_md_write_op_args_t *args,
                    u_int32_t                 op_size,
                    u_int32_t                *encoded_op,
                    pasis_md_write_op_res_t  *res)
{
    pasis_md_fid_t         *fid_array[2];
    pasis_md_hts_t         *lcwts_hts[2];
    pasis_md_hts_t         *latest_hts[2];
    pasis_md_hts_t         *candidate_hts[2];
    p_byte                  op_hash[PASIS_HASH_SIZE];
    pasis_error_t           error;
    pasis_md_link_args_t   *link_args;
    MDVersionedObject      *objects[2];
    char                    op_description[DEBUG_OP_DESC_SIZE];
    int                     already_executed, doing_repair, doing_fast_repair;
    p_uint32                real_node_id;

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4, "\nIN %s\n", __FUNCTION__);

    /* Setup crucial return fields right away */
    init_return_values(res, PASIS_MD_LINK);
    memset (objects, 0, sizeof(MDVersionedObject*) * 2);

    real_node_id = (node_id == PASIS_MAX_NUM_SERVERS ? args->part - 1: node_id);

    if (args->history_matrices_num != 2) {
        res->status = PASIS_MD_ERR_HIST_MATRIX_NUM;
        return;
    }

    link_args = &args->op_args.link;

    /* Lookup objects and lock them; will create objects (in memory) if they
     * DNE. (This simplifies locking and history retrieval.) */
    fid_array[0] = link_args->from_file_fid;
    fid_array[1] = link_args->to_dir_fid;
    lookup_and_lock_objects(real_node_id, 2, fid_array, PASIS_MD_DIR_BLK_SIZE,
                            objects);

    /* Calculate the operation hash early on, so the following code can test
     * whether this operation has already been executed */
    calculate_op_hash (op_size, encoded_op, op_hash);
    
    /* Validate the parent directory */
    /* This call does too much and needs to be split up.  Here is where we
     * classify the OHS, check the authenticators, check for stale RH, and if
     * we're doing repair we do copy-forward, and finally we do object syncing
     * if necessary.  */
    error = validate_write_op (real_node_id, 2, objects, op_hash,
                               args->history_matrices,
                               args->proof_fid_indices, args->proof_matrices,
                               lcwts_hts, latest_hts, candidate_hts,
                               &already_executed, &doing_repair,
                               &doing_fast_repair, PASIS_MD_DIR_BLK_SIZE);
    if (error) {
        printf ("%s: validating link failed (error=%d)\n",
                __FUNCTION__, error);
        goto exit;
    }

    if (doing_repair || already_executed == 1) {
        printf ("%s: NOT executing, since this is a repair or we already executed "
                "this operation\n", __FUNCTION__);
        res->op_status = (pasis_md_rpc_stat_t)OP_OK;
    } else {
        pasis_md_hts_t new_hts[2];

#if DEBUG_TIMESTAMPS > 0
        memset (op_description, 0, DEBUG_OP_DESC_SIZE);
        snprintf(op_description, DEBUG_OP_DESC_SIZE, "L %x %x %s",
                 *(p_uint32*)link_args->from_file_fid,
                 *(p_uint32*)link_args->to_dir_fid,
                 link_args->to_name);
#endif

        generate_new_timestamps(link_args->client_id, args->op_type,
                                op_description, op_hash, 2,
                                args->history_matrices, latest_hts, 1, new_hts);

        /* Execute link operation */
        res->op_status = (pasis_md_rpc_stat_t)
            pasis_md_op_link (args->mtime, objects[0], objects[1], new_hts,
                              lcwts_hts, link_args->to_name,
                              &res->op_res.link.from_file_attr,
                              &res->op_res.link.to_dir_attr);
        if (res->op_status) {
            error = PASIS_SUCCESS; /* redundant */
            goto exit;
        }
    }
    
 exit:
    /* Always return the current history, even on operation error */
    set_history_vectors(real_node_id, 2, objects, res->history_vectors);
    res->history_vec_num = 2;

    pasis_md_fid_unlock(2, objects);
    res->status = (pasis_md_rpc_stat_t)error;

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4, "\nOUT %s, error=%u\n",
                   __FUNCTION__, error);
}


void
pasis_md_work_unlink(pasis_md_write_op_args_t *args,
                     u_int32_t                 op_size,
                     u_int32_t                *encoded_op,
                     pasis_md_write_op_res_t  *res)
{
    pasis_md_fid_t         *fid_array[2];
    pasis_md_hts_t         *lcwts_hts[2];
    pasis_md_hts_t         *latest_hts[2];
    pasis_md_hts_t         *candidate_hts[2];
    p_byte                  op_hash[PASIS_HASH_SIZE];
    pasis_error_t           error;
    pasis_md_unlink_args_t *unlink_args;
    MDVersionedObject      *objects[2];
    char                    op_description[DEBUG_OP_DESC_SIZE];
    int                     already_executed, doing_repair, doing_fast_repair;
    p_uint32                real_node_id;

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4, "\nIN %s\n", __FUNCTION__);

    /* Setup crucial return fields right away */
    init_return_values(res, PASIS_MD_UNLINK);
    memset (objects, 0, sizeof(MDVersionedObject*) * 2);

    real_node_id = (node_id == PASIS_MAX_NUM_SERVERS ? args->part - 1: node_id);

    if (args->history_matrices_num != 2) {
        res->status = PASIS_MD_ERR_HIST_MATRIX_NUM;
        return;
    }

    unlink_args = &args->op_args.unlink;

    /* Lookup objects and lock them; will create objects (in memory) if they
     * DNE. (This simplifies locking and history retrieval.) */
    fid_array[0] = unlink_args->parent_dir_fid;
    fid_array[1] = unlink_args->name_fid;
    lookup_and_lock_objects(real_node_id, 2, fid_array, PASIS_MD_DIR_BLK_SIZE,
                            objects);

    /* Calculate the operation hash early on, so the following code can test
     * whether this operation has already been executed */
    calculate_op_hash (op_size, encoded_op, op_hash);
    
    /* Validate the parent directory */
    /* This call does too much and needs to be split up.  Here is where we
     * classify the OHS, check the authenticators, check for stale RH, and if
     * we're doing repair we do copy-forward, and finally we do object syncing
     * if necessary.  */
    error = validate_write_op (real_node_id, 2, objects, op_hash,
                               args->history_matrices,
                               args->proof_fid_indices, args->proof_matrices,
                               lcwts_hts, latest_hts, candidate_hts,
                               &already_executed, &doing_repair,
                               &doing_fast_repair, PASIS_MD_DIR_BLK_SIZE);
    if (error) {
        printf ("%s: validating unlink failed (error=%d)\n",
                __FUNCTION__, error);
        goto exit;
    }

    if (doing_repair || already_executed == 1) {
        printf ("%s: NOT executing, since this is a repair or we already executed "
                "this operation\n", __FUNCTION__);
        res->op_status = (pasis_md_rpc_stat_t)OP_OK;
    } else {
        pasis_md_hts_t new_hts[2];

#if DEBUG_TIMESTAMPS > 0
        memset (op_description, 0, DEBUG_OP_DESC_SIZE);
        snprintf(op_description, DEBUG_OP_DESC_SIZE, "U %x %x %s",
                 *(p_uint32*)unlink_args->parent_dir_fid,
                 *(p_uint32*)unlink_args->name_fid,
                 unlink_args->name);
#endif

        generate_new_timestamps(unlink_args->client_id, args->op_type,
                                op_description, op_hash, 2,
                                args->history_matrices, latest_hts, 1, new_hts);

        /* Execute unlink operation */
        res->op_status = (pasis_md_rpc_stat_t)
            pasis_md_op_unlink (args->mtime, objects[0], objects[1], new_hts,
                                lcwts_hts, unlink_args->name,
                                &res->op_res.unlink.parent_dir_attr);
        if (res->op_status) {
            error = PASIS_SUCCESS; /* redundant */
            goto exit;
        }
    }


#if 0
    printf("%s: successful UNLINK for fid %x, filename %s\n", __FUNCTION__,
           *(unsigned int*)unlink_args->name_fid, unlink_args->name);
#endif

 exit:
    /* Always return the current history, even on operation error */
    set_history_vectors(real_node_id, 2, objects, res->history_vectors);
    res->history_vec_num = 2;

    pasis_md_fid_unlock(2, objects);
    res->status = (pasis_md_rpc_stat_t)error;

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4, "\nOUT %s, error=%u\n",
                   __FUNCTION__, error);
}


void
pasis_md_work_getattr(pasis_md_read_op_args_t *args,
                      pasis_md_read_op_res_t  *res)
{
    pasis_error_t            error;
    pasis_md_getattr_args_t *getattr_args;
    MDVersionedObject       *objects[1];
    p_uint32                 real_node_id;
    MDVersion                query_version;

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4, "\nIN %s\n", __FUNCTION__);

    /* Setup crucial return fields right away */
    res->op_type = PASIS_MD_GETATTR;
    res->history_vec_num = 0;
    memset (&res->ts, 0, sizeof (pasis_md_hts_t));
    res->op_res.getattr.distribution.protocol.N = 0;
    res->op_status = (pasis_md_rpc_stat_t)PASIS_ERROR;

    getattr_args = &args->op_args.getattr;

    // Used when a single server is emulating many servers, by using the
    // incoming partition number as the node ID
    real_node_id = (node_id == PASIS_MAX_NUM_SERVERS ? args->part - 1 : node_id);

    /* Lock the object */
    lookup_and_lock_objects(real_node_id, 1, &getattr_args->fid,
                            PASIS_MD_DIR_BLK_SIZE, objects);

    /* if we received a history matrix in this query operation, then the client
     * has given us enough proof to object sync the object and satisfy their
     * request. -mabdelmalek 19/3/2005 */

    /* If the client gave us a history matrix, it means that the first phase of
     * an optimistic query failed, and we need to either perform fast or slow
     * repair. -mabdelmalek 19/3/2005  */
    if (args->history_matrices_num) {
        error = query_do_repair (real_node_id, args, 1,
                                 PASIS_MD_DIR_BLK_SIZE,
                                 objects[0], args->history_matrices);
        if (error) {
            printf("%s: query_do_repair failed (%d)\n", __FUNCTION__, error);
            goto exit;
        }
    }

    /* Construct a timestamp based on the client's requested version */
    if (args->read_flags & PASIS_MD_READ_LATEST_DATA) {
        objects[0]->get_latest(query_version);
    } else if (args->read_flags & PASIS_MD_READ_HTS_DATA) {
        query_version = MDVersion(&args->ts);
    } else {
        PASIS_DASSERT(0);
    }

    /* Check if we have the requested version */
    if (query_version.isZero() ||
        objects[0]->has_version(query_version) == false) {
        printf("%s: didn't find specified query version %lld\n", __FUNCTION__,
               args->ts.time);
        printf("%s: here are the versions we have: %s\n", __FUNCTION__,
               objects[0]->dump_history());
        error = PASIS_MD_ERR_INVAL_TIME;
        goto exit;
    }

    // Tell the client which version we're executing the query on
    md_hts_copy(&res->ts, &query_version.timestamp);


    /* Get the attributes */
    error = PASIS_SUCCESS;
    res->op_status = (pasis_md_rpc_stat_t)
        pasis_md_op_getattr (objects[0], &query_version,
                             &res->op_res.getattr.attr);
    if (res->op_status) {
        error = PASIS_SUCCESS; /* redundant, but for code readability */
        goto exit;
    }

    /* Second, copy the distribution into the result */
    memcpy (&res->op_res.getattr.distribution,
            &objects[0]->object_attrs.distribution,
            sizeof(pasis_md_loc_t));

 exit:
    /* Return history */
    set_history_vectors(real_node_id, 1, objects, res->history_vectors);
    res->history_vec_num = 1;

    pasis_md_fid_unlock (1, objects);
    res->status = (pasis_md_rpc_stat_t)error;

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4, "\nOUT %s, error=%u\n",
                   __FUNCTION__, error);
}


void
pasis_md_work_setattr (pasis_md_write_op_args_t *args,
                       u_int32_t                 op_size,
                       u_int32_t                *encoded_op,
                       pasis_md_write_op_res_t  *res)
{
    pasis_md_hts_t          *lcwts_hts[1];
    pasis_md_hts_t          *latest_hts[1];
    pasis_md_hts_t          *candidate_hts[1];
    MDVersionedObject       *objects[1];
    p_byte                   op_hash[PASIS_HASH_SIZE];
    pasis_error_t            error;
    pasis_md_setattr_args_t *setattr_args;
    char                     op_description[DEBUG_OP_DESC_SIZE];
    int                      already_executed, doing_repair, doing_fast_repair;
    p_uint32                 real_node_id;

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4, "\nIN %s\n", __FUNCTION__);

    /* Setup crucial return fields right away */
    init_return_values(res, PASIS_MD_SETATTR);
    memset (objects, 0, sizeof(MDVersionedObject*) * 1);

    real_node_id = (node_id == PASIS_MAX_NUM_SERVERS ? args->part - 1: node_id);

    if (args->history_matrices_num != 1) {
        res->status = PASIS_MD_ERR_HIST_MATRIX_NUM;
        return;
    }

    setattr_args = &args->op_args.setattr;

    /* Lookup objects and lock them; will create objects (in memory) if they
     * DNE. (This simplifies locking and history retrieval.) */
    lookup_and_lock_objects(real_node_id, 1, &setattr_args->fid,
                            PASIS_MD_DIR_BLK_SIZE, objects);

    /* Calculate the operation hash early on, so the following code can test
     * whether this operation has already been executed */
    calculate_op_hash (op_size, encoded_op, op_hash);

    /* Validate history matrices */
    /* This call does too much and needs to be split up.  Here is where we
     * classify the OHS, check the authenticators, check for stale RH, and if
     * we're doing repair we do copy-forward, and finally we do object syncing
     * if necessary.  */
    error = validate_write_op (real_node_id, 1, objects, op_hash,
                               args->history_matrices,
                               args->proof_fid_indices, args->proof_matrices,
                               lcwts_hts, latest_hts, candidate_hts,
                               &already_executed, &doing_repair,
                               &doing_fast_repair, PASIS_MD_DIR_BLK_SIZE);
    if (error) {
        printf ("%s: validating setattr failed (error=%d)\n",
                __FUNCTION__, error);
        goto exit;
    }

    if (doing_repair || already_executed == 1) {
        printf ("%s: NOT executing, since this is a repair or we already executed "
                "this operation\n", __FUNCTION__);
        res->op_status = (pasis_md_rpc_stat_t)OP_OK;
    } else {
        pasis_md_hts_t new_hts[1];

#if DEBUG_TIMESTAMPS > 0
    memset (op_description, 0, DEBUG_OP_DESC_SIZE);
    snprintf(op_description, DEBUG_OP_DESC_SIZE, "S %x",
             *(p_uint32*)setattr_args->fid);

#endif

        generate_new_timestamps(setattr_args->client_id, args->op_type,
                                op_description, op_hash, 1,
                                args->history_matrices, latest_hts, 1, new_hts);

        /* Execute setattr operation */
        res->op_status = (pasis_md_rpc_stat_t)
            pasis_md_op_setattr (objects[0], new_hts, lcwts_hts,
                                 &setattr_args->attr,
                                 setattr_args->guard_ctime,
                                 &res->op_res.setattr.attr);
        if (res->op_status) {
            error = PASIS_SUCCESS; /* redundant */
            goto exit;
        }
    }

#if 0
     printf("%s: successful setattr for fid 0x%x\n", __FUNCTION__,
           *(unsigned int*)setattr_args->fid);
#endif

exit:
    /* Always return the current history, even on operation error */
    set_history_vectors(real_node_id, 1, objects, res->history_vectors);
    res->history_vec_num = 1;

    pasis_md_fid_unlock(1, objects);
    res->status = (pasis_md_rpc_stat_t)error;

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4, "\nOUT %s, error=%u\n",
                   __FUNCTION__, error);
}


void
pasis_md_work_rename(pasis_md_write_op_args_t *args,
                     u_int32_t                 op_size,
                     u_int32_t                *encoded_op,
                     pasis_md_write_op_res_t  *res)
{
    pasis_md_fid_t         *fid_array[4];
    pasis_md_hts_t         *lcwts_hts[4];
    pasis_md_hts_t         *candidate_hts[4];
    pasis_md_hts_t         *latest_hts[4];
    MDVersionedObject      *objects[4];
    p_byte                  op_hash[PASIS_HASH_SIZE];
    pasis_error_t           error;
    pasis_md_rename_args_t *rename_args;
    p_uint32                real_node_id;

    int                     num_fids;
    MDVersionedObject      *null_obj = NULL;
    MDVersionedObject     **from_dir, **from_name, **to_dir=&null_obj,
        **to_name=&null_obj;

    char                    op_description[DEBUG_OP_DESC_SIZE];
    int                     already_executed, doing_repair, doing_fast_repair;


    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4, "\nIN %s\n", __FUNCTION__);

    /* Setup crucial return fields right away */
    init_return_values(res, PASIS_MD_RENAME);
    memset (objects, 0, sizeof(MDVersionedObject*) * 4);

    rename_args = &args->op_args.rename;

    real_node_id = (node_id == PASIS_MAX_NUM_SERVERS ? args->part - 1: node_id);

    PASIS_DPRINTF(PASIS_DEBUG_S4_MD_RENAME, 4, 
                  "%s: to_dir_valid=%d, to_name_valid=%d, num_fids=%d\n",
                  __FUNCTION__, rename_args->to_dir_valid, 
                  rename_args->to_name_valid, args->history_matrices_num);

    if (rename_args->to_dir_valid) {
        PASIS_DASSERT (md_fid_compare (rename_args->to_dir_fid,
                                       rename_args->from_dir_fid) != 0);
    }
    
    if (rename_args->to_name_valid) {
        PASIS_DASSERT (md_fid_compare (rename_args->to_name_fid,
                                       rename_args->from_name_fid) != 0);
    }

    if (args->history_matrices_num < 2 ||
        args->history_matrices_num > 4) {
        res->status = PASIS_MD_ERR_HIST_MATRIX_NUM;
        return;
    }

    /* Setup state for readdir, a little tricky due to the from and to dirs
     * potentially being the same, and since the to_file may not exist */
    num_fids = 0;

    /* Index 0 is for from_dir */
    fid_array[0] = rename_args->from_dir_fid;
    from_dir = &objects[0];
    num_fids++;

    if (rename_args->to_dir_valid) {
        /* Next index is for to_dir */
        fid_array[num_fids] = rename_args->to_dir_fid;
        to_dir = &objects[num_fids];
        num_fids++;
    }

    /* Next index is for from_name */
    fid_array[num_fids] = rename_args->from_name_fid;
    from_name = &objects[num_fids];
    num_fids++;

    if (rename_args->to_name_valid) {
        /* Next index is for to_name */
        fid_array[num_fids] = rename_args->to_name_fid;
        to_name = &objects[num_fids];
        num_fids++;
    }

    /* Lookup objects and lock them; will create objects (in memory) if they
     * DNE. (This simplifies locking and history retrieval.) */
    lookup_and_lock_objects(real_node_id, num_fids, fid_array,
                            PASIS_MD_DIR_BLK_SIZE, objects);
                       
    /* Calculate the operation hash early on, so the following code can test
     * whether this operation has already been executed */
    calculate_op_hash (op_size, encoded_op, op_hash);
    
    /* Validate the parent directory */
    /* This call does too much and needs to be split up.  Here is where we
     * classify the OHS, check the authenticators, check for stale RH, and if
     * we're doing repair we do copy-forward, and finally we do object syncing
     * if necessary.  */
    error = validate_write_op (real_node_id, num_fids, objects, op_hash,
                               args->history_matrices, args->proof_fid_indices,
                               args->proof_matrices, lcwts_hts, latest_hts,
                               candidate_hts, &already_executed, &doing_repair,
                               &doing_fast_repair, PASIS_MD_DIR_BLK_SIZE);
    if (error) {
        printf ("%s: validating for rename failed (error=%d)\n",
                __FUNCTION__, error);
        goto exit;
    }


    // If we're performing repair, then we've already done object syncing if we
    // needed to and we don't execute the operation.  Also, if we've already
    // executed the operation, then we obviously don't execute it again.
    if (doing_repair || already_executed == 1) {
        printf ("%s: NOT executing, since this is a repair or we already executed "
                "this operation\n", __FUNCTION__);
        res->op_status = (pasis_md_rpc_stat_t)OP_OK;
    } else {
        pasis_md_hts_t new_hts[4];
        char           op_description[DEBUG_OP_DESC_SIZE];

#if DEBUG_TIMESTAMPS > 0
        memset (op_description, 0, DEBUG_OP_DESC_SIZE);
        snprintf(op_description, DEBUG_OP_DESC_SIZE, "R %x %x %4s %4s",
                 *(p_uint32*)rename_args->from_dir_fid,
                 rename_args->to_dir_valid ?
                 *(p_uint32*)rename_args->to_dir_fid :
                 *(p_uint32*)rename_args->from_dir_fid,
                 rename_args->from_name, rename_args->to_name);
#endif

        generate_new_timestamps(rename_args->client_id, args->op_type,
                                op_description, op_hash, num_fids,
                                args->history_matrices, latest_hts, 1, new_hts);

        /* Execute rename operation */
        res->op_status = (pasis_md_rpc_stat_t)
                         pasis_md_op_rename (args->mtime, *from_dir, *to_dir,
                                             *from_name, *to_name,
                                             new_hts, lcwts_hts,
                                             rename_args->from_name,
                                             rename_args->to_name,
                                             &res->op_res.rename.from_dir_attr,
                                             &res->op_res.rename.to_dir_attr);
        if (res->op_status) {
            error = PASIS_SUCCESS; /* redundant */
            goto exit;
        }
    }

 exit:
    /* Always return the current history, even on operation error */
    set_history_vectors(real_node_id, num_fids, objects, res->history_vectors);
    res->history_vec_num = num_fids;

    pasis_md_fid_unlock(num_fids, objects);
    res->status = (pasis_md_rpc_stat_t)error;

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4, "\nOUT %s, error=%u\n",
                   __FUNCTION__, error);
}


void
pasis_md_work_lookup(pasis_md_read_op_args_t *args,
                     pasis_md_read_op_res_t  *res)
{
    pasis_error_t            error;
    pasis_md_lookup_args_t  *lookup_args;
    MDVersionedObject       *objects[1];
    p_uint32                 real_node_id;
    MDVersion                query_version;

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4, "\nIN %s\n", __FUNCTION__);

    /* Setup crucial return fields right away */
    res->op_type = PASIS_MD_LOOKUP;
    res->history_vec_num = 0;
    memset (&res->ts, 0, sizeof (pasis_md_hts_t));
    res->op_res.lookup.distribution.protocol.N = 0;
    res->op_status = (pasis_md_rpc_stat_t)PASIS_ERROR;

    lookup_args = &args->op_args.lookup;

    // Used when a single server is emulating many servers, by using the
    // incoming partition number as the node ID
    real_node_id = (node_id == PASIS_MAX_NUM_SERVERS ? args->part - 1 : node_id);

    /* Lock the object */
    lookup_and_lock_objects(real_node_id, 1, &lookup_args->parent_dir_fid,
                            PASIS_MD_DIR_BLK_SIZE, objects);

    /* if we received a history matrix in this query operation, then the client
     * has given us enough proof to object sync the object and satisfy their
     * request. -mabdelmalek 19/3/2005 */

    /* If the client gave us a history matrix, it means that the first phase of
     * an optimistic query failed, and we need to either perform fast or slow
     * repair. -mabdelmalek 19/3/2005  */
    if (args->history_matrices_num) {
        error = query_do_repair (real_node_id, args, 1,
                                 PASIS_MD_DIR_BLK_SIZE,
                                 objects[0], args->history_matrices);
        if (error) {
            printf("%s: query_do_repair failed (%d)\n", __FUNCTION__, error);
            goto exit;
        }
    }

    /* Construct a timestamp based on the client's requested version */
    if (args->read_flags & PASIS_MD_READ_LATEST_DATA) {
        objects[0]->get_latest(query_version);
    } else if (args->read_flags & PASIS_MD_READ_HTS_DATA) {
        query_version = MDVersion(&args->ts);
    } else {
        PASIS_DASSERT(0);
    }

    /* Check if we have the requested version */
    if (query_version.isZero() ||
        objects[0]->has_version(query_version) == false) {
        printf("%s: didn't find specified query version %lld\n", __FUNCTION__,
               args->ts.time);
        printf("%s: here are the versions we have: %s\n", __FUNCTION__,
               objects[0]->dump_history());
        error = PASIS_MD_ERR_INVAL_TIME;
        goto exit;
    }

    // Tell the client which version we're executing the query on
    md_hts_copy(&res->ts, &query_version.timestamp);

    /* Perform the lookup */
    error = PASIS_SUCCESS;
    res->op_status = (pasis_md_rpc_stat_t)
                     pasis_md_op_lookup (objects[0], &query_version,
                                         lookup_args->name,
                                         &res->op_res.lookup.fid,
                                         &res->op_res.lookup.dir_attr);
    if (res->op_status) {
        error = PASIS_SUCCESS; /* redundant */
        goto exit;
    }

    
    /* Lookup the distribution for the file */
    // We should be looking up the distribution in the object's attrs.  But for
    // now, same distribution everywhere, so just use the global one.
    memcpy (&res->op_res.lookup.distribution, distribution,
            sizeof(pasis_md_loc_t));

 exit:
    /* Return history */
    set_history_vectors(real_node_id, 1, objects, res->history_vectors);
    res->history_vec_num = 1;

#if 0
   printf("%s: lookup for filename %s returned error=%d, res->op_status=%d\n",
          __FUNCTION__, lookup_args->name, error, res->op_status);
#endif

    pasis_md_fid_unlock (1, objects);
    res->status = (pasis_md_rpc_stat_t)error;

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4, "\nOUT %s, error=%u\n",
                   __FUNCTION__, error);
}


void
pasis_md_work_readlink (pasis_md_read_op_args_t *args,
                        pasis_md_read_op_res_t  *res)
{
    unsigned int              dummy_len;
    pasis_error_t             error;
    pasis_md_readlink_args_t *readlink_args;
    MDVersionedObject        *objects[1];
    p_uint32                  real_node_id;
    MDVersion                 query_version;

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4, "\nIN %s\n", __FUNCTION__);

    /* Setup crucial return fields right away */
    res->op_type = PASIS_MD_READLINK;
    res->history_vec_num = 0;
    memset (&res->ts, 0, sizeof (pasis_md_hts_t));
    res->op_status = (pasis_md_rpc_stat_t)PASIS_ERROR;

    readlink_args = &args->op_args.readlink;

    // Used when a single server is emulating many servers, by using the
    // incoming partition number as the node ID
    real_node_id = (node_id == PASIS_MAX_NUM_SERVERS ? args->part - 1 : node_id);

    /* Lock the object */
    lookup_and_lock_objects(real_node_id, 1, &readlink_args->fid,
                            PASIS_MD_DIR_BLK_SIZE, objects);

    /* if we received a history matrix in this query operation, then the client
     * has given us enough proof to object sync the object and satisfy their
     * request. -mabdelmalek 19/3/2005 */

    /* If the client gave us a history matrix, it means that the first phase of
     * an optimistic query failed, and we need to either perform fast or slow
     * repair. -mabdelmalek 19/3/2005  */
    if (args->history_matrices_num) {
        error = query_do_repair (real_node_id, args, 1,
                                 PASIS_MD_DIR_BLK_SIZE,
                                 objects[0], args->history_matrices);
        if (error) {
            printf("%s: query_do_repair failed (%d)\n", __FUNCTION__, error);
            goto exit;
        }
    }

    /* Construct a timestamp based on the client's requested version */
    if (args->read_flags & PASIS_MD_READ_LATEST_DATA) {
        objects[0]->get_latest(query_version);
    } else if (args->read_flags & PASIS_MD_READ_HTS_DATA) {
        query_version = MDVersion(&args->ts);
    } else {
        PASIS_DASSERT(0);
    }

    /* Check if we have the requested version */
    if (query_version.isZero() ||
        objects[0]->has_version(query_version) == false) {
        printf("%s: didn't find specified query version %lld\n", __FUNCTION__,
               args->ts.time);
        printf("%s: here are the versions we have: %s\n", __FUNCTION__,
               objects[0]->dump_history());
        error = PASIS_MD_ERR_INVAL_TIME;
        goto exit;
    }

    // Tell the client which version we're executing the query on
    md_hts_copy(&res->ts, &query_version.timestamp);

    /* Read the symbolic link */
    error = PASIS_SUCCESS;
    res->op_status = (pasis_md_rpc_stat_t)
        pasis_md_op_readlink (objects[0], &query_version, &dummy_len,
                              res->op_res.readlink.name,
                              &res->op_res.readlink.fid_attr);
    if (res->op_status) {
        error = PASIS_SUCCESS;
        goto exit;
    }

 exit:
    /* Return history */
    set_history_vectors(real_node_id, 1, objects, res->history_vectors);
    res->history_vec_num = 1;

    pasis_md_fid_unlock (1, objects);
    res->status = (pasis_md_rpc_stat_t)error;
    
    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4, "\nOUT %s, error=%u\n",
                   __FUNCTION__, error);
}


void
pasis_md_work_readdir (pasis_md_read_op_args_t *args,
                       pasis_md_read_op_res_t  *res)
{
    pasis_error_t             error;
    pasis_md_readdir_args_t  *readdir_args;
    pasis_md_readdir_res_t   *readdir_res = NULL;
    pasis_md_dir_hdr_t       *hdr;
    MDVersionedObject        *objects[1];
    p_uint32                  real_node_id;
    MDVersion                 query_version;
    void *data = NULL;

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4, "\nIN %s\n", __FUNCTION__);

    readdir_args = &args->op_args.readdir;
    readdir_res  = &res->op_res.readdir;

    /* Setup crucial return fields right away */
    res->op_type = PASIS_MD_READDIR;
    res->history_vec_num = 0;
    memset (&res->ts, 0, sizeof (pasis_md_hts_t));
    readdir_res->data_size = 0;
    res->op_status = (pasis_md_rpc_stat_t)PASIS_ERROR;

    // Used when a single server is emulating many servers, by using the
    // incoming partition number as the node ID
    real_node_id = (node_id == PASIS_MAX_NUM_SERVERS ? args->part - 1 : node_id);

    /* Lock the object */
    lookup_and_lock_objects(real_node_id, 1, &readdir_args->fid,
                            PASIS_MD_DIR_BLK_SIZE, objects);

    /* if we received a history matrix in this query operation, then the client
     * has given us enough proof to object sync the object and satisfy their
     * request. -mabdelmalek 19/3/2005 */

    /* If the client gave us a history matrix, it means that the first phase of
     * an optimistic query failed, and we need to either perform fast or slow
     * repair. -mabdelmalek 19/3/2005  */
    if (args->history_matrices_num) {
        error = query_do_repair (real_node_id, args, 1,
                                 PASIS_MD_DIR_BLK_SIZE,
                                 objects[0], args->history_matrices);
        if (error) {
            printf("%s: query_do_repair failed (%d)\n", __FUNCTION__, error);
            goto exit;
        }
    }

    /* Construct a timestamp based on the client's requested version */
    if (args->read_flags & PASIS_MD_READ_LATEST_DATA) {
        objects[0]->get_latest(query_version);
    } else if (args->read_flags & PASIS_MD_READ_HTS_DATA) {
        query_version = MDVersion(&args->ts);
    } else {
        PASIS_DASSERT(0);
    }

    /* Check if we have the requested version */
    if (query_version.isZero() ||
        objects[0]->has_version(query_version) == false) {
        printf("%s: didn't find specified query version %lld\n", __FUNCTION__,
               args->ts.time);
        printf("%s: here are the versions we have: %s\n", __FUNCTION__,
               objects[0]->dump_history());
        error = PASIS_MD_ERR_INVAL_TIME;
        goto exit;
    }

    // Tell the client which version we're executing the query on
    md_hts_copy(&res->ts, &query_version.timestamp);

    /* Read the directory block */
    error = PASIS_SUCCESS;
    data = objects[0]->read(query_version, readdir_args->block_num);
    if (data == NULL) {
        res->op_status = (pasis_md_rpc_stat_t)OPERR_IO;
        error = PASIS_SUCCESS;
        goto exit;
    } else {
        memmove(readdir_res->data, data, PASIS_MD_DIR_BLK_SIZE);
        res->op_status = (pasis_md_rpc_stat_t)OP_OK;
    }

    hdr = (pasis_md_dir_hdr_t *) readdir_res->data;
    readdir_res->data_size = hdr->free_offset;

#if 1
    // Why is this hasing being done???  I don't think it's needed -mabdelmalek
    // 8/11/2005
    memset(readdir_res->hash, 0, PASIS_HASH_SIZE);
#else
    /* Hash the directory block */
    pasis_md_auth_hash (readdir_res->data_size, (p_byte*) hdr,
                        readdir_res->hash);
#endif

 exit:
    /* Return history */
    set_history_vectors(real_node_id, 1, objects, res->history_vectors);
    res->history_vec_num = 1;

    pasis_md_fid_unlock (1, objects);
    res->status = (pasis_md_rpc_stat_t)error;

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4, "\nOUT %s, error=%u\n",
                   __FUNCTION__, error);
}



static pasis_md_rpc_stat_t
execute_synth_query(p_uint32            num_fids,
                    MDVersionedObject  *objects[MD_MAX_OBJS_PER_REQ],
                    p_uint32            read_chunks[MD_MAX_OBJS_PER_REQ],
                    MDVersion           query_versions[MD_MAX_OBJS_PER_REQ],
                    p_uint32            res_size,
                    p_uint32            chunk_size,
                    p_byte             *res)
{
    pasis_md_rpc_stat_t result = (pasis_md_rpc_stat_t)OP_OK;
    bool first_time = 1;

    /* Read the specified chunks */
    for (p_uint32 i = 0; i < 32; ++i) {
        if (read_chunks[0] & (1 << i)) {
#if 0
            printf ("%s: reading chunk %d of object %x at TS %lld\n",
                    __FUNCTION__, i, *(p_uint32*)&objects[0]->oid.fid,
                    query_versions[0].timestamp.time);
#endif
            void* buffer = objects[0]->read(query_versions[0], i);

            if (first_time && res_size == chunk_size) {
                memmove(res, buffer, chunk_size);
                first_time = false;
            } else {
                memset (res, 0, res_size);
                memcpy (res, &query_versions[0].timestamp,
                        MIN(res_size, sizeof (pasis_md_hts_t)));
            }

        }
    }
    return (pasis_md_rpc_stat_t)OP_OK;
}


void
pasis_md_work_synth_query(pasis_fe_s4_ipc_t       *ipc_op,
                          pasis_md_read_op_args_t *args,
                          pasis_md_read_op_res_t  *res)
{
    pasis_error_t                error = PASIS_SUCCESS;
    pasis_md_synth_query_args_t *synth_query_args;
    pasis_md_synth_query_res_t  *synth_query_res = NULL;
    MDVersionedObject           *objects[MD_MAX_OBJS_PER_REQ];
    MDVersion                    query_version;
    p_uint32                     real_node_id;

    PASIS_TIMER_START ();

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4, "\nIN %s\n", __FUNCTION__);

    synth_query_args = &args->op_args.synth_query;
    synth_query_res  = &res->op_res.synth_query;

    /* Setup crucial return fields right away */
    res->op_type = PASIS_MD_SYNTH_QUERY;
    res->history_vec_num = 0;
    memset (&res->ts, 0, sizeof (pasis_md_hts_t));
    synth_query_res->res_size = 0;
    res->op_status = (pasis_md_rpc_stat_t)PASIS_ERROR;

    // Used when a single server is emulating many servers, by using the
    // incoming partition number as the node ID
    real_node_id = (node_id == PASIS_MAX_NUM_SERVERS ? args->part - 1 : node_id);

    // for now, query ops just take a single object
    PASIS_DASSERT(synth_query_args->num_fids == 1);

    /* Lock the object */
    lookup_and_lock_objects(real_node_id, 1, synth_query_args->fids,
                            synth_query_args->chunk_size, objects);

    /* if we received a history matrix in this query operation, then the client
     * has given us enough proof to object sync the object and satisfy their
     * request. -mabdelmalek 19/3/2005 */

    /* If the client gave us a history matrix, it means that the first phase of
     * an optimistic query failed, and we need to either perform fast or slow
     * repair. -mabdelmalek 19/3/2005  */
    if (args->history_matrices_num) {
        error = query_do_repair (real_node_id, args, 1,
                                 synth_query_args->chunk_size,
                                 objects[0], args->history_matrices);
        if (error) {
            printf("%s: query_do_repair failed (%d)\n", __FUNCTION__, error);
            goto exit;
        }
    }

    /* Construct a timestamp based on the client's requested version */
    if (args->read_flags & PASIS_MD_READ_LATEST_DATA) {
        objects[0]->get_latest(query_version);
    } else if (args->read_flags & PASIS_MD_READ_HTS_DATA) {
        query_version = MDVersion(&args->ts);
    } else {
        /* This is a read witness */
        error = PASIS_SUCCESS;
        goto exit;
    }

    /* Check if we have the requested version */
    if (query_version.isZero() ||
        objects[0]->has_version(query_version) == false) {
        printf("%s: didn't find specified query version %lld\n", __FUNCTION__,
               args->ts.time);
        printf("%s: here are the versions we have: %s\n", __FUNCTION__,
               objects[0]->dump_history());
        error = PASIS_MD_ERR_INVAL_TIME;
        goto exit;
    }

    // Tell the client which version we're executing the query on
    md_hts_copy(&res->ts, &query_version.timestamp);

    res->op_status = execute_synth_query(1, objects,
                                         synth_query_args->read_chunks,
                                         &query_version,
                                         synth_query_args->res_size,
                                         synth_query_args->chunk_size,
                                         synth_query_res->res);

    if (res->op_status) {
        res->op_status = (pasis_md_rpc_stat_t)OPERR_IO;
        error = PASIS_SUCCESS;
        goto exit;
    }

    /* Keep the CPU busy for "compute_time" ms */
    pasis_md_delay (synth_query_args->compute_time);

    synth_query_res->res_size = synth_query_args->res_size;

    if (error) goto exit;

 exit:
    /* Return history */
    set_history_vectors(real_node_id, 1, objects, res->history_vectors);
    res->history_vec_num = 1;

    pasis_md_fid_unlock (1, objects);
    res->status = (pasis_md_rpc_stat_t)error;

    PASIS_TIMER_STOP (stats.timers, PASIS_MD_WORK_SYNTH_QUERY);

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4,
                   "\nOUT %s, error=%u, op_status = %d\n",
                   __FUNCTION__, error, res->op_status);
}



inline static pasis_md_service_t
md_rpc_to_service (pasis_md_rpc_num_t rpc_num)
{
    switch (rpc_num) {
    case PASIS_MD_READ_HISTORY:
    case PASIS_MD_BARRIER:
    case PASIS_MD_OBJECT_SYNC:
        return PASIS_MD_GENERIC;
    case PASIS_MD_CREATE:
    case PASIS_MD_LINK:
    case PASIS_MD_UNLINK:
    case PASIS_MD_GETATTR:
    case PASIS_MD_SETATTR:
    case PASIS_MD_RENAME:
    case PASIS_MD_LOOKUP:
    case PASIS_MD_READLINK:
    case PASIS_MD_READDIR:
        return PASIS_MD_NFS;
    case PASIS_MD_SYNTH_QUERY:
    case PASIS_MD_SYNTH_UPDATE:
        return PASIS_MD_SYNTH;
    default:
        PASIS_DASSERT (0);
        return PASIS_MD_GENERIC;
    }
}


#if 0
static void
generate_synth_update_debug_ts()
{
#if DEBUG_TIMESTAMPS > 0
        /* Op description is in the form "SU <num FIDS> FID1 FID2 ... ARG_ SIZE*/
        memset (op_description, 0, DEBUG_OP_DESC_SIZE);
        j = snprintf(op_description, DEBUG_OP_DESC_SIZE, "SU %d ",
                     synth_update_args->num_fids);
        num_bytes_left = DEBUG_OP_DESC_SIZE - j;
        for (i = 0; i < synth_update_args->num_fids && num_bytes_left > 0; ++i) {
            char buffer[20];
            j = snprintf(buffer, sizeof(buffer), "%x ",
                         *(p_uint32*)synth_update_args->fids[i]);
            strncat(op_description, buffer, num_bytes_left);
            num_bytes_left -= j;
        }
        {char buffer[20];
        j = snprintf(buffer, sizeof(buffer), "%x", synth_update_args->modified_chunks[0]);
        strncat(op_description, buffer, num_bytes_left);
        num_bytes_left -= j;
        op_description[DEBUG_OP_DESC_SIZE - 1] = '\0';
        }
#endif    
}
#endif


static pasis_md_rpc_stat_t
execute_synth_update(p_uint32            num_fids,
                     MDVersionedObject  *objects[MD_MAX_OBJS_PER_REQ],
                     p_byte             *args,
                     p_uint32            args_size,
                     p_uint32            chunk_size,
                     p_uint32            modified_chunks[MD_MAX_OBJS_PER_REQ],
                     pasis_md_hts_t      new_hts[MD_MAX_OBJS_PER_REQ],
                     pasis_md_hts_t     *lcwts_hts[MD_MAX_OBJS_PER_REQ])
{
    PASIS_TIMER_START ();

    for (p_uint32 i = 0; i < num_fids; ++i) {
        for (p_uint32 j = 0; j < 32; ++j) {
            if (modified_chunks[i] & (1 << j)) {
#if 0
                printf ("%s: writing chunk %d (size %d) of object %x at TS "
                        "%lld (%lld)\n", __FUNCTION__, j,
                        synth_update_args->chunk_size,
                        *(p_uint32*)objects[i].oid.fid,
                        new_hts[i]->time, lcwts_hts[i]->time);
#endif

                void* buffer = objects[i]->get_memory_for_new_block();
                if (objects[i]->block_size() == args_size &&
                    args_size == chunk_size) {
                    memmove(buffer, args, args_size);
                } else {
                    memset(buffer, 0, objects[i]->block_size());
                    /* For each object, we sets its chunk data to:
                     * <conditioning TS> <chunk #> */
                    memcpy(buffer, lcwts_hts[i], MIN(objects[i]->block_size(),
                                                     sizeof(pasis_md_hts_t)));
                    if (objects[i]->block_size() >= sizeof(pasis_md_hts_t) + 1) {
                        ((p_uint8*)buffer)[sizeof(pasis_md_hts_t)] = j;
                    }
                }

                // Write the new chunk!  Damn, the Q/U interface is so clean!
                MDVersion new_version(&new_hts[i]), cond_version(lcwts_hts[i]);;
                objects[i]->write(new_version, cond_version, j, buffer);
            }
        }
    }
    
    PASIS_TIMER_STOP (stats.timers, PASIS_MD_WORK_EXECUTE_SYNTH_UPDATE);

    return (pasis_md_rpc_stat_t)OP_OK;
}


void
pasis_md_work_synth_update(pasis_md_write_op_args_t *args,
                           u_int32_t                 op_size,
                           u_int32_t                *encoded_op,
                           pasis_md_write_op_res_t  *res)
{
    pasis_md_hts_t               *lcwts_hts[MD_MAX_OBJS_PER_REQ];
    pasis_md_hts_t               *latest_hts[MD_MAX_OBJS_PER_REQ];
    pasis_md_hts_t               *candidate_hts[MD_MAX_OBJS_PER_REQ];
    p_byte                        op_hash[PASIS_HASH_SIZE];
    pasis_error_t                 error;
    pasis_md_synth_update_args_t *synth_update_args=&args->op_args.synth_update;
    pasis_md_synth_update_res_t  *synth_update_res = &res->op_res.synth_update;
    int                           already_executed;
    int                           doing_repair, doing_fast_repair;
    MDVersionedObject            *objects[MD_MAX_OBJS_PER_REQ];
    pasis_md_service_t            service;
    p_uint32                      real_node_id;
#if DEBUG_TIMESTAMPS > 0
    int                           num_bytes_left;
#endif

    PASIS_DPRINTF (PASIS_DEBUG_S4_MD_FUNC_NAMES, 4, "\nIN %s\n", __FUNCTION__);

    /* Setup crucial return fields right away */
    init_return_values(res, PASIS_MD_SYNTH_UPDATE);
    memset (objects, 0, sizeof(MDVersionedObject*) * MD_MAX_OBJS_PER_REQ);
    synth_update_res->res_size = 0;

    // Used when a single server is emulating many servers, by using the
    // incoming partition number as the node ID
    real_node_id = (node_id == PASIS_MAX_NUM_SERVERS ? args->part - 1 : node_id);

    if (args->history_matrices_num != synth_update_args->num_fids) {
        res->status = PASIS_MD_ERR_HIST_MATRIX_NUM;
        return;
    }

    service = md_rpc_to_service(args->op_type);

    /* Lookup objects and lock them; will create objects (in memory) if they
     * DNE. (This simplifies locking and history retrieval.) */
    lookup_and_lock_objects(real_node_id, synth_update_args->num_fids,
                            synth_update_args->fids,
                            synth_update_args->chunk_size, objects);

    /* Calculate the operation hash early on, so the following code can test
     * whether this operation has already been executed */
    calculate_op_hash(op_size, encoded_op, op_hash);

    /* This call does too much and needs to be split up.  Here is where we
     * classify the OHS, check the authenticators, check for stale RH, and if
     * we're doing repair we do copy-forward, and finally we do object syncing
     * if necessary.  */
    error = validate_write_op(real_node_id, synth_update_args->num_fids,
                              objects, op_hash, args->history_matrices,
                              args->proof_fid_indices, args->proof_matrices,
                              lcwts_hts, latest_hts, candidate_hts,
                              &already_executed, &doing_repair,
                              &doing_fast_repair, synth_update_args->chunk_size);
    if (error) {
        //printf("%s: validate_write_op failed with error %d\n", __FUNCTION__,
        //       error);
        goto exit;
    }

    // If we're performing repair, then we've already done object syncing if we
    // needed to and we don't execute the operation.  Also, if we've already
    // executed the operation, then we obviously don't execute it again.
    if (doing_repair || already_executed) {
        //printf ("%s: NOT executing, since this is a repair or we already executed "
        //        "this operation (%d, %d)\n", __FUNCTION__, doing_repair,
        //        already_executed);
        res->op_status = (pasis_md_rpc_stat_t)OP_OK;
    } else {
        pasis_md_hts_t new_hts[MD_MAX_OBJS_PER_REQ];
        char op_description[DEBUG_OP_DESC_SIZE];
        op_description[0] = '\0';

        generate_new_timestamps(synth_update_args->client_id, args->op_type,
                                op_description, op_hash,
                                synth_update_args->num_fids,
                                args->history_matrices, latest_hts,
                                synth_update_args->ts_increment, new_hts);

        res->op_status = execute_synth_update(synth_update_args->num_fids,
                                              objects, synth_update_args->args,
                                              synth_update_args->arg_size,
                                              synth_update_args->chunk_size,
                                              synth_update_args->modified_chunks,
                                              new_hts, lcwts_hts);

        if (res->op_status) {
            res->op_status = (pasis_md_rpc_stat_t)OPERR_IO;
            error = PASIS_SUCCESS;
            goto exit;
        }

        //printf("%s: node %u: first object has the following versions: %s\n",
        //       __FUNCTION__, real_node_id, objects[0]->dump_history());
    }

    /* Keep the CPU busy for "compute_time" ms */
    pasis_md_delay(synth_update_args->compute_time);

    if (doing_repair) {
        /* if doing repair, don't return data to client, since they were in
         * their read_history and may not have setup the pointers for returned
         * data yet... */
    } else {
        /* Zero the result data, and store timestamp of first FID in the beginning */
        synth_update_res->res_size = synth_update_args->res_size;
        memset(synth_update_res->res, 0, synth_update_res->res_size);
        memcpy(synth_update_res->res, lcwts_hts[0],
               MIN(synth_update_args->res_size, sizeof (pasis_md_hts_t)));
    }

 exit:
    /* Always return the current history, even on operation error */
    set_history_vectors(real_node_id, synth_update_args->num_fids, objects,
                        res->history_vectors);
    res->history_vec_num = synth_update_args->num_fids;

    pasis_md_fid_unlock(synth_update_args->num_fids, objects);

    res->status = (pasis_md_rpc_stat_t)error;

    PASIS_DPRINTF(PASIS_DEBUG_S4_MD_FUNC_NAMES, 4,
                  "\nOUT %s, error=%u, op_status = %d\n",
                  __FUNCTION__, error, res->op_status);
}


pasis_error_t
pasis_md_work_init (pasis_md_loc_t  *_distribution,
                    p_uint64         block_cache_size,
                    p_uint64         history_cache_size)
{
    distribution = _distribution;
    MDMemoryManager::init(block_cache_size, history_cache_size);

    return PASIS_MD_SUCCESS;
}


void
pasis_md_work_shutdown (void)
{
    MDMemoryManager::shutdown();
}


void
pasis_md_work_stats_init (void)
{
    /* Set up timers */
    stats.timers = Pasis_TimerInit ("Pasis_MD_Work_Timers",
                                    PASIS_MD_WORK_TIMER_NUM, 0);

    Pasis_TimerRegister (stats.timers, PASIS_MD_WORK_SYNTH_QUERY, "Synth_query"); 
    Pasis_TimerRegister (stats.timers, PASIS_MD_WORK_GET_HIST_GEN_AUTH,
                         "Get hist and gen auth");
    Pasis_TimerRegister (stats.timers, PASIS_MD_WORK_LOOKUP_AND_LOCK,
                         "Lookup and lock objects");
    Pasis_TimerRegister (stats.timers, PASIS_MD_WORK_VALIDATE_UPDATE,
                         "validate write op");
    Pasis_TimerRegister (stats.timers, PASIS_MD_WORK_EXECUTE_SYNTH_UPDATE,
                         "synth update execution");

    memset (&gethistgenauth_1_ts, 0, sizeof (pasis_timespec_t));
    memset (&gethistgenauth_2_ts, 0, sizeof (pasis_timespec_t));
}


void
pasis_md_work_stats_dump (int shutdown_flag)
{
    PASIS_PRINT_INFO ("Gethistgenauth phase 1 :  %u.%09u\n", 
                      gethistgenauth_1_ts.ts_sec, gethistgenauth_1_ts.ts_nsec);
    PASIS_PRINT_INFO ("Gethistgenauth phase 2 :  %u.%09u\n", 
                      gethistgenauth_2_ts.ts_sec, gethistgenauth_2_ts.ts_nsec);
    PASIS_PRINT_INFO ("Num get hist + gen auth:  %u\n", gethistgenauth_num);

    if (shutdown_flag) {
        Pasis_TimerShutdown (stats.timers);
    } else {
        Pasis_TimerPrintAndReset (stats.timers);
        gethistgenauth_1_ts.ts_sec = gethistgenauth_1_ts.ts_nsec = 0;
        gethistgenauth_2_ts.ts_sec = gethistgenauth_2_ts.ts_nsec = 0;
        gethistgenauth_num = 0;
    }
}


/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
