/*
 * The PASIS Query/Update Project
 * Copyright (c) 2007, Carnegie Mellon University.
 * All rights reserved.
 * http://www.pdl.cmu.edu/  (Parallel Data Lab at Carnegie Mellon)
 *
 * This software is being provided by the copyright holders under the
 * following license. By obtaining, using and/or copying this software,
 * you agree that you have read, understood, and will comply with the
 * following terms and conditions:
 *
 * Permission to reproduce, use, and prepare derivative works of this
 * software is granted provided the copyright and "No Warranty" statements
 * are included with all reproductions and derivative works and associated
 * documentation. This software may also be redistributed without charge
 * provided that the copyright and "No Warranty" statements are included
 * in all redistributions.
 *
 * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
 * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
 * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
 * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
 * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
 * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
 * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
 * OR DOCUMENTATION.
 */

#include "pasis_config.h"
#include "pasis_debug.h"
#include "pasis_types.h"

#include "pasis_md_dir.h"
#include "pasis_md_prot.h"
#include "pasis_md_xdr.h"

#include "rpc_common.h"


static inline void
xdr_adjust_req_iovec(rpc_req_t *req, u_int32_t *p)
{
    req->req_iov[req->req_iov_cur].iov_len = ((char *) p - 
                             (char *) req->req_iov[req->req_iov_cur].iov_base);
    req->req_iov_cur++;
    req->req_iov_len++;
    return;
}

static inline void
xdr_adjust_reply_iovec(rpc_req_t *req, u_int32_t *p)
{
    req->reply_iov[req->reply_iov_cur].iov_len = ((char *) p - 
                         (char *) req->reply_iov[req->reply_iov_cur].iov_base);
    req->reply_iov_cur++;
    req->reply_iov_len++;
    return;
}


static inline u_int32_t *
xdr_encode_fid (u_int32_t *p, 
                pasis_md_fid_t *fid)
{
    memcpy (p, fid->data, PASISIO_FILEID_SIZE);
    return (p + XDR_QUADLEN (PASISIO_FILEID_SIZE));
}


static inline u_int32_t *
xdr_encode_fid_no_copy (u_int32_t *op)
{
    return (op + XDR_QUADLEN (PASISIO_FILEID_SIZE));
}


static inline u_int32_t *
xdr_encode_hyper (u_int32_t *p,
                  uint64_t val)
{
    *p++ = htonl(val >> 32);
    *p++ = htonl(val & 0xFFFFFFFF);
    return p;
}


static inline u_int32_t *
xdr_encode_hyper_no_copy (u_int32_t *op)
{
    uint64_t val = *((uint64_t*) op);
    *op++ = htonl(val >> 32);
    *op++ = htonl(val & 0xFFFFFFFF);
    return op;
}


static inline u_int32_t *
xdr_encode_string (u_int32_t *p,
                   char* string)
{
    if (string) {
        u_int8_t* byte_p = (u_int8_t*)p;
        u_int8_t* aligned_p;

        while ((*byte_p++ = *string++))
            ;

        /* Zero bytes until next word alignment, for consistent hashing */
        aligned_p = (u_int8_t*)(p + XDR_QUADLEN (byte_p - (u_int8_t*)p));
        memset (byte_p, 0, aligned_p - byte_p);

        /* Keep p word-aligned */
        return (u_int32_t*) aligned_p;
    } else {
        /* write four null character (to be word-aligned) */
        *p++ = 0;
        return p;
    }
}


static inline u_int32_t *
xdr_encode_string_no_copy (u_int32_t *op)
{
    if (*op == 0) { /* null string */
        return op + 1;
    } else {
#if 0
        u_int8_t* byte_p = (u_int8_t*)op;
        u_int8_t* aligned_p;

        /* Find string length */
        while (*byte_p++)
            ;

        /* Keep p word-aligned */
        aligned_p = (u_int8_t*)(op + XDR_QUADLEN (byte_p - (u_int8_t*)op));
        return (u_int32_t*) aligned_p;
#endif
        printf ("%s: len=%d\n", __FUNCTION__, (int) strlen ((char *)(op)) + 1);
        return (op + XDR_QUADLEN (strlen ((char *)(op)) + 1));
    }
}


static inline u_int32_t *
xdr_encode_hts (u_int32_t *p,
                pasis_md_hts_t *hts)
{
    p = xdr_encode_hyper (p, hts->time);

#if USE_OBJECT_HASHES > 0
    memcpy (p, hts->object_hash, PASIS_HASH_SIZE);
    p += XDR_QUADLEN (PASIS_HASH_SIZE);
#endif

    memcpy (p, hts->op_hash, PASIS_HASH_SIZE);
    p += XDR_QUADLEN (PASIS_HASH_SIZE);

    memcpy (p, hts->matrix_hash, PASIS_HASH_SIZE);
    p += XDR_QUADLEN (PASIS_HASH_SIZE);

    *p++ = htonl (hts->client_id);

    /* Top two bits for op_type have to be 0, due to our XDRing */
    PASIS_DASSERT((hts->op_type >> 30) == 0);
    *p++ = htonl (hts->barrier_flag   << 31 |
                  hts->multiobject_op << 30 |
                  hts->op_type);

#if DEBUG_TIMESTAMPS > 0
    memcpy (p, hts->op_description, sizeof (hts->op_description));
    p += XDR_QUADLEN (sizeof (hts->op_description));
#endif

    return p;
}


static inline u_int32_t *
xdr_encode_protocol (u_int32_t *p,
                     pasis_md_protocol_t *proto)
{
    *p++ = htonl (proto->N);
    *p++ = htonl (proto->Qc);
    *p++ = htonl (proto->B);
    *p++ = htonl (proto->T);
    *p++ = htonl (proto->m);
    *p++ = htonl (proto->scheme_type);
    *p++ = htonl (proto->hash_type);
    *p++ = htonl (proto->cc_type);
    *p++ = htonl (proto->quorum_type);
    *p++ = htonl (proto->delta);
    
    return p;
}


static inline u_int32_t *
xdr_encode_protocol_no_copy (u_int32_t *op, u_int32_t *N)
{
    *op = htonl (*op); /* N */
    *N = *op;
    op++;
    *op = htonl (*op); /* Qc */
    op++;
    *op = htonl (*op); /* B */
    op++;
    *op = htonl (*op); /* T */
    op++;
    *op = htonl (*op); /* m */
    op++;
    *op = htonl (*op); /* scheme_type */
    op++;
    *op = htonl (*op); /* hash_type */
    op++;
    *op = htonl (*op); /* cc_type */
    op++;
    *op = htonl (*op); /* quorum_type */
    op++;
    *op = htonl (*op); /* delta */
    op++;

    return op;
}


static inline u_int32_t *
xdr_encode_linkage (u_int32_t *p,
                    pasis_md_linkage_t *linkage)
{
    *p++ = htonl (linkage->ip_address);
    *p++ = htonl (linkage->partition);
    return p;
}


static inline u_int32_t *
xdr_encode_linkage_no_copy (u_int32_t *op)
{
    *op = htonl (*op); /* ip_address */
    op++;
    *op = htonl (*op); /* partition */
    op++;
    return op;
}


static inline u_int32_t *
xdr_encode_authenticator (u_int32_t *p,
                          pasis_md_auth_t *authenticator)
{
    if (authenticator != NULL) {
        memcpy (p, authenticator->data, PASIS_HASH_SIZE);
    }
    p += XDR_QUADLEN (PASIS_HASH_SIZE);

    return p;
}


static inline u_int32_t *
xdr_encode_hash (u_int32_t *p,
                 p_byte    *hash)
{
    if (hash != NULL) {
        memcpy (p, hash, PASIS_HASH_SIZE);
    }
    p += XDR_QUADLEN (PASIS_HASH_SIZE);

    return p;
}


static inline u_int32_t *
xdr_encode_history_vector (u_int32_t *p,
                           pasis_md_hist_vector_t *vector)
{
    int i;

    PASIS_DASSERT (vector->history_num <= PASIS_MD_HIST_MAX_VERSIONS);
    
    *p++ = htonl (vector->node_id);
    *p++ = htonl (vector->history_num);
    for (i = 0; i < vector->history_num; i++) {
        p = xdr_encode_hts (p, &vector->histories[i]);
    }

    PASIS_DASSERT (vector->auth_num <= PASIS_MAX_NUM_SERVERS);

    if (DO_NOT_TRANSFER_AUTHENTICATORS) {
        /* no authenticators on network */
        *p++ = htonl (0); /* auth_num */
    } else {
        *p++ = htonl (vector->auth_num);
        for (i = 0; i < vector->auth_num; i++) {
            p = xdr_encode_authenticator (p, &vector->authenticators[i]);
        }
    }

    return p;
}


static inline u_int32_t *
xdr_encode_attr (u_int32_t *p,
                 pasis_md_attr_t *attr)
{
    *p++ = htonl (attr->mode);
    *p++ = htonl (attr->ref_cnt);
    *p++ = htonl (attr->uid);
    *p++ = htonl (attr->gid);
    *p++ = htonl (attr->fileid);
    *p++ = htonl (attr->mtime[0]);
    *p++ = htonl (attr->mtime[1]);
    *p++ = htonl (attr->atime[0]);
    *p++ = htonl (attr->atime[1]);
    *p++ = htonl (attr->ctime[0]);
    *p++ = htonl (attr->ctime[1]);
    p = xdr_encode_hyper (p, attr->blocks);
    p = xdr_encode_hyper (p, attr->size);

    return p;
}


static inline u_int32_t *
xdr_encode_attr_no_copy (u_int32_t *op)
{
    *op = htonl (*op); /* mode) */
    op++;
    *op = htonl (*op); /* ref_cnt) */
    op++;
    *op = htonl (*op); /* uid */
    op++;
    *op = htonl (*op); /* gid); */
    op++;
    *op = htonl (*op); /* fileid); */
    op++;
    *op = htonl (*op); /* mtime[0] */
    op++;
    *op = htonl (*op); /* mtime[1] */
    op++;
    *op = htonl (*op); /* atime[0] */
    op++;
    *op = htonl (*op); /* atime[1] */
    op++;
    *op = htonl (*op); /* ctime[0] */
    op++;
    *op = htonl (*op); /* ctime[1] */
    op++;

    op = xdr_encode_hyper_no_copy (op); /* blocks */
    op = xdr_encode_hyper_no_copy (op); /* size */

    return op;
}


static inline u_int32_t *
xdr_encode_sattr (u_int32_t *p,
                 pasis_md_sattr_t *sattr)
{
    *p++ = htonl (sattr->set_mode);
    *p++ = htonl (sattr->mode);
    *p++ = htonl (sattr->set_uid);
    *p++ = htonl (sattr->uid);
    *p++ = htonl (sattr->set_gid);
    *p++ = htonl (sattr->gid);
    *p++ = htonl (sattr->set_size);
    p = xdr_encode_hyper (p, sattr->size);
    *p++ = htonl (sattr->set_blocks);
    p = xdr_encode_hyper (p, sattr->blocks);
    *p++ = htonl (sattr->set_atime);
    *p++ = htonl (sattr->atime[0]);
    *p++ = htonl (sattr->atime[1]);
    *p++ = htonl (sattr->set_mtime);
    *p++ = htonl (sattr->mtime[0]);
    *p++ = htonl (sattr->mtime[1]);

    return p;
}


static inline u_int32_t *
xdr_encode_sattr_no_copy (u_int32_t *op)
{
    *op = htonl (*op); /* set_mode */
    op++;
    *op = htonl (*op); /* mode */
    op++;
    *op = htonl (*op); /* set_uid */
    op++;
    *op = htonl (*op); /* uid */
    op++;
    *op = htonl (*op); /* set_gid */
    op++;
    *op = htonl (*op); /* gid */
    op++;
    *op = htonl (*op); /* set_size */
    op++;
    op = xdr_encode_hyper_no_copy (op); /* size */
    *op = htonl (*op); /* set_blocks */
    op++;
    op = xdr_encode_hyper_no_copy (op); /* blocks */
    *op = htonl (*op); /* set_atime */
    op++;
    *op = htonl (*op); /* atime[0] */
    op++;
    *op = htonl (*op); /* atime[1] */
    op++;
    *op = htonl (*op); /* set_mtime */
    op++;
    *op = htonl (*op); /* mtime[0] */
    op++;
    *op = htonl (*op); /* mtime[1] */
    op++;

    return op;
}


static inline u_int32_t *
xdr_encode_distribution (u_int32_t *p,
                         pasis_md_loc_t *distribution)
{
    unsigned int i;

    p = xdr_encode_protocol (p, &distribution->protocol);

    PASIS_DASSERT(distribution->protocol.N <= PASIS_MAX_NUM_SERVERS);

    for (i = 0; i < distribution->protocol.N; i++) {
        p = xdr_encode_linkage (p, &distribution->linkages[i]);
    }

    return p;
}


static inline u_int32_t *
xdr_encode_distribution_no_copy (u_int32_t *op)
{
    unsigned int i, N;

    op = xdr_encode_protocol_no_copy (op, &N);

    for (i = 0; i < N; i++) {
        op = xdr_encode_linkage_no_copy (op);
    }

    return op;
}


static inline u_int32_t *
xdr_encode_hist_matrix (u_int32_t *p,
                        pasis_md_hist_matrix_t *matrix)
{
    int i, auth_num = 0;

    PASIS_DASSERT (matrix->ts_num <= PASIS_MD_HIST_MAX_VERSIONS);
    //#if DO_NOT_USE_AUTHENTICATORS == 0
    PASIS_DASSERT (matrix->auth_num <= PASIS_MAX_NUM_SERVERS);
    //#endif

    *p++ = htonl (matrix->ts_num);
    for (i = 0; i < matrix->ts_num; i++) {
        p = xdr_encode_hts (p, matrix->ts_vector[i]);
    }
    for (i = 0; i < matrix->ts_num; i++) {
        p = xdr_encode_hyper (p, matrix->ts_bitmap[i]);
    }

    //#if DO_NOT_USE_AUTHENTICATORS == 0
    /* I intermix the auth_exists and authenticators in the encoding, to save
     * an extra loop.  -mabdelmalek */
    *p++ = htonl (matrix->auth_num);
    for (i = 0; i < PASIS_MAX_NUM_SERVERS; i++) {
        PASIS_DASSERT (matrix->auth_exists[i] == 0 ||
                       matrix->auth_exists[i] == 1);
        *p++ = htonl (matrix->auth_exists[i]);
        if (matrix->auth_exists[i]) {
            p = xdr_encode_authenticator (p, matrix->authenticators[i]);
            auth_num++;
        }
    }
    PASIS_DASSERT (auth_num == matrix->auth_num);
    //#endif

    return p;
}


static inline void
xdr_encode_iovec (rpc_req_t *req,
                  unsigned int  iov_count,
                  struct iovec *iov)
{
    unsigned int i;

    //printf ("%s: count=%u\n", __FUNCTION__, iov_count);

    for (i = 0; i < iov_count; i++) {
        req->req_iov[req->req_iov_cur].iov_base = iov[i].iov_base;
        req->req_iov[req->req_iov_cur].iov_len = iov[i].iov_len;
        req->req_iov_cur++;
        //printf ("  - [%d] %u\n", i, iov[i].iov_len);
    }
    
    req->req_iov_len += i;
}

static inline u_int32_t *
xdr_decode_hyper(u_int32_t *p,
                 uint64_t *valp)
{
    /* May be decoding in place; so use auxillary variable so we don't destroy
     * the value by the first set. */
    uint64_t val;
    val  = ((uint64_t) ntohl(*p++)) << 32;
    val |= ntohl(*p++);
    *valp  = val;
    return p;
}


static inline u_int32_t *
xdr_decode_string (u_int32_t *p,
                   char* string)
{
    if (*p == 0) { /* a  null string was encoded */
        string[0] = '\0';
        return p + 1;
    } else {
        u_int8_t* byte_p = (u_int8_t*)p;
        u_int8_t* aligned_p;

        while ((*string++ = *byte_p++))
            ;

        /* Keep p word-aligned */
        aligned_p = (u_int8_t*)(p + XDR_QUADLEN (byte_p - (u_int8_t*)p));
        return (u_int32_t*) aligned_p;
    }
}


static inline u_int32_t *
xdr_decode_string_no_copy (u_int32_t *p,
                           char** string)
{
    if (*p == 0) { /* a null string was encoded */
        *string = NULL;
        return p + 1;
    } else {
        u_int8_t* byte_p = (u_int8_t*)p;
        u_int8_t* aligned_p;

        *string = (char*)p;

        /* Find string length */
        while (*byte_p++)
            ;

        /* Keep p word-aligned */
        aligned_p = (u_int8_t*)(p + XDR_QUADLEN (byte_p - (u_int8_t*)p));
        return (u_int32_t*) aligned_p;
    }
}


static inline u_int32_t *
xdr_decode_fid (u_int32_t *p, 
                pasis_md_fid_t *fid)
{
    memcpy (fid->data, p, PASISIO_FILEID_SIZE);
    return (p + XDR_QUADLEN (PASISIO_FILEID_SIZE));
}


static inline u_int32_t *
xdr_decode_fid_no_copy (u_int32_t *p, 
                        pasis_md_fid_t **fid)
{
    *fid = (pasis_md_fid_t*) p;
    return (p + XDR_QUADLEN (PASISIO_FILEID_SIZE));
}


static inline u_int32_t *
xdr_decode_hts (u_int32_t *p,
                pasis_md_hts_t *hts)
{
    p_uint32 t;

    p = xdr_decode_hyper (p, &hts->time);

#if USE_OBJECT_HASHES > 0
    memcpy (hts->object_hash, p, PASIS_HASH_SIZE);
    p += XDR_QUADLEN (PASIS_HASH_SIZE);
#endif

    memcpy (hts->op_hash, p, PASIS_HASH_SIZE);
    p += XDR_QUADLEN (PASIS_HASH_SIZE);

    memcpy (hts->matrix_hash, p, PASIS_HASH_SIZE);
    p += XDR_QUADLEN (PASIS_HASH_SIZE);
    
    hts->client_id = ntohl (*p++);

    t = ntohl (*p++);
    hts->barrier_flag   = t >> 31;
    hts->multiobject_op = t >> 30;
    hts->op_type        = t & 0x3fffffff;

#if DEBUG_TIMESTAMPS > 0
    memcpy (hts->op_description, p, sizeof (hts->op_description));
    p += XDR_QUADLEN (sizeof (hts->op_description));
#endif

    return p;
}


static inline u_int32_t *
xdr_decode_hts_no_copy (u_int32_t *p,
                        pasis_md_hts_t **hts)
{
    p_uint32 t;

    *hts = (pasis_md_hts_t *)p;

    p = xdr_decode_hyper (p, &(*hts)->time);

#if USE_OBJECT_HASHES > 0
    p += XDR_QUADLEN (PASIS_HASH_SIZE);
#endif
    p += XDR_QUADLEN (PASIS_HASH_SIZE);
    p += XDR_QUADLEN (PASIS_HASH_SIZE);

    (*hts)->client_id = ntohl (*p);
    p++;

    t = ntohl(*p++);
    (*hts)->barrier_flag   = t >> 31;
    (*hts)->multiobject_op = t >> 30;
    (*hts)->op_type        = t & 0x3fffffff;
    
#if DEBUG_TIMESTAMPS > 0
    p += XDR_QUADLEN (sizeof ((*hts)->op_description));
#endif

    return p;
}


static inline u_int32_t *
xdr_decode_protocol (u_int32_t *p,
                     pasis_md_protocol_t *proto)
{
    proto->N  = ntohl (*p++);
    proto->Qc = ntohl (*p++);
    proto->B  = ntohl (*p++);
    proto->T  = ntohl (*p++);
    proto->m  = ntohl (*p++);
    proto->scheme_type = ntohl (*p++);
    proto->hash_type = ntohl (*p++);
    proto->cc_type = ntohl (*p++);
    proto->quorum_type = ntohl (*p++);
    proto->delta = ntohl (*p++);

    return p;
}


static inline u_int32_t *
xdr_decode_linkage (u_int32_t *p,
                    pasis_md_linkage_t *linkage)
{
    linkage->ip_address = ntohl (*p++);
    linkage->partition = ntohl (*p++);
    return p;
}


static inline u_int32_t *
xdr_decode_distribution (u_int32_t *p,
                         pasis_md_loc_t *distribution)
{
    unsigned int i;

    p = xdr_decode_protocol (p, &distribution->protocol);

    PASIS_DASSERT(distribution->protocol.N <= PASIS_MAX_NUM_SERVERS);

    for (i = 0; i < distribution->protocol.N; i++) {
        p = xdr_decode_linkage (p, &distribution->linkages[i]);
    }

    return p;
}


static inline u_int32_t *
xdr_decode_authenticator (u_int32_t *p,
                          pasis_md_auth_t *authenticator)
{
    memcpy (authenticator->data, p, PASIS_HASH_SIZE);
    p += XDR_QUADLEN (PASIS_HASH_SIZE);

    return p;
}


static inline u_int32_t *
xdr_decode_authenticator_no_copy (u_int32_t *p, 
                                  pasis_md_auth_t **authenticator)
{
    *authenticator = (pasis_md_auth_t*) p;
    return (p + XDR_QUADLEN (PASIS_HASH_SIZE));
}


static inline u_int32_t *
xdr_decode_hash (u_int32_t *p,
                 p_byte    *hash)
{
    memcpy (hash, p, PASIS_HASH_SIZE);
    p += XDR_QUADLEN (PASIS_HASH_SIZE);

    return p;
}

static inline u_int32_t *
xdr_decode_hash_no_copy (u_int32_t  *p,
                         p_byte    **hash)
{
    *hash = (p_byte*) p;
    return (p + XDR_QUADLEN (PASIS_HASH_SIZE));
}


static inline u_int32_t *
xdr_decode_hist_matrix (u_int32_t *p,
                        pasis_md_hist_matrix_t *matrix)
{
    int i;

    matrix->ts_num = ntohl (*p++);
    PASIS_DASSERT (matrix->ts_num <= PASIS_MD_HIST_MAX_VERSIONS);

    for (i = 0; i < matrix->ts_num; i++) {
        /* Decode in place -- on storage-nodes */
        matrix->ts_vector[i] = (pasis_md_hts_t *)p;
        p = xdr_decode_hts_no_copy (p, &(matrix->ts_vector[i]));
    }
    for (i = 0; i < matrix->ts_num; i++) {
        p = xdr_decode_hyper (p, &(matrix->ts_bitmap[i]));
    }

    //#if DO_NOT_USE_AUTHENTICATORS == 0
    matrix->auth_num = ntohl (*p++);
    PASIS_DASSERT (matrix->auth_num <= PASIS_MAX_NUM_SERVERS);
    for (i = 0; i < PASIS_MAX_NUM_SERVERS; i++) {
        matrix->auth_exists[i] = ntohl (*p++);
        PASIS_DASSERT (matrix->auth_exists[i] == 0 ||
                       matrix->auth_exists[i] == 1);
        if (matrix->auth_exists[i]) {
            /* Decode in place on storage-nodes */
            p = xdr_decode_authenticator_no_copy(p, &matrix->authenticators[i]);
        }
    }
    //#else
    //    matrix->auth_num = 0;
    //#endif

    return p;
}


static inline u_int32_t *
xdr_decode_history_vector (u_int32_t *p,
                           pasis_md_hist_vector_t *vector)
{
    int i;
    
    vector->node_id = ntohl (*p++);
    vector->history_num = ntohl (*p++);
    PASIS_DASSERT (vector->history_num <= PASIS_MD_HIST_MAX_VERSIONS);

    for (i = 0; i < vector->history_num; i++) {
        p = xdr_decode_hts (p, &vector->histories[i]);
    }

    //#if DO_NOT_USE_AUTHENTICATORS == 0
    vector->auth_num = ntohl (*p++);
    PASIS_DASSERT (vector->auth_num <= PASIS_MAX_NUM_SERVERS);

    for (i = 0; i < vector->auth_num; i++) {
        p = xdr_decode_authenticator (p, &vector->authenticators[i]);
    }
    //#else
    //    vector->auth_num = 0;
    //#endif
    
    return p;
}

static inline u_int32_t *
xdr_decode_attr (u_int32_t *p,
                 pasis_md_attr_t *attr)
{
    attr->mode = ntohl (*p++);
    attr->ref_cnt = ntohl (*p++);
    attr->uid = ntohl (*p++);
    attr->gid = ntohl (*p++);
    attr->fileid = ntohl (*p++);
    attr->mtime[0] = ntohl (*p++);
    attr->mtime[1] = ntohl (*p++);
    attr->atime[0] = ntohl (*p++);
    attr->atime[1] = ntohl (*p++);
    attr->ctime[0] = ntohl (*p++);
    attr->ctime[1] = ntohl (*p++);
    p = xdr_decode_hyper (p, &attr->blocks);
    p = xdr_decode_hyper (p, &attr->size);

    return p;
}


static inline u_int32_t *
xdr_decode_sattr (u_int32_t *p,
                  pasis_md_sattr_t *sattr)
{
    sattr->set_mode = ntohl (*p++);
    sattr->mode = ntohl (*p++);
    sattr->set_uid = ntohl (*p++);
    sattr->uid = ntohl (*p++);
    sattr->set_gid = ntohl (*p++);
    sattr->gid = ntohl (*p++);
    sattr->set_size = ntohl (*p++);
    p = xdr_decode_hyper (p, &sattr->size);
    sattr->set_blocks = ntohl (*p++);
    p = xdr_decode_hyper (p, &sattr->blocks);
    sattr->set_atime = ntohl (*p++);
    sattr->atime[0] = ntohl (*p++);
    sattr->atime[1] = ntohl (*p++);
    sattr->set_mtime = ntohl (*p++);
    sattr->mtime[0] = ntohl (*p++);
    sattr->mtime[1] = ntohl (*p++);

    return p;
}



/* For client transmit stub -- send args */

u_int32_t *
pasis_md_xdr_enc_read_history_args (u_int32_t *p, 
                                    pasis_md_read_history_args_t *args)
{
    unsigned int i;

    *p++ = htonl (args->fid_num);

    PASIS_DASSERT (args->fid_num <= MD_MAX_OBJS_PER_REQ);
    for (i = 0; i < args->fid_num; ++i) {
        p = xdr_encode_fid (p, args->fid[i]);
    }

    return p;
}


u_int32_t *
pasis_md_xdr_enc_create_args (u_int32_t *p, 
                              pasis_md_create_args_t *args)
{
    p = xdr_encode_fid (p, args->parent_dir_fid);
    p = xdr_encode_fid (p, args->create_fid);
    p = xdr_encode_string (p, args->create_name);
    p = xdr_encode_attr (p, &args->create_attr);
    p = xdr_encode_string (p, args->symlink_name);
    p = xdr_encode_distribution (p, &args->distribution);
    *p++ = htonl (args->client_id);
    
    return p;
}


u_int32_t *
pasis_md_xdr_enc_link_args (u_int32_t *p, 
                            pasis_md_link_args_t *args)
{
    p = xdr_encode_fid (p, args->from_file_fid);
    p = xdr_encode_fid (p, args->to_dir_fid);
    p = xdr_encode_string (p, args->to_name);
    *p++ = htonl (args->client_id);
    return p;
}


u_int32_t *
pasis_md_xdr_enc_unlink_args (u_int32_t *p, 
                              pasis_md_unlink_args_t *args)
{
    p = xdr_encode_fid (p, args->parent_dir_fid);
    p = xdr_encode_fid (p, args->name_fid);
    p = xdr_encode_string (p, args->name);
    *p++ = htonl (args->client_id);
    return p;
}


u_int32_t *
pasis_md_xdr_enc_getattr_args (u_int32_t *p, 
                               pasis_md_getattr_args_t *args)
{
    p = xdr_encode_fid (p, args->fid);
    return p;
}


u_int32_t *
pasis_md_xdr_enc_setattr_args (u_int32_t *p, 
                               pasis_md_setattr_args_t *args)
{
    p = xdr_encode_fid (p, args->fid);
    p = xdr_encode_sattr (p, &args->attr);
    *p++ = htonl (args->guard_ctime[0]);
    *p++ = htonl (args->guard_ctime[1]);
    *p++ = htonl (args->client_id);

    return p;
}


u_int32_t *
pasis_md_xdr_enc_rename_args (u_int32_t *p, 
                              pasis_md_rename_args_t *args)
{
    p = xdr_encode_fid (p, args->from_dir_fid);
    *p++ = htonl (args->to_dir_valid);
    if (args->to_dir_valid) {
        p = xdr_encode_fid (p, args->to_dir_fid);
    }
    p = xdr_encode_fid (p, args->from_name_fid);
    *p++ = htonl (args->to_name_valid);
    if (args->to_name_valid) {
        p = xdr_encode_fid (p, args->to_name_fid);
    }
    p = xdr_encode_string (p, args->from_name);
    p = xdr_encode_string (p, args->to_name);
    *p++ = htonl (args->client_id);
    return p;
}


u_int32_t *
pasis_md_xdr_enc_lookup_args (u_int32_t *p, 
                              pasis_md_lookup_args_t *args)
{
    p = xdr_encode_fid (p, args->parent_dir_fid);
    p = xdr_encode_string (p, args->name);
    return p;
}


u_int32_t *
pasis_md_xdr_enc_readlink_args (u_int32_t *p, 
                                pasis_md_readlink_args_t *args)
{
    p = xdr_encode_fid (p, args->fid);
    return p;
}


u_int32_t *
pasis_md_xdr_enc_readdir_args (u_int32_t *p, 
                               pasis_md_readdir_args_t *args)
{
    p = xdr_encode_fid (p, args->fid);
    p = xdr_encode_hyper (p, args->block_num);
    return p;
}

u_int32_t *
pasis_md_xdr_enc_barrier_args (u_int32_t *p,
                               pasis_md_barrier_args_t *args)
{
    int i;
    
    *p++ = htonl (args->fid_num);

    for (i = 0; i < args->fid_num; i++) {
        p = xdr_encode_fid (p, args->fid[i]);
    }

    *p++ = htonl (args->chunk_size);
    *p++ = htonl (args->client_id);

    return p;
}

u_int32_t *
pasis_md_xdr_enc_object_sync_args (u_int32_t *p,
                                   pasis_md_object_sync_args_t *args)
{
    p = xdr_encode_fid (p, args->fid);
    *p++ = htonl (args->return_attrs);
    *p++ = htonl (args->start_block_num);
    *p++ = htonl (args->num_blocks_to_ret);

    return p;
}


u_int32_t *
pasis_md_xdr_enc_synth_query_args (u_int32_t *p, 
                                   pasis_md_synth_query_args_t *args)
{
    int i;

    *p++ = htonl (args->num_fids);
    for (i = 0; i < args->num_fids; ++i) {
        p = xdr_encode_fid (p, args->fids[i]);
    }
    *p++ = htonl (args->arg_size);
    *p++ = htonl (args->res_size);
    *p++ = htonl (args->chunk_size);
    for (i = 0; i < args->num_fids; ++i) {
        *p++ = htonl (args->read_chunks[i]);
    }
    *p++ = htonl (args->compute_time);

    return p;
}

u_int32_t *
pasis_md_xdr_enc_get_stats_args (u_int32_t *p, 
                                 pasis_md_get_stats_args_t *args)
{
    *p++ = htonl (args->res_max_num_buckets);

    return p;
}


u_int32_t *
pasis_md_xdr_enc_synth_update_args (u_int32_t *p, 
                                    pasis_md_synth_update_args_t *args)
{
    int i;

    *p++ = htonl (args->num_fids);
    for (i = 0; i < args->num_fids; ++i) {
        p = xdr_encode_fid (p, args->fids[i]);
    }
    *p++ = htonl (args->arg_size);
    *p++ = htonl (args->res_size);
    *p++ = htonl (args->chunk_size);
    for (i = 0; i < args->num_fids; ++i) {
        *p++ = htonl (args->modified_chunks[i]);
    }
    *p++ = htonl (args->compute_time);
    *p++ = htonl (args->ts_increment);
    *p++ = htonl (args->client_id);

    return p;
}


int
pasis_md_xdr_enc_read_op_args (rpc_req_t *req, u_int32_t *p, 
                               pasis_md_read_op_args_t *args)
{
    unsigned int i;

    *p++ = htonl (args->part);
    *p++ = htonl (args->read_flags);
    p = xdr_encode_hts (p, &args->ts);
    *p++ = htonl (args->op_type);
    
    switch (args->op_type) {
    case PASIS_MD_READ_HISTORY:
        p = pasis_md_xdr_enc_read_history_args (p, 
                                                &args->op_args.read_history);
        break;
    case PASIS_MD_GETATTR:
        p = pasis_md_xdr_enc_getattr_args (p, &args->op_args.getattr);
        break;
    case PASIS_MD_LOOKUP:
        p = pasis_md_xdr_enc_lookup_args (p, &args->op_args.lookup);
        break;
    case PASIS_MD_READLINK:
        p = pasis_md_xdr_enc_readlink_args (p, &args->op_args.readlink);
        break;
    case PASIS_MD_READDIR:
        p = pasis_md_xdr_enc_readdir_args (p, &args->op_args.readdir);
        break;
    case PASIS_MD_OBJECT_SYNC:
        p = pasis_md_xdr_enc_object_sync_args (p, &args->op_args.object_sync);
        break;
    case PASIS_MD_SYNTH_QUERY:
        p = pasis_md_xdr_enc_synth_query_args (p, &args->op_args.synth_query);
        break;
    case PASIS_MD_GET_STATS:
        p = pasis_md_xdr_enc_get_stats_args (p, &args->op_args.get_stats);
        break;
    default:
        PASIS_DASSERT (0);
        break;
    }

    *p++ = htonl (args->history_matrices_num);
    PASIS_DASSERT (args->history_matrices_num <= MD_MAX_OBJS_PER_REQ);
    for (i = 0; i < args->history_matrices_num; ++i) {
        p = xdr_encode_hist_matrix (p, &args->history_matrices[i]);
    }

    xdr_adjust_req_iovec (req, p);

    return 0;
}


int
pasis_md_xdr_enc_write_op_args (rpc_req_t *req, u_int32_t *p, 
                                pasis_md_write_op_args_t *args)
{
    unsigned int i;

    *p++ = htonl (args->part);
    *p++ = htonl (args->op_type);
    *p++ = htonl (args->mtime[0]);
    *p++ = htonl (args->mtime[1]);
    *p++ = htonl (args->repair_flag);

    switch (args->op_type) {
    case PASIS_MD_CREATE:
        p = pasis_md_xdr_enc_create_args (p, &args->op_args.create);
        break;
    case PASIS_MD_LINK:
        p = pasis_md_xdr_enc_link_args (p, &args->op_args.link);
        break;
    case PASIS_MD_UNLINK:
        p = pasis_md_xdr_enc_unlink_args (p, &args->op_args.unlink);
        break;
    case PASIS_MD_SETATTR:
        p = pasis_md_xdr_enc_setattr_args (p, &args->op_args.setattr);
        break;
    case PASIS_MD_RENAME:
        p = pasis_md_xdr_enc_rename_args (p, &args->op_args.rename);
        break;
    case PASIS_MD_BARRIER:
        p = pasis_md_xdr_enc_barrier_args (p, &args->op_args.barrier);
        break;
    case PASIS_MD_SYNTH_UPDATE:
        p = pasis_md_xdr_enc_synth_update_args (p, &args->op_args.synth_update);
        break;
    default:
        PASIS_PRINT_ERR ("%s: got bad op_type: %d\n", __FUNCTION__,
                         args->op_type);
        PASIS_DASSERT (0);
        break;
    }

    *p++ = htonl (args->history_matrices_num);
    PASIS_DASSERT (args->history_matrices_num <= MD_MAX_OBJS_PER_REQ);
    for (i = 0; i < args->history_matrices_num; ++i) {
        p = xdr_encode_hist_matrix (p, &args->history_matrices[i]);
    }

    for (i = 0; i < args->history_matrices_num; ++i) {
        *p++ = htonl (args->proof_fid_indices[i]);
        if (args->proof_fid_indices[i] != UINT_MAX) {
            p = xdr_encode_hist_matrix (p, &args->proof_matrices[i]);
        }
    }

    xdr_adjust_req_iovec (req, p);

    return 0;
}


/* For server recv stub -- recv args */
u_int32_t *
pasis_md_xdr_dec_read_history_args (u_int32_t *p, 
                                    pasis_md_read_history_args_t *args)
{
    unsigned int i;

    args->fid_num = ntohl (*p++);

    PASIS_DASSERT (args->fid_num <= MD_MAX_OBJS_PER_REQ);
    for (i = 0; i < args->fid_num; ++i) {
        p = xdr_decode_fid_no_copy (p, &args->fid[i]);
    }

    return p;
}


u_int32_t *
pasis_md_xdr_dec_create_args (u_int32_t *p, 
                              pasis_md_create_args_t *args)
{
    p = xdr_decode_fid_no_copy (p, &args->parent_dir_fid);
    p = xdr_decode_fid_no_copy (p, &args->create_fid);
    p = xdr_decode_string_no_copy (p, &args->create_name);
    p = xdr_decode_attr (p, &args->create_attr);
    p = xdr_decode_string_no_copy (p, &args->symlink_name);
    p = xdr_decode_distribution (p, &args->distribution);
    args->client_id = ntohl (*p++);

    return p;
}


u_int32_t *
pasis_md_xdr_dec_link_args (u_int32_t *p, 
                            pasis_md_link_args_t *args)
{
    p = xdr_decode_fid_no_copy (p, &args->from_file_fid);
    p = xdr_decode_fid_no_copy (p, &args->to_dir_fid);
    p = xdr_decode_string_no_copy (p, &args->to_name);
    args->client_id = ntohl (*p++);

    return p;
}


u_int32_t *
pasis_md_xdr_dec_unlink_args (u_int32_t *p, 
                              pasis_md_unlink_args_t *args)
{
    p = xdr_decode_fid_no_copy (p, &args->parent_dir_fid);
    p = xdr_decode_fid_no_copy (p, &args->name_fid);
    p = xdr_decode_string_no_copy (p, &args->name);
    args->client_id = ntohl (*p++);
    return p;
}


u_int32_t *
pasis_md_xdr_dec_getattr_args (u_int32_t *p, 
                               pasis_md_getattr_args_t *args)
{
    p = xdr_decode_fid_no_copy (p, &args->fid);
    return p;
}


u_int32_t *
pasis_md_xdr_dec_setattr_args (u_int32_t *p, 
                               pasis_md_setattr_args_t *args)
{
    p = xdr_decode_fid_no_copy (p, &args->fid);
    p = xdr_decode_sattr (p, &args->attr);
    args->guard_ctime[0] = ntohl (*p++);
    args->guard_ctime[1] = ntohl (*p++);
    args->client_id = ntohl (*p++);
    return p;
}


u_int32_t *
pasis_md_xdr_dec_rename_args (u_int32_t *p, 
                              pasis_md_rename_args_t *args)
{
    p = xdr_decode_fid_no_copy (p, &args->from_dir_fid);
    args->to_dir_valid = ntohl (*p++);
    if (args->to_dir_valid) {
        p = xdr_decode_fid_no_copy (p, &args->to_dir_fid);
    } else {
        args->to_dir_fid = NULL;
    }
    p = xdr_decode_fid_no_copy (p, &args->from_name_fid);
    args->to_name_valid = ntohl (*p++);
    if (args->to_name_valid) {
        p = xdr_decode_fid_no_copy (p, &args->to_name_fid);
    } else {
        args->to_name_fid = NULL;
    }
    p = xdr_decode_string_no_copy (p, &args->from_name);
    p = xdr_decode_string_no_copy (p, &args->to_name);
    args->client_id = ntohl (*p++);
    return p;
}


u_int32_t *
pasis_md_xdr_dec_barrier_args (u_int32_t *p,
                               pasis_md_barrier_args_t *args)
{
    int i;
    args->fid_num = ntohl (*p++);
    for (i = 0; i < args->fid_num; i++) {
        p = xdr_decode_fid_no_copy (p, &args->fid[i]);
    }
    args->chunk_size = ntohl (*p++);
    args->client_id = ntohl (*p++);
    
    return p;
}


u_int32_t *
pasis_md_xdr_dec_object_sync_args (u_int32_t *p,
                                   pasis_md_object_sync_args_t *args)
{
    p = xdr_decode_fid_no_copy (p, &args->fid);
    args->return_attrs = ntohl (*p++);
    args->start_block_num = ntohl (*p++);
    args->num_blocks_to_ret = ntohl (*p++);

    return p;
}


u_int32_t *
pasis_md_xdr_dec_lookup_args (u_int32_t *p, 
                              pasis_md_lookup_args_t *args)
{
    p = xdr_decode_fid_no_copy (p, &args->parent_dir_fid);
    p = xdr_decode_string_no_copy (p, &args->name);
    return p;
}


u_int32_t *
pasis_md_xdr_dec_readlink_args (u_int32_t *p, 
                                pasis_md_readlink_args_t *args)
{
    p = xdr_decode_fid_no_copy (p, &args->fid);
    return p;
}


u_int32_t *
pasis_md_xdr_dec_readdir_args (u_int32_t *p, 
                               pasis_md_readdir_args_t *args)
{
    p = xdr_decode_fid_no_copy (p, &args->fid);
    p = xdr_decode_hyper (p, &args->block_num);
    return p;
}


u_int32_t *
pasis_md_xdr_dec_synth_query_args (u_int32_t *p, 
                                   pasis_md_synth_query_args_t *args)
{
    int i;

    args->num_fids = ntohl (*p++);
    for (i = 0; i < args->num_fids; ++i) {
        p = xdr_decode_fid_no_copy (p, &args->fids[i]);
    }
    args->arg_size = ntohl (*p++);
    args->res_size = ntohl (*p++);
    args->chunk_size = ntohl (*p++);
    for (i = 0; i < args->num_fids; ++i) {
        args->read_chunks[i] = ntohl (*p++);
    }
    args->compute_time = ntohl (*p++);

    return p;
}


u_int32_t *
pasis_md_xdr_dec_get_stats_args (u_int32_t *p, 
                                 pasis_md_get_stats_args_t *args)
{
    args->res_max_num_buckets = ntohl (*p++);

    return p;
}


u_int32_t *
pasis_md_xdr_dec_synth_update_args (u_int32_t *p, 
                                    pasis_md_synth_update_args_t *args)
{
    int i;

    args->num_fids = ntohl (*p++);
    for (i = 0; i < args->num_fids; ++i) {
        p = xdr_decode_fid_no_copy (p, &args->fids[i]);
    }
    args->arg_size = ntohl (*p++);
    args->res_size = ntohl (*p++);
    args->chunk_size = ntohl (*p++);
    for (i = 0; i < args->num_fids; ++i) {
        args->modified_chunks[i] = ntohl (*p++);
    }
    args->compute_time = ntohl (*p++);
    args->ts_increment = ntohl (*p++);
    args->client_id = ntohl (*p++);
    args->args = (p_byte*)*p;

    return p;
}



int
pasis_md_xdr_dec_read_op_args (u_int32_t *p, 
                               pasis_md_read_op_args_t *args)
{
    unsigned int i;

    args->part = ntohl (*p++);
    args->read_flags = ntohl (*p++);
    p = xdr_decode_hts (p, &args->ts);
    args->op_type = ntohl (*p++);
    
    switch (args->op_type) {
    case PASIS_MD_READ_HISTORY:
        p =pasis_md_xdr_dec_read_history_args (p, &args->op_args.read_history);
        break;
    case PASIS_MD_GETATTR:
        p = pasis_md_xdr_dec_getattr_args (p, &args->op_args.getattr);
        break;
    case PASIS_MD_LOOKUP:
        p = pasis_md_xdr_dec_lookup_args (p, &args->op_args.lookup);
        break;
    case PASIS_MD_READLINK:
        p = pasis_md_xdr_dec_readlink_args (p, &args->op_args.readlink);
        break;
    case PASIS_MD_READDIR:
        p = pasis_md_xdr_dec_readdir_args (p, &args->op_args.readdir);
        break;
    case PASIS_MD_OBJECT_SYNC:
        p = pasis_md_xdr_dec_object_sync_args (p, &args->op_args.object_sync);
        break;
    case PASIS_MD_SYNTH_QUERY:
        p = pasis_md_xdr_dec_synth_query_args (p, &args->op_args.synth_query);
        break;
    case PASIS_MD_GET_STATS:
        p = pasis_md_xdr_dec_get_stats_args (p, &args->op_args.get_stats);
        break;
    default:
        PASIS_DASSERT (0);
        break;
    }

    if (args->op_type == PASIS_MD_SYNTH_QUERY) {
        args->op_args.synth_query.args = (p_byte *)p;
    }

    args->history_matrices_num = ntohl (*p++);
    PASIS_DASSERT (args->history_matrices_num <= MD_MAX_OBJS_PER_REQ);
    for (i = 0; i < args->history_matrices_num; ++i) {
        p = xdr_decode_hist_matrix (p, &args->history_matrices[i]);
    }

    return 0;
}


int
pasis_md_xdr_dec_write_op_args (u_int32_t *p, 
                                pasis_md_write_op_args_t *args,
                                u_int32_t *op_size,
                                u_int32_t **op)
{
    unsigned int i;

    args->part = ntohl (*p++);
    args->op_type = ntohl (*p++);
    args->mtime[0] = ntohl (*p++);
    args->mtime[1] = ntohl (*p++);
    args->repair_flag = ntohl (*p++);

    if (op != NULL) *op = p;
        
    switch (args->op_type) {
    case PASIS_MD_CREATE:
        p = pasis_md_xdr_dec_create_args (p, &args->op_args.create);
        break;
    case PASIS_MD_LINK:
        p = pasis_md_xdr_dec_link_args (p, &args->op_args.link);
        break;
    case PASIS_MD_UNLINK:
        p = pasis_md_xdr_dec_unlink_args (p, &args->op_args.unlink);
        break;
    case PASIS_MD_SETATTR:
        p = pasis_md_xdr_dec_setattr_args (p, &args->op_args.setattr);
        break;
    case PASIS_MD_RENAME:
        p = pasis_md_xdr_dec_rename_args (p, &args->op_args.rename);
        break;
    case PASIS_MD_BARRIER:
        p = pasis_md_xdr_dec_barrier_args (p, &args->op_args.barrier);
        break;
    case PASIS_MD_SYNTH_UPDATE:
        p = pasis_md_xdr_dec_synth_update_args (p, &args->op_args.synth_update);
        break;
    default:
        printf("%s: invalid op_type: %d\n", __FUNCTION__, args->op_type);
        PASIS_DASSERT (0);
        break;
    }

    if (op_size != NULL) *op_size = p - *op;

    args->history_matrices_num = ntohl (*p++);
    PASIS_DASSERT (args->history_matrices_num <= MD_MAX_OBJS_PER_REQ);
    for (i = 0; i < args->history_matrices_num; ++i) {
        p = xdr_decode_hist_matrix (p, &args->history_matrices[i]);
    }

    for (i = 0; i < args->history_matrices_num; ++i) {
        args->proof_fid_indices[i] = ntohl (*p++);
        if (args->proof_fid_indices[i] != UINT_MAX) {
            p = xdr_decode_hist_matrix (p, &args->proof_matrices[i]);
        }
    }

    if (args->op_type == PASIS_MD_SYNTH_UPDATE) {
        args->op_args.synth_update.args = (p_byte *)p;
    }

    return 0;
}


/* For server recv stub -- send args */

u_int32_t *
pasis_md_xdr_enc_read_history_res (u_int32_t *p, 
                                   pasis_md_read_history_res_t *res)
{
    int i;

    PASIS_DASSERT (res->result_num <= MD_MAX_OBJS_PER_REQ);

    *p++ = htonl (res->result_num);

    for (i = 0; i < res->result_num; i++) {
        *p++ = htonl (res->statuses[i]);
    }

    return p;
}


u_int32_t *
pasis_md_xdr_enc_create_res (u_int32_t *p, 
                             pasis_md_create_res_t *res)
{
    p = xdr_encode_attr (p, &res->create_attr);
    p = xdr_encode_attr (p, &res->parent_attr);

    return p;
}


u_int32_t *
pasis_md_xdr_enc_link_res (u_int32_t *p, 
                           pasis_md_link_res_t *res)
{
    p = xdr_encode_attr (p, &res->from_file_attr);
    p = xdr_encode_attr (p, &res->to_dir_attr);
    return p;
}


u_int32_t *
pasis_md_xdr_enc_unlink_res (u_int32_t *p, 
                             pasis_md_unlink_res_t *res)
{
    p = xdr_encode_attr (p, &res->parent_dir_attr);
    return p;
}


u_int32_t *
pasis_md_xdr_enc_getattr_res (u_int32_t *p, 
                              pasis_md_getattr_res_t *res)
{
    p = xdr_encode_attr (p, &res->attr);
    p = xdr_encode_distribution (p, &res->distribution);
    return p;
}


u_int32_t *
pasis_md_xdr_enc_setattr_res (u_int32_t *p, 
                              pasis_md_setattr_res_t *res)
{
    p = xdr_encode_attr (p, &res->attr);
    return p;
}


u_int32_t *
pasis_md_xdr_enc_rename_res (u_int32_t *p, 
                             pasis_md_rename_res_t *res)
{
    p = xdr_encode_attr (p, &res->from_dir_attr);
    p = xdr_encode_attr (p, &res->to_dir_attr);
    return p;
}


u_int32_t *
pasis_md_xdr_enc_barrier_res (u_int32_t *p,
                              pasis_md_barrier_res_t *res)
{
    return p;
}


u_int32_t *
pasis_md_xdr_enc_object_sync_res (rpc_req_t *req, u_int32_t *p,
                                  pasis_md_object_sync_res_t *res)
{
    p_uint32 i;

    p = xdr_encode_attr (p, &res->attr);
    *p++ = htonl (res->block_size);
    *p++ = htonl (res->num_total_blocks);
    *p++ = htonl (res->num_returned_blocks);

    xdr_adjust_reply_iovec (req, p);

    PASIS_DASSERT(res->data.server.iovec_count == res->num_returned_blocks);
    for (i = 0; i < res->num_returned_blocks; ++i) {
        req->reply_iov[req->reply_iov_cur].iov_base = res->data.server.iovec[i].iov_base;
        req->reply_iov[req->reply_iov_cur].iov_len  = res->block_size;
        req->reply_iov_cur++;
        req->reply_iov_len++;
    }

    return p;
}


u_int32_t *
pasis_md_xdr_enc_lookup_res (u_int32_t *p, 
                             pasis_md_lookup_res_t *res)
{
    p = xdr_encode_fid (p, &res->fid);
    p = xdr_encode_attr (p, &res->dir_attr);
    p = xdr_encode_distribution (p, &res->distribution);
    return p;
}


u_int32_t *
pasis_md_xdr_enc_readlink_res (u_int32_t *p, 
                               pasis_md_readlink_res_t *res)
{
    p = xdr_encode_string (p, res->name);
    p = xdr_encode_attr (p, &res->fid_attr);
    return p;
}


u_int32_t *
pasis_md_xdr_enc_readdir_res (rpc_req_t *req, u_int32_t *p, 
                              pasis_md_readdir_res_t *res)
{
    p = xdr_encode_attr (p, &res->dir_attr);
    p = xdr_encode_hash (p, res->hash);
    *p++ = htonl (res->data_size);

    xdr_adjust_reply_iovec (req, p);

    /* Data goes here... */
    req->reply_iov[req->reply_iov_cur].iov_base = res->data;
    req->reply_iov[req->reply_iov_cur].iov_len = res->data_size;
    req->reply_iov_cur++;
    req->reply_iov_len++;

    return p;
}


u_int32_t *
pasis_md_xdr_enc_synth_query_res (rpc_req_t *req, u_int32_t *p, 
                                  pasis_md_synth_query_res_t *res)
{
    *p++ = htonl (res->res_size);

    xdr_adjust_reply_iovec (req, p);
    
    if (res->res_size > 0) {
        req->reply_iov[req->reply_iov_cur].iov_base = res->res;
        req->reply_iov[req->reply_iov_cur].iov_len = res->res_size;
        req->reply_iov_cur++;
        req->reply_iov_len++;
    }

    return p;
}


u_int32_t *
pasis_md_xdr_enc_get_stats_res (rpc_req_t *req, u_int32_t *p, 
                                pasis_md_get_stats_res_t *res)
{
    *p++ = htonl (res->num_buckets);
    *p++ = htonl (res->bucket_size);

    xdr_adjust_reply_iovec (req, p);
    
    if (res->num_buckets > 0) {
        req->reply_iov[req->reply_iov_cur].iov_base = res->res;
        req->reply_iov[req->reply_iov_cur].iov_len =
            res->num_buckets * sizeof(p_uint32);
        req->reply_iov_cur++;
        req->reply_iov_len++;
    }

    return p;
}


u_int32_t *
pasis_md_xdr_enc_synth_update_res (rpc_req_t *req, u_int32_t *p, 
                                   pasis_md_synth_update_res_t *res)
{
    *p++ = htonl (res->res_size);

    xdr_adjust_reply_iovec (req, p);
    
    if (res->res_size > 0) {
        req->reply_iov[req->reply_iov_cur].iov_base = res->res;
        req->reply_iov[req->reply_iov_cur].iov_len = res->res_size;
        req->reply_iov_cur++;
        req->reply_iov_len++;
    }

    return p;
}


int
pasis_md_xdr_enc_read_op_res (rpc_req_t *req, u_int32_t *p, 
                              pasis_md_read_op_res_t *res)
{
    unsigned int i;

    *p++ = htonl (res->status);
    *p++ = htonl (res->op_status);
    *p++ = htonl (res->op_type);

    p = xdr_encode_hts (p, &res->ts);

    *p++ = htonl (res->history_vec_num);
    PASIS_DASSERT (res->history_vec_num <= MD_MAX_OBJS_PER_REQ);
    for (i = 0; i < res->history_vec_num; i++) {
        p = xdr_encode_history_vector (p, &res->history_vectors[i]);
    }

    switch (res->op_type) {
    case PASIS_MD_READ_HISTORY:
        p = pasis_md_xdr_enc_read_history_res (p, &res->op_res.read_history);
        xdr_adjust_reply_iovec (req, p);
        break;
    case PASIS_MD_GETATTR:
        p = pasis_md_xdr_enc_getattr_res (p, &res->op_res.getattr);
        xdr_adjust_reply_iovec (req, p);
        break;
    case PASIS_MD_LOOKUP:
        p = pasis_md_xdr_enc_lookup_res (p, &res->op_res.lookup);
        xdr_adjust_reply_iovec (req, p);
        break;
    case PASIS_MD_READLINK:
        p = pasis_md_xdr_enc_readlink_res (p, &res->op_res.readlink);
        xdr_adjust_reply_iovec (req, p);
        break;
    case PASIS_MD_READDIR:
        p = pasis_md_xdr_enc_readdir_res (req, p, &res->op_res.readdir);
        /* enc_readdir_res adjusts the reply iovec accordingly */
        break;
    case PASIS_MD_OBJECT_SYNC:
        p = pasis_md_xdr_enc_object_sync_res (req, p, &res->op_res.object_sync);
        /* enc_object_sync_res adjusts the reply iovec accordingly */
        break;
    case PASIS_MD_SYNTH_QUERY:
        p = pasis_md_xdr_enc_synth_query_res (req, p, &res->op_res.synth_query);
        /* enc_synth_query_res adjusts the reply iovec accordingly */
        break;
    case PASIS_MD_GET_STATS:
        p = pasis_md_xdr_enc_get_stats_res (req, p, &res->op_res.get_stats);
        /* enc_get_stats_res adjusts the reply iovec accordingly */
        break;
    default:
        PASIS_DASSERT (0);
        break;
    }

    return 0;
}


int
pasis_md_xdr_enc_write_op_res (rpc_req_t *req, u_int32_t *p, 
                               pasis_md_write_op_res_t *res)
{
    int i;

    *p++ = htonl (res->status);
    *p++ = htonl (res->op_status);
    *p++ = htonl (res->op_type);

    *p++ = htonl (res->history_vec_num);
    PASIS_DASSERT (res->history_vec_num <= MD_MAX_OBJS_PER_REQ);
    for (i = 0; i < res->history_vec_num; i++) {
        p = xdr_encode_history_vector (p, &res->history_vectors[i]);
    }

    switch (res->op_type) {
    case PASIS_MD_CREATE:
        p = pasis_md_xdr_enc_create_res (p, &res->op_res.create);
        xdr_adjust_reply_iovec (req, p);
        break;
    case PASIS_MD_LINK:
        p = pasis_md_xdr_enc_link_res (p, &res->op_res.link);
        xdr_adjust_reply_iovec (req, p);
        break;
    case PASIS_MD_UNLINK:
        p = pasis_md_xdr_enc_unlink_res (p, &res->op_res.unlink);
        xdr_adjust_reply_iovec (req, p);
        break;
    case PASIS_MD_SETATTR:
        p = pasis_md_xdr_enc_setattr_res (p, &res->op_res.setattr);
        xdr_adjust_reply_iovec (req, p);
        break;
    case PASIS_MD_RENAME:
        p = pasis_md_xdr_enc_rename_res (p, &res->op_res.rename);
        xdr_adjust_reply_iovec (req, p);
        break;
    case PASIS_MD_BARRIER:
        p = pasis_md_xdr_enc_barrier_res (p, &res->op_res.barrier);
        xdr_adjust_reply_iovec (req, p);
        break;
    case PASIS_MD_SYNTH_UPDATE:
        p = pasis_md_xdr_enc_synth_update_res (req, p, &res->op_res.synth_update);
        /* enc_synth_query_res adjusts the reply iovec accordingly */
        break;
    default:
        PASIS_DASSERT (0);
        break;
    }

    return 0;
}



/* For client recv stub -- recv results */
u_int32_t *
pasis_md_xdr_dec_read_history_res (u_int32_t *p, 
                                   pasis_md_read_history_res_t *res)
{
    int i;

    res->result_num = ntohl (*p++);

    PASIS_DASSERT (res->result_num <= MD_MAX_OBJS_PER_REQ);

    for (i = 0; i < res->result_num; i++) {
        res->statuses[i] = ntohl (*p++);
    }

    return p;
}


u_int32_t *
pasis_md_xdr_dec_create_res (u_int32_t *p, 
                             pasis_md_create_res_t *res)
{
    p = xdr_decode_attr (p, &res->create_attr);
    p = xdr_decode_attr (p, &res->parent_attr);

    return p;
}


u_int32_t *
pasis_md_xdr_dec_link_res (u_int32_t *p, 
                           pasis_md_link_res_t *res)
{
    p = xdr_decode_attr (p, &res->from_file_attr);
    p = xdr_decode_attr (p, &res->to_dir_attr);
    return p;
}


u_int32_t *
pasis_md_xdr_dec_unlink_res (u_int32_t *p, 
                             pasis_md_unlink_res_t *res)
{
    p = xdr_decode_attr (p, &res->parent_dir_attr);
    return p;
}


u_int32_t *
pasis_md_xdr_dec_getattr_res (u_int32_t *p, 
                              pasis_md_getattr_res_t *res)
{
    p = xdr_decode_attr (p, &res->attr);
    p = xdr_decode_distribution (p, &res->distribution);
    return p;
}


u_int32_t *
pasis_md_xdr_dec_setattr_res (u_int32_t *p, 
                              pasis_md_setattr_res_t *res)
{
    p = xdr_decode_attr (p, &res->attr);
    return p;
}


u_int32_t *
pasis_md_xdr_dec_rename_res (u_int32_t *p, 
                             pasis_md_rename_res_t *res)
{
    p = xdr_decode_attr (p, &res->from_dir_attr);
    p = xdr_decode_attr (p, &res->to_dir_attr);
    return p;
}


u_int32_t *
pasis_md_xdr_dec_barrier_res (u_int32_t *p,
                              pasis_md_barrier_res_t *res)
{
    return p;
}


u_int32_t *
pasis_md_xdr_dec_object_sync_res (u_int32_t *p,
                                  pasis_md_object_sync_res_t *res)
{
    p_uint32 i;

    p = xdr_decode_attr (p, &res->attr);
    res->block_size = ntohl (*p++);
    res->num_total_blocks = ntohl (*p++);
    res->num_returned_blocks = ntohl (*p++);

    memcpy (res->data.client, p, res->num_returned_blocks * res->block_size);
    p += res->num_returned_blocks * res->block_size;

    return p;
}


u_int32_t *
pasis_md_xdr_dec_lookup_res (u_int32_t *p, 
                             pasis_md_lookup_res_t *res)
{
    /* FID */
    p = xdr_decode_fid (p, &res->fid);
    p = xdr_decode_attr (p, &res->dir_attr);
    p = xdr_decode_distribution (p, &res->distribution);
    return p;
}


u_int32_t *
pasis_md_xdr_dec_readlink_res (u_int32_t *p, 
                               pasis_md_readlink_res_t *res)
{
    p = xdr_decode_string (p, res->name);
    p = xdr_decode_attr (p, &res->fid_attr);
    return p;
}


u_int32_t *
pasis_md_xdr_dec_readdir_res (u_int32_t *p, 
                              pasis_md_readdir_res_t *res)
{
    p = xdr_decode_attr (p, &res->dir_attr);
    p = xdr_decode_hash (p, res->hash);
    res->data_size = ntohl (*p++);

    if (res->data) {
        memcpy (res->data, p, res->data_size);
    } else {
        /* If res->data is NULL, then the Readdir request has already completed,
         *   and we must be getting back an early-terminated response.  res->data
         * has been freed so we can't write into it (and it's useless anyways,
         * since we've already finished the operation. */
    }

    p += res->data_size >> 2;

    return p;
}


u_int32_t *
pasis_md_xdr_dec_synth_query_res (u_int32_t *p, 
                                  pasis_md_synth_query_res_t *res)
{
    res->res_size = ntohl (*p++);
    
    /* Data follows -- unfortunately we must do a copy here */
    memcpy (res->res, p, res->res_size);
    p += res->res_size >> 2;

    return p;
}


u_int32_t *
pasis_md_xdr_dec_get_stats_res (u_int32_t *p, 
                                pasis_md_get_stats_res_t *res)
{
    res->num_buckets = ntohl (*p++);
    res->bucket_size = ntohl (*p++);
    
    /* Data follows -- unfortunately we must do a copy here */
    if (res->num_buckets > 0) {
        memcpy (res->res, p, res->num_buckets * sizeof(p_uint32));
        p += res->num_buckets; /* p's type is pointer to word */
    }

    return p;
}


u_int32_t *
pasis_md_xdr_dec_synth_update_res (u_int32_t *p, 
                                   pasis_md_synth_update_res_t *res)
{
    res->res_size = ntohl (*p++);
    
    /* Data follows -- unfortunately we must do a copy here */
    memcpy (res->res, p, res->res_size);
    p += res->res_size >> 2;

    return p;
}


int
pasis_md_xdr_dec_read_op_res (u_int32_t *p, 
                              pasis_md_read_op_res_t *res)
{
    unsigned int i;

    res->status = ntohl (*p++);
    res->op_status = ntohl (*p++);
    res->op_type = ntohl (*p++);
    
    p = xdr_decode_hts (p, &res->ts);

    res->history_vec_num = ntohl (*p++);
    PASIS_DASSERT (res->history_vec_num <= MD_MAX_OBJS_PER_REQ);

    for (i = 0; i < res->history_vec_num; i++) {
        p = xdr_decode_history_vector (p, &res->history_vectors[i]);
    }

    switch (res->op_type) {
    case PASIS_MD_READ_HISTORY:
        p = pasis_md_xdr_dec_read_history_res (p, &res->op_res.read_history);
        break;
    case PASIS_MD_GETATTR:
        p = pasis_md_xdr_dec_getattr_res (p, &res->op_res.getattr);
        break;
    case PASIS_MD_LOOKUP:
        p = pasis_md_xdr_dec_lookup_res (p, &res->op_res.lookup);
        break;
    case PASIS_MD_READLINK:
        p = pasis_md_xdr_dec_readlink_res (p, &res->op_res.readlink);
        break;
    case PASIS_MD_READDIR:
        p = pasis_md_xdr_dec_readdir_res (p, &res->op_res.readdir);
        break;
    case PASIS_MD_OBJECT_SYNC:
        p = pasis_md_xdr_dec_object_sync_res (p, &res->op_res.object_sync);
        break;
    case PASIS_MD_SYNTH_QUERY:
        p = pasis_md_xdr_dec_synth_query_res (p, &res->op_res.synth_query);
        break;
    case PASIS_MD_GET_STATS:
        p = pasis_md_xdr_dec_get_stats_res (p, &res->op_res.get_stats);
        break;
    default:
        PASIS_DASSERT (0);
        break;
    }

    return TRUE;
}


int
pasis_md_xdr_dec_write_op_res (u_int32_t *p, 
                               pasis_md_write_op_res_t *res)
{
    int i;

    res->status = ntohl (*p++);
    res->op_status = ntohl (*p++);
    res->op_type = ntohl (*p++);

    res->history_vec_num = ntohl (*p++);
    PASIS_DASSERT (res->history_vec_num <= MD_MAX_OBJS_PER_REQ);

    for (i = 0; i < res->history_vec_num; i++) {
        p = xdr_decode_history_vector (p, &res->history_vectors[i]);
    }
    
    switch (res->op_type) {
    case PASIS_MD_CREATE:
        p = pasis_md_xdr_dec_create_res (p, &res->op_res.create);
        break;
    case PASIS_MD_LINK:
        p = pasis_md_xdr_dec_link_res (p, &res->op_res.link);
        break;
    case PASIS_MD_UNLINK:
        p = pasis_md_xdr_dec_unlink_res (p, &res->op_res.unlink);
        break;
    case PASIS_MD_SETATTR:
        p = pasis_md_xdr_dec_setattr_res (p, &res->op_res.setattr);
        break;
    case PASIS_MD_RENAME:
        p = pasis_md_xdr_dec_rename_res (p, &res->op_res.rename);
        break;
    case PASIS_MD_BARRIER:
        p = pasis_md_xdr_dec_barrier_res (p, &res->op_res.barrier);
        break;
    case PASIS_MD_SYNTH_UPDATE:
        p = pasis_md_xdr_dec_synth_update_res (p, &res->op_res.synth_update);
        break;

    default:
        PASIS_DASSERT (0);
        break;
    }

    return TRUE;
}


/* For server history code -- reencode operation */
u_int32_t *
pasis_md_xdr_enc_create_args_no_copy (u_int32_t *op)
{
    printf ("%s (%d): op=%p\n", __FUNCTION__, __LINE__, op);
    op = xdr_encode_fid_no_copy (op);
    printf ("%s (%d): op=%p\n", __FUNCTION__, __LINE__, op);
    op = xdr_encode_fid_no_copy (op);
    printf ("%s (%d): op=%p\n", __FUNCTION__, __LINE__, op);
    op = xdr_encode_string_no_copy (op);
    printf ("%s (%d): op=%p\n", __FUNCTION__, __LINE__, op);
    op = xdr_encode_attr_no_copy (op);
    printf ("%s (%d): op=%p\n", __FUNCTION__, __LINE__, op);
    op = xdr_encode_string_no_copy (op);
    printf ("%s (%d): op=%p\n", __FUNCTION__, __LINE__, op);
    op = xdr_encode_distribution_no_copy (op);
    printf ("%s (%d): op=%p\n", __FUNCTION__, __LINE__, op);
    
    return op;
}


u_int32_t *
pasis_md_xdr_enc_link_args_no_copy (u_int32_t *op)
{
    op = xdr_encode_fid_no_copy (op);
    op = xdr_encode_fid_no_copy (op);
    op = xdr_encode_string_no_copy (op);
    return op;
}


u_int32_t *
pasis_md_xdr_enc_unlink_args_no_copy (u_int32_t *op)
{
    op = xdr_encode_fid_no_copy (op);
    op = xdr_encode_fid_no_copy (op);
    op = xdr_encode_string_no_copy (op);
    return op;
}


u_int32_t *
pasis_md_xdr_enc_setattr_args_no_copy (u_int32_t *op)
{
    op = xdr_encode_fid_no_copy (op);
    op = xdr_encode_sattr_no_copy (op);
    *op = htonl (*op); /* guard_ctime[0] */
    op++;
    *op = htonl (*op);/* guard_ctime[1]); */
    op++;

    return op;
}


u_int32_t *
pasis_md_xdr_enc_rename_args_no_copy (u_int32_t *op)
{
    u_int32_t to_name_valid, to_dir_valid;

    op = xdr_encode_fid_no_copy (op);
    *op = to_dir_valid = htonl (*op); /* to_name_fid_valid; */
    op++;
    if (to_dir_valid) {
        op = xdr_encode_fid_no_copy (op);
    }
    op = xdr_encode_fid_no_copy (op);
    *op = to_name_valid = htonl (*op); /* to_name_fid_valid; */
    op++;
    if (to_name_valid) {
        op = xdr_encode_fid_no_copy (op); /* to_name_fid */
    }
    op = xdr_encode_string_no_copy (op);
    op = xdr_encode_string_no_copy (op);
    return op;
}


u_int32_t *
pasis_md_xdr_enc_barrier_args_no_copy (u_int32_t *op)
{
    u_int32_t i, fid_num;

    *op = fid_num = htonl (*op); /* fid_num */
    op++;
    for (i = 0; i < fid_num; i++) {
        op = xdr_encode_fid_no_copy (op);
    }
    return op;
}


u_int32_t *
pasis_md_xdr_enc_object_sync_args_no_copy (u_int32_t *op)
{
    PASIS_DASSERT (0); /* we don't use this... */
}



/*
 * Local variables:
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
